package com.perdemia.permissionanalyzer.application.export;

import gnu.trove.list.array.TIntArrayList;

import java.io.BufferedWriter;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.sql.PreparedStatement;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.commons.mail.EmailException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.e4.ui.workbench.IWorkbench;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.swt.program.Program;
import org.eclipse.swt.widgets.Display;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.E4LifeCycle;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.AclPartContentProvider;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.AllMembersPartContentProvider;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.AllMembersPartContentProviderConfig;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.EffectivePermissionsPartContentProvider;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.EffectivePermissionsPartContentProviderConfig;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.MainFileTableContentProvider;
import com.perdemia.permissionanalyzer.application.jface.dialogs.ReportDialog;
import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.GroupOwnershipColumnLabelProvider;
import com.perdemia.permissionanalyzer.application.logging.GUILogHandler;
import com.perdemia.permissionanalyzer.application.parts.FilterPermissionPart;
import com.perdemia.permissionanalyzer.application.preferences.PreferenceHelper;
import com.perdemia.permissionanalyzer.application.util.ApplicationUtil;
import com.perdemia.permissionanalyzer.application.util.EmailHelper;
import com.perdemia.permissionanalyzer.application.util.FilterConfig;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.FilterHelper.ObjectType;
import com.perdemia.permissionanalyzer.application.util.ImageFactory;
import com.perdemia.permissionanalyzer.application.util.ImageFactory.AppImage;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper.Edition;
import com.perdemia.permissionanalyzer.application.util.PermissionUtil;
import com.perdemia.permissionanalyzer.application.util.ProgramParameters;
import com.perdemia.permissionanalyzer.data.AclItem;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.EffectivePermissions;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Filter;
import com.perdemia.permissionanalyzer.data.Filter.FilterType;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.PermissionLookup;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SimplePermission;
import com.perdemia.permissionanalyzer.data.Report;
import com.perdemia.permissionanalyzer.data.Report.FileType;
import com.perdemia.permissionanalyzer.data.Report.ReportType;
import com.perdemia.permissionanalyzer.data.ReportHistory;
import com.perdemia.permissionanalyzer.data.util.MembershipCache;
import com.perdemia.permissionanalyzer.scanner.acl.AclScanner;

public final class ExportHelper {
	private static Logger logger = LoggerFactory.getLogger(ExportHelper.class.getName());
	private static final String MAIN_TEMPLATE = "MAIN TEMPLATE";
	private static final String FILE_TEMPLATE = "FILE LOOP";
	private static final String ACE_TEMPLATE = "ACE LOOP";
	private static final String MEMBER_TEMPLATE = "MEMBER LOOP";
	private static final String GROUP_TEMPLATE = "GROUP LOOP";
	
	private static final long MB_BYTE_SIZE = 1048576;
	private static final char UTF_8_WITH_BOM_MARKER = '\ufeff';
	
	/**
	 * Exports a report based on the {@link Report#getFileType()} and {@link Report#getReportType()} and the related filters.
	 * 
	 * @param report report to export, the file is exported to {@link Report#getFilePath()}
	 * @param silentMode used when reports are generated using command line. This will prevent Message dialogs.
	 * @param startFolder optional start point to retrieve files. The column FileInfo.PARENT_FILE_IDS is used to filter. 
	 * 			Example value "1,34,735", meaning only get sub-folders of the folder 1 > 34 > 735
	 *  
	 */
	public static void exportReport(final Report report, final FileInfo startFolder, final boolean exportSubdirectories, final boolean showOpenReportQuestion) {
		final long startTime = System.currentTimeMillis();
		System.out.println("Start: " + System.currentTimeMillis());
		final PreferenceHelper prefs = PreferenceHelper.getInstance();
				
		// start with a progress monitor
		final ProgressMonitorDialog progressDialog = new ProgressMonitorDialog(Display.getDefault().getActiveShell());
		try {
			progressDialog.run(true, true, new IRunnableWithProgress() {

				@Override
				public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {				
					FilterHelper filterHelper = new FilterHelper();
					boolean memberMaxWarningLogged = false;
					
					// keep track of the number of characters written to a file (in order to split CSV files)
					long characterCountWrittenToFile = 0;
					int csvFileCounter = 1;
					
					// get the report target path
					String reportFilePath = report.getFilePath();
					
					boolean printToFile = !reportFilePath.isEmpty();
					boolean isPolicy = report.getEmailThreshold() > 0;
					
					// date format in the path?
					int start = reportFilePath.indexOf("[date:");
					if (start > -1) {
						int end = reportFilePath.indexOf(']', start);
						String dateFormat = reportFilePath.substring(start + 6, end);
						SimpleDateFormat dateFormatter = new SimpleDateFormat(dateFormat); 
						reportFilePath = replaceField(reportFilePath, "date:" + dateFormat, dateFormatter.format(new Date()), false, false, report);
					}
					
					// computer name in the path?
					reportFilePath = replaceField(reportFilePath, "computerName", AclScanner.getLocalMachineName(), false, false, report);
					reportFilePath = replaceField(reportFilePath, "ComputerName", AclScanner.getLocalMachineName(), false, false, report);
										
					// first delete any existing report
					try {
						if (printToFile) {
							java.io.File reportFile = new java.io.File(reportFilePath);
							if (reportFile.exists()) {
								reportFile.delete();
							}
							reportFile.getParentFile().mkdirs();
						}
					} catch (Exception ex) {
						logger.error("Error while creating the report file.", ex);
						SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while creating the report file: " + ex.getMessage());
						return;
					}
					
					// get the file templates
					String currentTemplateName = null;					
					String templateContent = "";
					String jqueryJSContent = "";
					String clipboardJSContent = "";
					String jqueryDataTableJSContent = "";
					String jqueryDataTableCSSContent = "";
					String reportCSSContent = "";
					
					try {
						// template file
						if (report.getTemplatePath() != null && report.getTemplatePath().length() > 0) {
							java.io.File templateFile = new java.io.File(report.getTemplatePath());
							templateContent = new String(Files.readAllBytes(templateFile.toPath()));
							
						} else {
							// find the path to the embedded template file
							String extension = "";
							switch (report.getFileType()) {
								case HTML: extension = ".html"; break;
								case HTML_JQUERY: extension = "-jquery.html"; break;
								case CSV: extension = ".csv"; break;
							}
							
							// determine the template file name based on the report type 
							switch (report.getReportType()) {
								case GROUPS_WITH_PERMISSIONS_AND_DIRECT_MEMBERS:
								case MATCHING_GROUPS_AND_DIRECT_MEMBERS:
									currentTemplateName = "template-direct_group_members" + extension;
									break;
									
								case GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS:
								case MATCHING_GROUPS_AND_NESTED_MEMBERS:
									currentTemplateName = "template-nested_group_members" + extension;
									break;
									
								case MATCHING_USERS_AND_DIRECT_GROUP_MEMBERSHIPS:
									currentTemplateName = "template-users_and_direct_group_memberships" + extension;
									break;
								
								case MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS:
									currentTemplateName = "template-users_and_nested_group_memberships" + extension;
									break;
									
								default:
									currentTemplateName = "template-" + report.getReportType().toString().toLowerCase();
									
									if (report.isUseSimplePresentation() && report.getFileType().equals(FileType.HTML)) {
										currentTemplateName += "_simple";
									}
									
									currentTemplateName += extension;
									break;
							}
							
							String templatePath;
							if (prefs.isCalculateOwnershipRatioInOverview() && report.getMemberFilters().size() > 1
									&& !report.getReportType().equals(ReportType.MATCHING_GROUPS_AND_DIRECT_MEMBERS) && !report.getReportType().equals(ReportType.MATCHING_GROUPS_AND_NESTED_MEMBERS)
									&& !report.getReportType().equals(ReportType.GROUPS_WITH_PERMISSIONS_AND_DIRECT_MEMBERS) && !report.getReportType().equals(ReportType.GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS)  ) {
								templatePath = "templates/with_ownership_ratio/" + currentTemplateName;
							} else {
								templatePath = "templates/" + currentTemplateName;
							}
							templateContent = ApplicationUtil.getFileContent(templatePath);
							
							// CSV separator
							if (report.getFileType().equals(FileType.CSV)) {
								templateContent = templateContent.replace(",", prefs.getCsvSeparator());
							}
						}
						
						// Jquery JS
						currentTemplateName = "templates/resources/jquery.js";
						jqueryJSContent = ApplicationUtil.getFileContent(currentTemplateName);
						
						// Clipboard JS
						currentTemplateName = "templates/resources/clipboard.min.js";
						clipboardJSContent = ApplicationUtil.getFileContent(currentTemplateName);
						
						// Jquery DataTable JS
						currentTemplateName = "templates/resources/jquery.dataTables.min.js";
						jqueryDataTableJSContent = ApplicationUtil.getFileContent(currentTemplateName);
						
						// Jquery DataTable CSS
						currentTemplateName = "templates/resources/jquery.dataTables.min.css";
						jqueryDataTableCSSContent = ApplicationUtil.getFileContent(currentTemplateName);
						
						// Report CSS
						currentTemplateName = "templates/resources/styles.css";
						reportCSSContent = ApplicationUtil.getFileContent(currentTemplateName);
						
					} catch (Exception ex) {
						logger.error("Error while reading the template content using path " + currentTemplateName, ex);
						SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error reading template", "Error while reading the template content using path " + currentTemplateName + ":\r\n\r\n" + ex.getMessage());
						
						return;
					}
					
					BufferedWriter writer = null;
					int memberCount = 0;
					int fileCount = 0;
					
					try {
						// determine a human-readable summary of the report filters
						ReportFilters reportFilters = new ReportFilters();
						reportFilters.setMemberFilters(report.getMemberFilters());
						reportFilters.setMemberFilterSet(report.getMemberFilterSet());
						reportFilters.setPermissionFilters(report.getPermissionFilters());
						reportFilters.setPermissionFilterSet(report.getPermissionFilterSet());
						reportFilters.setFolderFilters(report.getFolderFilters());
						reportFilters.setFolderFilterSet(report.getFolderFilterSet());
						reportFilters.setLinkToMemberFilterSet(true);
						reportFilters.setLinkToPermissionFilterSet(true);
						reportFilters.setLinkToFolderFilterSet(true);
						List<FilterRow> filterRows = determineFilterRows(reportFilters);
						
						// add the filter rows to the template content
						templateContent = replaceField(templateContent, "filterInfo", filterRowsToHTML(filterRows), false, false, report);
						
						// get the template for all the 'blocks' used in the loops
						String mainTemplate = getTemplateBlock(MAIN_TEMPLATE, templateContent);
						String fileTemplate = getTemplateBlock(FILE_TEMPLATE, templateContent);
						String aceTemplate = getTemplateBlock(ACE_TEMPLATE, templateContent);
						String memberTemplate = getTemplateBlock(MEMBER_TEMPLATE, templateContent);
						String groupTemplate = getTemplateBlock(GROUP_TEMPLATE, templateContent);
						
						// input for the ALL_MEMBERS report
						List<Member> allMembers = null;
						AllMembersPartContentProvider allMembersContentProvider = null;
						
						// input for the other reports
						TIntArrayList uniqueFileIds = null;
																		
						// Apply the filter config of this report
						filterHelper.setMemberFilterList(report.getMemberFilters());
						filterHelper.setPermissionFilterList(report.getPermissionFilters());
						filterHelper.setFolderFilterList(report.getFolderFilters());
						filterHelper.applyNewFilter(false);
						
						if (report.getReportType().equals(ReportType.ALL_MEMBERS)) {
							monitor.setTaskName("Finding matching members for the report");
							
							// use the same content as the all members tab
							allMembersContentProvider = new AllMembersPartContentProvider(null);
							AllMembersPartContentProviderConfig providerConfig = new AllMembersPartContentProviderConfig("", filterHelper);
							allMembersContentProvider.inputChanged(null, null, providerConfig);
							allMembers = allMembersContentProvider.getMembers();
							memberCount = allMembers.size();
							mainTemplate = replaceField(mainTemplate, "memberCount", String.valueOf(memberCount), false, false, report);
						
						} else if (report.getReportType().equals(ReportType.MATCHING_GROUPS_AND_DIRECT_MEMBERS) || report.getReportType().equals(ReportType.MATCHING_GROUPS_AND_NESTED_MEMBERS)) {
							monitor.setTaskName("Finding matching groups and members for the report");
							
							// get all GROUPS members that match the filter criteria
							int limit = 0;
							if (LicenseHelper.getEdition().equals(Edition.TRIAL)) {
								limit = 30;
							}
							allMembers = Member.getByPreparedStatement(filterHelper.buildFilterQuery(ObjectType.MEMBER, false, false, limit, "FILTER_MEMBER.DOMAIN_ID, FILTER_MEMBER.NAME_LOWERCASE", "FILTER_MEMBER.IS_GROUP = true"));
							
							memberCount = allMembers.size();
							mainTemplate = replaceField(mainTemplate, "memberCount", String.valueOf(memberCount), false, false, report);
							
						} else if (report.getReportType().equals(ReportType.MATCHING_USERS_AND_DIRECT_GROUP_MEMBERSHIPS) || report.getReportType().equals(ReportType.MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS)) {
							monitor.setTaskName("Finding matching users and group memberships for the report");
							
							// get all USERS members that match the filter criteria
							int limit = 0;
							if (LicenseHelper.getEdition().equals(Edition.TRIAL)) {
								limit = 30;
							}
							allMembers = Member.getByPreparedStatement(filterHelper.buildFilterQuery(ObjectType.MEMBER, false, false, limit, "FILTER_MEMBER.DOMAIN_ID, FILTER_MEMBER.NAME_LOWERCASE", "FILTER_MEMBER.IS_GROUP = false"));
							
							memberCount = allMembers.size();
							mainTemplate = replaceField(mainTemplate, "memberCount", String.valueOf(memberCount), false, false, report);
							
						} else if (report.getReportType().equals(ReportType.GROUPS_WITH_PERMISSIONS_AND_DIRECT_MEMBERS) || report.getReportType().equals(ReportType.GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS)) {
							monitor.setTaskName("Finding groups with permissions and their members for the report");
							
							// get all GROUPS that actually have permission in the current folder tree
							// use the same content as the all members tab and only use the groups
							allMembersContentProvider = new AllMembersPartContentProvider(null);
							
							// content provider config to filter on groups 
							AllMembersPartContentProviderConfig providerConfig = new AllMembersPartContentProviderConfig("", filterHelper);
							providerConfig.setAdditionalFilter("FILTER_MEMBER.IS_GROUP = TRUE");
							
							if (LicenseHelper.getEdition().equals(Edition.TRIAL)) {
								providerConfig.setLimit(30);
							}
							
							allMembersContentProvider.inputChanged(null, null, providerConfig);
							allMembers = allMembersContentProvider.getMembers();
							memberCount = allMembers.size();
							mainTemplate = replaceField(mainTemplate, "memberCount", String.valueOf(memberCount), false, false, report);
							
						} else {
							// apply the report filters in the FilterHelper
							monitor.setTaskName("Executing query on the database");
														
							// execute the query and extract all the applicable file ids
							PreparedStatement ps;
							
							if (startFolder != null) {
								if (exportSubdirectories) {
									ps = filterHelper.buildFilterQuery(ObjectType.FILE, false, false, 0, "FILEINFO.PARENT_FILE_IDS ASC, FILEINFO.NAME_LOWERCASE ASC", "FILEINFO.PARENT_FILE_IDS = ? OR FILEINFO.PARENT_FILE_IDS LIKE ?", startFolder.getParentFileIds(), startFolder.getParentFileIds() + ",%");
								} else {
									// select the given directory and its parents
									String[] ids = startFolder.getParentFileIds().split(",");
									ps = filterHelper.buildFilterQuery(ObjectType.FILE, false, false, 0, "FILEINFO.PARENT_FILE_IDS ASC, FILEINFO.NAME_LOWERCASE ASC", "FILEINFO.ID IN " + Arrays.asList(ids).toString().replace("[", "(").replace("]", ")"));
								}
							} else {															
								ps = filterHelper.buildFilterQuery(ObjectType.FILE, false, false, 0, "FILEINFO.PARENT_FILE_IDS ASC, FILEINFO.NAME_LOWERCASE ASC", null);
							}
													
							uniqueFileIds = filterHelper.getApplicableFileIds(ps, false, true);
							fileCount = uniqueFileIds.size();
							
							mainTemplate = replaceField(mainTemplate, "fileCount", String.valueOf(fileCount), false, false, report);
						}
						
						if (printToFile) {
						
							// apply generic placeholders
							mainTemplate = replaceField(mainTemplate, "jqueryJS", jqueryJSContent, false, false, report);
							mainTemplate = replaceField(mainTemplate, "jqueryDataTableJS", jqueryDataTableJSContent, false, false, report);
							mainTemplate = replaceField(mainTemplate, "jqueryDataTableCSS", jqueryDataTableCSSContent, false, false, report);
							mainTemplate = replaceField(mainTemplate, "reportCSS", reportCSSContent, false, false, report);
							mainTemplate = replaceField(mainTemplate, "clipboardJS", clipboardJSContent, false, false, report);
							mainTemplate = replaceField(mainTemplate, "logoBase64", prefs.getLogoBase64(), false, false, report);
							mainTemplate = replaceField(mainTemplate, "logoWidth", String.valueOf(prefs.getLogoWidth()), false, false, report);
							mainTemplate = replaceField(mainTemplate, "logoHeight", String.valueOf(prefs.getLogoHeight()), false, false, report);
							mainTemplate = replaceField(mainTemplate, "logoTop", String.valueOf(prefs.getLogoLocationTop()), false, false, report);
							mainTemplate = replaceField(mainTemplate, "logoRight", String.valueOf(prefs.getLogoLocationRight()), false, false, report);
							
							// find the date format
							start = mainTemplate.indexOf("[date:");
							if (start > -1) {
								int end = mainTemplate.indexOf(']', start);
								String dateFormat = mainTemplate.substring(start + 6, end);
								SimpleDateFormat dateFormatter = new SimpleDateFormat(dateFormat); 
								mainTemplate = replaceField(mainTemplate, "date:" + dateFormat, dateFormatter.format(new Date()), true, false, report);
							}
							
							mainTemplate = replaceField(mainTemplate, "reportName", report.getName(), true, false, report);
							mainTemplate = replaceField(mainTemplate, "reportDescription", report.getDescription(), true, false, report);
							mainTemplate = replaceField(mainTemplate, "reportType", report.getReportType().name().toLowerCase(), true, false, report);
													
							// determine the header and footer of the template. Header is everything before the 'start directory loop' and the footer is everything after the 'end directory loop'
							String fileTemplateTag = "${" + FILE_TEMPLATE + "}";
							String aceTemplateTag = "${" + ACE_TEMPLATE + "}";
							String groupTemplateTag = "${" + GROUP_TEMPLATE + "}";
							String memberTemplateTag = "${" + MEMBER_TEMPLATE + "}";
							
							String header;
							String footer;
							if (mainTemplate.indexOf(fileTemplateTag) > -1) {
								header = mainTemplate.substring(0, mainTemplate.indexOf(fileTemplateTag));
								footer = mainTemplate.substring(mainTemplate.indexOf(fileTemplateTag) + fileTemplateTag.length());
								
							} else if (mainTemplate.indexOf(groupTemplateTag) > -1) {
								header = mainTemplate.substring(0, mainTemplate.indexOf(groupTemplateTag));
								footer = mainTemplate.substring(mainTemplate.indexOf(groupTemplateTag) + groupTemplateTag.length());
								
							} else {
								header = mainTemplate.substring(0, mainTemplate.indexOf(memberTemplateTag));
								footer = mainTemplate.substring(mainTemplate.indexOf(memberTemplateTag) + memberTemplateTag.length());
							}
							
							// open an BufferedWriter to the target file and print the initial template content to the file
							writer = Files.newBufferedWriter(Paths.get(reportFilePath), Charset.forName("UTF-8"), StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.CREATE);
							
							// write the UTF-8 representation of the BOM to get UTF-8 with BOM
							// that way Excel will recognize UTF-8 characters in a CSV export!
							writer.write(UTF_8_WITH_BOM_MARKER);
							
							//writer = new FileWriter(reportFilePath);
							writer.write(header);
							
							boolean maxDepthWarningDisplayed = false;
							
							// -------------------------------------------
							// ALL MEMBERS REPORT
							// -------------------------------------------
							if (report.getReportType().equals(ReportType.ALL_MEMBERS)) {
														
								monitor.beginTask("Building report", allMembers.size());
								
								for (int counter = 0; counter < allMembers.size(); counter++) {
									Member member = allMembers.get(counter);
									
									if (monitor.isCanceled()) {
										return;
									}
									monitor.subTask("Processing member " + member.getName(true));
									
									// load all the explicit ACE's that belong to this member
									// the updateChildCount method of the content provider will fill member.getAclItemList()
									allMembersContentProvider.updateChildCount(member, 0);
									
									if (member.getAclItemList().size() > 0 && member.getAclItemList().get(0).getMemberId() != null) {
										String appliedMemberTemplate = replaceField(memberTemplate, "memberClass", member.isGroup() ? "g" : "u", false, true, report);
										appliedMemberTemplate = replaceField(appliedMemberTemplate, "viaMember", PreferenceHelper.getInstance().isUseUserDisplayNameInReport() ? member.getDisplayName(true, false) : member.getName(true), true, true, report);
										appliedMemberTemplate = replaceField(appliedMemberTemplate, "aceCount", String.valueOf(member.getAclItemList().size()), true, true, report);
										
										StringBuffer aceList = new StringBuffer();
										for (AclItem aclItem : member.getAclItemList()) {
											// double check on the Trial license, the database may contain scanned data from another installation of Permission Analyzer
											if (LicenseHelper.getEdition().equals(Edition.TRIAL) && aclItem.getFile(true).getParentFileIds() != null) {
												int commaCount = aclItem.getFile(true).getParentFileIds().length() - aclItem.getFile(true).getParentFileIds().replace(",", "").length();
												if (commaCount > 3) {
													if (!maxDepthWarningDisplayed) {
														logger.warn("File depth in the report not allowed for the trial license");
														maxDepthWarningDisplayed = true;
													}
													continue;
												}
											}
											
											// replace ACE fields
											aclItem.setViaMember(member);
											
											String appliedAceTemplate = replaceField(aceTemplate, "memberClass", member.isGroup() ? "g" : "u", false, true, report);
											appliedAceTemplate = applyAclItemToTemplate(appliedAceTemplate, aclItem, aclItem.getFile(true), report, null, String.valueOf(member.getAclItemList().size()), filterHelper); 
											aceList.append(appliedAceTemplate.toString());
										}
										
										// apply the list of ACE's in the member template and write to the writer
										appliedMemberTemplate = appliedMemberTemplate.replace(aceTemplateTag, aceList.toString());
										
										writer.write(appliedMemberTemplate);
										characterCountWrittenToFile += appliedMemberTemplate.length();
										
										//if (counter % 10 == 0) {
										//	writer.flush();
										//}
									}
									
									monitor.worked(1);
								}
								
							// -------------------------------------------
							// GROUP REPORT
							// -------------------------------------------
							} else if (report.getReportType().equals(ReportType.MATCHING_GROUPS_AND_DIRECT_MEMBERS) || report.getReportType().equals(ReportType.MATCHING_GROUPS_AND_NESTED_MEMBERS)
									|| report.getReportType().equals(ReportType.GROUPS_WITH_PERMISSIONS_AND_DIRECT_MEMBERS) || report.getReportType().equals(ReportType.GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS)) {
								
								monitor.beginTask("Building report", allMembers.size());
								
								for (int counter = 0; counter < allMembers.size(); counter++) {
									Member group = allMembers.get(counter);
									
									if (monitor.isCanceled()) {
										return;
									}
									monitor.subTask("Processing group " + group.getName(true));
									
									// load all the matching group members
									// we first only included matching group members, but the customer wants to exclude AD groups and users, which means the AD object would also disappear as local group member.
									// for now we display every local group member
									//List<Member> members = MembershipCache.getFiltedGroupMembers(group, true, true, report.getReportType().equals(ReportType.MATCHING_GROUPS_AND_NESTED_MEMBERS), true, filterHelper);
									List<Member> members = MembershipCache.getGroupMembers(group, true, true, report.getReportType().equals(ReportType.MATCHING_GROUPS_AND_NESTED_MEMBERS) || report.getReportType().equals(ReportType.GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS));
									
									String appliedGroupTemplate = replaceField(groupTemplate, "groupName", group.getName(true), true, true, report);
									appliedGroupTemplate = replaceField(appliedGroupTemplate, "groupMemberCount", String.valueOf(members.size()), true, true, report);
									appliedGroupTemplate = replaceField(appliedGroupTemplate, "description", group.getDisplayName(false, false), true, true, report);
									
									if (members.isEmpty() && report.getFileType().equals(FileType.CSV)) {
										// create a dummy CSV record
										Member dummy = new Member("None");
										members = new ArrayList<Member>();
										members.add(dummy);
									}
									
									StringBuffer memberList = new StringBuffer();
									for (Member member : members) {
										// replace group member fields
										String appliedMemberTemplate = replaceField(memberTemplate, "groupName", group.getName(true), true, true, report);
										appliedMemberTemplate = replaceField(appliedMemberTemplate, "groupMemberCount", String.valueOf(members.size()), true, true, report);
										appliedMemberTemplate = replaceField(appliedMemberTemplate, "description", group.getDisplayName(false, false), true, true, report);
										appliedMemberTemplate = replaceField(appliedMemberTemplate, "memberName", PreferenceHelper.getInstance().isUseUserDisplayNameInReport() ? member.getDisplayName(true, false) : member.getName(true), true, true, report);
										appliedMemberTemplate = replaceField(appliedMemberTemplate, "memberDisplayName", member.getDisplayName(false, false), true, true, report);
										appliedMemberTemplate = replaceField(appliedMemberTemplate, "memberIsEnabled", member.isEnabled() ? "yes" : "no", true, true, report);
										appliedMemberTemplate = replaceField(appliedMemberTemplate, "viaGroup", member.getViaGroupName() != null ? member.getViaGroupName() : "", true, true, report);
										appliedMemberTemplate = replaceField(appliedMemberTemplate, "memberClass", member.isGroup() ? "g" : "u", false, true, report);
										memberList.append(appliedMemberTemplate.toString());
									}
									
									// apply the list of ACE's in the member template and write to the writer
									appliedGroupTemplate = appliedGroupTemplate.replace(memberTemplateTag, memberList.toString());
									
									writer.write(appliedGroupTemplate);
									characterCountWrittenToFile += appliedGroupTemplate.length();
									
									//if (counter % 10 == 0) {
									//	writer.flush();
									//}
																		
									monitor.worked(1);
								}
								
							// -------------------------------------------
							// USER REPORT
							// -------------------------------------------
							} else if (report.getReportType().equals(ReportType.MATCHING_USERS_AND_DIRECT_GROUP_MEMBERSHIPS) || report.getReportType().equals(ReportType.MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS)) {
								
								monitor.beginTask("Building report", allMembers.size());
								
								for (int counter = 0; counter < allMembers.size(); counter++) {
									Member user = allMembers.get(counter);
									
									if (monitor.isCanceled()) {
										return;
									}
									monitor.subTask("Processing user " + user.getName(true));
									
									// load all the matching memberOf groups
									List<Member> groups = Member.getMemberOfList(user, null, report.getReportType().equals(ReportType.MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS), 0, "3", 0);
									
									String appliedMemberTemplate = replaceField(memberTemplate, "username", user.getName(true), true, true, report);
									appliedMemberTemplate = replaceField(appliedMemberTemplate, "groupCount", String.valueOf(groups.size()), true, true, report);
									appliedMemberTemplate = replaceField(appliedMemberTemplate, "description", user.getDisplayName(false, false), true, true, report);
									
									if (groups.isEmpty() && report.getFileType().equals(FileType.CSV)) {
										// create a dummy CSV record
										Member dummy = new Member("None");
										groups = new ArrayList<Member>();
										groups.add(dummy);
									}
									
									StringBuffer groupList = new StringBuffer();
									for (Member group : groups) {
										// replace user/group fields
										String appliedGroupTemplate = replaceField(groupTemplate, "username", user.getName(true), true, true, report);
										appliedGroupTemplate = replaceField(appliedGroupTemplate, "groupCount", String.valueOf(groups.size()), true, true, report);
										appliedGroupTemplate = replaceField(appliedGroupTemplate, "description", user.getDisplayName(false, false), true, true, report);
										appliedGroupTemplate = replaceField(appliedGroupTemplate, "userIsEnabled", user.isEnabled() ? "yes" : "no", true, true, report);
										appliedGroupTemplate = replaceField(appliedGroupTemplate, "groupName", group.getDisplayName(true, false), true, true, report);
										appliedGroupTemplate = replaceField(appliedGroupTemplate, "groupDisplayName", group.getDisplayName(false, false), true, true, report);
										appliedGroupTemplate = replaceField(appliedGroupTemplate, "groupIsEnabled", group.isEnabled() ? "yes" : "no", true, true, report);
										appliedGroupTemplate = replaceField(appliedGroupTemplate, "viaGroup", group.getViaGroupName() != null ? group.getViaGroupName() : "", true, true, report);
										groupList.append(appliedGroupTemplate.toString());
									}
									
									// apply the list of ACE's in the member template and write to the writer
									appliedMemberTemplate = appliedMemberTemplate.replace(groupTemplateTag, groupList.toString());
									
									writer.write(appliedMemberTemplate);
									characterCountWrittenToFile += appliedMemberTemplate.length();
									
									//if (counter % 10 == 0) {
									//	writer.flush();
									//}
																		
									monitor.worked(1);
								}
							
							// -------------------------------------------
							// FILE BASED REPORT
							// -------------------------------------------
							} else {
								// use the content provider of the Table viewer in the FilePart
								// the provider loads all the parent directories to build up the complete file path and the ACL information
								MainFileTableContentProvider fileContentProvider = new MainFileTableContentProvider(null); //no viewer
								fileContentProvider.inputChanged(null, null, uniqueFileIds);
								
								// use the content provider 
								EffectivePermissionsPartContentProvider effectivePermissionsProvider = new EffectivePermissionsPartContentProvider(null);
								EffectivePermissionsPartContentProviderConfig effectivePermissionsProviderConfig = new EffectivePermissionsPartContentProviderConfig(null, true, "", true, filterHelper);
								
								if (fileCount > 10000 && !E4LifeCycle.isSilentMode) {
									if (!SyncMessageDialog.showMessage(MessageDialog.CONFIRM, "Large data set", "You are about to export " + fileCount + " directories and files, are you sure you want to continue?")) {
										return;
									}
								}
								
								monitor.beginTask("Building report", fileCount);
								
								// process all the file ids and add the permission info to the export file
								for (int counter = 0; counter < fileCount; counter++) {
									if (monitor.isCanceled()) {
										return;
									}
									
									// load file information
									FileInfo file = fileContentProvider.loadFileInfo(counter, filterHelper);
									
									// double check on the Trial license, the database may contain scanned data from another installation of Permission Analyzer
									if (LicenseHelper.getEdition().equals(Edition.TRIAL) && file.getParentFileIds() != null) {
										int commaCountStartFolder = startFolder != null ? startFolder.getParentFileIds().length() - startFolder.getParentFileIds().replace(",", "").length() : 0;
										int commaCount = file.getParentFileIds().length() - file.getParentFileIds().replace(",", "").length();
										if (commaCount - commaCountStartFolder > 3) {
											if (!maxDepthWarningDisplayed) {
												logger.warn("The file depth in the report is not allowed for the trial license");
												SyncMessageDialog.showMessage(MessageDialog.WARNING, "Warning", "The file depth in the report is limited to 3 levels for the trial license, the report will not be complete.");
												maxDepthWarningDisplayed = true;
											}
											continue;
										}
									}
									
									if (file == null) {
										logger.warn("FileInfo with index " + counter + " not found in the database.");
										continue;
									}
									
									// -------------------------------------------
									// DETERMINE THE MEMBER COUNT FOR IN THE REPORT
									// -------------------------------------------
									int fileMemberCount = 0;
									List<AclItem> aclItems = file.getAclItems();
									List<Member> usersOnFolder = null; // used for the FOLDERS_AND_USERS_EFFECTIVE report and the member count
									boolean isAclReport = report.getReportType().equals(ReportType.ACL) || report.getReportType().equals(ReportType.ACL_DIRECT_GROUP_MEMBERS) || report.getReportType().equals(ReportType.ACL_NESTED_GROUP_MEMBERS);
									boolean aclWithGroupMembers = report.getReportType().equals(ReportType.ACL_NESTED_GROUP_MEMBERS) || report.getReportType().equals(ReportType.ACL_DIRECT_GROUP_MEMBERS);
																										
									// memberCount for the ACL report's
									if (isAclReport) {
										for (AclItem aclItem : aclItems) {
											fileMemberCount++;
											
											// ACL report with group members?
											if (aclWithGroupMembers && aclItem.getMember(true).isGroup()) {
												// get the (nested) members of this group
												aclItem.getMember(true).setMembers(MembershipCache.getFiltedGroupMembers(aclItem.getMember(true), filterHelper.getFilterConfig(false).isIncludeUsers(), filterHelper.getFilterConfig(false).isIncludeGroups(), 
														report.getReportType().equals(ReportType.ACL_NESTED_GROUP_MEMBERS), true, true, filterHelper));
												fileMemberCount += aclItem.getMember(true).getMembers().size();
											}
										}
									} else if (report.getReportType().equals(ReportType.FOLDERS_AND_USERS_EFFECTIVE)) {
										// File report with users?
										if (!aclItems.isEmpty()) {
											StringBuilder additionalMemberFilterBuilder = new StringBuilder("FILTER_MEMBER.IS_GROUP = FALSE AND FILTER_MEMBERSHIP.GROUP_ID IN (");
											List<Object> additionalMemberFilterValues = new ArrayList<Object>();
											for (AclItem aclItem : aclItems) {
												additionalMemberFilterBuilder.append("?");
												additionalMemberFilterValues.add(aclItem.getMemberId());
												additionalMemberFilterBuilder.append(",");
											}
											additionalMemberFilterBuilder.deleteCharAt(additionalMemberFilterBuilder.length()-1);
											additionalMemberFilterBuilder.append(")");
											
											// IncludeUsers and IncludeGroups checkboxes are handled in buildFilterQuery 
											PreparedStatement searchMembersPS = filterHelper.buildFilterQuery(ObjectType.MEMBER, false, false, 0, "NAME_LOWERCASE", additionalMemberFilterBuilder.toString(), additionalMemberFilterValues.toArray());
											usersOnFolder = Member.getByPreparedStatement(searchMembersPS);
											
										} else {
											// use an empty list
											usersOnFolder = new ArrayList<Member>();
										}
										
										fileMemberCount = usersOnFolder.size();
									} else {
										fileMemberCount = aclItems.size();
									}
									
									if (fileMemberCount >= 1000) {
										if (!memberMaxWarningLogged) {
											logger.warn("Found " + fileMemberCount + " members found for the report " + report.getName() + " and file path " + file.determineFullPath());
											if (!SyncMessageDialog.showMessage(MessageDialog.QUESTION, "Warning", "Found " + fileMemberCount + " members on one directory (" + file.determineFullPath() + "). You might want to refine your filters, perhaps excluding certain groups or disable group memberships. Do you want to continue with the export?", true)) {
												return;
											}
											memberMaxWarningLogged = true;
										}
									}
									
									// fill the file info place holders
									String appliedFileTemplate = applyFileToTemplate(fileTemplate, file, report, String.valueOf(fileMemberCount), null, filterHelper);
																
									// -----------------------------------------------------------------------
									// REPORT TYPE ACL, ACL_DIRECT_GROUP_MEMBERS OR ACL_NESTED_GROUP_MEMBERS
									// -----------------------------------------------------------------------
									if (isAclReport) {
										StringBuilder aceList = new StringBuilder();
										
										// determine the 'inherited from' folder path (using FileAclContentProvider), we only have the 'inherited from' folder id
										AclPartContentProvider.determineInheritedFromFolderPath(aclItems, file);
										
										effectivePermissionsProviderConfig.setSelectedFile(file);
										effectivePermissionsProvider.setConfig(effectivePermissionsProviderConfig);
										
										if (!aclItems.isEmpty()) {
											// loop through the ACE's of this file and apply them to the template
											for (AclItem aclItem : aclItems) {
												
												// replace ACE fields
												String appliedAceTemplate = applyAclItemToTemplate(aceTemplate, aclItem, file, report, String.valueOf(fileMemberCount), null, filterHelper); 
												
												if (aclWithGroupMembers) {
													StringBuilder memberList = new StringBuilder();
													
													if (aclItem.getMember(true).isGroup() && aclItem.getMember(true).getMembers().size() > 0) {
														// add all members of the group to the report
														for (int memberIndex = 0; memberIndex < aclItem.getMember(true).getMembers().size(); memberIndex++) {
															// the index starts with 1 because the effectivePermissionsProvider also counts the
															// ACE's of the given members and has always a dummy ACE "No explicit ACE"!
															effectivePermissionsProvider.updateElement(aclItem.getMember(true), memberIndex + 1);
															
															Member aceMember = aclItem.getMember(true).getMembers().get(memberIndex);
															String prefix = report.getFileType().equals(FileType.HTML) ? "groupMember" : "";
															
															String appliedMemberTemplate = replaceField(memberTemplate, "groupMemberName", PreferenceHelper.getInstance().isUseUserDisplayNameInReport() ? aceMember.getDisplayName(true, false) : aceMember.getName(true), true, true, report);
																														
															if (report.getFileType().equals(FileType.CSV)) {
																appliedMemberTemplate = replaceField(appliedMemberTemplate, "groupMemberClass", aceMember.isGroup() ? "yes" : "no", false, true, report);
															} else {
																appliedMemberTemplate = replaceField(appliedMemberTemplate, "groupMemberClass", aceMember.isGroup() ? "g" : "u", false, true, report);
															}
														
															if (aceMember.getViaGroupName() != null && !aceMember.getViaGroupName().isEmpty()) {
																appliedMemberTemplate = replaceField(appliedMemberTemplate, "viaMember", aceMember.getViaGroupName(), true, true, report);
															} else {
																appliedMemberTemplate = replaceField(appliedMemberTemplate, "viaMember", aclItem.getMember(true).getName(true), true, true, report);
															}
																														
															appliedMemberTemplate = replaceField(appliedMemberTemplate, "memberIsEnabled", aceMember.isEnabled() ? "yes" : "no", false, false, report);
															
															// replace the ACE field again for CSV and the jquery datatable report
															appliedMemberTemplate = applyAclItemToTemplate(appliedMemberTemplate, aclItem, file, report, String.valueOf(fileMemberCount), null, filterHelper); 
															appliedMemberTemplate = applyEffectivePermissionsToTemplate(aceMember.getEffectivePermissions(), null, appliedMemberTemplate, file.isDirectory(), report, prefix);
															
															memberList.append(appliedMemberTemplate);
														}
													} else {
														//if we export a datatable we need to at least add the ACE once
														// in case of CSV we check if we want to mention the group itself
														if (report.getFileType().equals(FileType.HTML) || !aclItem.getMember(true).isGroup() || filterHelper.getFilterConfig(true).isIncludeGroups()) {
														
															String appliedMemberTemplate = replaceField(memberTemplate, "viaMember", "", true, true, report);
															
															if (report.getFileType().equals(FileType.HTML)) {
																// for the HTML report we don't use these fields 
																appliedMemberTemplate = replaceField(appliedMemberTemplate, "groupMemberName", "", true, true, report);
																appliedMemberTemplate = replaceField(appliedMemberTemplate, "groupMemberClass", "", false, true, report);
															} else {
																appliedMemberTemplate = replaceField(appliedMemberTemplate, "groupMemberName", aclItem.getMember(true).getDisplayName(true, false), true, true, report);
																appliedMemberTemplate = replaceField(appliedMemberTemplate, "groupMemberClass", aclItem.getMember(true).isGroup() ? "yes" : "no", false, true, report);
															}
															
															appliedMemberTemplate = applyAclItemToTemplate(appliedMemberTemplate, aclItem, file, report, String.valueOf(fileMemberCount), null, filterHelper);
															appliedMemberTemplate = applyEffectivePermissionsToTemplate(null, null, appliedMemberTemplate, file.isDirectory(), report, "groupMember");
															memberList.append(appliedMemberTemplate);
														}
														//} else {
															// we have removed the group line from the CSV template (cleared the ACE LOOP) because
														//}
													}
													
													// apply the list of ACE's in the file template and write to the writer
													appliedAceTemplate = appliedAceTemplate.replace(memberTemplateTag, memberList.toString());
												} else {
													appliedAceTemplate = appliedAceTemplate.replace(memberTemplateTag, "");
												}
																							
												aceList.append(appliedAceTemplate);
											}
										} else {
											// clear ACE fields, and at least mention the root folder
											String appliedAceTemplate = applyAclItemToTemplate(aceTemplate, null, file, report, String.valueOf(fileMemberCount), null, filterHelper); 
											
											if (aclWithGroupMembers) {
												String prefix = report.getFileType().equals(FileType.HTML) ? "groupMember" : "";
												String appliedMemberTemplate = replaceField(memberTemplate, "groupMemberName", "", true, true, report);
												appliedMemberTemplate = replaceField(appliedMemberTemplate, "groupMemberClass", "", false, true, report);
												appliedMemberTemplate = replaceField(appliedMemberTemplate, "viaMember", "", true, true, report);
												appliedMemberTemplate = applyAclItemToTemplate(appliedMemberTemplate, null, file, report, String.valueOf(fileMemberCount), null, filterHelper); 
												appliedMemberTemplate = applyEffectivePermissionsToTemplate(null, null, appliedMemberTemplate, file.isDirectory(), report, prefix);
												appliedAceTemplate = appliedAceTemplate.replace(memberTemplateTag, appliedMemberTemplate);
											} else {
												appliedAceTemplate = appliedAceTemplate.replace(memberTemplateTag, "");
											}
											aceList.append(appliedAceTemplate);
										}
										
										// apply the list of ACE's in the file template and write to the writer
										appliedFileTemplate = appliedFileTemplate.replace(aceTemplateTag, aceList.toString());
									
									// -----------------------------------------------------------------------
									// REPORT TYPE FOLDERS_AND_USERS_EFFECTIVE
									// -----------------------------------------------------------------------
									} else if (report.getReportType().equals(ReportType.FOLDERS_AND_USERS_EFFECTIVE)) {
										StringBuilder usersOnFolderList = new StringBuilder();
										
										if (!usersOnFolder.isEmpty()) {
											// loop through the ACE's of this file and apply them to the template
											for (Member user : usersOnFolder) {
												
												// get the memberOf list
												List<Integer> memberOfIdList = MembershipCache.getMemberOf(user);
												user.setMemberOfList(memberOfIdList);
												
												// determine the applicable ACE's for this user
												List<AclItem> memberAclItems = new ArrayList<AclItem>();
												String viaGroups = "";
												
												for (AclItem aclItem : aclItems) {
													// match on the member of one of its group memberships?
													if (aclItem.getMemberId().equals(user.getId()) || memberOfIdList.contains(aclItem.getMemberId())) {
														memberAclItems.add(aclItem);
														
														// mention group membership?
														if (!aclItem.getMemberId().equals(user.getId())) {
															String groupName = aclItem.getMember(true).getName(false) + ", ";
															if (!viaGroups.contains(groupName)) {
																viaGroups += groupName;
															}
														}
													}
												}
												
												// remove the last comma and space
												if (!viaGroups.isEmpty()) {
													viaGroups = viaGroups.substring(0, viaGroups.length() - 2);
												} else {
													viaGroups = "(direct)";
												}
												
												// determine the effective permissions for this user
												EffectivePermissions effectivePermissions = PermissionUtil.determineEffectivePermissions(memberAclItems);
												
												// reset the effective permissions on the file, because this affects the file icon
												// the LoadFileInfo method doesn't match the icon because we only include users
												file.setEffectivePermissions(null);
												
												// replace ACE fields and the file info fields mentioned in the ACE template!
												String appliedAceTemplate = applyFileToTemplate(aceTemplate, file, report, String.valueOf(fileMemberCount), null, filterHelper);
												appliedAceTemplate = applyEffectivePermissionsToTemplate(effectivePermissions, user, appliedAceTemplate, file.isDirectory(), report, "");
												appliedAceTemplate = replaceField(appliedAceTemplate, "viaMember", viaGroups, true, true, report);
																							
												usersOnFolderList.append(appliedAceTemplate);
											}
										} else {
											// clear ACE fields, and at least mention the root folder ??? Don't mention the folder for now
											// String appliedAceTemplate = applyFileToTemplate(aceTemplate, file, report, String.valueOf(fileMemberCount), null, filterHelper);
											// appliedAceTemplate = applyEffectivePermissionsToTemplate(null, null, appliedAceTemplate, file.isDirectory(), report, "");
											// appliedAceTemplate = replaceField(appliedAceTemplate, "viaMember", "", true, true, report);
											// usersOnFolderList.append(appliedAceTemplate);
										}
										
										// apply the list of ACE's in the file template and write to the writer
										appliedFileTemplate = appliedFileTemplate.replace(aceTemplateTag, usersOnFolderList.toString());
											
									// ---------------------------------------------------------------
									// REPORT TYPE EFFECTIVE (based on file.getEffectivePermissions())
									// ---------------------------------------------------------------
									} else if (report.getReportType().equals(ReportType.EFFECTIVE)) {								
										appliedFileTemplate = applyEffectivePermissionsToTemplate(file.getEffectivePermissions(), null, appliedFileTemplate, file.isDirectory(), report, "");										
									}
									
									writer.write(appliedFileTemplate);
									characterCountWrittenToFile += appliedFileTemplate.length();
																			
									//if (counter % 50 == 0) {
									//	writer.flush();
									//}
									
									if (prefs.getMaxFileExportMegaBytes() > 0 && characterCountWrittenToFile > (prefs.getMaxFileExportMegaBytes() * MB_BYTE_SIZE)) {
										// reset the count
										characterCountWrittenToFile = 0;
										csvFileCounter++;
										
										// close the current CSV file
										writer.write(footer);
										writer.close();
										
										// open a new BufferedWriter to the new target file and print the initial template content to the file
										String newFilePath = reportFilePath.substring(0,  reportFilePath.length() - (report.getFileType().toString().length() + 1)) + "_" + csvFileCounter + "." + report.getFileType().toString().toLowerCase();
										writer = Files.newBufferedWriter(Paths.get(newFilePath), Charset.forName("UTF-8"), StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.CREATE);
										
										// write the UTF-8 representation of the BOM to get UTF-8 with BOM
										// that way Excel will recognize UTF-8 characters in a CSV export!
										writer.write(UTF_8_WITH_BOM_MARKER);
										
										//writer = new FileWriter(reportFilePath);
										writer.write(header);
									}
									
									monitor.worked(1);
								} // process next file
							}
							
							writer.write(footer);
							writer.close();
						} // end printToFile
						
						if (!monitor.isCanceled()) {
							// send e-mail?
							if (report.getEmailAddress() != null && report.getEmailAddress().length() > 0 && (fileCount >= report.getEmailThreshold() || report.getLastStatusMessage() == null || !report.getLastStatusMessage().equalsIgnoreCase("ok"))) {
								String msg;
								String subject;
								if (isPolicy) {
									msg = prefs.getPolicyEmailTemplate();
									subject = prefs.getPolicyEmailSubject();
								} else {
									msg = prefs.getReportEmailTemplate();
									subject = prefs.getReportEmailSubject();
								}
								
								monitor.beginTask("Sending e-mail", IProgressMonitor.UNKNOWN);
								//try {
									EmailHelper.sendMail(subject, msg, report);
								//} catch (EmailException ex) {
								//	logger.error("Error while e-mailing the report: " + ex.getMessage(), ex);
								//	SyncMessageDialog.showMessage(MessageDialog.WARNING, "Warning", "The report has been exported, but the e-mail could not be sent: " + GUILogHandler.getNestedMessage(ex));
								//}
							}
							
							System.out.println("End: " + System.currentTimeMillis());
							System.out.println("Done: " + (System.currentTimeMillis() - startTime)  + " ms");
						}
						
						// existing report? (instead of a plain export)
						if (report.getName().length() > 0) {
							report.setLastRunDate(new Date());
							report.setLastStatusMessage("OK");
							report.setLastFilePath(reportFilePath);
							if (report.getReportType().equals(ReportType.ALL_MEMBERS)) {
								report.setLastCount(memberCount);
							} else {
								report.setLastCount(fileCount);
							}
							report.save();
							
							// keep a history record
							ReportHistory reportHistory = new ReportHistory(report);
							reportHistory.save();
						}
						
						// comparison report?
						if (report.isCreateComparisonReport()) {
							ReportComparator.compareLastTwoReportRuns(report, true, true, true, true, false, false, monitor);
						}
						
						if (printToFile && showOpenReportQuestion && SyncMessageDialog.showMessage(MessageDialog.QUESTION, (isPolicy ? "Policy" : "Report") + " ready", "The " + (isPolicy ? "policy" : "report") + " has been built. Do you want to open the report?")) {
							Program.launch(reportFilePath);								
						} else if (!printToFile && showOpenReportQuestion) {
							SyncMessageDialog.showMessage(MessageDialog.INFORMATION, (isPolicy ? "Policy" : "Report")  + " completed", "The " + (isPolicy ? "policy" : "report") + " has been refreshed successfully.");
						}
						
					} catch (Exception ex) {
						if (ex instanceof EmailException) {
							logger.error("Error while e-mailing the report: " + ex.getMessage(), ex);
							SyncMessageDialog.showMessage(MessageDialog.WARNING, "Warning", "The report has been exported, but the e-mail could not be sent: " + ex.getMessage());
						} else {
							logger.error("Error while exporting report: " + ex.getMessage(), ex);
							SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while exporting the report: " + GUILogHandler.getNestedMessage(ex));
						}
						
						if (report != null && report.getName().length() > 0) {
							report.setLastRunDate(new Date());
							report.setLastFilePath(reportFilePath);
							if (report.getReportType().equals(ReportType.ALL_MEMBERS)) {
								report.setLastCount(memberCount);
							} else {
								report.setLastCount(fileCount);
							}
							
							if (ex instanceof EmailException) {
								report.setLastStatusMessage("Error while e-mailing the report: " + ex.getMessage());
							} else {
								report.setLastStatusMessage(GUILogHandler.getNestedMessage(ex));
							}
							
							try {
								report.save();
								
								// keep a history record
								ReportHistory reportHistory = new ReportHistory(report);
								reportHistory.save();
								
							} catch (DBException e2) {
								// bad luck
								logger.error("Error while updating report status", e2);
							}
						}												
					} finally {
						if (writer != null) {
							try {
								writer.close();
							} catch (IOException ex) {
								logger.warn("Could not close export file for report: " + ex.getMessage());
							}
						}
					}
				}			
			});
		} catch (InvocationTargetException | InterruptedException e) {
			SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while processing report: " + e.getMessage());
			logger.error("Error while processing report", e);
		}
	}
	
	/**
	 * Applies a {@link EffectivePermissions} object to the given template.
	 * 
	 * @param effectivePermissions
	 * @param template
	 * @param isDirectory
	 * @param report
	 * @return
	 */
	private static String applyEffectivePermissionsToTemplate(EffectivePermissions effectivePermissions, Member member, String template, boolean isDirectory, Report report, String prefix) {
		// effective permission
		if (effectivePermissions != null) {
			template = replaceField(template, prefix + "permissionText", PermissionUtil.effectivePermissionsToString(effectivePermissions, isDirectory), false, true, report);
		} else {
			template = replaceField(template, prefix + "permissionText", "", false, true, report);
		}
		
		// Special permission images
		int permissionIndex = 1;
		for (final Integer permission : PermissionUtil.getCustomPermissionOrderList()) {
			String permissionStyle = "";
			
			if (effectivePermissions != null) {
				// get related AclItem
				AclItem aclItem = effectivePermissions.getPermissions().get(permission);
				
				if (aclItem != null) {									
					// deny and/or explicit
					if (PermissionLookup.containsSpecialPermission(aclItem.getPermissionId(), permission) && aclItem.isDeny()) {	
						permissionStyle = "deny";
					}
					
					if (!aclItem.isFlagInheritedAce()) {	
						permissionStyle += ((!permissionStyle.isEmpty() ? " " : "") + "ex");
					}
					
					if (PermissionLookup.containsSpecialPermission(aclItem.getPermissionId(), permission)) {
						if (report.getFileType().equals(FileType.CSV)) {
							template = replaceField(template, prefix + "permission" + permissionIndex + "Value", "yes" + (!permissionStyle.isEmpty() ? " (" + permissionStyle + ")" : ""), false, true, report);
						} else {
							// style class
							template = replaceField(template, prefix + "permission" + permissionIndex + "Value", "p" + (permissionIndex < 10 ? "0" : "") + permissionIndex, false, true, report);
						}
						
						// permission is present
						permissionStyle += (" p" + (permissionIndex < 10 ? "0" : "") + permissionIndex);
						
					} else {
						template = replaceField(template, prefix + "permission" + permissionIndex + "Value", "", false, true, report);
					}
					
				} else {
					// special permission not present for this file
					template = replaceField(template, prefix + "permission" + permissionIndex + "Value", "", false, true, report);
				}
			} else {
				template = replaceField(template, prefix + "permission" + permissionIndex + "Value", "", false, true, report);
			}
			
			template = replaceField(template, prefix + "permission" + permissionIndex + "Style", permissionStyle, false, true, report);
			
			permissionIndex++;
		}
		
		// member info?
		if (member != null) {
			if (report.getFileType().equals(FileType.CSV)) {
				template = replaceField(template, "memberClass", member.isGroup() ? "yes" : "no", false, true, report);
			} else {
				template = replaceField(template, "memberClass", member.isGroup() ? "g" : "u", false, true, report);
			}
			template = replaceField(template, "memberName", PreferenceHelper.getInstance().isUseUserDisplayNameInReport() ? member.getDisplayName(true, false) : member.getName(true), true, true, report);
			template = replaceField(template, "memberDisplayName", member.getDisplayName(false, false), true, true, report);
			template = replaceField(template, "memberIsEnabled", member.isEnabled() ? "yes" : "no", true, true, report);
		} else {
			// empty values
			template = replaceField(template, "memberClass", "", false, true, report);
			template = replaceField(template, "memberName", "", true, true, report);
			template = replaceField(template, "memberDisplayName", "", true, true, report);
			template = replaceField(template, "memberIsEnabled", "", true, true, report);
		}
		
		return template;
	}
	
	/**
	 * Applies the file info to the given template, meaning place holders like filePath, icon and style classes are added to the template.
	 * 
	 * @param template
	 * @param file
	 * @param report
	 * @return the templates with the applied file info
	 */
	private static String applyFileToTemplate(String template, FileInfo file, Report report, String memberCount, String aceCount, FilterHelper filterHelper) {
		PreferenceHelper prefs = PreferenceHelper.getInstance();
		
		// apply the file path to the template
		// the file object may be null in case of "No ACE's found for this member" in the All Members tab
		String appliedFileTemplate = replaceField(template, "filePath", file != null ? file.determineFullPath() : "", true, true, report);
		
		// owner
		appliedFileTemplate = replaceField(appliedFileTemplate, "owner", file != null && file.getOwner() != null ? file.getOwner().getName(true) : "", true, true, report);
		
		// owner
		appliedFileTemplate = replaceField(appliedFileTemplate, "inheritAcl", file != null && file.isDaclProtected() ? "no" : "yes", false, true, report);
				
		// total ownership ratio
		if (prefs.isCalculateOwnershipRatioInOverview() && filterHelper.getMemberFilterList().size() > 1) {
			StringBuilder totalOwnershipRatio = new StringBuilder();
			totalOwnershipRatio.append(GroupOwnershipColumnLabelProvider.OWNERSHIP_RATIO_FORMAT.format(((double)file.getAllMatchingMembersOnFolder() / (double)file.getAllMembersOnFolder() * 100)));
			totalOwnershipRatio.append("% (");
			totalOwnershipRatio.append(file.getAllMatchingMembersOnFolder());
			totalOwnershipRatio.append(" / ");
			totalOwnershipRatio.append(file.getAllMembersOnFolder());
			totalOwnershipRatio.append(")");
			
			appliedFileTemplate = replaceField(appliedFileTemplate, "totalOwnershipRatio", totalOwnershipRatio.toString(), false, true, report);
		} else {
			appliedFileTemplate = replaceField(appliedFileTemplate, "totalOwnershipRatio", "", false, true, report);
		}
		
		// group ownership ratio
		if (prefs.isCalculateOwnershipRatioInOverview() && filterHelper.isApplyMemberFilter()) {
			StringBuilder groupOwnershipRatio = new StringBuilder();
			groupOwnershipRatio.append(GroupOwnershipColumnLabelProvider.OWNERSHIP_RATIO_FORMAT.format(((double)file.getAllMatchingMembersOnFolder() / (double)filterHelper.getAllFilteredMembers() * 100)));
			groupOwnershipRatio.append("% (");
			groupOwnershipRatio.append(file.getAllMatchingMembersOnFolder());
			groupOwnershipRatio.append(" / ");
			groupOwnershipRatio.append(filterHelper.getAllFilteredMembers());
			groupOwnershipRatio.append(")");
			
			appliedFileTemplate = replaceField(appliedFileTemplate, "groupOwnershipRatio", groupOwnershipRatio.toString(), false, true, report);
		} else {
			appliedFileTemplate = replaceField(appliedFileTemplate, "groupOwnershipRatio", "", false, true, report);
		}
		
		if (report.getFileType().equals(FileType.CSV)) {
			appliedFileTemplate = replaceField(appliedFileTemplate, "directoryFile", file != null && file.isDirectory() ? "yes" : "no", false, true, report); //used in CSV report
		} else {
			appliedFileTemplate = replaceField(appliedFileTemplate, "directoryFile", file != null && file.isDirectory() ? "d" : "f", false, true, report);
		}
		
		// Determine the file icon
		String fileClass;
		if (file != null && file.isDirectory()) {
			fileClass = PermissionUtil.getFileImage(null, file).name().toString().toLowerCase();
			
		} else {
			if (file != null) {
				int extensionIndex = file.getName().lastIndexOf(".");
				String extension = "";
				
				if (extensionIndex != -1) {
					extension = file.getName().substring(extensionIndex + 1).toLowerCase();
				}
				
				if (ImageFactory.FILE_EXTENSIONS.containsKey(extension)) {
					fileClass = ImageFactory.FILE_EXTENSIONS.get(extension);
				} else {
					fileClass = "file";
				}
			} else {
				fileClass = "file";
			}
		}
		
		// file or directory image
		appliedFileTemplate = replaceField(appliedFileTemplate, "fileClass", fileClass, false, true, report);
	
		// on file level (for the HTML report without jquery datatable)
		appliedFileTemplate = replaceField(appliedFileTemplate, "memberCount", memberCount, false, true, report);
		
		// aceCount used in the All Members report
		appliedFileTemplate = replaceField(appliedFileTemplate, "aceCount", aceCount, false, true, report);
			
		return appliedFileTemplate;
	}
	
	/**
	 * Applies the ace info to the given template, meaning place holders like filePath, icon, member info and permissions are added to the template.
	 * 
	 * @param template
	 * @param file
	 * @param report
	 * @return the templates with the applied ace info
	 * @throws DBException 
	 */
	private static String applyAclItemToTemplate(String template, AclItem aclItem, FileInfo file, Report report, String memberCount, String aceCount, FilterHelper filterHelper) throws DBException {
		// replace the filePath and memberCount as well for the CSV export and jquery datatable
		String appliedAceTemplate = applyFileToTemplate(template, file, report, memberCount, aceCount, filterHelper);
		
		if (aclItem != null) {
			// member info (from cache)
			Member aclMember = aclItem.getMember(true);
					
			// replace ACE fields
			if (report.getFileType().equals(FileType.CSV)) {
				appliedAceTemplate = replaceField(appliedAceTemplate, "aceExplicit", !aclItem.isFlagInheritedAce() ? "yes" : "no", false, true, report); 
				appliedAceTemplate = replaceField(appliedAceTemplate, "aceDeny", aclItem.isDeny() ? "yes" : "no", false, true, report);
				appliedAceTemplate = replaceField(appliedAceTemplate, "memberClass", aclMember.isGroup() ? "yes" : "no", false, true, report);
			} else {
				appliedAceTemplate = replaceField(appliedAceTemplate, "aceExplicit", !aclItem.isFlagInheritedAce() ? "ex" : "", false, true, report); 
				appliedAceTemplate = replaceField(appliedAceTemplate, "aceDeny", aclItem.isDeny() ? "deny" : "", false, true, report);
				appliedAceTemplate = replaceField(appliedAceTemplate, "memberClass", aclMember.isGroup() ? "g" : "u", false, true, report);
			}
			
			appliedAceTemplate = replaceField(appliedAceTemplate, "memberName", PreferenceHelper.getInstance().isUseUserDisplayNameInReport() ? aclMember.getDisplayName(true, false) : aclMember.getName(true), true, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "memberDisplayName", aclMember.getDisplayName(false, false), true, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "memberIsEnabled", aclMember.isEnabled() ? "yes" : "no", true, true, report);
			//boolean showInheritanceInPermission = report.getReportType().equals(ReportType.ACL_NESTED_GROUP_MEMBERS) || report.getReportType().equals(ReportType.ACL_DIRECT_GROUP_MEMBERS);
			boolean showInheritanceInPermission = true;
			appliedAceTemplate = replaceField(appliedAceTemplate, "permissionText", PermissionUtil.permissionsToString(aclItem, showInheritanceInPermission), false, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "viaMember", aclItem.getViaMember() != null ? aclItem.getViaMember().getDisplayName(true, false) : "(direct)", true, true, report);
					
			// Special permission images
			int permissionIndex = 1;
			for (final Integer permission : PermissionUtil.getCustomPermissionOrderList()) {
				
				String permissionValue = report.getFileType().equals(FileType.CSV) ? "no" : ""; //default value
				if (PermissionLookup.containsSpecialPermission(aclItem.getPermissionId(), permission)) {
					if (report.getFileType().equals(FileType.CSV)) {
						permissionValue = "yes";
					} else {
						permissionValue = "p" + (permissionIndex < 10 ? "0" : "") + permissionIndex;
					}
					//appliedAceTemplate = replaceField(appliedAceTemplate, "permission" + permissionIndex + "Value", "p" + (permissionIndex < 10 ? "0" : "") + permissionIndex, false, report);
				}
				
				// use as value to search for in jquery table
				appliedAceTemplate = replaceField(appliedAceTemplate, "permission" + permissionIndex + "Value", permissionValue, false, true, report);
				
				// use value as styling for CSS
				appliedAceTemplate = replaceField(appliedAceTemplate, "permission" + permissionIndex + "Style", permissionValue, false, true, report);
				
				permissionIndex++;
			}
			
			appliedAceTemplate = replaceField(appliedAceTemplate, "flags", PermissionUtil.aclEntryFlagsToString(aclItem, file != null && file.isDirectory()), false, true, report);
			
			// inherited from
			if (!aclItem.isFlagInheritedAce()) {
				appliedAceTemplate = replaceField(appliedAceTemplate, "inheritedFrom", "(explicitly assigned)", false, true, report);
			} else if (aclItem.getInheritedFromFolder() != null && aclItem.getInheritedFromFolder().length() > 0) {
				appliedAceTemplate = replaceField(appliedAceTemplate, "inheritedFrom", aclItem.getInheritedFromFolder(), true, true, report);
			} else {
				appliedAceTemplate = replaceField(appliedAceTemplate, "inheritedFrom", "Unknown parent folder", false, true, report);
			}
		} else {
			
			appliedAceTemplate = replaceField(appliedAceTemplate, "aceExplicit", "", false, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "aceDeny", "", false, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "memberClass", "", false, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "memberName", "", true, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "permissionText", "", false, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "viaMember", "", true, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "flags", "", false, true, report);
			appliedAceTemplate = replaceField(appliedAceTemplate, "inheritedFrom", "", false, true, report);
			int permissionIndex = 1;
			for (final Integer permission : PermissionUtil.getCustomPermissionOrderList()) {			
				appliedAceTemplate = replaceField(appliedAceTemplate, "permission" + permissionIndex + "Value", "", false, true, report);
				appliedAceTemplate = replaceField(appliedAceTemplate, "permission" + permissionIndex + "Style", "", false, true, report);				
				permissionIndex++;
			}
		}
		
		return appliedAceTemplate;
	}
	
	/**
	 * <p>
	 * The template files contain separate blocks to indicate the templates used in a loop, like a directory or ACE. The blocks are marked as follow: 
	 * <pre>
	 * &lt;!-- START ACE LOOP --&gt;
	 * 	 &lt;my html&gt;
	 * &lt;!-- END ACE LOOP --&gt;
	 * </pre>
	 * </p>
	 * 
	 * @param blockName like "ACE LOOP"
	 * @param content template content loaded from the html file
	 * 
	 * @return html found between the start and end tags of the given blockName
	 */
	private static String getTemplateBlock(String blockName, String content) {
		String templateBlock = "";
		int start = content.indexOf("<!-- START " + blockName + " -->");
		int end = content.indexOf("<!-- END " + blockName + " -->");
		
		if (start > -1) {
			start += ("<!-- START " + blockName + " -->").length() + 2;
			templateBlock = content.substring(start, end);
			
			// remove tabs in front of the template
			while (templateBlock.startsWith("\t")) {
				templateBlock = templateBlock.substring(1);
			}
		}
		
		return templateBlock;			
	}
	
	/**
	 * Builds a HTML IMG tag based on a {@link AppImage}. The name of the enum value is used as image name, it is lower cased and .png will be added. The src link assumes that the images are located 
	 * in a submap called 'img'
	 * 
	 * @param image AppImage to link
	 * 
	 * @return HTML IMG tag like &lt;img src="img/folder_16x16.png"&gt;
	 */
	private static String getImgTag(AppImage image) {
		return "<img src=\"img/" + image.toString().toLowerCase() + ".png\">";
	}
	
	private static String replaceField(String text, String field, String value, boolean escape, boolean isJsonField, Report report) {
		if (value == null) {
			value = "";
		}
	
		if (escape && !isJsonField && (report.getFileType().equals(FileType.HTML) || report.getFileType().equals(FileType.HTML_JQUERY))) {
			value = StringEscapeUtils.escapeHtml(value);
		}
		
		if (escape && isJsonField && (report.getFileType().equals(FileType.HTML) || report.getFileType().equals(FileType.HTML_JQUERY))) {
			// escape the Strings that will be used in the JSON array
			value = value.replace("\\", "\\\\");
			value = value.replace("\"", "\\\"");
		}
		
		if (escape && report.getFileType().equals(FileType.CSV)) {
			value = StringEscapeUtils.escapeCsv(value);
		}
				
		return text.replace("[" + field + "]", value);
	}
	
	/**
	 * Loops through the list of {@link FilterRow} objects and prints the description text and image to the export file. This list may contain FilterSets which act as parents/containers
	 * for the a set of filters. This is represented in the export file as tree.
	 *  
	 * @param filterRows
	 */
	private static String filterRowsToHTML(List<FilterRow> filterRows) {
		StringBuilder filters = new StringBuilder();
		boolean processingPermissionList = false;
		int baseIndent = 3;
		filters.append("<ul>\r\n");
		
		for (FilterRow row : filterRows) {
			
			if (processingPermissionList && !row.getLabel().contains("(deny)") && !row.getLabel().contains("(allow)")) {
				// permission list is ended (another filter is encountered)
				filters.append("</ul>");
				processingPermissionList = false;
			}
			
			// create <li>
			filters.append(tab(baseIndent + 1));
			filters.append("<li class=\"filter ");
			filters.append(row.getImage().toString().toLowerCase());
			filters.append("\">");
			filters.append(row.getLabel());
			filters.append("</li>\r\n");
			
			if (row.getLabel().contains("the following permissions:")) {
				// start a new list of permissions
				filters.append("<ul class=\"permission-list\">");
				processingPermissionList = true;
			}
			
			if (row.getChildren() != null) {
				// filter set, start a new <ul>
				filters.append(tab(baseIndent + 2));
				filters.append("<ul>\r\n");
				
				for (FilterRow childRow : row.getChildren()) {
					
					if (processingPermissionList && !row.getLabel().contains("(deny)") && !row.getLabel().contains("(allow)")) {
						// permission list is ended (another filter is encountered)
						filters.append("</ul>");
						processingPermissionList = false;
					}
					
					// create <li>
					filters.append(tab(baseIndent + 3));
					filters.append("<li class=\"filter ");
					filters.append(childRow.getImage().toString().toLowerCase());
					filters.append("\">");
					filters.append(childRow.getLabel());
					filters.append("</li>\r\n");
					
					if (row.getLabel().contains("the following permissions:")) {
						// start a new list of permissions
						filters.append("<ul class=\"permission-list\">");
						processingPermissionList = true;
					}
				}
				
				if (processingPermissionList) {
					// end the permission list, because is has not yet been ended
					filters.append("</ul>");
					processingPermissionList = false;
				}
				
				filters.append(tab(baseIndent + 2));
				filters.append("</ul>\r\n");
			}
		}	
		
		if (processingPermissionList) {
			// end the permission list, because is has not yet been ended
			filters.append("</ul>");
			processingPermissionList = false;
		}
		
		filters.append(tab(baseIndent));
		filters.append("</ul>");
	
		return filters.toString();
	}
	
	private static String tab(int count) {
		return new String(new char[count]).replace("\0", "\t");
	}
	
	/**
	 * Builds a list of {@link FilterRow} objects that represents a human readable 'summary' of the selected filters. It may contain a tree structured in case
	 * the report contains references to a filter set. In that case the filter set will be a parent (with image {@link AppImage#SELECTION_16X16}) and the an attribute
	 * children that contains the filters that belong to the filter set. Each filter is represented as {@link FilterRow} object to give it a human readable description and image.
	 * 
	 * @param reportFilters
	 * @return list of FilterRow objects with human readable descriptions and images.
	 * @see {@link ReportDialog.ReportFilterContentProvider#getElements(Object)}
	 * @see {@link #exportReport(Report)}
	 */
	public static List<FilterRow> determineFilterRows(ReportFilters reportFilters) {
		FilterRow rootRow = new FilterRow("root", null, true);
		List<Filter> selectedMemberFilters;
		List<Filter> selectedPermissionFilters;
		List<Filter> selectedFolderFilters;
		FilterRow memberFilterSetRow = null;
		FilterRow permissionFilterSetRow = null;
		FilterRow folderFilterSetRow = null;
		
		//-----------------------------------------------------------
		// use filters from the FilterSets or the filters directly assigned to the report
		//-----------------------------------------------------------
		if (reportFilters.isLinkToMemberFilterSet() && reportFilters.getMemberFilterSet() != null) {
			selectedMemberFilters = reportFilters.getMemberFilterSet().getFilters();
			memberFilterSetRow = new FilterRow(reportFilters.getMemberFilterSet().getName(), AppImage.SELECTION_16X16, true);
			rootRow.getChildren().add(memberFilterSetRow);
		} else {
			selectedMemberFilters = reportFilters.getMemberFilters();
		}
		if (reportFilters.isLinkToPermissionFilterSet() && reportFilters.getPermissionFilterSet() != null) {
			selectedPermissionFilters = reportFilters.getPermissionFilterSet().getFilters();
			permissionFilterSetRow = new FilterRow(reportFilters.getPermissionFilterSet().getName(), AppImage.SELECTION_16X16, true);
			rootRow.getChildren().add(permissionFilterSetRow);
		} else {
			selectedPermissionFilters = reportFilters.getPermissionFilters();
		}
		if (reportFilters.isLinkToFolderFilterSet() && reportFilters.getFolderFilterSet() != null) {
			selectedFolderFilters = reportFilters.getFolderFilterSet().getFilters();
			folderFilterSetRow = new FilterRow(reportFilters.getFolderFilterSet().getName(), AppImage.SELECTION_16X16, true);
			rootRow.getChildren().add(folderFilterSetRow);
		} else {
			selectedFolderFilters = reportFilters.getFolderFilters();
		}
		
		//-----------------------------------------------------------
		// get the CONFIG filters
		//-----------------------------------------------------------
		Filter memberConfigFilter = null;
		Filter permissionConfigFilter = null;
		Filter folderConfigFilter = null;
		
		for (Filter filter : selectedMemberFilters) {
			if (filter.getType().equals(FilterType.MEMBER_FILTER_CONFIG)) {
				memberConfigFilter = filter;
				break;
			}
		}
		
		for (Filter filter : selectedPermissionFilters) {
			if (filter.getType().equals(FilterType.PERMISSION_FILTER_CONFIG)) {
				permissionConfigFilter = filter;
				break;
			}
		}
		
		for (Filter filter : selectedFolderFilters) {
			if (filter.getType().equals(FilterType.FOLDER_FILTER_CONFIG)) {
				folderConfigFilter = filter;
				break;
			}
		}
		
		FilterConfig filterConfig = new FilterConfig(memberConfigFilter, permissionConfigFilter, folderConfigFilter);
		
		//-----------------------------------------------------------
		// Member filter rows
		//-----------------------------------------------------------
		FilterRow addToRow;
		if (memberFilterSetRow != null) {
			addToRow = memberFilterSetRow;
		} else {
			addToRow = rootRow;
		}
		
		if (filterConfig.isIncludeGroupMembership()) {
			addToRow.getChildren().add(new FilterRow("Include (nested) group membership", AppImage.GROUP_MEMBERSHIP_16X16));
		} else {
			addToRow.getChildren().add(new FilterRow("Don't include group membership", AppImage.GROUP_16X16));
		}
		
		//if (!filterConfig.isIncludeGroups() || !filterConfig.isIncludeUsers()) {
			if (!filterConfig.isIncludeGroups()) {
				addToRow.getChildren().add(new FilterRow("Exclude groups", AppImage.GROUP_16X16));
			} 
			if (!filterConfig.isIncludeUsers()) {
				addToRow.getChildren().add(new FilterRow("Exclude users", AppImage.USER_16X16));
			}
		//}
		
		//if (!filterConfig.isIncludeBuiltIn() || !filterConfig.isIncludeADGroupsAndUsers()) {
			if (!filterConfig.isIncludeBuiltIn()) {
				addToRow.getChildren().add(new FilterRow("Exclude built-in groups", AppImage.BUILTIN_16X16));
			}
			if (!filterConfig.isIncludeADGroupsAndUsers()) {
				addToRow.getChildren().add(new FilterRow("Exclude LDAP users and groups", AppImage.AD_GROUPS_USERS_16X16));
			}
			if (!filterConfig.isIncludeDisabledUsers()) {
				addToRow.getChildren().add(new FilterRow("Exclude disabled users", AppImage.DISABLED_USER_20X20));
			}
		//}
		
		for (Filter filter : selectedMemberFilters) {
			String includeExclude = filter.isInclude() ? "Include " : "Exclude ";
			switch (filter.getType()) {
				case MEMBER_GROUP_ID: addToRow.getChildren().add(new FilterRow(includeExclude + "all members from the group " + filter.getName(), filter.isInclude() ? AppImage.INCLUDE_GROUPS_16X16 : AppImage.EXCLUDE_GROUPS_16X16)); break;
				case MEMBER_GROUP_NAME: addToRow.getChildren().add(new FilterRow(includeExclude + "all members from the groups with the name \"" + filter.getName() + "\"", filter.isInclude() ? AppImage.INCLUDE_WILDCARD_16X16 : AppImage.EXCLUDE_WILDCARD_16X16)); break;
				case MEMBER_ID: addToRow.getChildren().add(new FilterRow(includeExclude + "member " + filter.getName(), filter.isInclude() ? AppImage.INCLUDE_GROUPS_16X16 : AppImage.EXCLUDE_GROUPS_16X16)); break;
				case MEMBER_NAME: addToRow.getChildren().add(new FilterRow(includeExclude + "all members with the account name \"" + filter.getName() + "\"", filter.isInclude() ? AppImage.INCLUDE_WILDCARD_16X16 : AppImage.EXCLUDE_WILDCARD_16X16)); break;
				case MEMBER_LDAP_OU: addToRow.getChildren().add(new FilterRow(includeExclude + "all members in the LDAP OU " + filter.getName(), filter.isInclude() ? AppImage.INCLUDE_LDAP_OU_16X16 : AppImage.EXCLUDE_LDAP_OU_16X16)); break;
				case OWNER_ID: addToRow.getChildren().add(new FilterRow(includeExclude + "all files with the owner " + filter.getName(), filter.isInclude() ? AppImage.INCLUDE_OWNER_16X16 : AppImage.EXCLUDE_OWNER_16X16)); break;
				case MEMBER_REPORT: addToRow.getChildren().add(new FilterRow(includeExclude + "all members from the report " + filter.getName(), filter.isInclude() ? AppImage.INCLUDE_REPORT_16X16 : AppImage.EXCLUDE_REPORT_16X16)); break;
				default: break;
			}
		}
		
		//-----------------------------------------------------------
		// Permission filter rows
		//-----------------------------------------------------------
		if (permissionFilterSetRow != null) {
			addToRow = permissionFilterSetRow;
		} else {
			addToRow = rootRow;
		}
		
		if (filterConfig.isHideInheritedPermissions()) {
			addToRow.getChildren().add(new FilterRow("Hide permissions that are inherited from a parent folder", AppImage.BROWSE_16X16));
		}
		
		// 6 simple permissions, 13 special permissions, * 2 because the list contains allow and deny filters and + 1 for the configuration filter
		boolean hasFilterOnPermission = selectedPermissionFilters.size() > 1 && // not only the FILTER_CONFIG type 
									(selectedPermissionFilters.size() != (6 * 2 + 13 * 2 + 1) || filterConfig.isAceNeedAllPermissions());
		
		if (hasFilterOnPermission) {
			String includeExclude = filterConfig.isIncludePermissions() ? "Include" : "Exclude";
			String anyAll = filterConfig.isAceNeedAllPermissions() ? "all" : "any";
			addToRow.getChildren().add(new FilterRow(includeExclude + " ACE's that have " + anyAll + " of the following permissions:", filterConfig.isIncludePermissions() ? AppImage.INCLUDE_PERMISSIONS_16X16 : AppImage.EXCLUDE_PERMISSIONS_16X16));
			
			String permissionText = "";
			for (Filter filter : selectedPermissionFilters) {					
				if (filter.getType().equals(FilterType.PERMISSION_ALLOW) || filter.getType().equals(FilterType.PERMISSION_DENY)) {
					int permissionIndex = Integer.parseInt(filter.getValue());
					
					if (permissionIndex < FilterPermissionPart.SIMPLE_PERMISSIONS_INDEX_OFFSET) {
						// special permission
						permissionText = PermissionUtil.specialPermissionToString(permissionIndex, false, null);
					} else {
						// simple permission
						permissionText = PermissionUtil.simplePermissionToString(SimplePermission.values()[permissionIndex - FilterPermissionPart.SIMPLE_PERMISSIONS_INDEX_OFFSET]);
					}
				}
				if (filter.getType().equals(FilterType.PERMISSION_ALLOW)) {
					addToRow.getChildren().add(new FilterRow(permissionText + " (allow)", AppImage.PERM_READ_ACL_16X16));
				} else if (filter.getType().equals(FilterType.PERMISSION_DENY)) {
					addToRow.getChildren().add(new FilterRow(permissionText + " (deny)", AppImage.PERM_READ_ACL_16X16));
				}
			}
		}
		
		//-----------------------------------------------------------
		// Folder filter rows
		//-----------------------------------------------------------
		if (folderFilterSetRow != null) {
			addToRow = folderFilterSetRow;
		} else {
			addToRow = rootRow;
		}
						
		if (filterConfig.isIncludeFolders() != filterConfig.isIncludeFiles()) {
			if (filterConfig.isIncludeFolders()) {
				addToRow.getChildren().add(new FilterRow("Only include folders", AppImage.FOLDER_16X16));
			} else {
				addToRow.getChildren().add(new FilterRow("Only include files", AppImage.FILE_TXT_READ_16X16));
			}
		}
		
		if (filterConfig.isIncludeFilesThatBlockInheritance()) {
			addToRow.getChildren().add(new FilterRow("Only include files and folder that block the ACL inheritance", AppImage.PERM_READ_ACL_16X16));
		}
		
		if (filterConfig.isInvertFolders()) {
			addToRow.getChildren().add(new FilterRow("Invert folders, exclude folders that match the filter criteria", AppImage.INVERT_FOLDERS_24X24));
		}
		
		for (Filter filter : selectedFolderFilters) {
			String includeExclude = filter.isInclude() ? "Include " : "Exclude ";
			if (filter.getType().equals(FilterType.FOLDER_PATH)) {
				addToRow.getChildren().add(new FilterRow(includeExclude + "folder " + filter.getName(), filter.isInclude() ? AppImage.INCLUDE_FOLDER_16X16 : AppImage.EXCLUDE_FOLDER_16X16));
				
			} else if (filter.getType().equals(FilterType.FOLDER_NAME)) {
				addToRow.getChildren().add(new FilterRow(includeExclude + "all folders with the name \"" + filter.getName() + "\"", filter.isInclude() ? AppImage.INCLUDE_WILDCARD_16X16 : AppImage.EXCLUDE_WILDCARD_16X16));	
			}
		}
		
		return rootRow.getChildren();
	}
	
	public static void exportScheduledReports(final IWorkbench workbench) {
		if (Member.isUserGroupCountAllowed() && !LicenseHelper.getEdition().equals(Edition.SCAN_AGENT)) {
			List<Report> reportsList = null;
			
			if (ProgramParameters.isAllReports()) {
				try {
					reportsList = Report.where("EMAIL_THRESHOLD = 0", null);
				} catch (DBException e) {
					logger.error("Error while retrieving report information for all the reports", e);
				} 
			} else if (ProgramParameters.isAllPolicies()) {
				try {
					reportsList = Report.where("EMAIL_THRESHOLD > 0", null);
				} catch (DBException e) {
					logger.error("Error while retrieving report information for all the policy reports", e);
				} 
			} else {
				reportsList = new ArrayList<Report>();
				for (String reportName : ProgramParameters.getReports()) {
					try {						
						Report report = Report.findFirst("LOWER(NAME) = ?", reportName.toLowerCase());
						
						if (report != null) {
							reportsList.add(report);
						} else {
							logger.error("No report found with the name: " + reportName);
						}
					} catch (DBException e) {
						logger.error("Error while retrieving report information for report " + reportName, e);
					}
				}
			}
			
			// done, close the application
			for (final Report report : reportsList) {
				Display.getDefault().syncExec(new Runnable() {
					public void run() {
						exportReport(report, null, true, false);
						logger.info("Scheduled report exported: " + report.getName());
					}
				});
			}
		}

		// done, close the application
		Display.getDefault().syncExec(new Runnable() {
			public void run() {
				if (workbench != null) {
					workbench.close();
				} else {
					System.exit(0);
				}
			}
		});
	}
}
