package com.perdemia.permissionanalyzer.application.handlers.toolbar;

import gnu.trove.list.array.TIntArrayList;

import java.lang.reflect.InvocationTargetException;
import java.sql.PreparedStatement;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.Iterator;
import java.util.List;

import javax.inject.Named;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.e4.core.di.annotations.CanExecute;
import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.e4.core.di.annotations.Optional;
import org.eclipse.e4.ui.model.application.ui.basic.MPart;
import org.eclipse.e4.ui.workbench.modeling.EModelService;
import org.eclipse.e4.ui.workbench.modeling.EPartService;
import org.eclipse.e4.ui.workbench.modeling.ESelectionService;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.logging.GUILogHandler;
import com.perdemia.permissionanalyzer.application.parts.AllMembersPart;
import com.perdemia.permissionanalyzer.application.parts.FilePart;
import com.perdemia.permissionanalyzer.application.parts.TracePart;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper.Edition;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Filter;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.Report;
import com.perdemia.permissionanalyzer.data.util.MembershipCache;

public class ApplyFilterHandler {
	private static final Logger logger = LoggerFactory.getLogger(ApplyFilterHandler.class.getName());
	/** data for the TreeViewer */
	private List<FileInfo> rootFiles = null;
	private String selectedParentIds = null;
	private boolean reloadFromDB;
	private Integer itemCount;
	private boolean filterHasBeenApplied;
	
	/**
	 * Clears the cache in the FilterHelper and retrieves new FileInfo id's using {@link FilterHelper#buildFilterQuery(com.perdemia.permissionanalyzer.application.util.FilterHelper.ObjectType, boolean, int, String, Object...)}.
	 * 
	 * @param app
	 * @param partService
	 * @param modelService
	 * @param selectionService
	 * @param onlyApplyWhenEmpty in case this parameters length is greater than zero, the FileInfo id's are only refreshed in case the Tree or Table input is not loaded yet.
	 * 			This parameter is used by {@link UseTableViewHandler} and {@link UseTreeViewHandler}.
	 */
	@Execute
	public void execute(final EPartService partService, final EModelService modelService, final ESelectionService selectionService,
			@Optional @Named("com.perdemia.permissionanalyzer.application.commands.applyfilter.param.onlyapplywhenempty") final String onlyApplyWhenEmpty){
		
		ProgressMonitorDialog progressDialog = new ProgressMonitorDialog(Display.getDefault().getActiveShell());
		final MPart fileTreePart = partService.findPart(FilePart.ID);
		
		if (fileTreePart != null) {
			final FilePart fileTreePartObject = (FilePart) fileTreePart.getObject();
			final TreeViewer treeViewer = fileTreePartObject.getTreeViewer();
			final TableViewer tableViewer = fileTreePartObject.getTableViewer();
	
			// indicates if data is loaded for a TreeViewer or TableViewer
			final boolean useTreeViewer = treeViewer.getTree().isVisible();
			
			// remember current selection in order to re-select after refresh
			if (useTreeViewer) {
				StructuredSelection sel = (StructuredSelection) treeViewer.getSelection();
				if (!sel.isEmpty()) {
					selectedParentIds = ((FileInfo) sel.getFirstElement()).getParentFileIds();
				}
			}
									
			if (onlyApplyWhenEmpty != null) {
				if (useTreeViewer) {
					reloadFromDB = treeViewer.getInput() == null || ((List<FileInfo>)treeViewer.getInput()).size() == 0;
				} else {
					reloadFromDB = tableViewer.getTable().getItemCount() == 0;
				}
			} else {
				reloadFromDB = true;
			}
			
			// reset input to release resources
			if (onlyApplyWhenEmpty == null && reloadFromDB) {
				treeViewer.setInput(new ArrayList<FileInfo>());
				tableViewer.setInput(null);
				tableViewer.setItemCount(0);
			}
			
			final FilterHelper filterHelper = FilterHelper.getInstance();
			
			try {
				progressDialog.run(true, false, new IRunnableWithProgress(){
	
					@Override
					public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
					
						if (reloadFromDB) {
							try {
								// clear the filtered group member cache
								MembershipCache.clearFiltedGroupMembersCache();
								
								monitor.beginTask("Applying filter...", IProgressMonitor.UNKNOWN);
											
								// load all the applicable file id's based on the selected filter
								filterHelper.applyNewFilter(true);
								PreparedStatement ps = filterHelper.buildFilterQuery(FilterHelper.ObjectType.FILE, useTreeViewer, false, 0, null, null);
								filterHasBeenApplied = filterHelper.isFilterHasBeenApplied();
								
								if (ps != null) {
									TIntArrayList uniqueFileIds = filterHelper.getApplicableFileIds(ps, useTreeViewer, true);
									itemCount = uniqueFileIds.size();
									
									if (useTreeViewer) {
										filterHelper.setApplicableFileIdsIncludingParents(uniqueFileIds);
									} else {
										filterHelper.setApplicableFileIds(uniqueFileIds);
									}
									
									logger.info("Found " + itemCount + " items.");
									
									// load data for TreeViewer?
									if (useTreeViewer) {
										// load the root files
										rootFiles = FileInfo.where("PARENT_ID IS NULL", null);
										
										// check which root files are applicable
										Iterator<FileInfo> iter = rootFiles.iterator();
										while (iter.hasNext()) {
											if (!filterHelper.getApplicableFileIdsIncludingParents().contains(iter.next().getId())) {
												iter.remove();
											}
										}
									} else {
										itemCount = filterHelper.getApplicableFileIds().size();
									}
								} else {
									// load data for TreeViewer?
									if (useTreeViewer) {
										filterHelper.setApplicableFileIdsIncludingParents(null);
										rootFiles = FileInfo.where("PARENT_ID IS NULL", null);
										itemCount = FileInfo.getCount();
										
									} else {
										// load input data for the TableViewer
										filterHelper.setApplicableFileIds(null);
										itemCount = FileInfo.getCount();
									}
								}
								
								filterHelper.setItemCount(itemCount);
								
								// database actions have been performed, now update the GUI
								Display.getDefault().asyncExec(new Runnable() {
									
									public void run() {
										try {
											if (useTreeViewer) {
												treeViewer.getTree().setItemCount(rootFiles.size());
												treeViewer.setInput(rootFiles);											
											} else {
												tableViewer.setItemCount(itemCount);
												tableViewer.setInput(filterHelper.getApplicableFileIds());
											}
											
											// select the first item in the tree or table so that the TracePart and AclPart get refreshed through the SelectionService
											if (itemCount > 0) {
											
												// selection service is only triggered when the view has the focus
												if (useTreeViewer) {
													StructuredSelection structuredSelection = null;
													
													// re-select previous selected item?
													if (selectedParentIds != null && useTreeViewer) {
														String[] parentIds = selectedParentIds.split(",");
														List<FileInfo> files = rootFiles;
														Iterator<String> iter = Arrays.asList(parentIds).iterator();
														while (iter.hasNext()) {
															int parentId = Integer.parseInt(iter.next());
															boolean foundParent = false;
															
															if (files == null) {
																break;
															}
															
															for (FileInfo file : files) {
																if (file.getId().equals(parentId)) {
																	// last item? Just select the tree item and break
																	if (!iter.hasNext()) {
																		structuredSelection = new StructuredSelection(file);
																		treeViewer.setSelection(structuredSelection, true);
																		//selectionService.setSelection(treeViewer.getSelection());
																	} else {
																		// expand this parent
																		treeViewer.expandToLevel(file, 1);
																		foundParent = true;
																		files = file.getChildren();
																	}
																	break;
																}
															}
															
															if (!foundParent) {
																break;
															}
														}
													} else {
														// select root file
														if (!rootFiles.isEmpty()) {
															structuredSelection = new StructuredSelection(rootFiles.get(0));
														}
													}
													
													partService.activate(fileTreePart, true);
													treeViewer.getControl().setFocus();
													
													if (structuredSelection != null) {
														treeViewer.setSelection(structuredSelection, true);
													}
													
												} else {												
													partService.activate(fileTreePart, true);
													tableViewer.getTable().setFocus();
													tableViewer.getTable().select(0);
													tableViewer.getTable().notifyListeners(SWT.SELECTED, new Event());
												}
												
											} else {
												if (fileTreePartObject.isInitialized()) {
													MessageDialog.openInformation(null, "No data found", "No files or folders found.");
												}
												selectionService.setSelection(StructuredSelection.EMPTY);
											}
											
											String reportName = "";
											if (filterHelper.getSelectedReport() != null) {
												String entity = filterHelper.getSelectedReport().isPolicy() ? "policy" : "report";
												reportName = filterHelper.getSelectedReport() != null ? "    (" + entity + ": " + filterHelper.getSelectedReport().getName() + ")" : "";
											}
											
											if (useTreeViewer) {
												treeViewer.getTree().getColumn(0).setText("Path - " + itemCount + " items found" + reportName);
											} else {
												tableViewer.getTable().getColumn(0).setText("Path - " + itemCount + " items found" + reportName);
											}
											
											// update the last run date of the report
											Report selectedReportOrPolicy = filterHelper.getSelectedReport();
											if (selectedReportOrPolicy != null) {
												selectedReportOrPolicy.setLastRunDate(new Date());
												selectedReportOrPolicy.setLastCount(itemCount);
												selectedReportOrPolicy.setLastStatusMessage("OK");
												selectedReportOrPolicy.save();
											}
											
										} catch (Exception ex) {
											logger.error("Error while applying filter", ex);
											SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while applying filter: " + ex.getMessage());
											
											// update the last run date of the report
											Report selectedReportOrPolicy = filterHelper.getSelectedReport();
											if (selectedReportOrPolicy != null) {
												selectedReportOrPolicy.setLastRunDate(new Date());
												selectedReportOrPolicy.setLastStatusMessage(GUILogHandler.getNestedMessage(ex));
												try {
													selectedReportOrPolicy.save();
												} catch (DBException e) {
													logger.error("Could not update the report status.", e);
												}
											}
										}
										
										// activate or hide the member scope part, depending on whether or not multiple members have been found in the filter results
										//MPart memberScopePart = partService.findPart(MemberScopePart.ID);
										MPart allMembersPart = partService.findPart(AllMembersPart.ID);
										MPart tracePart = partService.findPart(TracePart.ID);
										TracePart tracePartObject = (TracePart)tracePart.getObject();
										
										// filtered on one member?
										Filter singleMemberFilter = filterHelper.isFilterOnSingleMember();
																			
										if (filterHasBeenApplied && singleMemberFilter == null) {
											//partService.activate(memberScopePart);
											
											// reset the Trace view
											if (tracePartObject != null) {
												tracePartObject.applySelectedMember(null);
											} else {
												TracePart.setSelectedMember(null);
											}
											
										} else {
											// don't use the memberScopePart
											//partService.hidePart(memberScopePart);
											
											// show the member in the trace view? 
											if (singleMemberFilter != null) {
												
												// retrieve the member details, one of the two filters must contain the member id
												Integer memberId = Integer.parseInt(singleMemberFilter.getValue());
												
												try {
													Member member = Member.getById(memberId);
													if (tracePartObject != null) {
														// refresh the UI
														tracePartObject.applySelectedMember(member);
													} else {
														// just update the member field
														TracePart.setSelectedMember(member);
													}
													
												} catch (DBException e) {
													logger.error("Could not retrieve the member details for the Trace view.", e);
												}
											}
										}
										
										if (allMembersPart.getObject() != null) {
											((AllMembersPart)allMembersPart.getObject()).clearAllMembers();
										}
									}
								});
								
								
							} catch (Exception ex) {
								logger.error("Error while applying filter", ex);
								SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while applying filter: " + ex.getMessage());
							}
						}
					}
				});
			} catch (InvocationTargetException | InterruptedException ex) {
				logger.error("Error while applying filter", ex);
				SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while applying filter: " + ex.getMessage());
			}
		} else {
			logger.warn("Could not apply filter, fileTreePart is null");
		}
	}
	
	@CanExecute
	public boolean canExecute() {
		return Member.isUserGroupCountAllowed() && !LicenseHelper.getEdition().equals(Edition.SCAN_AGENT);
	}
}
