package com.perdemia.permissionanalyzer.application.handlers.toolbar;

import java.util.List;

import org.eclipse.e4.core.di.annotations.Execute;
import org.eclipse.e4.ui.model.application.ui.basic.MPart;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.dnd.TextTransfer;
import org.eclipse.swt.dnd.Transfer;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.parts.AclPart;
import com.perdemia.permissionanalyzer.application.parts.AllMembersPart;
import com.perdemia.permissionanalyzer.application.util.FilterConfig;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.PermissionUtil;
import com.perdemia.permissionanalyzer.data.AclItem;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.PermissionLookup;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SpecialPermission;
import com.perdemia.permissionanalyzer.data.util.MembershipCache;

public class CopyAclItemsToClipboardHandler {
	private static final Logger logger = LoggerFactory.getLogger(CopyAclItemsToClipboardHandler.class.getName());

	@Execute
	public void execute(Shell shell, MPart mPart){
		if (mPart != null && mPart.getObject() != null) {
			List<AclItem> aclItems;
			int itemCount = 0;
			boolean expandGroups = false;
			boolean applyFilter = false;
			
			if (mPart.getObject() instanceof AclPart) {
				int result = SyncMessageDialog.showQuestionWithCancelMessage("Expand groups?", "Do you want to expand groups and include their members?");
				
				if (result == IDialogConstants.CANCEL_ID) {
					return;
				}
				expandGroups = result == IDialogConstants.YES_ID;
				
				AclPart part = (AclPart) mPart.getObject();
				aclItems = part.getAclItems();
				applyFilter = part.isApplyFilter();
				
			} else {
				AllMembersPart part = (AllMembersPart) mPart.getObject();
				
				if (part.getSelectedMember() == null) {
					SyncMessageDialog.showMessage(MessageDialog.ERROR, "No member selected", "Please select a member in the tab");
					return;
				} else {
					aclItems = part.getSelectedMember().getAclItemList();
				}
			}
			
			// build the CSV headers
			StringBuilder membersText = new StringBuilder("File path,Account name,Display name,Permission,");;
			
			// special permissions headers
			for (final Integer permission : PermissionUtil.getCustomPermissionOrderList()) {
				//skip SYNCHRONIZE
				if (permission.intValue() != SpecialPermission.SYNCHRONIZE.ordinal()) {
					membersText.append(PermissionUtil.specialPermissionToString(permission, false, null));
					membersText.append(",");
				}
			}
			membersText.append("Allow,Inheritance flags,From folder");
			
			if (expandGroups) {
				membersText.append(",Via group");
			}
			
			membersText.append("\r\n");
			
			// build the CSV content
			for (AclItem aclItem : aclItems) {
				itemCount++;
				
				String folderPath = aclItem.getFile(true).determineFullPath();
				boolean isDirectory = aclItem.getFile().isDirectory();
				
				toCSV(aclItem, aclItem.getMember(true), folderPath, isDirectory, "", membersText);
				
				if (aclItem.getMember(false).isGroup() && expandGroups) {
					FilterConfig filterConfig = FilterHelper.getInstance().getFilterConfig(false);
					
					// add all members of the group
					try {
						List<Member> groupMembers;
						if (applyFilter) {
							// only select the members that match the filter criteria
							groupMembers = MembershipCache.getFiltedGroupMembers(aclItem.getMember(false), filterConfig.isIncludeUsers(), filterConfig.isIncludeGroups(), true, false, false, FilterHelper.getInstance());
						} else {
							groupMembers = MembershipCache.getGroupMembers(aclItem.getMember(false), filterConfig.isIncludeUsers(), filterConfig.isIncludeGroups(), true);
						}
						
						for (Member groupMember : groupMembers) {
							itemCount++;
							toCSV(aclItem, groupMember, folderPath, isDirectory, aclItem.getMember(true).getName(true), membersText);
						}
					} catch (DBException ex) {
						logger.error("Error while getting group members for " + aclItem.getMember(false).getName(true), ex);
						SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while getting group members for " + aclItem.getMember(false).getName(true) + ": " + ex.getMessage());
					}
					
				}
			}
			
			Clipboard clipboard = new Clipboard(Display.getDefault());
			TextTransfer textTransfer = TextTransfer.getInstance();
	        clipboard.setContents(new String[]{membersText.toString()}, new Transfer[]{textTransfer});
	        clipboard.dispose();
	        SyncMessageDialog.showMessage(MessageDialog.INFORMATION, "Done", "Copied " + itemCount + " items to the clipboard.");
		}
	}
	
	private void toCSV(AclItem aclItem, Member member, String folderPath, boolean isDirectory, String viaGroup, StringBuilder output) {
		output.append("\"");
		output.append(folderPath);
		output.append("\",");
		output.append(member.getName(true));
		output.append(",\"");
		output.append(member.getDisplayName(false, false));
		output.append("\",");
		output.append(PermissionUtil.permissionsToString(aclItem, true));
		output.append(",");
		
		// special permissions
		for (final Integer permission : PermissionUtil.getCustomPermissionOrderList()) {
			//skip SYNCHRONIZE
			if (permission.intValue() != SpecialPermission.SYNCHRONIZE.ordinal()) {
				if (PermissionLookup.containsSpecialPermission(aclItem.getPermissionId(), permission)) {
					output.append("yes");
				} else {
					output.append("no");
				}
				output.append(",");
			}
		}
		
		output.append(aclItem.isDeny() ? "no" : "yes");
		output.append(",\"");
		output.append(PermissionUtil.aclEntryFlagsToString(aclItem, isDirectory));
		output.append("\",\"");
		if (!aclItem.isFlagInheritedAce()) {
			output.append("(explicitly assigned)");
		} else if (aclItem.getInheritedFromFolder() != null && aclItem.getInheritedFromFolder().length() > 0) {
			output.append(aclItem.getInheritedFromFolder());
		} else {
			output.append("Unknown parent folder");
		}
		
		output.append("\",");
		output.append(viaGroup);
		output.append("\r\n");
	}
}
