package com.perdemia.permissionanalyzer.application.jface.contentproviders;

import java.sql.PreparedStatement;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ILazyTreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.util.ApplicationUtil;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.FilterHelper.ObjectType;
import com.perdemia.permissionanalyzer.data.AclItem;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.util.MembershipCache;

public class AllMembersPartContentProvider implements ILazyTreeContentProvider {
	private static Logger logger = LoggerFactory.getLogger(AllMembersPartContentProvider.class.getName());	
	
	private TreeViewer viewer;
	private AllMembersPartContentProviderConfig config;
	private List<Member> members = null;	
	private List<Member> filteredMembers = null;
	
	public AllMembersPartContentProvider(TreeViewer viewer){
		this.viewer = viewer;
	}
	
	@Override
	public void dispose() {
	}
	
	@SuppressWarnings("unchecked")
	@Override
	public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		if (newInput != null) {
			config = (AllMembersPartContentProviderConfig) newInput;
			
			try {
				if (members == null) {
					members  = Member.getByPreparedStatement(config.getFilterHelper().buildFilterQuery(ObjectType.MEMBERS_IN_FILE_TREE, false, false, config.getLimit(), "FILTER_MEMBER.NAME_LOWERCASE", config.getAdditionalFilter()));
				}
				
				String search = config.getMemberSearch();
				if (search.length() > 0) {
					filteredMembers.clear();
					for (Member member : members) {
						// match display name or member name?
						if ((member.getDisplayName(false, true).toLowerCase().contains(search.toLowerCase()) || member.getName(false).toLowerCase().contains(search.toLowerCase()))) {
							filteredMembers.add(member);
						}
					}
				} else {
					filteredMembers = (List<Member>) ((ArrayList<Member>)members).clone();
				}
			} catch (DBException ex) {
				logger.error("Error while getting all members from the results", ex);
				SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while getting all members from the results: " + ex.getMessage());
				
				// rethrow
				throw new RuntimeException("Error while getting all members from the results: " + ex.getMessage(), ex);
			}
		} else {
			members = null;
			filteredMembers = null;
		}
	}

	@Override
	public Object getParent(Object element) {
		return null;
	}

	@Override
	public void updateElement(Object parent, int index) {
		if (logger.isTraceEnabled()){
			logger.trace("Entering updateElement: " + parent + ", " + index);
		}
		
		Member memberElement = null;
		AclItem aclItemElement = null;
		
		try {
			if (parent instanceof Member) {
				aclItemElement = ((Member) parent).getAclItemList().get(index);
				// reference back to the member, see AllMembersPart.getSelectedMember()
				aclItemElement.setViaMember((Member) parent);
				
			} else {
				memberElement = filteredMembers.get(index);
			}
					
			if (aclItemElement != null) {
				// just replace the AclItem in the tree and set the child count to 0
				viewer.replace(parent, index, aclItemElement);
				viewer.setHasChildren(aclItemElement, false);
				
			} else {
				
				if (logger.isTraceEnabled()){
					logger.trace("Replacing tree element with index " + index + ": " + memberElement);
				}
					
				viewer.replace(parent, index, memberElement);
				viewer.setHasChildren(memberElement, true);
			}
		} catch (Exception ex) {
			SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while updating element of " + parent + ": " + ex.getMessage());
		}
	}

	@Override
	public void updateChildCount(Object element, int currentChildCount) {
		if (logger.isTraceEnabled()){
			logger.trace("Entering updateChildCount: " + element + ", " + currentChildCount);
		}
		
		ApplicationUtil.setBusyCursor();
		
		int childSize = 0;
		
		if (element instanceof Member) {
			Member memberElement = (Member) element;
			
			// get ALL applicable AclItems for this member
			try {
				String groupIdFilter = MembershipCache.getMemberOf(memberElement).toString().replace("[", "(").replace("]", ")");
				
				// add the current member id as well
				String comma = "";
				if (groupIdFilter.length() > 2) {
					comma = ",";
				}
				groupIdFilter = groupIdFilter.replace(")", comma + memberElement.getId() + ")");
				
				String hideInheritedPermissions = config.getFilterHelper().getFilterConfig(true).isHideInheritedPermissions() ? "ACLITEM.FLAG_INHERITED_ACE = FALSE AND " : "";
				
				PreparedStatement ps = config.getFilterHelper().buildFilterQuery(FilterHelper.ObjectType.ACL_ITEM, false, false, 0, "ACLITEM.FILEINFO_ID", hideInheritedPermissions + "ACLITEM.MEMBER_ID IN " + groupIdFilter);
				memberElement.setAclItemList(AclItem.getByPreparedStatement(ps));
			
				if (memberElement.getAclItemList().size() == 0) {
					AclItem dummyAcl = new AclItem();
					dummyAcl.setMember(new Member("No explicit ACE's found"));
					memberElement.getAclItemList().add(dummyAcl);
				}
				
				childSize = memberElement.getAclItemList().size();
			} catch (DBException ex) {
				logger.error("Error while getting ALL ACE's for " + memberElement.getName(true), ex);
				SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while getting ALL ACE's for " + memberElement.getName(true) + ": " + ex.getMessage());
			}
		} else {
			childSize = filteredMembers.size();
		}
		
		if (logger.isDebugEnabled()) {
			logger.debug("Setting child count " + childSize + " for element " + element.toString());
		}
		
		if (viewer != null) {
			viewer.setChildCount(element, childSize);
		}
		
		ApplicationUtil.setDefaultCursor();
		
		if (logger.isDebugEnabled()) {
			logger.debug("Exiting updateChildCount");
		}
	}

	public List<Member> getMembers() {
		return members;
	}

	public List<Member> getFilteredMembers() {
		return filteredMembers;
	}
}