package com.perdemia.permissionanalyzer.application.jface.contentproviders;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ILazyTreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.parts.EffectivePermissionsPart;
import com.perdemia.permissionanalyzer.application.preferences.PreferenceHelper;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.FilterHelper.ObjectType;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper.Edition;
import com.perdemia.permissionanalyzer.application.util.PermissionUtil;
import com.perdemia.permissionanalyzer.data.AclItem;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Filter;
import com.perdemia.permissionanalyzer.data.Report;
import com.perdemia.permissionanalyzer.data.Filter.FilterType;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.util.MemberCache;
import com.perdemia.permissionanalyzer.data.util.MembershipCache;

public class MainFileTreeContentProvider implements ILazyTreeContentProvider {
	private static Logger logger = LoggerFactory.getLogger(MainFileTreeContentProvider.class.getName());	
	private TreeViewer viewer;
	private List<FileInfo> rootFiles;
	private FilterHelper filterHelper;
	private static boolean errorDisplayed;
	
	public MainFileTreeContentProvider(TreeViewer viewer, FilterHelper filterHelper){
		this.viewer = viewer;
		this.filterHelper = filterHelper;
	}
	
	@Override
	public void dispose() {
	}

	@SuppressWarnings("unchecked")
	@Override
	public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		rootFiles = (List<FileInfo>) newInput;
		filterHelper.setFilesDisplayedInTree(0);
		errorDisplayed = false;
	}

	@Override
	public Object getParent(Object element) {
		if (element instanceof FileInfo) {
			FileInfo parentFile = ((FileInfo) element).getParent();
			if (parentFile != null) {
				logger.debug("returning parent with id " + parentFile.getId());
				return parentFile;
			} else {
				logger.debug("returning rootFiles as parent");
				return rootFiles;
			}
		} else {
			return rootFiles;
		}
	}

	@Override
	public void updateElement(Object parent, int index) {
		if (logger.isTraceEnabled()){
			logger.trace("Entering updateElement");
		}
		
		FileInfo element;
		if (parent instanceof FileInfo) {
			element = (FileInfo)((FileInfo) parent).getChildren().get(index);
		} else {
			element = rootFiles.get(index);
		}
		
		// get the owner
		if (element.getOwnerId() != null) {
			try {
				element.setOwner(MemberCache.getMember(element.getOwnerId()));
			} catch (DBException e) {
				logger.error("Error while getting the owner.", e);
			}
		}
		
		// load the related AclItems for this FileInfo
		if (element.getAclItems() == null) {
			filterHelper.setFilesDisplayedInTree(filterHelper.getFilesDisplayedInTree() + 1);
			try {
				List<AclItem> aclItems;
				PreparedStatement ps = filterHelper.buildFilterQuery(FilterHelper.ObjectType.ACL_ITEM, false, false, 0, "ACLITEM.MEMBER_ID", "ACLITEM.FILEINFO_ID = ?", element.getId());
				//if (ps != null) {
					aclItems = AclItem.getByPreparedStatement(ps);
				//} else {
					// no filter applied, just retrieve all the ACE of this file
				//	aclItems = AclItem.where("FILEINFO_ID = ?", new Object[]{element.getId()});
				//}
			
				element.setAclItems(aclItems);
				element.setEffectivePermissions(PermissionUtil.determineEffectivePermissions(element.getAclItems()));
			} catch (DBException ex) {
				SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while retrieving ACE's for the file/folder " + element.getName() + ": " + ex.getMessage());
			}
			
			if (PreferenceHelper.getInstance().isCalculateOwnershipRatioInOverview()) {
				determineOwnershipRatio(filterHelper, element);
			}
				
			if (logger.isTraceEnabled()){
				logger.trace("Replacing tree element with index " + index + ": " + element);
			}
			viewer.replace(parent, index, element);
			updateChildCount(element, -1);
		}
	}

	@Override
	public void updateChildCount(Object element, int currentChildCount) {
		if (logger.isTraceEnabled()){
			logger.trace("Entering updateChildCount");
		}
		int length = 0;
		if (element instanceof FileInfo) {
			FileInfo fileElement = (FileInfo) element;
						
			//load the children of this FileInfo element
			if (fileElement.isDirectory()) {
				try {
					// load the child files
					List<FileInfo> childFiles = FileInfo.where("FILEINFO.PARENT_ID = ?", new Object[]{fileElement.getId()});
					
					// remember the reference to the parent file
					for (FileInfo childFile : childFiles) {
						childFile.setParent(fileElement);
					}
					
					// remove child files that are not applicable
					if (filterHelper.getApplicableFileIdsIncludingParents() != null) {
						Iterator<FileInfo> iter = childFiles.iterator();
						while (iter.hasNext()) {
							if (!filterHelper.getApplicableFileIdsIncludingParents().contains(iter.next().getId())) {
								iter.remove();
							}
						}
					}
					
					Collections.sort(childFiles);
					
					fileElement.setChildren(childFiles);
					length = childFiles.size();
					
				} catch (DBException ex) {
					SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while loading children of the folder " + fileElement.getName() + ": " + ex.getMessage());
				}
			}
		} else if(element == rootFiles) {
			length = rootFiles.size();
			
		} else {
			logger.warn("Element is not of type FileInfo or rootFiles in FileContentProvider.updateChildCount: " + element.getClass());
		}
		
		if (logger.isDebugEnabled()) {
			logger.debug("Setting child count " + length + " for element " + element.toString());
		}
		
		viewer.setChildCount(element, length);
		
		if (logger.isDebugEnabled()) {
			logger.debug("Exiting updateChildCount");
		}
	}
	
	/**
	 * This method calculates the ownership ratio:
	 * 
	 * <ul>
	 * <li><strong>allMatchingMembersOnFolder</strong> = all matching members from the selected directory (the effective permissions tab)</li>
	 * <li><strong>allFilteredMembers</strong> = all members from the filter (static value after a filter has been applied)</li>
	 * <li><strong>allMembersOnFolder</strong> = all members from the file ACL, without using the filters.</li>
	 * </ul>
	 * 
	 * These values result in the following two ratio's:
	 * <ul>
	 * <li><strong>Total Ownership ratio</strong>: allMatchingMembersOnFolder  / allFilteredMembers</li> 
	 * <li><strong>Group Ownership ratio</strong>: allMatchingMembersOnFolder  / allMembersOnFolder</li>
	 * </ul>
	 * This method is called either from the {@link EffectivePermissionsPart} tab when a file is selected, or in the 
	 * {@link MainFileTreeContentProvider} and {@link MainFileTableContentProvider} to be displayed in the folder tree/table.
	 * This depends on the preferences, the last scenario has more impact on the performance because this calculation is done for
	 * every folder/file in the tree/table.
	 * 
	 * @see {@link FilterHelper#getAllFilteredMembers()}
	 */
	public static void determineOwnershipRatio(FilterHelper filterHelper, FileInfo file) {
		ResultSet rs = null;
		
		boolean noMemberships = !filterHelper.getFilterConfig(false).isIncludeGroupMembership();
		boolean hasMemberFilter = filterHelper.isApplyMemberFilter();
		
		if (hasMemberFilter) {
			int allMatchingMembersOnFolder = 0;
			int allMembersOnFolder = 0;
			
			try {
				// ----------------------------------------
				// all matching members on the current folder
				// ----------------------------------------
				List<Member> directMembers = new ArrayList<Member>();
				
				// get the distinct list with members
				for (AclItem aclItem : file.getAclItems()) {
					// already in the list
					boolean memberFound = false;
					for (Member member : directMembers) {
						if (member.getId().equals(aclItem.getMemberId())) {
							memberFound = true;
							break;
						}
					}
					
					if (!memberFound) {
						directMembers.add(aclItem.getMember(true));
					}
				}
				
				file.setDirectMembers(directMembers);
				
				if (noMemberships) {
					allMatchingMembersOnFolder = directMembers.size();
					
				} else {
					// use a query to get all (nested) members for this folder that match the filter criteria
					if (directMembers.size() > 0) {
						StringBuilder additionalMemberFilterBuilder = new StringBuilder("MEMBERSHIP.GROUP_ID IN (");
						Object[] additionalMemberFilterValues = new Object[directMembers.size()];
						Iterator<Member> iter = directMembers.iterator();
						int index = 0;
						while (iter.hasNext()) {
							additionalMemberFilterBuilder.append("?");
							additionalMemberFilterValues[index] = iter.next().getId();
							if (iter.hasNext()) {
								additionalMemberFilterBuilder.append(",");
							}
							index++;
						}
						additionalMemberFilterBuilder.append(")");
						
						Integer memberCount = MembershipCache.getMembershipCount(Arrays.asList(additionalMemberFilterValues).toString(), true);
						if (memberCount == null) {
							PreparedStatement membersThatMatchFilterPS = filterHelper.buildFilterQuery(ObjectType.MEMBER_COUNT, false, false, 0, null, additionalMemberFilterBuilder.toString(), additionalMemberFilterValues);
							rs = DBObject.executePreparedStatement(membersThatMatchFilterPS, false);
							if (rs.next()){
								memberCount = rs.getInt(1);
							}
							rs.close();
							
							MembershipCache.cacheMembershipCount(Arrays.asList(additionalMemberFilterValues).toString(), memberCount, true);						
						}
						
						allMatchingMembersOnFolder += memberCount;
					}							
				}
				
				// ---------------------------------------------------------------
				// all members on the current folder (without using the filters)
				// ---------------------------------------------------------------
				List<AclItem> allFolderAclItems = AclItem.where("FILEINFO_ID = ?", new Object[]{file.getId()});
				List<Member> allDirectMembers = new ArrayList<Member>();
				
				// get the distinct list with members
				for (AclItem aclItem : allFolderAclItems) {
					// already in the list
					boolean memberFound = false;
					for (Member member : allDirectMembers) {
						if (member.getId().equals(aclItem.getMemberId())) {
							memberFound = true;
							break;
						}
					}
					
					if (!memberFound) {
						allDirectMembers.add(aclItem.getMember(true));
					}
				}
				
				// although we don't use the filters, the customer does want to apply the exclude filters! Do this manually
				List<Integer> excludeMemberIds = new ArrayList<Integer>();
				List<String> excludeMemberNames = new ArrayList<String>();
				for (Filter filter : filterHelper.getMemberFilterListApplied()) {
					if (!filter.isInclude()) {
						if (filter.getType().equals(FilterType.MEMBER_ID)) {
							excludeMemberIds.add(Integer.parseInt(filter.getValue()));
							
						} else if (filter.getType().equals(FilterType.MEMBER_NAME)) {
							excludeMemberNames.add(filter.getValue().replace("%", "").replace("*", "").toLowerCase());
							
						} else if (filter.getType().equals(FilterType.MEMBER_GROUP_ID)) {
							List<Member> members = MembershipCache.getGroupMembers(new Member("", Integer.parseInt(filter.getValue())), true, true, false);
							for (Member member : members) {
								excludeMemberIds.add(member.getId());
							}
						} else if (filter.getType().equals(FilterType.MEMBER_REPORT)) {
							Report report = Report.getById(Integer.parseInt(filter.getValue()));
							List<Member> reportMembers = report.getMembersInFolderTree();
							for (Member member : reportMembers) {
								excludeMemberIds.add(member.getId());
							}
						}
					}
				}
				
				// filter out the members
				Iterator<Member> memberIter = allDirectMembers.iterator();
				while (memberIter.hasNext()) {
					Member member = memberIter.next();
					if (excludeMemberIds.contains(member.getId())) {
						memberIter.remove();
					} else {
						// check for member name
						for (String excludeName : excludeMemberNames) {
							if (member.getDisplayName(true, false).toLowerCase().contains(excludeName)) {
								memberIter.remove();
								break;
							}
						}
					}
				}
				
				file.setAllDirectMembers(allDirectMembers);
				
				if (noMemberships) {
					allMembersOnFolder = allDirectMembers.size();
					
				} else {
					// use a query to get all (nested) members for this folder (without using the filter criteria)
					if (allDirectMembers.size() > 0) {
						StringBuilder additionalMemberFilterBuilder = new StringBuilder("MEMBERSHIP.GROUP_ID IN (");
						Object[] additionalMemberFilterValues = new Object[allDirectMembers.size() + excludeMemberIds.size()];
						Iterator<Member> iter = allDirectMembers.iterator();
						int index = 0;
						while (iter.hasNext()) {
							additionalMemberFilterBuilder.append("?");
							additionalMemberFilterValues[index] = iter.next().getId();
							if (iter.hasNext()) {
								additionalMemberFilterBuilder.append(",");
							}
							index++;
						}
						additionalMemberFilterBuilder.append(")");
						
						if (excludeMemberIds.size() > 0) {
							// apply the exclude member id's
							additionalMemberFilterBuilder.append(" AND MEMBERSHIP.MEMBER_ID NOT IN (");
							
							Iterator<Integer> excludeIdsIter = excludeMemberIds.iterator();
							
							while (excludeIdsIter.hasNext()) {
								additionalMemberFilterBuilder.append("?");
								additionalMemberFilterValues[index] = excludeIdsIter.next();
								if (excludeIdsIter.hasNext()) {
									additionalMemberFilterBuilder.append(",");
								}
								index++;
							}
							additionalMemberFilterBuilder.append(")");
						}
						
						Integer memberCount = MembershipCache.getMembershipCount(Arrays.asList(additionalMemberFilterValues).toString(), false);
						if (memberCount == null) {
							rs = DBObject.executeQuery("select count(distinct member_id) from permissionanalyzer.membership where " + additionalMemberFilterBuilder.toString(), additionalMemberFilterValues);
							if (rs.next()){
								memberCount = rs.getInt(1);
							}
							rs.close();
													
							MembershipCache.cacheMembershipCount(Arrays.asList(additionalMemberFilterValues).toString(), memberCount, false);						
						}
						
						allMembersOnFolder += memberCount;
					}									
				}
				
				file.setAllMatchingMembersOnFolder(allMatchingMembersOnFolder);
				file.setAllMembersOnFolder(allMembersOnFolder);
				
			} catch (Exception ex) {
				if (!errorDisplayed) {
					errorDisplayed = true;
					logger.error("Error while calculating ownership ratio's for the file/folder " + file.getName(), ex);
					SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while calculating ownership ratio's for the file/folder " + file.getName() + ": " + ex.getMessage());
				}
			} finally {
				try {
					if (rs != null && !rs.isClosed()) {
						rs.close();
					}
				}catch (Exception ex) {}
			}	
		}
	}
}