package com.perdemia.permissionanalyzer.application.jface.dialogs;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.E4LifeCycle;
import com.perdemia.permissionanalyzer.application.preferences.PreferenceHelper;
import com.perdemia.permissionanalyzer.application.util.ImageFactory;
import com.perdemia.permissionanalyzer.application.util.ImageFactory.AppImage;

public class AlertDialog extends DialogWithTopBar {
	private static Logger logger = LoggerFactory.getLogger(AlertDialog.class.getName());
	
	private String title;
	private String message;
	private AppImage image;
	private Button hideNextTimeButton;
	private boolean hideNextTime = true;
	private boolean confirmMode;
	private AlertType type;
	
	public enum AlertType {
		OPEN_SCAN_VIEW, APPLY_FILTER_BUTTON, CONFIGURE_LDAP_CONNECTION, FOLLOW_LDAP_REFERRALS
	}
	
	/**
	 * Constructor for the AlertDialog
	 * 
	 * @param shell 
	 * @param title the title shown in the header
	 * @param message message of the alert
	 * @param image an image of 48X48
	 * @param confirmMode if true the dialog will display Yes/No buttons, otherwise a Close button is displayed
	 */
	public AlertDialog(Shell shell, String title, String message, AppImage image, boolean confirmMode, AlertType type) {
		super(shell);
		this.title = title;
		this.message = message;
		this.image = image;
		this.confirmMode = confirmMode;
		this.type = type;
	}

	@Override
	protected void configureShell(Shell shell) {
		super.configureShell(shell);
		shell.setImage(ImageFactory.getInstance().getImage(AppImage.APP_ICON_16X16));
		if (confirmMode) {
			shell.setSize(600, 280);
		} else {
			shell.setSize(450, 280);
		}			
		shell.setText(title);
	}
	
	@Override
	protected Control createContents(Composite parent) {
		if (PreferenceHelper.getInstance().getShowAlertDialogs().get(type.ordinal())) {
			return super.createContents(parent);
		} else {
			// don't build the dialog
			return new Composite(parent, 0);
		}
	}
	
	@Override
	public int open() {
		if (!E4LifeCycle.isSilentMode && PreferenceHelper.getInstance().getShowAlertDialogs().get(type.ordinal())) {
			return super.open();
		} else {
			return 0;
		}
	}
	
	@Override
	protected Control createDialogArea(Composite container) {
		Composite composite = (Composite) super.createDialogArea(container, title, ImageFactory.getInstance().getImage(image));
		container.setBackgroundMode(SWT.INHERIT_DEFAULT);
				
		GridLayout gridLayout = new GridLayout(1, false);
		gridLayout.marginTop = 30;
		gridLayout.marginLeft = 30;
		gridLayout.marginRight = 30;
		composite.setLayout(gridLayout);
		
		Label descriptionLabel = new Label(composite, SWT.NONE | SWT.WRAP);
		GridData gridData = new GridData();
		if (confirmMode) {
			gridData.widthHint = 520;
		} else {
			gridData.widthHint = 400;
		}
		descriptionLabel.setLayoutData(gridData);
		descriptionLabel.setText(message);
				
		return composite;
	}

	@Override
	protected void createButtonsForButtonBar(Composite parent) {
		GridLayout gridLayout = (GridLayout) parent.getLayout();
		gridLayout.makeColumnsEqualWidth = false;
		((GridData)parent.getLayoutData()).horizontalAlignment = GridData.FILL;
		
		hideNextTimeButton = new Button(parent, SWT.CHECK);
		hideNextTimeButton.setText("Do not show this message again");
		hideNextTimeButton.setSelection(true);
		GridData gridData = new GridData();
		
		if (confirmMode) {
			gridData.widthHint = 360;
		} else {
			gridData.widthHint = 300;
		}
		hideNextTimeButton.setLayoutData(gridData);
		hideNextTimeButton.addSelectionListener(new SelectionAdapter(){
			@Override
			public void widgetSelected(SelectionEvent e) {
				hideNextTime = hideNextTimeButton.getSelection();
			}
		});
		((GridLayout) parent.getLayout()).numColumns++; //also triggered by the createButton calls
		
		if (confirmMode) {
			createButton(parent, IDialogConstants.YES_ID, "Yes", true);
			createButton(parent, IDialogConstants.NO_ID, "No", false);
		} else {
			createButton(parent, IDialogConstants.CLOSE_ID, "Close", true);
		}
	}
	
	@Override
	protected void buttonPressed(int buttonId) {
		if (hideNextTime) {
			try {
				PreferenceHelper.getInstance().getShowAlertDialogs().set(type.ordinal(), false);
				PreferenceHelper.getInstance().storePreferences(false);
			} catch (Exception e) {
				logger.error("Error while storing preference for AlertDialog", e);
			}
		}
		setReturnCode(buttonId);
		close();
	}
}
