package com.perdemia.permissionanalyzer.application.jface.dialogs;

import java.util.List;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Spinner;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Tree;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.export.FilterRow;
import com.perdemia.permissionanalyzer.application.export.ReportFilters;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.ReportFilterContentProvider;
import com.perdemia.permissionanalyzer.application.model.LdapConnectionInfo;
import com.perdemia.permissionanalyzer.application.swt.ExplainLabel;
import com.perdemia.permissionanalyzer.application.swt.ExplainLabel.ExplainType;
import com.perdemia.permissionanalyzer.application.swt.ImageCombo;
import com.perdemia.permissionanalyzer.application.util.ApplicationUtil;
import com.perdemia.permissionanalyzer.application.util.ColorFactory;
import com.perdemia.permissionanalyzer.application.util.ColorFactory.AppColor;
import com.perdemia.permissionanalyzer.application.util.Constants;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.ImageFactory;
import com.perdemia.permissionanalyzer.application.util.ImageFactory.AppImage;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper.Edition;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Filter;
import com.perdemia.permissionanalyzer.data.FilterSet;
import com.perdemia.permissionanalyzer.data.Report;
import com.perdemia.permissionanalyzer.data.Report.FileType;
import com.perdemia.permissionanalyzer.data.Report.ReportType;
import com.perdemia.permissionanalyzer.data.ReportFilterSet;

public class ReportDetailsDialog extends DialogWithTopBar {
	private static Logger logger = LoggerFactory.getLogger(ReportDetailsDialog.class.getName());
	
	/** Options for this dialog.
	 * <p>
	 * <ul>
	 * <li>The CREATE option displays the details, a name text box and a Add button.</li>
	 * <li>The SAVE option expects a selectedReport and will write the data to that object. The name is displayed but disabled.</li>
	 * <li>The EXPORT option hides the name and will only show the details and an export button.</li>
	 * </ul>
	 * <p>
	 * */
	public enum ReportDialogMode {
		CREATE, SAVE, EXPORT
	}
	private ReportDialogMode dialogMode;
	
	/**
	 * This dialog can be used for Reports and Polices, both entities require the same fields except that a policy has a threshold value.
	 */
	private boolean isPolicy;
		
	/** UI controls */
	private TreeViewer filterTreeViewer;
	private Text reportNameText;
	private Text reportDescriptionText;
	private Composite expandGroupsComposite;
	private Button dontExpandGroupsButton;
	private Button expandGroupsDirectMembersButton;
	private Button expandGroupsNestedMembersButton;
	
	private ImageCombo reportTypeCombo;
	private Combo fileTypeCombo;
	private Button useSimplePresentationCheckbox;
	private Text filePathText;
	private Text templatePathText;
	private Text emailAddressText;
	private Spinner emailThresholdSpinner;
	private Button createComparisonButton;
	
	private Composite linkToFilterSetComposite;
	private Button linkToMemberFilterSetButton;
	private Button linkToPermissionFilterSetButton;
	private Button linkToFolderFilterSetButton;
	private Label noFilterSetsLabel;
	private Button selectedFileButton;
	private Button exportSubdirectoriesButton;
	
	/** Selected report used when loading a report into the Report perspective. */
	private Report selectedReport = new Report();
	
	/** Selected FileInfo in case of an export */
	private FileInfo selectedFile;
	private boolean exportSelectedFile;
	private boolean exportSubdirectories;
	
	/** Filters and FilterSets that belong to the selected report (or GUI in case a new report is created) */
	private ReportFilters reportFilters = new ReportFilters();
	
	/**
	 * Constructor used for creating a new report or policy
	 * 
	 * @param shell
	 * @param dialogMode
	 */
	public ReportDetailsDialog(Shell shell, Report report) {
		super(shell);
		
		// load the filter sets config
		try {
			reportFilters.setMemberFilters(report.getMemberFilters());
			reportFilters.setPermissionFilters(report.getPermissionFilters());
			reportFilters.setFolderFilters(report.getFolderFilters());
			reportFilters.setMemberFilterSet(report.getMemberFilterSet());
			reportFilters.setPermissionFilterSet(report.getPermissionFilterSet());
			reportFilters.setFolderFilterSet(report.getFolderFilterSet());
			
		} catch (DBException ex) {
			logger.error("Error while loading the report filters from the database for report " + report.getName(), ex);
			MessageDialog.openError(null, "Error", "Error while loading the report filters from the database: " + ex.getMessage());
		}
		
		dialogMode = ReportDialogMode.SAVE;
		selectedReport = report;
		isPolicy = report.getEmailThreshold() > 0;
	}
	
	
	/**
	 * Constructor used for creating a new report or policy
	 * 
	 * @param shell
	 * @param dialogMode
	 */
	public ReportDetailsDialog(Shell shell, boolean isPolicy, FilterSet memberFilterSet, FilterSet permissionFilterSet, FilterSet folderFilterSet) {
		this(shell, isPolicy, null, memberFilterSet, permissionFilterSet, folderFilterSet);
	}
	
	/**
	 * Constructor used to save an existing report
	 * 
	 * @param shell
	 * @param selectedReport report to override
	 */
	public ReportDetailsDialog(Shell shell, boolean isPolicy, Report report, FilterSet memberFilterSet, FilterSet permissionFilterSet, FilterSet folderFilterSet) {
		super(shell);
		
		this.isPolicy = isPolicy; 
		if (report != null) {
			dialogMode = ReportDialogMode.SAVE;
			selectedReport = report;
		} else {
			dialogMode = ReportDialogMode.CREATE;
		}
		
		// input for the TreeViewer
		reportFilters.setMemberFilterSet(memberFilterSet);
		reportFilters.setPermissionFilterSet(permissionFilterSet);
		reportFilters.setFolderFilterSet(folderFilterSet);
		
		// the filters for this report are the ones selected in the GUI
		reportFilters.setMemberFilters(FilterHelper.getInstance().getMemberFilterList());
		reportFilters.setPermissionFilters(FilterHelper.getInstance().getPermissionFilterList());
		reportFilters.setFolderFilters(FilterHelper.getInstance().getFolderFilterList());
	}
	
	/**
	 * Constructor used to export the current data or selected file
	 * 
	 * @param shell
	 */
	public ReportDetailsDialog(Shell shell, FileInfo file){
		super(shell);
		dialogMode = ReportDialogMode.EXPORT;
		selectedFile = file;
		
		// the filters for this report are the ones selected in the GUI
		reportFilters.setMemberFilters(FilterHelper.getInstance().getMemberFilterListApplied());
		reportFilters.setPermissionFilters(FilterHelper.getInstance().getPermissionFilterListApplied());
		reportFilters.setFolderFilters(FilterHelper.getInstance().getFolderFilterListApplied());
	}
		
	@Override
	protected boolean isResizable()  {
	  return true;
	}
	
	@Override
	protected void configureShell(Shell shell) {
		super.configureShell(shell);
		if (isPolicy) {
			shell.setImage(ImageFactory.getInstance().getImage(AppImage.POLICY_16X16));
		} else {
			shell.setImage(ImageFactory.getInstance().getImage(AppImage.REPORT_16X16));
		}
		shell.setSize(1000, 750);
	}
	
	@Override
	protected Control createDialogArea(final Composite container) {		
		ApplicationUtil.setBusyCursor();
		int labelWidth = dialogMode.equals(ReportDialogMode.EXPORT) ? 270 : 240;
		
		container.setData(Constants.CSS_CLASS_NAME_KEY, "MPart");
		container.setBackgroundMode(SWT.INHERIT_DEFAULT);
		
		String title = "";
		Image image = null;
		switch (dialogMode) {
			case CREATE:
				if (isPolicy) {
					title = "Create a new policy";
					image = ImageFactory.getInstance().getImage(AppImage.CREATE_POLICY_48X48);
				} else {
					title = "Create a new report";
					image = ImageFactory.getInstance().getImage(AppImage.CREATE_REPORT_48X48);
				}
				break;
			case SAVE:
				if (isPolicy) {
					title = "Policy details";
					image = ImageFactory.getInstance().getImage(AppImage.POLICY_48X48);
				} else {
					title = "Report details";
					image = ImageFactory.getInstance().getImage(AppImage.REPORT_48X48);
				}
				break;
			case EXPORT:
				title = "Export";
				image = ImageFactory.getInstance().getImage(AppImage.REPORT_48X48);
				break;
		}
		Composite composite = (Composite) super.createDialogArea(container, title, image);
		container.getShell().setText(title);
						
		// set width, height and layout
		GridLayout gridLayout = new GridLayout(2, false);
		gridLayout.marginTop = 7;
		gridLayout.marginLeft = 7;
		gridLayout.marginRight = 7;
		composite.setLayout(gridLayout);			
		
		// TRIAL banner
		if (LicenseHelper.getEdition().equals(Edition.TRIAL) || LicenseHelper.getEdition().equals(Edition.BASIC)) {
			Composite trialComposite = new Composite(composite, SWT.NONE);
			GridData gridData = new GridData(GridData.FILL_HORIZONTAL);
			gridData.horizontalSpan = 3;
			gridData.heightHint = 40;
			//gridData.verticalIndent = 20;
			trialComposite.setLayoutData(gridData);
			gridLayout = new GridLayout(1, false);		
			trialComposite.setLayout(gridLayout);
			trialComposite.setBackground(ColorFactory.getInstance().getColor(AppColor.TRIAL_BANNER));
			
			Label trialLabel = new Label(trialComposite, SWT.NONE);
			trialLabel.setText("The trial version will only export a depth of 3 subdirectories and only 30 groups in the \"Groups and members\" report.");
			gridData = new GridData(GridData.FILL_HORIZONTAL);
			gridData.horizontalIndent = 20;
			gridData.verticalAlignment = SWT.CENTER;
			gridData.grabExcessVerticalSpace = true;
			trialLabel.setLayoutData(gridData);
			trialLabel.setData(Constants.CSS_CLASS_NAME_KEY, "h2");
		}
		
		new ExplainLabel(composite, "If you have selected a filter selection as a filter, then that selection will show up as an option when creating the report. Using the selection will create a reference to the filter selection within the report and any modifications to the filter selection will result in all reports automatically applying the modified selection. Unchecking the option in the report will result in the report saving a copy of the filters and not change according to the filter selection.", ExplainType.REPORT_DETAILS, 2);
		
		if (!dialogMode.equals(ReportDialogMode.EXPORT)) {
			Label reportNameLabel = new Label(composite, SWT.NONE);
			if (isPolicy) {
				reportNameLabel.setText("Policy name:");
			} else {
				reportNameLabel.setText("Report name:");
			}
			GridData gridData = new GridData();
			gridData.widthHint = labelWidth;
			reportNameLabel.setLayoutData(gridData);
			reportNameText = new Text(composite, SWT.BORDER);
			reportNameText.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
			
		} else if (selectedFile != null) {
			// export mode with a file selected
			Label selectedFileLabel = new Label(composite, SWT.NONE);
			selectedFileLabel.setText("Only export selected file and children:");
			GridData gridData = new GridData();
			gridData.widthHint = labelWidth;
			selectedFileLabel.setLayoutData(gridData);
			selectedFileButton = new Button(composite, SWT.CHECK);
			selectedFileButton.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
			selectedFileButton.setText(selectedFile.determineFullPath());
			selectedFileButton.setSelection(true);
			selectedFileButton.addSelectionListener(new SelectionAdapter() {
				@Override
				public void widgetSelected(SelectionEvent e) {
					exportSubdirectoriesButton.setEnabled(selectedFileButton.getSelection());
				}
			});
			
			Label exportSubdirectoriesLabel = new Label(composite, SWT.NONE);
			gridData = new GridData();
			gridData.widthHint = labelWidth;
			exportSubdirectoriesLabel.setLayoutData(gridData);
			exportSubdirectoriesButton = new Button(composite, SWT.CHECK);
			exportSubdirectoriesButton.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
			exportSubdirectoriesButton.setText("Export subdirectories and files");
			exportSubdirectoriesButton.setSelection(true);
		}
		
		Label reportDescriptionLabel = new Label(composite, SWT.NONE);
		reportDescriptionLabel.setText("Report description:");
		reportDescriptionText = new Text(composite, SWT.BORDER | SWT.MULTI | SWT.WRAP | SWT.V_SCROLL);
		GridData gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.heightHint = 3 * reportDescriptionText.getLineHeight();
		reportDescriptionText.setLayoutData(gridData);
		
		Label reportTypeLabel = new Label(composite, SWT.NONE);
		reportTypeLabel.setText("Report type:");
				
		reportTypeCombo = new ImageCombo(composite, SWT.BORDER | SWT.READ_ONLY);
		reportTypeCombo.setVisibleItemCount(20);
		reportTypeCombo.add(ImageFactory.getInstance().getImage(AppImage.FOLDER_16X16), Report.reportTypeToString(ReportType.EFFECTIVE));
		reportTypeCombo.add(ImageFactory.getInstance().getImage(AppImage.FOLDER_USER_16X16), Report.reportTypeToString(ReportType.FOLDERS_AND_USERS_EFFECTIVE));
		reportTypeCombo.add(ImageFactory.getInstance().getImage(AppImage.PERM_READ_ACL_16X16), Report.reportTypeToString(ReportType.ACL));
		//reportTypeCombo.add(Report.reportTypeToString(ReportType.ACL_DIRECT_GROUP_MEMBERS));
		//reportTypeCombo.add(Report.reportTypeToString(ReportType.ACL_NESTED_GROUP_MEMBERS));
		reportTypeCombo.add(ImageFactory.getInstance().getImage(AppImage.EFFECTIVE_PERMISSIONS_16X16), Report.reportTypeToString(ReportType.ALL_MEMBERS));
		reportTypeCombo.add(ImageFactory.getInstance().getImage(AppImage.GROUP_16X16), "Groups that match the filter criteria and their members");
		//reportTypeCombo.add(Report.reportTypeToString(ReportType.MATCHING_GROUPS_AND_NESTED_MEMBERS));
		reportTypeCombo.add(ImageFactory.getInstance().getImage(AppImage.GROUP_16X16), "Groups that have permissions in the folder tree and their members");
		//reportTypeCombo.add(Report.reportTypeToString(ReportType.GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS));
		reportTypeCombo.add(ImageFactory.getInstance().getImage(AppImage.USER_16X16), "Users that match the filter criteria and their group memberships");
		//reportTypeCombo.add(Report.reportTypeToString(ReportType.MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS));
		reportTypeCombo.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		reportTypeCombo.addSelectionListener(new SelectionAdapter(){
			@Override
			public void widgetSelected(SelectionEvent e) {
				int i = reportTypeCombo.getSelectionIndex();
				
				dontExpandGroupsButton.setEnabled(i == 2);
				expandGroupsDirectMembersButton.setEnabled(i == 2 || i >= 4);
				expandGroupsNestedMembersButton.setEnabled(i == 2 || i >= 4);
				useSimplePresentationCheckbox.setEnabled(i < 4);
				if (!useSimplePresentationCheckbox.isEnabled()) {
					useSimplePresentationCheckbox.setSelection(false);
				}
				
				if (!dontExpandGroupsButton.isEnabled()) {
					dontExpandGroupsButton.setSelection(false);
				}
				if (!expandGroupsDirectMembersButton.isEnabled()) {
					expandGroupsDirectMembersButton.setSelection(false);
				}
				if (!expandGroupsNestedMembersButton.isEnabled()) {
					expandGroupsNestedMembersButton.setSelection(false);
				}
				
				// nothing selected yet? Select the default
				if (!dontExpandGroupsButton.getSelection() && !expandGroupsDirectMembersButton.getSelection() && !expandGroupsNestedMembersButton.getSelection() && expandGroupsDirectMembersButton.isEnabled()) {
					expandGroupsDirectMembersButton.setSelection(true);
				}
			}
		});
				
		// Expand groups radio buttons
		expandGroupsComposite = new Composite(composite, SWT.NONE);
		gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.horizontalSpan = 2;
		expandGroupsComposite.setLayoutData(gridData);
		expandGroupsComposite.setLayout(new GridLayout(4, false));
		Label expandGroupsLabel = new Label(expandGroupsComposite, SWT.NONE);
		expandGroupsLabel.setText("Expand groups:");
		gridData = new GridData();
		gridData.widthHint = labelWidth;
		gridData.horizontalIndent = -5;
		expandGroupsLabel.setLayoutData(gridData);

		dontExpandGroupsButton = new Button(expandGroupsComposite, SWT.RADIO);		
		dontExpandGroupsButton.setText("Do not expand");
		gridData = new GridData();
		gridData.widthHint = 120;
		dontExpandGroupsButton.setLayoutData(gridData);
				
		expandGroupsDirectMembersButton = new Button(expandGroupsComposite, SWT.RADIO);
		expandGroupsDirectMembersButton.setText("Expand direct memberships");
		gridData = new GridData();
		gridData.widthHint = 190;
		expandGroupsDirectMembersButton.setLayoutData(gridData);
				
		expandGroupsNestedMembersButton = new Button(expandGroupsComposite, SWT.RADIO);
		expandGroupsNestedMembersButton.setText("Expand nested memberships");
		gridData = new GridData();
		gridData.widthHint = 200;
		expandGroupsNestedMembersButton.setLayoutData(gridData);
		
		Composite fileTypeComposite = new Composite(composite, SWT.NONE);
		gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.horizontalSpan = 2;
		fileTypeComposite.setLayoutData(gridData);
		fileTypeComposite.setLayout(new GridLayout(4, false));
				
		Label fileTypeLabel = new Label(fileTypeComposite, SWT.NONE);
		fileTypeLabel.setText("File type:");
		gridData = new GridData();
		gridData.widthHint = labelWidth;
		gridData.horizontalIndent = -5;
		fileTypeLabel.setLayoutData(gridData);
		
		fileTypeCombo = new Combo(fileTypeComposite, SWT.BORDER | SWT.READ_ONLY);
		fileTypeCombo.add(Report.fileTypeToString(FileType.HTML));
		fileTypeCombo.add(Report.fileTypeToString(FileType.CSV));
		fileTypeCombo.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
				
		useSimplePresentationCheckbox = new Button(fileTypeComposite, SWT.CHECK);
		useSimplePresentationCheckbox.setText("Use simple presentation");
		
		Button infoSimplePresentationButton = new Button(fileTypeComposite, SWT.PUSH);
		infoSimplePresentationButton.setImage(ImageFactory.getInstance().getImage(AppImage.INFO_16X16));
		infoSimplePresentationButton.addSelectionListener(new SelectionAdapter(){
			@Override
			public void widgetSelected(SelectionEvent e) {
				MessageDialog.openInformation(getShell(), "Help", "The simple presentation will leave out most of the HTML columns, like the special permissions, ACL flags and inheritance information. It will also hide the filter information and column checkboxes.");
			}
		});
		
		Composite filePathComposite = new Composite(composite, SWT.NONE);
		gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.horizontalSpan = 2;
		filePathComposite.setLayoutData(gridData);
		filePathComposite.setLayout(new GridLayout(4, false));
		Label filePathLabel = new Label(filePathComposite, SWT.NONE);
		filePathLabel.setText("Target file path:");
		filePathLabel.setToolTipText("Use a date placeholder to use the current date in the file path, like [date:yyyy-MM-dd HHmm]");
		gridData = new GridData();
		gridData.widthHint = labelWidth;
		gridData.horizontalIndent = -5;
		filePathLabel.setLayoutData(gridData);
		filePathText = new Text(filePathComposite, SWT.BORDER);
		filePathText.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		filePathText.setToolTipText("Use a date placeholder to use the current date in the file path, like [date:yyyy-MM-dd HHmm]");
				
		Button browseFileButton = new Button(filePathComposite, SWT.PUSH);
		browseFileButton.setImage(ImageFactory.getInstance().getImage(AppImage.BROWSE_16X16));
		browseFileButton.addSelectionListener(new SelectionAdapter(){
			@Override
			public void widgetSelected(SelectionEvent e) {
				FileDialog dialog = new FileDialog(container.getShell(), SWT.SAVE);
			    
			    if (fileTypeCombo.getText().equals(Report.fileTypeToString(FileType.CSV))) {
			    	dialog.setFilterNames(new String[] { "CSV Files", "All Files (*.*)" });
			    	dialog.setFilterExtensions(new String[] { "*.csv", "*.*" });
			    } else {
			    	dialog.setFilterNames(new String[] { "HTML Files", "All Files (*.*)" });
			    	dialog.setFilterExtensions(new String[] { "*.html", "*.*" });
			    }
			    String filePath = dialog.open();
			    if (filePath != null) {
			    	filePathText.setText(filePath);
			    }
			}
		});
		
		Button insertDateButton = new Button(filePathComposite, SWT.PUSH);
		insertDateButton.setImage(ImageFactory.getInstance().getImage(AppImage.DATE_16X16));
		insertDateButton.addSelectionListener(new SelectionAdapter(){
			@Override
			public void widgetSelected(SelectionEvent e) {
		    	String prefix = filePathText.getText().substring(0, filePathText.getCaretPosition());
		    	String suffix = filePathText.getText().substring(filePathText.getCaretPosition());
				filePathText.setText(prefix + "[date:yyyy-MM-dd HHmm]" + suffix);
				filePathText.setFocus();
			}
		});
		
		Composite templatePathComposite = new Composite(composite, SWT.NONE);
		gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.horizontalSpan = 2;
		gridData.verticalIndent = -10;
		gridData.heightHint = 45;
		templatePathComposite.setLayoutData(gridData);
		templatePathComposite.setLayout(new GridLayout(4, false));
		Label templatePathLabel = new Label(templatePathComposite, SWT.NONE);
		templatePathLabel.setText("Custom template path:");
		gridData = new GridData();
		gridData.widthHint = labelWidth;
		gridData.horizontalIndent = -5;
		templatePathLabel.setLayoutData(gridData);
		templatePathText = new Text(templatePathComposite, SWT.BORDER);
		templatePathText.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
				
		Button browseTemplateButton = new Button(templatePathComposite, SWT.PUSH);
		browseTemplateButton.setImage(ImageFactory.getInstance().getImage(AppImage.BROWSE_16X16));
		browseTemplateButton.addSelectionListener(new SelectionAdapter(){
			@Override
			public void widgetSelected(SelectionEvent e) {
				FileDialog dialog = new FileDialog(container.getShell(), SWT.OPEN);
			    dialog.setFilterNames(new String[] { fileTypeCombo.getText() + " Files", "All Files (*.*)" });
			    dialog.setFilterExtensions(new String[] { "*.html;*.htm", "*.*" }); 
			    String filePath = dialog.open();
			    if (filePath != null) {
			    	templatePathText.setText(filePath);
			    }
			}
		});
		
		Button infoTemplateButton = new Button(templatePathComposite, SWT.PUSH);
		infoTemplateButton.setImage(ImageFactory.getInstance().getImage(AppImage.INFO_16X16));
		infoTemplateButton.addSelectionListener(new SelectionAdapter(){
			@Override
			public void widgetSelected(SelectionEvent e) {
				MessageDialog.openInformation(getShell(), "Help", "Use a custom template to modify the layout of the HTML or CSV file. If you leave this field empty, then the appplication will use a the default template. \r\n\r\nYou can find the default templates in the zip file <app_dir>\\plugins\\Permission_Analyzer_<version>.jar, it contains a folder called \"templates\".");
			}
		});
		
		Label emailAddressLabel = new Label(composite, SWT.NONE);
		emailAddressLabel.setText("E-mail recipient:");
		emailAddressLabel.setToolTipText("An e-mail will be sent after the report export.");
		emailAddressText = new Text(composite, SWT.BORDER);
		emailAddressText.setToolTipText("An e-mail will be sent after the report export.");
		emailAddressText.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
		
		if (isPolicy) {
			Label emailThresholdLabel = new Label(composite, SWT.NONE);
			emailThresholdLabel.setText("Policy alert threshold (file/folder count):   ");
			emailThresholdLabel.setToolTipText("Enter the minimal file/folder count that determines whether or not a policy has failed.");
			gridData = new GridData();
			//gridData.widthHint = labelWidth;
			gridData.verticalIndent = 15;
			gridData.heightHint = 25;
			emailThresholdLabel.setLayoutData(gridData);
			
			emailThresholdSpinner = new Spinner(composite, SWT.BORDER);
			emailThresholdSpinner.setMinimum(1);
			emailThresholdSpinner.setMaximum(Integer.MAX_VALUE);
			gridData = new GridData();
			gridData.widthHint = 50;
			emailThresholdSpinner.setLayoutData(gridData);
			emailThresholdSpinner.setToolTipText("Enter the minimal file/folder count that determines whether or not a policy has failed.");
		}
		
		if (!dialogMode.equals(ReportDialogMode.EXPORT)) {
			if (!isPolicy) {
				// Comparison checkbox
				Composite createComparisonComposite = new Composite(composite, SWT.NONE);
				gridData = new GridData(GridData.FILL_HORIZONTAL);
				gridData.horizontalSpan = 2;
				gridData.verticalIndent = 5;
				gridData.heightHint = 35;
				createComparisonComposite.setLayoutData(gridData);
				createComparisonComposite.setLayout(new GridLayout(3, false));
				
				Label createComparisonLabel = new Label(createComparisonComposite, SWT.NONE);
				createComparisonLabel.setText("Create comparison report:");
				createComparisonLabel.setToolTipText("When a report is generated, the application can automatically create a comparison report comparing the HTML / CSV file of the previous export with the current one.");
				gridData = new GridData();
				gridData.widthHint = labelWidth;
				gridData.horizontalIndent = -5;
				createComparisonLabel.setLayoutData(gridData);
				
				createComparisonButton = new Button(createComparisonComposite, SWT.CHECK);
				createComparisonButton.setToolTipText("When a report is generated, the application can automatically create a comparison report comparing the HTML / CSV file of the previous export with the current one.");
				createComparisonButton.setLayoutData(new GridData());
				
				Button createComparisonInfoButton = new Button(createComparisonComposite, SWT.PUSH);
				createComparisonInfoButton.setImage(ImageFactory.getInstance().getImage(AppImage.INFO_16X16));
				createComparisonInfoButton.addSelectionListener(new SelectionAdapter(){
					@Override
					public void widgetSelected(SelectionEvent e) {
						MessageDialog.openInformation(getShell(), "Help", "When a report is generated, the application can automatically create a comparison report comparing the HTML / CSV file of the previous export with the current one.\r\n\r\nMake sure that the report output path contains a date/timestamp so that previous HTML / CSV files are not overwritten.\r\n\r\nYou can always create a report comparison by selecting two report runs in the report history dialog.");
					}
				});
			}
			
			// link to filter sets (three check boxes, max)
			linkToFilterSetComposite = new Composite(composite, SWT.NONE);
			gridData = new GridData(GridData.FILL_HORIZONTAL);
			gridData.horizontalSpan = 2;
			linkToFilterSetComposite.setLayoutData(gridData);
			linkToFilterSetComposite.setLayout(new GridLayout(4, false));
			Label linkToFilterSetLabel = new Label(linkToFilterSetComposite, SWT.NONE);
			linkToFilterSetLabel.setText("Link to filter sets:");
			gridData = new GridData();
			gridData.widthHint = labelWidth;
			gridData.horizontalIndent = -5;
			linkToFilterSetLabel.setLayoutData(gridData);
	
			linkToMemberFilterSetButton = new Button(linkToFilterSetComposite, SWT.CHECK);
			linkToMemberFilterSetButton.setImage(ImageFactory.getInstance().getImage(AppImage.SELECTION_16X16));
			linkToMemberFilterSetButton.setLayoutData(new GridData());
			linkToMemberFilterSetButton.addSelectionListener(new SelectionAdapter() {
				@Override
				public void widgetSelected(SelectionEvent e) {
					reportFilters.setLinkToMemberFilterSet(linkToMemberFilterSetButton.getSelection());
					filterTreeViewer.refresh();
					filterTreeViewer.expandAll();
				}
			});
		
			linkToPermissionFilterSetButton = new Button(linkToFilterSetComposite, SWT.CHECK);
			linkToPermissionFilterSetButton.setImage(ImageFactory.getInstance().getImage(AppImage.SELECTION_16X16));
			linkToPermissionFilterSetButton.setLayoutData(new GridData());
			linkToPermissionFilterSetButton.addSelectionListener(new SelectionAdapter() {
				@Override
				public void widgetSelected(SelectionEvent e) {
					reportFilters.setLinkToPermissionFilterSet(linkToPermissionFilterSetButton.getSelection());
					filterTreeViewer.refresh();
					filterTreeViewer.expandAll();
				}
			});
		
			linkToFolderFilterSetButton = new Button(linkToFilterSetComposite, SWT.CHECK);
			linkToFolderFilterSetButton.setImage(ImageFactory.getInstance().getImage(AppImage.SELECTION_16X16));
			linkToFolderFilterSetButton.setLayoutData(new GridData());
			linkToFolderFilterSetButton.addSelectionListener(new SelectionAdapter() {
				@Override
				public void widgetSelected(SelectionEvent e) {
					reportFilters.setLinkToFolderFilterSet(linkToFolderFilterSetButton.getSelection());
					filterTreeViewer.refresh();
					filterTreeViewer.expandAll();
				}
			});
			
			noFilterSetsLabel = new Label(linkToFilterSetComposite, SWT.NONE);
			noFilterSetsLabel.setText("No filter sets have been selected");
			noFilterSetsLabel.setLayoutData(new GridData());
		}
		
		// -----------------------------------------
		// TreeViewer with Filters and FilterSets
		// -----------------------------------------
		Label filtersLabel = new Label(composite, SWT.NONE);
		filtersLabel.setText("Filters for this report:");
		filtersLabel.setData(Constants.CSS_CLASS_NAME_KEY, "h2");
		gridData = new GridData(SWT.FILL, SWT.NONE, true, false, 1, 1);
		gridData.horizontalSpan = 2;
		gridData.verticalIndent = 10;
		gridData.heightHint = 20;
		filtersLabel.setLayoutData(gridData);
		
		filterTreeViewer = new TreeViewer(composite, SWT.BORDER | SWT.FULL_SELECTION);
		Tree tree = filterTreeViewer.getTree();
		tree.setLinesVisible(true);
		filterTreeViewer.setAutoExpandLevel(TreeViewer.ALL_LEVELS);
		gridData = new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1);
		gridData.horizontalSpan = 2;
		tree.setLayoutData(gridData);

		filterTreeViewer.setLabelProvider(new LabelProvider(){
			@Override
			public String getText(Object element) {
				return ((FilterRow) element).getLabel();
			}
			
			@Override
			public Image getImage(Object element) {
				return ImageFactory.getInstance().getImage(((FilterRow) element).getImage());
			}
		});
		
		filterTreeViewer.setContentProvider(new ReportFilterContentProvider());
				
		// load the existing (given) report
		if (!dialogMode.equals(ReportDialogMode.EXPORT)) {
			loadReport();
		} else {
			// export mode
			reportTypeCombo.select(0);
			fileTypeCombo.select(0);
			filterTreeViewer.setInput(reportFilters);
			filterTreeViewer.refresh();
			
			// enable / disable the radio buttons for expanding groups?
			reportTypeCombo.notifyListeners(SWT.Selection, new Event());
		}
	
		ApplicationUtil.setDefaultCursor();
		
		return composite;
	}
	
	@Override
	protected void createButtonsForButtonBar(Composite parent) {		
		if (dialogMode.equals(ReportDialogMode.CREATE)) {
			createButton(parent, IDialogConstants.OK_ID, "Create", true);
			
		} else if (dialogMode.equals(ReportDialogMode.SAVE)) {
			createButton(parent, IDialogConstants.OK_ID, "Save", true);
			
		} else if (dialogMode.equals(ReportDialogMode.EXPORT)) {
			createButton(parent, IDialogConstants.OK_ID, "Export", true);
		}
		createButton(parent, IDialogConstants.CANCEL_ID, "Cancel", false);
	}
	
	protected void buttonPressed(int buttonId) {
		setReturnCode(buttonId);
		
		if (buttonId == IDialogConstants.OK_ID) {
			
			// save report information in case an export is created
			if (dialogMode.equals(ReportDialogMode.EXPORT)) {
				selectedReport = new Report();
				getReportInfoFromUI(selectedReport);
				selectedReport.setMemberFilters(reportFilters.getMemberFilters());
				selectedReport.setPermissionFilters(reportFilters.getPermissionFilters());
				selectedReport.setFolderFilters(reportFilters.getFolderFilters());
				exportSelectedFile = selectedFileButton.getSelection();
				exportSubdirectories = exportSubdirectoriesButton.getSelection();
				
				if (selectedReport.getFilePath().isEmpty()) {
					SyncMessageDialog.showMessage(MessageDialog.ERROR, "No file path", "Please enter a file path");
					return;
				}
			} else {
				// save or create report in database, return if an error occurred
				if (!saveReport()) {
					return;
				}
			}
		}
				
		super.buttonPressed(buttonId);
		close();
	}
	
	/**
	 * Creates a new or saves an existing Report in the database including a copy of the Filter objects or a reference to the FilterSets
	 * @throws DBException 
	 */
	private boolean saveReport() {
		boolean createReport = dialogMode.equals(ReportDialogMode.CREATE); 
		Report report;
		
		if (createReport) {
			report = new Report();
		} else {
			report = selectedReport;
		}
		
		boolean nameChanged = !reportNameText.getText().equalsIgnoreCase(report.getName()); 
		
		//validate fields
		if (reportNameText.getText().length() == 0) {
			MessageDialog.openError(null, "Details not complete", "Please enter a report name.");
			return false;
		}
		
		try {
			//check if the name already exists
			if (createReport || nameChanged) {
				List<Report> reports = Report.getAll();
				for (Report nextReport : reports) {
					if (nextReport.getName().equalsIgnoreCase(reportNameText.getText())) {
						// report name already exists
						MessageDialog.openError(null, "Name already exists", "The report name " + reportNameText.getText() + " already exists in the list.");
						return false;
					}
				}
			}
			
			getReportInfoFromUI(report);
		
		
			report.save();
			
			// load current FilterSets that belong to the report (if we deal with an existing report)
			if (report.getId() != null) {
								
				// Delete current filter set references
				ReportFilterSet.delete("REPORT_ID = ?", new Object[]{report.getId()});
				
				// delete current filters attached to this report
				Filter.delete("REPORT_ID = ?", report.getId());
			}
			
			// save the related filters and filter sets
			if (reportFilters.getMemberFilterSet() != null && linkToMemberFilterSetButton.getSelection()) {
				ReportFilterSet reportFilter = new ReportFilterSet(report.getId(), reportFilters.getMemberFilterSet().getId());
				reportFilter.save();
				
			} else {
				for (Filter filter : reportFilters.getMemberFilters()) {
					// create a copy of each filter and connect it to this report
					Filter filterClone = filter.clone();
					filterClone.setReportId(report.getId());
					filterClone.save();
				}
			}
			
			// permission filters
			if (reportFilters.getPermissionFilterSet() != null && linkToPermissionFilterSetButton.getSelection()) {
				ReportFilterSet reportFilter = new ReportFilterSet(report.getId(), reportFilters.getPermissionFilterSet().getId());
				reportFilter.save();
				
			} else {
				for (Filter filter : reportFilters.getPermissionFilters()) {
					// create a copy of each filter and connect it to this report
					Filter filterClone = filter.clone();
					filterClone.setReportId(report.getId());
					filterClone.save();
				}
			}
			
			// folder filters
			if (reportFilters.getFolderFilterSet() != null && linkToFolderFilterSetButton.getSelection()) {
				ReportFilterSet reportFilter = new ReportFilterSet(report.getId(), reportFilters.getFolderFilterSet().getId());
				reportFilter.save();
				
			} else {
				for (Filter filter : reportFilters.getFolderFilters()) {
					// create a copy of each filter and connect it to this report
					Filter filterClone = filter.clone();
					filterClone.setReportId(report.getId());
					filterClone.save();
				}
			}
			
			selectedReport = report;
		} catch (DBException ex) {
			MessageDialog.openError(null, "Database error", "Error while saving the report: " + ex.getMessage());
			return false;
		}
		
		return true;
	}
	
	/**
	 * Extracts information from the UI fields and updates the given {@link LdapConnectionInfo} object.
	 * 
	 */
	private void getReportInfoFromUI(Report report) throws NumberFormatException {
		report.setName(reportNameText != null ? reportNameText.getText() : "");
		report.setDescription(reportDescriptionText != null ? reportDescriptionText.getText() : "");
		
		switch (reportTypeCombo.getSelectionIndex()) {
			case 0:
				report.setReportType(ReportType.EFFECTIVE);
				break;
			case 1:
				report.setReportType(ReportType.FOLDERS_AND_USERS_EFFECTIVE);
				break;
			case 2:
				if (expandGroupsDirectMembersButton.getSelection()) {
					report.setReportType(ReportType.ACL_DIRECT_GROUP_MEMBERS);
				} else if (expandGroupsNestedMembersButton.getSelection()) {
					report.setReportType(ReportType.ACL_NESTED_GROUP_MEMBERS);
				} else {
					report.setReportType(ReportType.ACL);
				}
				break;
			case 3:
				report.setReportType(ReportType.ALL_MEMBERS);
				break;
			case 4:
				if (expandGroupsDirectMembersButton.getSelection()) {
					report.setReportType(ReportType.MATCHING_GROUPS_AND_DIRECT_MEMBERS);
				} else {
					report.setReportType(ReportType.MATCHING_GROUPS_AND_NESTED_MEMBERS);
				}
				break;
			case 5:
				if (expandGroupsDirectMembersButton.getSelection()) {
					report.setReportType(ReportType.GROUPS_WITH_PERMISSIONS_AND_DIRECT_MEMBERS);
				} else {
					report.setReportType(ReportType.GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS);
				}
				break;
			case 6:
				if (expandGroupsDirectMembersButton.getSelection()) {
					report.setReportType(ReportType.MATCHING_USERS_AND_DIRECT_GROUP_MEMBERSHIPS);
				} else {
					report.setReportType(ReportType.MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS);
				}
				break;
		}
				
		for (FileType fileType : FileType.values()) {
			if (Report.fileTypeToString(fileType).equals(fileTypeCombo.getItem(fileTypeCombo.getSelectionIndex()))) {
				report.setFileType(fileType);
			}
		}
		
		report.setUseSimplePresentation(useSimplePresentationCheckbox.getSelection());
		
		// check if the target path has an extension
		if (report.getFileType().equals(FileType.HTML) || report.getFileType().equals(FileType.HTML_JQUERY)) {
			if (!filePathText.getText().isEmpty() && !filePathText.getText().toLowerCase().endsWith(".htm") && !filePathText.getText().toLowerCase().endsWith(".html")) {
				filePathText.setText(filePathText.getText() + ".html");
			}
		} else {
			if (!filePathText.getText().isEmpty() && !filePathText.getText().toLowerCase().endsWith(".csv")) {
				filePathText.setText(filePathText.getText() + ".csv");
			}
		}
		report.setFilePath(filePathText.getText());
		report.setTemplatePath(templatePathText.getText());
		report.setEmailAddress(emailAddressText.getText());
		
		if (!isPolicy && !dialogMode.equals(ReportDialogMode.EXPORT)) {
			report.setCreateComparisonReport(createComparisonButton.getSelection());
		}
		
		if (isPolicy) {
			report.setEmailThreshold(emailThresholdSpinner.getSelection());
		}
	}
	
	/**
	 * Loads the {@link #selectedReport} into the GUI fields (Report properties and the TreeViewer with Filters / FilterSets that belong to the selected report).
	 * 
	 * @param useFiltersFromFilterParts this parameter determines what filter to use when a FilterSet reference is unchecked in this dialog. In that case a set of filters is added to the
	 * report instead of using a reference to a FilterSet. When the parameter is true (used for dialog mode SAVE), the Dialog will use the current filters selected in the GUI using
	 * {@link FilterHelper#getMemberFilterList()}, {@link FilterHelper#getPermissionFilterList()} and {@link FilterHelper#getFolderFilterList()}. If false then the Dialog will use
	 * the filters from the selected FilterSets
	 */
	private void loadReport() {
		reportNameText.setText(selectedReport.getName());
		reportDescriptionText.setText(selectedReport.getDescription());
		
		reportTypeCombo.select(0);
		ReportType reportType = selectedReport.getReportType();
		switch(reportType) {
			case EFFECTIVE: reportTypeCombo.select(0); break;
			case FOLDERS_AND_USERS_EFFECTIVE: reportTypeCombo.select(1); break;
			case ACL: case ACL_DIRECT_GROUP_MEMBERS: case ACL_NESTED_GROUP_MEMBERS: reportTypeCombo.select(2); break;
			case ALL_MEMBERS: reportTypeCombo.select(3); break;
			case MATCHING_GROUPS_AND_DIRECT_MEMBERS: case MATCHING_GROUPS_AND_NESTED_MEMBERS: reportTypeCombo.select(4); break;
			case GROUPS_WITH_PERMISSIONS_AND_DIRECT_MEMBERS: case GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS: reportTypeCombo.select(5); break;
			case MATCHING_USERS_AND_DIRECT_GROUP_MEMBERSHIPS: case MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS: reportTypeCombo.select(6); break;
		}
		
		dontExpandGroupsButton.setSelection(reportType.equals(ReportType.ACL));
		expandGroupsDirectMembersButton.setSelection(reportType.equals(ReportType.ACL_DIRECT_GROUP_MEMBERS) || reportType.equals(ReportType.GROUPS_WITH_PERMISSIONS_AND_DIRECT_MEMBERS) 
				|| reportType.equals(ReportType.MATCHING_GROUPS_AND_DIRECT_MEMBERS) || reportType.equals(ReportType.MATCHING_USERS_AND_DIRECT_GROUP_MEMBERSHIPS));
		expandGroupsNestedMembersButton.setSelection(reportType.equals(ReportType.ACL_NESTED_GROUP_MEMBERS) || reportType.equals(ReportType.GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS) 
				|| reportType.equals(ReportType.MATCHING_GROUPS_AND_NESTED_MEMBERS) || reportType.equals(ReportType.MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS));
		
		// enable / disable the radio buttons for expanding groups?
		reportTypeCombo.notifyListeners(SWT.Selection, new Event());
		
		fileTypeCombo.select(0);
		for (int counter = 0; counter < fileTypeCombo.getItemCount(); counter++) {
			if (fileTypeCombo.getItem(counter).equals(Report.fileTypeToString(selectedReport.getFileType()))) {
				fileTypeCombo.select(counter);
				break;
			}
		}
		
		useSimplePresentationCheckbox.setSelection(selectedReport.isUseSimplePresentation());
		filePathText.setText(selectedReport.getFilePath());
		templatePathText.setText(selectedReport.getTemplatePath() != null ? selectedReport.getTemplatePath() : "");
		emailAddressText.setText(selectedReport.getEmailAddress() != null ? selectedReport.getEmailAddress() : "");
		
		if (!isPolicy) {
			createComparisonButton.setSelection(selectedReport.isCreateComparisonReport());
		}
		
		if (isPolicy) {
			emailThresholdSpinner.setSelection(selectedReport.getEmailThreshold());
		}
			
		// Label saying No Filterset selected?
		if (reportFilters.getMemberFilterSet() == null && reportFilters.getPermissionFilterSet() == null && reportFilters.getFolderFilterSet() == null) {
			((GridData) noFilterSetsLabel.getLayoutData()).exclude = false;
			noFilterSetsLabel.setVisible(true);
		} else {
			((GridData) noFilterSetsLabel.getLayoutData()).exclude = true;
			noFilterSetsLabel.setVisible(false);
		}
		
		// checkbox for member FilterSet?
		if (reportFilters.getMemberFilterSet() != null) {
			linkToMemberFilterSetButton.setText(reportFilters.getMemberFilterSet().getName());
			linkToMemberFilterSetButton.pack(true);
			linkToMemberFilterSetButton.setVisible(true);
			linkToMemberFilterSetButton.setSelection(true);
			reportFilters.setLinkToMemberFilterSet(true);
			((GridData)linkToMemberFilterSetButton.getLayoutData()).exclude = false;
		} else {
			((GridData)linkToMemberFilterSetButton.getLayoutData()).exclude = true;
			linkToMemberFilterSetButton.setSelection(false);
			linkToMemberFilterSetButton.setVisible(false);
			reportFilters.setLinkToMemberFilterSet(false);
		}
		
		// checkbox for permission FilterSet?
		if (reportFilters.getPermissionFilterSet() != null) {
			linkToPermissionFilterSetButton.setText(reportFilters.getPermissionFilterSet().getName());
			linkToPermissionFilterSetButton.pack(true);
			linkToPermissionFilterSetButton.setSelection(true);
			linkToPermissionFilterSetButton.setVisible(true);
			reportFilters.setLinkToPermissionFilterSet(true);
			((GridData)linkToPermissionFilterSetButton.getLayoutData()).exclude = false;
		} else {
			((GridData)linkToPermissionFilterSetButton.getLayoutData()).exclude = true;
			linkToPermissionFilterSetButton.setSelection(false);
			linkToPermissionFilterSetButton.setVisible(false);
			reportFilters.setLinkToPermissionFilterSet(false);
		}
		
		// checkbox for folder FilterSet?
		if (reportFilters.getFolderFilterSet() != null) {
			linkToFolderFilterSetButton.setText(reportFilters.getFolderFilterSet().getName());
			linkToFolderFilterSetButton.pack(true);
			linkToFolderFilterSetButton.setSelection(true);
			linkToFolderFilterSetButton.setVisible(true);
			reportFilters.setLinkToFolderFilterSet(true);
			((GridData)linkToFolderFilterSetButton.getLayoutData()).exclude = false;
		} else {
			((GridData)linkToFolderFilterSetButton.getLayoutData()).exclude = true;
			linkToFolderFilterSetButton.setSelection(false);
			linkToFolderFilterSetButton.setVisible(false);
			reportFilters.setLinkToFolderFilterSet(false);
		}
		
		linkToFilterSetComposite.layout(true);
		filterTreeViewer.setInput(reportFilters);
		filterTreeViewer.refresh();
	}

	public Report getSelectedReport() {
		return selectedReport;
	}

	public void setSelectedReport(Report selectedMember) {
		this.selectedReport = selectedMember;
	}

	public ReportFilters getReportFilters() {
		return reportFilters;
	}

	public void setReportFilters(ReportFilters reportFilters) {
		this.reportFilters = reportFilters;
	}

	public boolean isExportSelectedFile() {
		return exportSelectedFile;
	}

	public void setExportSelectedFile(boolean exportSelectedFile) {
		this.exportSelectedFile = exportSelectedFile;
	}


	public boolean isExportSubdirectories() {
		return exportSubdirectories;
	}
}
