package com.perdemia.permissionanalyzer.application.jface.dialogs;

import gnu.trove.list.array.TIntArrayList;

import java.lang.reflect.InvocationTargetException;
import java.sql.PreparedStatement;
import java.util.Iterator;
import java.util.List;

import javax.inject.Inject;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.e4.ui.services.EMenuService;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.ColumnViewerToolTipSupport;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.TreeViewerColumn;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.contentproviders.MainFileTreeContentProvider;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.PathColumnLabelProvider;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.PermissionLabelProvider;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.SpecialPermissionLabelProvider;
import com.perdemia.permissionanalyzer.application.swt.TreeViewerWithMonitor;
import com.perdemia.permissionanalyzer.application.util.Constants;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.ImageFactory;
import com.perdemia.permissionanalyzer.application.util.ImageFactory.AppImage;
import com.perdemia.permissionanalyzer.application.util.PermissionUtil;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Filter;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SpecialPermission;

public class ReportPreviewDialog extends DialogWithTopBar {
	private static Logger logger = LoggerFactory.getLogger(ReportPreviewDialog.class.getName());
	private List<FileInfo> rootFiles = null;
	private Integer itemCount;
	private FilterHelper filterHelper = new FilterHelper();
	
	/** UI controls */
	private Label countLabel;
	private TreeViewer treeViewer;
	
	/** Dialog settings and values */
	private List<Filter> memberFilterList;
	private List<Filter> permissionFilterList;
	private List<Filter> fileFilterList;
	private boolean isPolicy;
	
	public ReportPreviewDialog(Shell shell, List<Filter> memberFilterList, List<Filter> permissionFilterList, List<Filter> fileFilterList, boolean isPolicy) {
		super(shell);
		this.memberFilterList = memberFilterList;
		this.permissionFilterList = permissionFilterList;
		this.fileFilterList = fileFilterList;
		this.isPolicy = isPolicy;
	}
	
	@Override
	protected boolean isResizable() {
		return true;
	}
	
	@Override
	protected void configureShell(Shell shell) {
		super.configureShell(shell);
		if (isPolicy) {
			shell.setText("Policy preview");
			shell.setImage(ImageFactory.getInstance().getImage(AppImage.POLICY_16X16));
		} else {
			shell.setText("Report preview");
			shell.setImage(ImageFactory.getInstance().getImage(AppImage.REPORT_16X16));
		}
		shell.setSize(1000, 700);
	}

	@Override
	protected Control createDialogArea(Composite container) {
		Composite composite;
		
		if (isPolicy) {
			composite = (Composite) super.createDialogArea(container, "Policy preview", ImageFactory.getInstance().getImage(AppImage.POLICY_48X48));
		} else {
			composite = (Composite) super.createDialogArea(container, "Report preview", ImageFactory.getInstance().getImage(AppImage.REPORT_48X48));
		}
		container.setBackgroundMode(SWT.INHERIT_DEFAULT);
						
		// set width, height and layout
		GridLayout gridLayout = new GridLayout(1, false);
		gridLayout.marginTop = 20;
		gridLayout.marginLeft = 7;
		gridLayout.marginRight = 7;
		composite.setLayout(gridLayout);
		
		countLabel = new Label(composite, SWT.NONE);
		countLabel.setData(Constants.CSS_CLASS_NAME_KEY, "h2");
		GridData gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.heightHint = 30;
		countLabel.setLayoutData(gridData);
		
		treeViewer = new TreeViewerWithMonitor(composite, SWT.VIRTUAL | SWT.H_SCROLL | SWT.V_SCROLL | SWT.BORDER | SWT.FULL_SELECTION);
		treeViewer.getTree().setLinesVisible(true);
		treeViewer.getTree().setHeaderVisible(true);
		treeViewer.setUseHashlookup(true);
		treeViewer.setContentProvider(new MainFileTreeContentProvider(treeViewer, filterHelper));
		
		//layout data for the tree
		gridData = new GridData(SWT.FILL, SWT.FILL, true, true);
		treeViewer.getTree().setLayoutData(gridData);
		
		// Path column
		TreeViewerColumn column = new TreeViewerColumn(treeViewer, SWT.NONE);
		column.getColumn().setText("Path");
		column.getColumn().setWidth(300);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new PathColumnLabelProvider(false));
				
		// Permission column
		column = new TreeViewerColumn(treeViewer, SWT.NONE);
		column.getColumn().setText("Permission");
		column.getColumn().setWidth(200);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new PermissionLabelProvider());
				
		// Special permission columns
		for (final Integer permission : PermissionUtil.getCustomPermissionOrderList()) {
			//skip SYNCHRONIZE
			if (permission.intValue() != SpecialPermission.SYNCHRONIZE.ordinal()) {
				column = new TreeViewerColumn(treeViewer, SWT.NONE);
				column.getColumn().setWidth(30);
				column.getColumn().setResizable(false);
				column.getColumn().setToolTipText(PermissionUtil.specialPermissionToString(permission, false, null));
				column.getColumn().setImage(PermissionUtil.getSpecialPermissionImage(permission, false, true));
				column.setLabelProvider(new SpecialPermissionLabelProvider(permission, false, true));
			}
		}
		
		//enable tooltips for this tree viewer
		ColumnViewerToolTipSupport.enableFor(treeViewer);
			
		loadTreeViewer();
		
		return composite;
	}

	@Override
	protected void createButtonsForButtonBar(Composite parent) {
		createButton(parent, IDialogConstants.CLOSE_ID, "Close", false);
	}
	
	@Override
	protected void buttonPressed(int buttonId) {
		close();
	}
	
	private void loadTreeViewer() {
		filterHelper.setMemberFilterList(memberFilterList);
		filterHelper.setPermissionFilterList(permissionFilterList);
		filterHelper.setFolderFilterList(fileFilterList);
		
		try {
			new ProgressMonitorDialog(getShell()).run(true, false, new IRunnableWithProgress(){
				@Override
				public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
					try {							
						monitor.beginTask("Applying filter...", IProgressMonitor.UNKNOWN);
									
						// load all the applicable file id's based on the selected filter
						filterHelper.applyNewFilter(false);
						PreparedStatement ps = filterHelper.buildFilterQuery(FilterHelper.ObjectType.FILE, true, false, 0, null, null);
						
						if (ps != null) {
							TIntArrayList uniqueFileIds = filterHelper.getApplicableFileIds(ps, true, true);
							itemCount = uniqueFileIds.size();
							filterHelper.setApplicableFileIdsIncludingParents(uniqueFileIds);
							
							// load the root files
							rootFiles = FileInfo.where("PARENT_ID IS NULL", null);
							
							// check which root files are applicable
							Iterator<FileInfo> iter = rootFiles.iterator();
							while (iter.hasNext()) {
								if (!filterHelper.getApplicableFileIdsIncludingParents().contains(iter.next().getId())) {
									iter.remove();
								}
							}
						}
						
						filterHelper.setItemCount(itemCount);
						
						// database actions have been performed, now update the GUI
						Display.getDefault().asyncExec(new Runnable() {
							
							public void run() {
								try {
									treeViewer.getTree().setItemCount(rootFiles.size());
									treeViewer.setInput(rootFiles);											
									treeViewer.refresh();
									
									// select the first item in the tree or table so that the TracePart and AclPart get refreshed through the SelectionService
									if (itemCount == 0) {
										MessageDialog.openInformation(null, "No data found", "No files or folders found.");
										close();
									} else {
										countLabel.setText("Found " + itemCount + " folders/files:");
										countLabel.pack();
									}
								} catch (Exception ex) {
									logger.error("Error while applying filter", ex);
									SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while applying filter: " + ex.getMessage());
								}									
							}
						});
						
						
					} catch (Exception ex) {
						logger.error("Error while applying filter", ex);
						SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while applying filter: " + ex.getMessage());
					}
				}
			});
		} catch (InvocationTargetException | InterruptedException ex) {
			logger.error("Error while applying filter", ex);
			SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while applying filter: " + ex.getMessage());
		}
	}
}