package com.perdemia.permissionanalyzer.application.jface.dialogs;

import java.util.List;

import com.perdemia.permissionanalyzer.application.util.ImageFactory;
import com.perdemia.permissionanalyzer.application.util.ImageFactory.AppImage;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Filter;
import com.perdemia.permissionanalyzer.data.Filter.FilterType;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ILazyTreeContentProvider;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TreePath;
import org.eclipse.jface.viewers.TreeSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Shell;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class SearchFolderDialog extends DialogWithTopBar {
	private static Logger logger = LoggerFactory.getLogger(SearchFolderDialog.class.getName());
	
	/** ID's of the custom buttons */
	public final static int ADD_TO_INCLUDE_ID = 3;
	public final static int ADD_TO_EXCLUDE_ID = 4;
	
	/** UI controls */
	private TreeViewer folderTreeViewer;
	
	/** Reference to the Filter ListViewer in the FilterFolderPart */
	private TableViewer filterListViewer;
	
	/** Dialog settings and values */
	private FileInfo selectedFolder;
	private String selectedPath = "";
	private boolean useInclude;
	private boolean useExclude;
	
	public SearchFolderDialog(Shell shell) {
		super(shell);
	}
	
	public SearchFolderDialog(Shell shell, TableViewer filterListViewer, boolean useInclude, boolean useExclude) {
		super(shell);
		this.filterListViewer = filterListViewer;
		this.useInclude = useInclude;
		this.useExclude = useExclude;
	}

	@Override
	protected boolean isResizable() {
		return true;
	}
	
	@Override
	protected void configureShell(Shell shell) {
		super.configureShell(shell);
		if (filterListViewer != null) {
			shell.setText("Include or exclude folders");
		} else {
			shell.setText("Select a folder");
		}
		shell.setImage(ImageFactory.getInstance().getImage(AppImage.SEARCH_16X16));
		shell.setSize(850, 700);
	}

	@Override
	protected Control createDialogArea(Composite container) {
		Composite composite = (Composite) super.createDialogArea(container, "Include or exclude folders", ImageFactory.getInstance().getImage(AppImage.FOLDER_48X48));
		container.setBackgroundMode(SWT.INHERIT_DEFAULT);
						
		// set width, height and layout
		GridLayout gridLayout = new GridLayout(1, false);
		gridLayout.marginTop = 7;
		gridLayout.marginLeft = 7;
		gridLayout.marginRight = 7;
		composite.setLayout(gridLayout);
				
		// FileInfo TreeViewer
		int flags = SWT.BORDER | SWT.V_SCROLL | SWT.VIRTUAL;
		if (filterListViewer != null) {
			flags |= SWT.MULTI;
		}
		folderTreeViewer = new TreeViewer(composite, flags);
		folderTreeViewer.setUseHashlookup(true);
		
		GridData gridData = new GridData(GridData.FILL_BOTH);
		folderTreeViewer.getTree().setLayoutData(gridData);
		folderTreeViewer.setContentProvider(new FolderContentProvider(folderTreeViewer));
		folderTreeViewer.setLabelProvider(new FolderLabelProvider());
		try {
			folderTreeViewer.setInput(FileInfo.where("parent_id is null", null));
		} catch (DBException e) {
			MessageDialog.openError(null, "Error", "Error while loading the list of root folders: " + e.getMessage());
			logger.error("Error while intializing the SearchFolderDialog with root folders.", e);
		}
		
		return composite;
	}

	@Override
	protected void createButtonsForButtonBar(Composite parent) {
		GridLayout gridLayout = (GridLayout) parent.getLayout();
		gridLayout.makeColumnsEqualWidth = false;
		
		if (filterListViewer != null) {
			// this dialog is being used to add Files to the include/exclude filter lists
			if (useInclude) {
				Button includeButton = createButton(parent, ADD_TO_INCLUDE_ID, "Include selected folders", false);
				includeButton.setImage(ImageFactory.getInstance().getImage(AppImage.INCLUDE_FOLDER_16X16));
				((GridData) includeButton.getLayoutData()).widthHint = 240;
				includeButton.pack(true);
			}
			
			if (useExclude) {
				Button excludeButton = createButton(parent, ADD_TO_EXCLUDE_ID, "Exclude selected folders", false);
				excludeButton.setImage(ImageFactory.getInstance().getImage(AppImage.EXCLUDE_FOLDER_16X16));
				((GridData) excludeButton.getLayoutData()).widthHint = 240;
				excludeButton.pack(true);
			}
		} else {
			// this dialog is used to browse a single FileInfo
			createButton(parent, IDialogConstants.OK_ID, "OK", false);
		}
		
		createButton(parent, IDialogConstants.CLOSE_ID, "Close", false);
	}

	protected void buttonPressed(int buttonId) {
		TreeSelection selection = (TreeSelection) folderTreeViewer.getSelection();
		
		if (buttonId == ADD_TO_INCLUDE_ID || buttonId == ADD_TO_EXCLUDE_ID) {

			// loop through all selected items
			for (TreePath treePath : selection.getPaths()) {
				
				// determine the complete file path
				StringBuilder path = new StringBuilder();
				for (int counter = 0; counter < treePath.getSegmentCount(); counter++) {
					if (path.length() > 0) {
						path.append("\\");
					}
					path.append(((FileInfo)treePath.getSegment(counter)).getName());
				}
				
				selectedPath = path.toString();
				
				FileInfo selectedFile = (FileInfo)treePath.getSegment(treePath.getSegmentCount() - 1);
				String filterValue = selectedFile.getParentFileIds();
				Filter filter = new Filter(FilterType.FOLDER_PATH, buttonId == ADD_TO_INCLUDE_ID, filterValue, path.toString());

				// does this filter already exists?
				boolean exists = false;
				List<Filter> currentFilters = (java.util.List<Filter>) filterListViewer.getInput();
				for (Filter curFilter : currentFilters) {
					if (curFilter.getType().equals(FilterType.FOLDER_PATH) && (curFilter.getValue().equals(filterValue))) {
						MessageDialog.openError(null, "Already exists", "The folder \"" + path.toString() + "\" already exists in the list of filters.");
						exists = true;
						break;
					}
				}

				if (!exists) {
					currentFilters.add(filter);
					MessageDialog.openInformation(null, "Folder added", "The folder has been added as new filter.");
				}

				setReturnCode(buttonId);
			}
			filterListViewer.refresh();

		} else if (buttonId == IDialogConstants.OK_ID) {

			// ok button
			selectedFolder = (FileInfo)selection.getFirstElement();
			
			// determine the complete file path
			if (selection.getPaths().length > 0) {
				StringBuilder path = new StringBuilder();
				for (int counter = 0; counter < selection.getPaths()[0].getSegmentCount(); counter++) {
					if (path.length() > 0) {
						path.append("\\");
					}
					path.append(((FileInfo)selection.getPaths()[0].getSegment(counter)).getName());
				}
				selectedPath = path.toString();
			}			
			
			super.buttonPressed(buttonId);

		} else if (buttonId == IDialogConstants.CLOSE_ID) {
			// don't set a return code, because it may already have been set by the include/exclude button
			close();
		} else {
			super.buttonPressed(buttonId);
		}
	}

	public FileInfo getSelectedFolder() {
		return selectedFolder;
	}
	
	public String getSelectedPath() {
		return selectedPath;
	}
}

class FolderContentProvider implements ILazyTreeContentProvider {
	private static Logger logger = LoggerFactory.getLogger(FolderContentProvider.class.getName());
	private TreeViewer viewer;
	private List<FileInfo> rootElements;

	public FolderContentProvider(TreeViewer viewer) {
		this.viewer = viewer;
	}

	@Override
	public void dispose() {}

	@Override
	public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		rootElements = (List<FileInfo>) newInput;
	}

	@Override
	public Object getParent(Object element) {
		FileInfo file = (FileInfo) element;
		if (file.getParent() != null) {
			return file.getParent();
		} else {
			return rootElements;
		}
	}

	@Override
	public void updateElement(Object parent, int index) {
		FileInfo element;
		if (parent instanceof FileInfo) {
			element = (FileInfo) ((FileInfo) parent).getChildren().get(index);
		} else {
			element = rootElements.get(index);
		}

		if (logger.isTraceEnabled()) {
			logger.trace("Replacing tree element with index " + index + ": " + element);
		}
		viewer.replace(parent, index, element);
		viewer.setHasChildren(element, true);
		updateChildCount(element, -1);
	}

	@Override
	public void updateChildCount(Object element, int currentChildCount) {
		int length = 0;
		if (element instanceof FileInfo) {
			FileInfo file = (FileInfo) element;
			// load the children of this FileInfo element
			try {
				file.setChildren(FileInfo.where("parent_id = ?", new Object[]{file.getId()}));
			} catch (DBException e) {
				MessageDialog.openError(null, "Database error", "Error while getting child files for file " + file.getName() + ".\r\n\r\n" + e.getMessage());
				logger.error("Error while getting child files for file " + file.getName(), e);
			}
			length = (file.getChildren() != null ? file.getChildren().size() : 0);

		} else if (element == rootElements) {
			length = rootElements.size();

		} else {
			logger.warn("Element is not of type FileInfo or rootElements in FileInfo.updateChildCount: " + element.getClass());
		}

		if (logger.isDebugEnabled()) {
			logger.debug("Setting child count " + length + " for element " + element.toString());
		}
		viewer.setChildCount(element, length);
	}
}

class FolderLabelProvider extends LabelProvider {
	@Override
	public Image getImage(Object element) {
		return ImageFactory.getInstance().getImage(AppImage.FOLDER_16X16);
	}

	@Override
	public String getText(Object element) {
		return ((FileInfo) element).getName();
	}
}