package com.perdemia.permissionanalyzer.application.jface.dialogs;

import java.sql.ResultSet;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;

import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ComboViewer;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CLabel;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.TraverseEvent;
import org.eclipse.swt.events.TraverseListener;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.labelproviders.TableLabelProvider;
import com.perdemia.permissionanalyzer.application.parts.FilterMemberPart;
import com.perdemia.permissionanalyzer.application.swt.DelayedKeyListenerText;
import com.perdemia.permissionanalyzer.application.util.ApplicationUtil;
import com.perdemia.permissionanalyzer.application.util.ImageFactory;
import com.perdemia.permissionanalyzer.application.util.ImageFactory.AppImage;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.ScanLogItem;

/** This dialog is used to search a member in the MemberFilterComposite, or to select a member in the TracePart or to display the matching members of a wild card filter
 * in the MemberFilterComposite. In the last two cases this dialog won't display Include and Exclude buttons.
 */
public class ViewScanLogDialog extends DialogWithTopBar {
	private static Logger logger = LoggerFactory.getLogger(ViewScanLogDialog.class.getName());
	
	/** Filter data */
	private java.util.List<ScanLogItem> searchResultList = new ArrayList<ScanLogItem>();
	
	/** UI controls */
	private Text searchText;
	private TableViewer searchResultTableViewer;
	private ComboViewer hostnameCombo;
	private Button showInfoLevelCheckbox;
	private Button showWarningLevelCheckbox;
	private Button showErrorLevelCheckbox;
	private Button hideAccessDeniedCheckbox;
	private Button hideUnresolvableSidsCheckbox;
	
	/** Date format in the search result list, for each scan log record */
	private DateFormat dateFormatter = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.MEDIUM, Locale.getDefault());
	
	/**
	 * Constructor used in the {@link FilterMemberPart}. This constructor contains a filterListViewer and filterType which are used
	 * to add Members to a filter list as Include or Exclude items.  
	 * 
	 * @param shell
	 * @param filterListViewer
	 * @param filterType
	 * @param onlyShowGroups
	 */
	public ViewScanLogDialog(Shell shell){
		super(shell);
				
		setShellStyle(SWT.DIALOG_TRIM | SWT.RESIZE | SWT.MAX | SWT.APPLICATION_MODAL);		
	}
		
	@Override
	protected boolean isResizable()  {
	  return true;
	}
	
	@Override
	protected void configureShell(Shell shell) {
		super.configureShell(shell);
		shell.setImage(ImageFactory.getInstance().getImage(AppImage.LOG_FILE_16X16));
		shell.setSize(1200, 900);
	}
	
	@Override
	protected Control createDialogArea(Composite container) {		
		ApplicationUtil.setBusyCursor();
		
		Composite composite = (Composite) super.createDialogArea(container, "Remote scan log files", ImageFactory.getInstance().getImage(AppImage.LOG_FILE_48X48));
		container.setBackgroundMode(SWT.INHERIT_DEFAULT);
		container.getShell().setText("Remote scan log files");
				
		// set width, height and layout
		GridLayout gridLayout = new GridLayout(3, false);
		gridLayout.marginTop = 7;
		gridLayout.marginLeft = 7;
		gridLayout.marginRight = 7;
		composite.setLayout(gridLayout);			
		
		// Search label
		CLabel searchLabel = new CLabel(composite, SWT.NONE);
		searchLabel.setText("Search the log records (use * as wildcard):");
		GridData gridData = new GridData();
		gridData.verticalAlignment = SWT.BOTTOM;
		gridData.horizontalAlignment = SWT.LEFT;
		gridData.horizontalSpan = 3;
		searchLabel.setLayoutData(gridData);
					
		// Search textbox
		searchText = DelayedKeyListenerText.create(composite, SWT.BORDER, new KeyAdapter() {
			@Override
			public void keyReleased(KeyEvent e) {
				if (e.keyCode == SWT.ARROW_DOWN) {
					searchResultTableViewer.getTable().setFocus();
				} else {
					searchLogRecords(false);
				}					
			}			
		});
		searchText.setMessage("search");
		gridData = new GridData(GridData.FILL_HORIZONTAL);
		searchText.setLayoutData(gridData);
		searchText.addTraverseListener(new TraverseListener() {
			@Override
			public void keyTraversed(TraverseEvent e) {
				if (e.detail == SWT.TRAVERSE_RETURN) {
					searchLogRecords(true);
				}
			}
		});
		
		// server dropdown
		hostnameCombo = new ComboViewer(composite, SWT.BORDER | SWT.READ_ONLY);
		hostnameCombo.setContentProvider(ArrayContentProvider.getInstance());
		hostnameCombo.setLabelProvider(new LabelProvider());
		gridData = new GridData();
		hostnameCombo.getCombo().setLayoutData(gridData);
		
		List<String> servers = new ArrayList<String>();
		servers.add("All servers");
		ResultSet rs = null;
		try {
			rs = DBObject.executeQuery("SELECT DISTINCT SERVER FROM PERMISSIONANALYZER.SCANLOGITEM", new Object[0]);
			
			while (rs.next()) {
				servers.add(rs.getString(1));
			}
		} catch (Exception e) {
			logger.error("Error while retrieving the list of servers for the scan log dialog", e);
		}
		finally {
			DBObject.closeResultSet(rs);
		}
		
		hostnameCombo.setInput(servers);		
		hostnameCombo.setSelection(new StructuredSelection(hostnameCombo.getElementAt(0)), true);
		hostnameCombo.addSelectionChangedListener(new ISelectionChangedListener() {
			@Override
			public void selectionChanged(SelectionChangedEvent event) {
				searchLogRecords(true);						
			}					
		});
		
		
		// Search button
		Button searchMembersButton = new Button(composite, SWT.PUSH);
		searchMembersButton.setImage(ImageFactory.getInstance().getImage(AppImage.FILTER_20X20));
		searchMembersButton.setToolTipText("Search records");
		searchMembersButton.setLayoutData(new GridData());
		searchMembersButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				searchLogRecords(true);
			}
		});
		
		// filter checkboxes
		Composite filterComposite = new Composite(composite, SWT.NONE);
		filterComposite.setLayout(new GridLayout(5, false));
		gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.horizontalSpan = 2;
		filterComposite.setLayoutData(gridData);
		
		showInfoLevelCheckbox = new Button(filterComposite, SWT.CHECK);
		showInfoLevelCheckbox.setText("Show info");
		showInfoLevelCheckbox.setSelection(true);
		showInfoLevelCheckbox.setImage(ImageFactory.getInstance().getImage(AppImage.STATUS_INFO_16X16));
		showInfoLevelCheckbox.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				searchLogRecords(true);
			}
		});
		
		showWarningLevelCheckbox = new Button(filterComposite, SWT.CHECK);
		showWarningLevelCheckbox.setText("Show warning");
		showWarningLevelCheckbox.setSelection(true);
		showWarningLevelCheckbox.setImage(ImageFactory.getInstance().getImage(AppImage.STATUS_WARNING_16X16));
		showWarningLevelCheckbox.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				searchLogRecords(true);
			}
		});
		
		showErrorLevelCheckbox = new Button(filterComposite, SWT.CHECK);
		showErrorLevelCheckbox.setText("Show error");
		showErrorLevelCheckbox.setSelection(true);
		showErrorLevelCheckbox.setImage(ImageFactory.getInstance().getImage(AppImage.STATUS_ERROR_16X16));
		showErrorLevelCheckbox.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				searchLogRecords(true);
			}
		});
		
		// exclude log records checkboxes		
		hideAccessDeniedCheckbox = new Button(filterComposite, SWT.CHECK);
		hideAccessDeniedCheckbox.setText("Hide access denied errors");
		hideAccessDeniedCheckbox.setImage(ImageFactory.getInstance().getImage(AppImage.EXCLUDE_FILTER_20X20));
		hideAccessDeniedCheckbox.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				searchLogRecords(true);
			}
		});
		
		hideUnresolvableSidsCheckbox = new Button(filterComposite, SWT.CHECK);
		hideUnresolvableSidsCheckbox.setText("Hide unresolvable sid errors");
		hideUnresolvableSidsCheckbox.setImage(ImageFactory.getInstance().getImage(AppImage.EXCLUDE_FILTER_20X20));
		hideUnresolvableSidsCheckbox.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				searchLogRecords(true);
			}
		});
					
		searchResultTableViewer = new TableViewer(composite, SWT.BORDER | SWT.V_SCROLL | SWT.FULL_SELECTION | SWT.VIRTUAL);	
		searchResultTableViewer.setUseHashlookup(true);
		searchResultTableViewer.getTable().setHeaderVisible(true);
		searchResultTableViewer.getTable().setLinesVisible(true);
		gridData = new GridData(GridData.FILL_BOTH);
		gridData.horizontalSpan = 3;
		searchResultTableViewer.getTable().setLayoutData(gridData);
		
		TableViewerColumn timeColumn = new TableViewerColumn(searchResultTableViewer, SWT.NONE);
		timeColumn.getColumn().setWidth(200);
		timeColumn.getColumn().setText("Date / time");
		
		TableViewerColumn serverColumn = new TableViewerColumn(searchResultTableViewer, SWT.NONE);
		serverColumn.getColumn().setWidth(150);
		serverColumn.getColumn().setText("Server");
		
		TableViewerColumn messageColumn = new TableViewerColumn(searchResultTableViewer, SWT.NONE);
		messageColumn.getColumn().setWidth(750);
		messageColumn.getColumn().setText("Message");

		searchResultTableViewer.setContentProvider(ArrayContentProvider.getInstance());
		searchResultTableViewer.setLabelProvider(new TableLabelProvider() {
			@Override
			public String getColumnText(Object element, int columnIndex) {
				ScanLogItem msg = (ScanLogItem) element;
				switch (columnIndex) {
				case 0:
					return dateFormatter.format(msg.getLogDate());
				case 1:
					return msg.getServer();
				case 2:
					return msg.getMessage();
				default:
					return "";
				}
			}

			@Override
			public Image getColumnImage(Object element, int columnIndex) {
				if (columnIndex == 0) {
					ScanLogItem msg = (ScanLogItem) element;
					switch (msg.getSeverity()) {
					case TRACE:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_TRACE_16X16);
					case DEBUG:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_TRACE_16X16);
					case INFO:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_INFO_16X16);
					case WARNING:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_WARNING_16X16);
					case ERROR:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_ERROR_16X16);
					default:
						return null;
					}
				} else {
					return null;
				}
			}
		});

		searchResultTableViewer.addDoubleClickListener(new IDoubleClickListener() {
			@Override
			public void doubleClick(DoubleClickEvent event) {
				if (searchResultTableViewer.getTable().getSelectionIndex() != -1) {
					StructuredSelection selection = (StructuredSelection) searchResultTableViewer.getSelection();
					ScanLogItem logItem = (ScanLogItem) selection.getFirstElement();
					String details = logItem.getMessage();
					if (logItem.getDetails() != null && logItem.getDetails().length() > 0) {
						details += ":\r\n\r\n" + logItem.getDetails();
					}
					TextboxDialog detailsDialog = new TextboxDialog(null, "Details", "Error details:", details);
					detailsDialog.open();
				}
			}
		});
		
		
		searchResultTableViewer.setItemCount(searchResultList.size());
		searchResultTableViewer.setInput(searchResultList);
		
		// initialize the search
		searchLogRecords(true);
		
		ApplicationUtil.setDefaultCursor();
		return composite;
	}
	
	@Override
	protected void createButtonsForButtonBar(Composite parent) {
		GridLayout gridLayout = (GridLayout) parent.getLayout();
		gridLayout.makeColumnsEqualWidth = false;
		((GridData)parent.getLayoutData()).horizontalAlignment = GridData.FILL;
		
		CLabel explainLabel = new CLabel(parent, SWT.NONE);
		explainLabel.setText("Double click a record to view possible details.");
		explainLabel.setImage(ImageFactory.getInstance().getImage(AppImage.SEARCH_16X16));
		((GridLayout) parent.getLayout()).numColumns++; //also triggered by the createButton calls
			
		// filler
		Composite fillComposite = new Composite(parent, SWT.NONE);
		GridData gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.heightHint = 10;
		fillComposite.setLayoutData(gridData);
		((GridLayout) parent.getLayout()).numColumns++; //also triggered by the createButton calls
		
		createButton(parent, IDialogConstants.CANCEL_ID, "Close", false);
	}
	
	/**
	 * Create the Job that refreshes the Member list
	 * 
	 * @param explicitUserSearchAction true in case the user pressed the search button or ENTER in the search text box
	 */
	private void searchLogRecords(boolean explicitUserSearchAction) {
		String search = searchText.getText().toLowerCase().replace('*', '%');
		
		// only search when the user explicit requested the search OR when the length is greater than 2 and doesn't start with % OR the search list is based on the baseList
		if (explicitUserSearchAction ||	(search.length() > 2 && !search.startsWith("%"))) {
			
			searchResultList.clear();
			
			if (!search.endsWith("%")) {
				search += "%";
			}
		
			try {
				ApplicationUtil.setBusyCursor();
				
				StringBuilder whereClause = new StringBuilder();
				List<String> whereParameters = new ArrayList<String>();
				if (hostnameCombo.getCombo().getSelectionIndex() > 0) {
					// always remove the ' and ', we'll remove it at the end
					whereClause.append(" and server = ?");
					whereParameters.add(hostnameCombo.getCombo().getText());
				}
				
				if (!showInfoLevelCheckbox.getSelection()) {
					whereClause.append(" and severity <> ");
					whereClause.append(ScanLogItem.Severity.INFO.ordinal());
				}
				
				if (!showWarningLevelCheckbox.getSelection()) {
					whereClause.append(" and severity <> ");
					whereClause.append(ScanLogItem.Severity.WARNING.ordinal());
				}
				
				if (!showErrorLevelCheckbox.getSelection()) {
					whereClause.append(" and severity <> ");
					whereClause.append(ScanLogItem.Severity.ERROR.ordinal());
				}
				
				if (hideAccessDeniedCheckbox.getSelection()) {
					whereClause.append(" and message not like '%Access denied' ");
				}
				
				if (hideUnresolvableSidsCheckbox.getSelection()) {
					whereClause.append(" and message not like 'Could not translate SID%' ");				
				}
				
				if (search.length() > 1) {
					whereClause.append(" and (lower(message) like ? or lower(details) like ?) ");
					whereParameters.add(search);
					whereParameters.add(search);
				}
				
				if (whereClause.length() > 0) {
					searchResultList = ScanLogItem.where(whereClause.toString().substring(4), whereParameters.toArray(), "SERVER, LOG_DATE");
				} else {
					searchResultList = ScanLogItem.getAll();
				}
				
				if (searchResultList.isEmpty() && explicitUserSearchAction){
					MessageDialog.openInformation(Display.getCurrent().getActiveShell(), "No results", "No scan log records found.");
				}
				
			} catch (DBException ex) {
				logger.error("Error while searching for scan log records: " + ex.getMessage(), ex);
				MessageDialog.openError(null, "Error", "Error while searching for scan log records: " + ex.getMessage());
			}
			
			searchResultTableViewer.setInput(searchResultList);
			searchResultTableViewer.setItemCount(searchResultList.size());
			searchResultTableViewer.refresh();
			
			ApplicationUtil.setDefaultCursor();
		}		
	}
}
