package com.perdemia.permissionanalyzer.application.parts;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;

import javax.annotation.PostConstruct;
import javax.inject.Inject;

import org.eclipse.core.commands.ParameterizedCommand;
import org.eclipse.core.runtime.Platform;
import org.eclipse.e4.core.commands.ECommandService;
import org.eclipse.e4.core.commands.EHandlerService;
import org.eclipse.e4.ui.di.Focus;
import org.eclipse.e4.ui.model.application.MApplication;
import org.eclipse.e4.ui.model.application.ui.advanced.MPerspective;
import org.eclipse.e4.ui.services.EMenuService;
import org.eclipse.e4.ui.workbench.IWorkbench;
import org.eclipse.e4.ui.workbench.modeling.EModelService;
import org.eclipse.e4.ui.workbench.modeling.EPartService;
import org.eclipse.e4.ui.workbench.modeling.ESelectionService;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.viewers.ColumnViewerToolTipSupport;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.TreeViewerColumn;
import org.eclipse.swt.SWT;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.E4LifeCycle;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.MainFileTableContentProvider;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.MainFileTreeContentProvider;
import com.perdemia.permissionanalyzer.application.jface.dialogs.AlertDialog;
import com.perdemia.permissionanalyzer.application.jface.dialogs.AlertDialog.AlertType;
import com.perdemia.permissionanalyzer.application.jface.dialogs.TextboxDialog;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.GroupOwnershipColumnLabelProvider;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.InheritAclColumnLabelProvider;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.OwnerColumnLabelProvider;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.PathColumnLabelProvider;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.PermissionLabelProvider;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.SpecialPermissionLabelProvider;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.TotalOwnershipColumnLabelProvider;
import com.perdemia.permissionanalyzer.application.jface.wizard.firstuse.FirstUseWizard;
import com.perdemia.permissionanalyzer.application.preferences.PreferenceHelper;
import com.perdemia.permissionanalyzer.application.swt.ExplainLabel;
import com.perdemia.permissionanalyzer.application.swt.ExplainLabel.ExplainType;
import com.perdemia.permissionanalyzer.application.swt.TreeViewerWithMonitor;
import com.perdemia.permissionanalyzer.application.util.ApplicationUtil;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.ImageFactory.AppImage;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper.Edition;
import com.perdemia.permissionanalyzer.application.util.PermissionUtil;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SpecialPermission;

public class FilePart {
	private static final Logger logger = LoggerFactory.getLogger(FilePart.class.getName());
	public static final String ID = "com.perdemia.permissionanalyzer.application.parts.filetree";
	private TreeViewerWithMonitor treeViewer;
	private TableViewer tableViewer;
	private boolean initialized;
	
	@Inject private ESelectionService selectionService;
	@Inject private ECommandService commandService;
	@Inject private EModelService modelService;
	@Inject private EHandlerService handlerService;
	@Inject private EPartService partService;
	@Inject private MApplication application;
	@Inject private IWorkbench workbench;
	
	@PostConstruct
	public void createComposite(Composite parent, EMenuService menuService) {
		GridLayout gridLayout = new GridLayout(1, false);
		gridLayout.verticalSpacing = 5;
		gridLayout.horizontalSpacing = 0;
		parent.setLayout(gridLayout);
		
		new ExplainLabel(parent, "The file tree displays an aggregated view of all the permissions that match the filter. It shows a label with the relevant permission and a number of columns showing which special permissions apply e.g. permissions of various members, as each row is a sum of all retrieved permissions. Use the tabs at the bottom of the screen to view more details on the selected directory or file.", ExplainType.FILE_TREE);
		
		createTreeViewer(parent, menuService);
		createTableViewer(parent, menuService);
								
		// initialize the tree content
		// wait until this part is finished so that MPart.getObject returns this object instead of null (see first two lines in ApplyFilterHandler)
		if (!E4LifeCycle.isSilentMode) {
			Display.getCurrent().timerExec(1000, new Runnable() {
				public void run() {
					ParameterizedCommand applyFilterCommand = commandService.createCommand("com.perdemia.permissionanalyzer.application.commands.applyfilter", null);
					handlerService.executeHandler( applyFilterCommand );
					initialized = true;
					
					showFirstUseDialogs();
					
					// check for updates
					// TODO enable and test auto update
					//Map<String, Object> params = new HashMap<String, Object>();
					//params.put("com.perdemia.permissionanalyzer.application.commands.checkforupdates.param.silentmode", "true");
					//ParameterizedCommand checkForUpdatesCommand = commandService.createCommand("com.perdemia.permissionanalyzer.application.commands.checkforupdates", params);
					//handlerService.executeHandler( checkForUpdatesCommand );
										
					// is this a scan agent?
					if (LicenseHelper.getEdition().equals(Edition.SCAN_AGENT)) {
						// open the scan perspective
						ApplicationUtil.setBusyCursor();
						MPerspective scanPerspective = (MPerspective) modelService.find("com.perdemia.permissionanalyzer.application.perspective.scan", application);
						partService.switchPerspective(scanPerspective);
						ApplicationUtil.setDefaultCursor();
					}
				}
			});
		}
	}
	
	/**
	 * Creates the TreeViewer with all the files and permission columns. This TreeViewer is lazy loaded, passes its selection with the SelectionService and has a context menu
	 * that opens a folder in Explorer.
	 * 
	 * @param parent parent Composite
	 */
	private void createTreeViewer(Composite parent, EMenuService menuService) {
		treeViewer = new TreeViewerWithMonitor(parent, SWT.VIRTUAL | SWT.H_SCROLL | SWT.V_SCROLL | SWT.BORDER | SWT.FULL_SELECTION);
		treeViewer.getTree().setLinesVisible(true);
		treeViewer.getTree().setHeaderVisible(true);
		treeViewer.setUseHashlookup(true);
		treeViewer.setContentProvider(new MainFileTreeContentProvider(treeViewer, FilterHelper.getInstance()));
		
		//layout data for the tree
		GridData gridData = new GridData(SWT.FILL, SWT.FILL, true, true);
		treeViewer.getTree().setLayoutData(gridData);
		
		// Path column
		TreeViewerColumn column = new TreeViewerColumn(treeViewer, SWT.NONE);
		column.getColumn().setText("Path");
		column.getColumn().setWidth(500);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new PathColumnLabelProvider(false));
				
		// Permission column
		column = new TreeViewerColumn(treeViewer, SWT.NONE);
		column.getColumn().setText("Permission");
		column.getColumn().setWidth(200);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new PermissionLabelProvider());
				
		// Special permission columns
		for (final Integer permission : PermissionUtil.getCustomPermissionOrderList()) {
			//skip SYNCHRONIZE
			if (permission.intValue() != SpecialPermission.SYNCHRONIZE.ordinal()) {
				column = new TreeViewerColumn(treeViewer, SWT.NONE);
				column.getColumn().setWidth(30);
				column.getColumn().setResizable(false);
				column.getColumn().setToolTipText(PermissionUtil.specialPermissionToString(permission, false, null));
				column.getColumn().setImage(PermissionUtil.getSpecialPermissionImage(permission, false, true));
				column.setLabelProvider(new SpecialPermissionLabelProvider(permission, false, true));
			}
		}
		
		// Owner column
		column = new TreeViewerColumn(treeViewer, SWT.NONE);
		column.getColumn().setText("Owner");
		column.getColumn().setWidth(200);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new OwnerColumnLabelProvider());
								
		// Inherit ACL from parent column
		column = new TreeViewerColumn(treeViewer, SWT.NONE);
		column.getColumn().setText("Inherit parent ACL");
		column.getColumn().setWidth(120);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new InheritAclColumnLabelProvider());
		
		if (PreferenceHelper.getInstance().isCalculateOwnershipRatioInOverview()) {
			// Total ownership ratio
			column = new TreeViewerColumn(treeViewer, SWT.NONE);
			column.getColumn().setText("Total ownership ratio");
			column.getColumn().setWidth(150);
			column.getColumn().setResizable(true);
			column.setLabelProvider(new TotalOwnershipColumnLabelProvider());
			
			// Group ownership ratio
			column = new TreeViewerColumn(treeViewer, SWT.NONE);
			column.getColumn().setText("Group ownership ratio");
			column.getColumn().setWidth(150);
			column.getColumn().setResizable(true);
			column.setLabelProvider(new GroupOwnershipColumnLabelProvider());
		}
		
		//enable tooltips for this tree viewer
		ColumnViewerToolTipSupport.enableFor(treeViewer);
		
		//register this viewer with the selection service, so details views can react on selections
		treeViewer.addSelectionChangedListener(new ISelectionChangedListener() {
			@Override
			public void selectionChanged(SelectionChangedEvent event) {
				selectionService.setSelection(treeViewer.getSelection());
			}
		}); 
		
		// context menu for the Tree
		menuService.registerContextMenu(treeViewer.getControl(), "permission analyzer.popupmenu.file");
	}
	
	/**
	 * Creates the TableViewer with all the files and permission columns. This TableViewer is lazy loaded, passes its selection with the SelectionService and has a context menu
	 * that opens a folder in Explorer. The TableViewer is default invisible, because the user can toggle between the TreeViewer and this TableViewer.
	 * 
	 * @param parent parent Composite
	 */
	private void createTableViewer(Composite parent, EMenuService menuService) {
		tableViewer = new TableViewer(parent, SWT.VIRTUAL | SWT.H_SCROLL | SWT.V_SCROLL | SWT.BORDER | SWT.FULL_SELECTION);
		tableViewer.getTable().setLinesVisible(true);
		tableViewer.getTable().setHeaderVisible(true);
		tableViewer.setUseHashlookup(true);
		tableViewer.setContentProvider(new MainFileTableContentProvider(tableViewer));
		
		//layout data for the tree
		GridData gridData = new GridData(SWT.FILL, SWT.FILL, true, true);
		gridData.exclude = true; //TreeViewer is taking all the space until the user switches to Table view
		tableViewer.getTable().setLayoutData(gridData);
		
		// default hidden, because the TreeViewer is visible
		tableViewer.getTable().setVisible(false);
		
		// Path column
		TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.NONE);
		column.getColumn().setText("Path");
		column.getColumn().setWidth(800);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new PathColumnLabelProvider(true));
			
		// Permission column
		column = new TableViewerColumn(tableViewer, SWT.NONE);
		column.getColumn().setText("Permission");
		column.getColumn().setWidth(200);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new PermissionLabelProvider());
				
		// Special permission columns
		for (final Integer permission : PermissionUtil.getCustomPermissionOrderList()) {
			//skip SYNCHRONIZE
			if (permission.intValue() != SpecialPermission.SYNCHRONIZE.ordinal()) {
				column = new TableViewerColumn(tableViewer, SWT.NONE);
				column.getColumn().setWidth(30);
				column.getColumn().setResizable(false);
				column.getColumn().setToolTipText(PermissionUtil.specialPermissionToString(permission, false, null));
				column.getColumn().setImage(PermissionUtil.getSpecialPermissionImage(permission, false, true));
				column.setLabelProvider(new SpecialPermissionLabelProvider(permission, false, true));
			}
		}
		
		// Owner column
		column = new TableViewerColumn(tableViewer, SWT.NONE);
		column.getColumn().setText("Owner");
		column.getColumn().setWidth(200);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new OwnerColumnLabelProvider());
		
		// Inherit ACL from parent column
		column = new TableViewerColumn(tableViewer, SWT.NONE);
		column.getColumn().setText("Inherit parent ACL");
		column.getColumn().setWidth(120);
		column.getColumn().setResizable(true);
		column.setLabelProvider(new InheritAclColumnLabelProvider());
		
		if (PreferenceHelper.getInstance().isCalculateOwnershipRatioInOverview()) {
			// Total ownership ratio
			column = new TableViewerColumn(tableViewer, SWT.NONE);
			column.getColumn().setText("Total ownership ratio");
			column.getColumn().setWidth(150);
			column.getColumn().setResizable(true);
			column.setLabelProvider(new TotalOwnershipColumnLabelProvider());
			
			// Group ownership ratio
			column = new TableViewerColumn(tableViewer, SWT.NONE);
			column.getColumn().setText("Group ownership ratio");
			column.getColumn().setWidth(150);
			column.getColumn().setResizable(true);
			column.setLabelProvider(new GroupOwnershipColumnLabelProvider());
		}
		
		//enable tooltips for this table viewer
		ColumnViewerToolTipSupport.enableFor(tableViewer);
		
		//register this viewer with the selection service, so details views can react on selections
		tableViewer.addSelectionChangedListener(new ISelectionChangedListener() {
			@Override
			public void selectionChanged(SelectionChangedEvent event) {
				selectionService.setSelection(tableViewer.getSelection());
			}
		}); 
		
		// context menu for the Table
		menuService.registerContextMenu(tableViewer.getControl(), "permission analyzer.popupmenu.file");
	}
	
	private void showFirstUseDialogs() {
		PreferenceHelper prefHelper = PreferenceHelper.getInstance();
		
		// License agreement
		if (prefHelper.isShowLicenseAgreement()) {
			try{
				File licenseFile = new File(Platform.getLocation().toFile().getParent() + File.separator +  "License agreement.txt");
				String licenseText = new String(Files.readAllBytes(licenseFile.toPath()));
				
				int result = new TextboxDialog(null, "License Agreement", "", licenseText, "Don't agree", "Agree").open();
				if (result != IDialogConstants.OK_ID) {
					System.exit(0);
				} else {
					prefHelper.setShowLicenseAgreement(false);
					prefHelper.storePreferences(false);
				}
			} catch (IOException e) {
				logger.error("Could not load license file", e);
			}
		}
		
		if (!LicenseHelper.getEdition().equals(Edition.SCAN_AGENT)) {
			// First use wizard
			if (prefHelper.isShowFirstUseWizard()) {
				new FirstUseWizard(null).open();
			}
					
			if (treeViewer.getTree().getItemCount() == 0) {
				// no items found, open Scan View?
				AlertDialog dialog = new AlertDialog(null, "Scan your network", "The database contains no file info, do you want to open the Scan view?", AppImage.FILL_DATABASE_48X48, true, AlertType.OPEN_SCAN_VIEW);
				int result = dialog.open();
				if (result == IDialogConstants.YES_ID) {
					// open the scan perspective
					ApplicationUtil.setBusyCursor();
					MPerspective scanPerspective = (MPerspective) modelService.find("com.perdemia.permissionanalyzer.application.perspective.scan", application);
					partService.switchPerspective(scanPerspective);
					ApplicationUtil.setDefaultCursor();
				}
			} else {
				// show reminder for the Run button to apply filters?
				new AlertDialog(null, "Apply your filters", "Do not forget to apply your filters using the Run button in the toolbar or the shortcut F5.", AppImage.RUN_48X48, false, AlertType.APPLY_FILTER_BUTTON).open();
			}
		}
	}
	
	/**
	 * Passing the focus request to the viewer's control.
	 */
	@Focus
	public void setFocus() {
		treeViewer.getControl().setFocus();
		if (FilterHelper.getInstance().isReloadReportView()) {
			FilterHelper.getInstance().setReloadReportView(false);
			ParameterizedCommand applyFilterCommand = commandService.createCommand("com.perdemia.permissionanalyzer.application.commands.applyfilter", null);
			handlerService.executeHandler( applyFilterCommand );
			new AlertDialog(null, "Apply your filters", "Do not forget to apply your filters using the Run button in the toolbar or the shortcut F5.", AppImage.RUN_48X48, false, AlertType.APPLY_FILTER_BUTTON).open();
		}
	}
	
	public TreeViewerWithMonitor getTreeViewer() {
		return treeViewer;
	}

	public TableViewer getTableViewer() {
		return tableViewer;
	}
	
	public boolean isInitialized() {
		return initialized;
	}

	public IWorkbench getWorkbench() {
		return workbench;
	}
}