package com.perdemia.permissionanalyzer.application.parts;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.text.DateFormat;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Scanner;
import java.util.regex.Pattern;

import javax.annotation.PostConstruct;
import javax.inject.Inject;

import org.eclipse.core.commands.ParameterizedCommand;
import org.eclipse.e4.core.commands.ECommandService;
import org.eclipse.e4.core.commands.EHandlerService;
import org.eclipse.e4.ui.di.Focus;
import org.eclipse.e4.ui.model.application.ui.basic.MPart;
import org.eclipse.e4.ui.services.IStylingEngine;
import org.eclipse.e4.ui.workbench.IWorkbench;
import org.eclipse.e4.ui.workbench.modeling.EPartService;
import org.eclipse.jface.dialogs.IDialogConstants;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CLabel;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Table;
import org.mihalis.opal.notify.Notifier;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.export.ExportHelper;
import com.perdemia.permissionanalyzer.application.jface.dialogs.AlertDialog;
import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.jface.dialogs.TextboxDialog;
import com.perdemia.permissionanalyzer.application.jface.dialogs.ViewScanLogDialog;
import com.perdemia.permissionanalyzer.application.jface.dialogs.AlertDialog.AlertType;
import com.perdemia.permissionanalyzer.application.jface.labelproviders.TableLabelProvider;
import com.perdemia.permissionanalyzer.application.logging.GUILogHandler;
import com.perdemia.permissionanalyzer.application.model.DirectoryInfo;
import com.perdemia.permissionanalyzer.application.model.LdapOUInfo;
import com.perdemia.permissionanalyzer.application.model.StatusListener;
import com.perdemia.permissionanalyzer.application.preferences.PreferenceHelper;
import com.perdemia.permissionanalyzer.application.swt.AnimatedImage;
import com.perdemia.permissionanalyzer.application.swt.ImageComposite;
import com.perdemia.permissionanalyzer.application.swt.ShellHelper;
import com.perdemia.permissionanalyzer.application.util.ApplicationUtil;
import com.perdemia.permissionanalyzer.application.util.ColorFactory;
import com.perdemia.permissionanalyzer.application.util.ColorFactory.AppColor;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.ImageFactory;
import com.perdemia.permissionanalyzer.application.util.ImageFactory.AppImage;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper.Edition;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper;
import com.perdemia.permissionanalyzer.application.util.ProgramParameters;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.ScanLogItem;
import com.perdemia.permissionanalyzer.data.ScanLogItem.Severity;
import com.perdemia.permissionanalyzer.data.util.DBUtil;
import com.perdemia.permissionanalyzer.data.util.DomainCache;
import com.perdemia.permissionanalyzer.data.util.FileCache;
import com.perdemia.permissionanalyzer.data.util.MemberCache;
import com.perdemia.permissionanalyzer.data.util.MembershipCache;
import com.perdemia.permissionanalyzer.data.util.RdnCache;
import com.perdemia.permissionanalyzer.data.util.TableCountCache;
import com.perdemia.permissionanalyzer.scanner.acl.AclScanner;
import com.perdemia.permissionanalyzer.scanner.groups.NestedGroupHelper;
import com.perdemia.permissionanalyzer.scanner.ldap.LdapConnector;
import com.perdemia.permissionanalyzer.scanner.ldap.LdapFactory;

public class ScanResultsPart {
	public static final String ID = "com.perdemia.permissionanalyzer.application.parts.scanresults";
	private static Logger logger = LoggerFactory.getLogger(ScanResultsPart.class.getName());
	private static final Color ERROR_COLOR = new Color(null, 255, 0, 85);
	private static final Color BLACK_COLOR = new Color(null, 0, 0, 0);
	
	@Inject EPartService partService;
	@Inject IStylingEngine engine;
	@Inject IWorkbench workbench;
	@Inject ECommandService commandService;
	@Inject EHandlerService handlerService;
	
	/** UI fields */
	private Button cancelScanButton; 
	private Button scanButton;
	
	/** Status table fields */
	private TableViewer statusTableViewer;
	private List<ScanLogItem> scanLogItemList = new ArrayList<ScanLogItem>();
	private DateFormat dateFormatter = DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.MEDIUM, Locale.getDefault());
	private CLabel generalStatusLabel;
	private Label statusLabel;
	private AnimatedImage scanAnimation;
	private Image[] analyzeFoldersImages;
	private Image[] analyzeUsersImages;
	
	/** Listener used to display info messages in the GUI during the scan process */
	private StatusListener statusListener;

	private PreferenceHelper preferences = PreferenceHelper.getInstance();
	
	/**
	 * Constructor, initializes the result list with the results of the last scan
	 */
	public ScanResultsPart() {
		readStatusListFromDB();
	}	

	@PostConstruct
	public void createComposite(final Composite parent) {
		parent.setLayout(new GridLayout(1, true));
		parent.setBackground(ColorFactory.getInstance().getColor(AppColor.PART_BACKGROUND));
		
		createStatusComposite(parent);

		// status listener to display message in the status table
		statusListener = new StatusListener() {
			private boolean cancelled;
			private int step = 0;
			private int numberOfDirectories;
			private int numberOfFiles;
			private int numberOfUsers;
			private int numberOfGroups;
						
			@Override
			public void newStatusMessage(final ScanLogItem message) {
				scanLogItemList.add(message);
				try {
					message.save();
				} catch (DBException e) {
					// no logger.error because we will end up in an infinitive loop because of the GUILogHandler that is currently active
					// (which will notify all listeners using newStatusMessage (this method!)
					System.out.println("Error while saving log record to database: " + e.getMessage());
					e.printStackTrace();
				}
				
				Display.getDefault().syncExec(new Runnable() {
					public void run() {
						//if (message.getLevel().ordinal() >= STATUS_LEVEL.INFO.ordinal()) {
							if (scanLogItemList.size() > 0) {
								statusTableViewer.refresh(statusTableViewer.getElementAt(scanLogItemList.size() - 1));
							} else {
								statusTableViewer.refresh();
							}
							statusTableViewer.getTable().setTopIndex(scanLogItemList.size());
							if (message.getSeverity().equals(Severity.ERROR)) {
								statusLabel.setText("An error occured during the scan. Please see the status table for more details.");
								statusLabel.setForeground(ERROR_COLOR);
							} else {
								statusLabel.setText(message.getMessage());
								statusLabel.setForeground(BLACK_COLOR);
							}
						//}
					}
				});
				
				if (message.getSeverity().equals(Severity.ERROR) && message.getMessage().toLowerCase().contains("referral")) {
					Display.getDefault().syncExec(new Runnable() {
						public void run() {
							// we have a possible error concerning follow referral URLs
							PreferenceHelper prefs = PreferenceHelper.getInstance();
							String msg;
							if (prefs.isFollowReferrals()) {
								msg = "The preference 'Follow LDAP referrals' is enabled, disabling this option might solve the error. \r\n\r\nDo you want to disable the option?";
							} else {
								msg = "The preference 'Follow LDAP referrals' is disabled, enabling this option might solve the error. \r\n\r\nDo you want to enable the option?";
							}
							AlertDialog dialog = new AlertDialog(null, "Follow LDAP referrals", "The application detected an error regarding LDAP referrals (references to other domain controllers). " + msg + " You can always modify the setting in the Advanced tab of the aplication preferences (F6).", AppImage.CONNECTION_48X48, true, AlertType.FOLLOW_LDAP_REFERRALS);
							int result = dialog.open();
							if (result == IDialogConstants.YES_ID) {
								// toggle the preference "Follow referrals"
								prefs.setFollowReferrals(!prefs.isFollowReferrals());
								prefs.storePreferences(false);
								cancelled = true;
								SyncMessageDialog.showMessage(MessageDialog.INFORMATION, "preference saved", "The preference has been saved. Please restart the scan.");
							}
						}
					});
				}
			}

			@Override
			public boolean isCancelled() {
				return cancelled;
			}

			@Override
			public void cancel() {
				cancelled = true;
			}

			@Override
			public void updateStatus(final String message) {
				Display.getDefault().syncExec(new Runnable() {
					public void run() {
						statusLabel.setText(message);
						statusLabel.setForeground(BLACK_COLOR);
					}
				});

			}

			@Override
			public void directoryScanned(final DirectoryInfo directory, final boolean success) {
				Display.getDefault().syncExec(new Runnable() {
					public void run() {
						ScanFoldersPart foldersPart = getScanFoldersPart();
						if (foldersPart != null) {
							foldersPart.directoryScanned(directory, success);
						}
					}
				});
			}

			@Override
			public void ldapOUScanned(final LdapOUInfo ouInfo, final boolean success) {
				Display.getDefault().syncExec(new Runnable() {
					public void run() {
						ScanLdapPart ldapPart = getScanLdapPart();
						if (ldapPart != null) {
							ldapPart.ldapOUScanned(ouInfo, success);
						}
					}
				});
			}

			@Override
			public void reset() {
				cancelled = false;
				step = 0;
				numberOfDirectories = 0;
				numberOfFiles = 0;
				numberOfUsers = 0;
				numberOfGroups = 0;
			}
			
			@Override
			public void proceed() {
				cancelled = false;
			}

			@Override
			public void directoriesScanned(final int numberOfDirectories, final int numberOfFiles) {
				this.numberOfDirectories = numberOfDirectories;
				this.numberOfFiles = numberOfFiles;
			}

			@Override
			public void usersScanned(final int numberOfUsers) {
				this.numberOfUsers += numberOfUsers;
			}
			
			@Override
			public void groupsScanned(final int numberOfGroups) {
				this.numberOfGroups += numberOfGroups;
			}

			@Override
			public void nextStep() {
				step++;
				Display.getDefault().syncExec(new Runnable() {
					public void run() {
						switch (step) {
							case 1: generalStatusLabel.setText("Step 1 of 5: Scan LDAP servers"); break;
							case 2: generalStatusLabel.setText("Step 2 of 5: Determine nested group memberships"); break;
							case 3: generalStatusLabel.setText("Step 3 of 5: Search local groups"); break;
							case 4: generalStatusLabel.setText("Step 4 of 5: Scan file systems"); break;
							case 5: generalStatusLabel.setText("Step 5 of 5: Run database statistics"); break;
							case 6: generalStatusLabel.setText("No scan in progress"); break;
						}
						
						if (step == 6) {
							logger.info(MessageFormat.format("Scan finished, processed {0} directories, {1} files, {2} users and {3} groups.", numberOfDirectories, numberOfFiles, numberOfUsers, numberOfGroups));
							Notifier.notify("Permission Analyzer", MessageFormat.format("Scan finished, processed <b>{0}</b> directories, <b>{1}</b> files, <b>{2}</b> users and <b>{3}</b> groups   .", numberOfDirectories, numberOfFiles, numberOfUsers, numberOfGroups));
							ShellHelper.forceActive(Display.getCurrent().getActiveShell());
						}
					}
				});
			}

			public int getNumberOfDirectories() {
				return numberOfDirectories;
			}

			public int getNumberOfFiles() {
				return numberOfFiles;
			}

			public int getNumberOfUsers() {
				return numberOfUsers;
			}

			public int getNumberOfGroups() {
				return numberOfGroups;
			}
		};

		// general status label (step 1 of 5 info etc...)
		generalStatusLabel = new CLabel(parent, SWT.NONE);
		generalStatusLabel.setText("No scan in progress");
		GridData gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.horizontalSpan = 2;
		gridData.horizontalIndent = 10;
		generalStatusLabel.setLayoutData(gridData);
		engine.setClassname(generalStatusLabel, "h2");
		
		// button bar
		Composite buttonBar = new Composite(parent, SWT.NONE);
		gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.horizontalSpan = 2;
		buttonBar.setLayoutData(gridData);
		buttonBar.setLayout(new GridLayout(6, false));

		final Button exportStatusListButton = new Button(buttonBar, SWT.NONE);
		exportStatusListButton.setLayoutData(new GridData(SWT.LEFT, SWT.CENTER, false, false, 1, 1));
		exportStatusListButton.setText("Export status list");
		exportStatusListButton.setImage(ImageFactory.getInstance().getImage(AppImage.EXPORT_16X16));
		exportStatusListButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				FileDialog dialog = new FileDialog(parent.getShell(), SWT.SAVE);
				dialog.setOverwrite(true);
				dialog.setFilterExtensions(new String[] { "*.csv" });
				String filePath = dialog.open();

				if (filePath != null) {
					saveStatusListToFile(filePath);
				}
			}
		});
		
		final Button viewRemoteLogsButton = new Button(buttonBar, SWT.NONE);
		gridData = new GridData(SWT.LEFT, SWT.CENTER, false, false, 1, 1);
		viewRemoteLogsButton.setLayoutData(gridData);
		viewRemoteLogsButton.setText("View remote scan logs");
		viewRemoteLogsButton.setImage(ImageFactory.getInstance().getImage(AppImage.LOG_FILE_16X16));
		viewRemoteLogsButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				ViewScanLogDialog dialog = new ViewScanLogDialog(parent.getShell());
				dialog.open();
			}
		});

		// Image animation for the scanning process
		ImageComposite analyzeFoldersImageComposite = new ImageComposite(buttonBar, ImageFactory.getInstance().getImage(AppImage.ANALYZE_FOLDERS_48X48_1));
		((GridData)analyzeFoldersImageComposite.getLayoutData()).horizontalIndent = 30;

		// images for the folder animation
		analyzeFoldersImages = new Image[] { ImageFactory.getInstance().getImage(AppImage.ANALYZE_FOLDERS_48X48_1), ImageFactory.getInstance().getImage(AppImage.ANALYZE_FOLDERS_48X48_2), ImageFactory.getInstance().getImage(AppImage.ANALYZE_FOLDERS_48X48_3),
				ImageFactory.getInstance().getImage(AppImage.ANALYZE_FOLDERS_48X48_4) };

		// images for the user animation
		analyzeUsersImages = new Image[] { ImageFactory.getInstance().getImage(AppImage.ANALYZE_USERS_48X48_1), ImageFactory.getInstance().getImage(AppImage.ANALYZE_USERS_48X48_2), ImageFactory.getInstance().getImage(AppImage.ANALYZE_USERS_48X48_3),
				ImageFactory.getInstance().getImage(AppImage.ANALYZE_USERS_48X48_4) };

		scanAnimation = new AnimatedImage(analyzeFoldersImageComposite, analyzeUsersImages);

		statusLabel = new Label(buttonBar, SWT.NONE);
		gridData = new GridData(GridData.FILL_HORIZONTAL);
		gridData.horizontalIndent = 10;
		statusLabel.setLayoutData(gridData);
		statusLabel.setText("Ready to start scanning...");

		cancelScanButton = new Button(buttonBar, SWT.NONE);
		cancelScanButton.setLayoutData(new GridData(SWT.RIGHT, SWT.CENTER, false, false, 1, 1));
		cancelScanButton.setText("Stop scan");
		cancelScanButton.setEnabled(false);
		cancelScanButton.setImage(ImageFactory.getInstance().getImage(AppImage.CANCEL_SCAN_16X16));
		cancelScanButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				statusListener.cancel();
				cancelScanButton.setEnabled(false);
			}
		});

		scanButton = new Button(buttonBar, SWT.NONE);
		scanButton.setLayoutData(new GridData(SWT.RIGHT, SWT.CENTER, false, false, 1, 1));
		scanButton.setText("Scan");
		scanButton.setImage(ImageFactory.getInstance().getImage(AppImage.FILL_DATABASE_16X16));
		scanButton.addSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				startScan();
			}
		});
	}
	
	public void createStatusComposite(Composite parent) {
		Composite statusComposite = new Composite(parent, SWT.NONE);
		GridData gridData = new GridData(GridData.FILL_BOTH);
		//gridData.horizontalSpan = 2;
		statusComposite.setLayoutData(gridData);
		statusComposite.setLayout(new GridLayout(1, false));

		/*
		TitledSeparator statusSeparator = new TitledSeparator(statusComposite, SWT.NONE);
		statusSeparator.setText("Status");
		statusSeparator.setImage(ImageFactory.getInstance().getImage(AppImage.GROUP_24X24));
		statusSeparator.setSize(50, 800);
		statusSeparator.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
*/
		
		statusTableViewer = new TableViewer(statusComposite, SWT.BORDER | SWT.FULL_SELECTION);
		Table table = statusTableViewer.getTable();
		table.setLinesVisible(true);
		table.setHeaderVisible(true);
		GridData gd_table = new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1);
		//gd_table.horizontalIndent = 30;
		table.setLayoutData(gd_table);

		TableViewerColumn timeColumn = new TableViewerColumn(statusTableViewer, SWT.NONE);
		timeColumn.getColumn().setWidth(200);
		timeColumn.getColumn().setText("Date / time");

		TableViewerColumn messageColumn = new TableViewerColumn(statusTableViewer, SWT.NONE);
		messageColumn.getColumn().setWidth(1200);
		messageColumn.getColumn().setText("Message");

		statusTableViewer.setContentProvider(ArrayContentProvider.getInstance());
		statusTableViewer.setLabelProvider(new TableLabelProvider() {
			@Override
			public String getColumnText(Object element, int columnIndex) {
				ScanLogItem msg = (ScanLogItem) element;
				switch (columnIndex) {
				case 0:
					return dateFormatter.format(msg.getLogDate());
				case 1:
					return msg.getMessage();
				case 2:
					return msg.getDetails();
				default:
					return "";
				}
			}

			@Override
			public Image getColumnImage(Object element, int columnIndex) {
				if (columnIndex == 0) {
					ScanLogItem msg = (ScanLogItem) element;
					switch (msg.getSeverity()) {
					case TRACE:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_TRACE_16X16);
					case DEBUG:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_TRACE_16X16);
					case INFO:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_INFO_16X16);
					case WARNING:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_WARNING_16X16);
					case ERROR:
						return ImageFactory.getInstance().getImage(AppImage.STATUS_ERROR_16X16);
					default:
						return null;
					}
				} else {
					return null;
				}
			}
		});
		
		statusTableViewer.addDoubleClickListener(new IDoubleClickListener() {
			@Override
			public void doubleClick(DoubleClickEvent event) {
				if (statusTableViewer.getTable().getSelectionIndex() != -1) {
					StructuredSelection selection = (StructuredSelection) statusTableViewer.getSelection();
					ScanLogItem logItem = (ScanLogItem) selection.getFirstElement();
					String details = logItem.getMessage();
					if (logItem.getDetails() != null && logItem.getDetails().length() > 0) {
						details += ":\r\n\r\n" + logItem.getDetails();
					}
					TextboxDialog detailsDialog = new TextboxDialog(null, "Details", "Error details:", details);
					detailsDialog.open();
				}
			}
		});

		statusTableViewer.setInput(scanLogItemList);
	}

	private String encodeForCSV(String value) {
		if (value == null) {
			return "";
		} else {
			value = value.replaceAll(Pattern.quote("\""), "^");
			value = value.replaceAll(Pattern.quote("\n"), "|");
			value = value.replaceAll(Pattern.quote("\r"), "~");
			value = value.replaceAll(Pattern.quote(","), "_");
			return value;
		}
	}

	private String decodeFromCSV(String value) {
		value = value.trim().replaceAll(Pattern.quote("^"), "\"");
		value = value.replaceAll(Pattern.quote("|"), "\n");
		value = value.replaceAll(Pattern.quote("~"), "\r");
		value = value.replaceAll(Pattern.quote("_"), ",");
		return value;
	}

	/**
	 * Initializes the {@link #scanLogItemList} with objects using a File
	 * @deprecated not used anymore, log messages are read from the database. Perhaps this method will become useful for an Import button.
	 */
	private void readStatusListFromFile() {
		Scanner scanner = null;
		try {
			
			Path path = Paths.get(ApplicationUtil.getWorkspacePath() + "Last_status_messages.csv");
			if (Files.exists(path)) {
				scanner = new Scanner(path, "UTF-8");
				boolean firstLine = true;
				String line;
				while (scanner.hasNextLine()) {
					line = scanner.nextLine();
					if (firstLine) {
						firstLine = false;
						continue;
					}
					String[] columns = line.split(",");
					ScanLogItem scanLogItem = new ScanLogItem();
					scanLogItem.setLogDate(dateFormatter.parse(decodeFromCSV(columns[0])));
					scanLogItem.setSeverity(Severity.valueOf(decodeFromCSV(columns[1])));
					scanLogItem.setMessage(decodeFromCSV(columns[2]));
					if (columns.length > 3) {
						scanLogItem.setDetails(decodeFromCSV(columns[3]));
					}
					scanLogItemList.add(scanLogItem);
				}
			}
		} catch (Exception ex) {
			logger.error("Error while restoring status messages from last scan: " + ex.getMessage(), ex);
			StringWriter stacktrace = new StringWriter();
			ex.printStackTrace(new PrintWriter(stacktrace));
			scanLogItemList.add(new ScanLogItem(Severity.ERROR, "Error while restoring status messages from last scan: " + ex.getMessage(), stacktrace.toString()));
		} finally {
			if (scanner != null) {
				scanner.close();
			}
		}
	}
	
	/**
	 * Initializes the {@link #scanLogItemList} with objects using the database
	 * 
	 */
	private void readStatusListFromDB() {
		try {
			scanLogItemList.addAll(ScanLogItem.where("server = ?", new Object[]{AclScanner.getLocalMachineName()}, "log_date"));
		} catch (Exception ex) {
			logger.error("Error while restoring status messages from last scan: " + ex.getMessage(), ex);
			StringWriter stacktrace = new StringWriter();
			ex.printStackTrace(new PrintWriter(stacktrace));
			scanLogItemList.add(new ScanLogItem(Severity.ERROR, "Error while restoring status messages from last scan: " + ex.getMessage(), stacktrace.toString()));
		}
	}

	/**
	 * Save the list of objects in {@link #scanLogItemList} to a File
	 */
	private void saveStatusListToFile(String filePath) {
		BufferedWriter writer = null;
		try {		
			Path path = Paths.get(filePath);
			writer = Files.newBufferedWriter(path, StandardCharsets.UTF_8, StandardOpenOption.CREATE, StandardOpenOption.WRITE, StandardOpenOption.TRUNCATE_EXISTING);
			writer.write("Date,Severity,Message,Details");
	        writer.newLine();
	        for (ScanLogItem message : scanLogItemList) {
	        	writer.write(encodeForCSV(dateFormatter.format(message.getLogDate())));
	        	writer.write(",");
	        	writer.write(message.getSeverity().toString());
	        	writer.write(",");
	        	writer.write(encodeForCSV(message.getMessage()));
	        	writer.write(",");
	        	writer.write(encodeForCSV(message.getDetails()));
	        	writer.newLine();
	        }
	        writer.flush();
		
		} catch (Exception ex) {
			logger.error("Error while writing status messages to file " + filePath + ": " + ex.getMessage(), ex);
			SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while writing status messages to file " + filePath + ": " + ex.getMessage());
		} finally {
			if (writer != null) {
				try {
					writer.close();
				} catch (IOException e) {
					logger.error("Error while closing writer in saveStatusListToFile: " + e.getMessage(), e);
				}
			}
		}
	}
	
	private ScanFoldersPart getScanFoldersPart() {
		MPart part = partService.findPart(ScanFoldersPart.ID);
		if (part != null) {
			return (ScanFoldersPart) part.getObject();
		} else {
			return null;
		}
	}
	
	private ScanLdapPart getScanLdapPart() {
		MPart part = partService.findPart(ScanLdapPart.ID);
		if (part != null) {
			return (ScanLdapPart) part.getObject();
		} else {
			return null;
		}
	}
	
	@Focus
	public void setFocus() {
		statusTableViewer.getTable().setFocus();
	}
	
	public void startScan() {
		Thread myTread = new Thread(new Runnable() {
			@Override
			public void run() {
				long curTime = System.currentTimeMillis();
				try {
					scanLogItemList.clear();
					statusListener.reset();
					GUILogHandler.addListener(statusListener);
					Member.clearStatistics();
					ScanLogItem.delete("server = ?", AclScanner.getLocalMachineName());
					ApplicationUtil.setScanInProgress(true);
					
					Display.getDefault().syncExec(new Runnable() {
						public void run() {
							statusTableViewer.refresh();
							getScanFoldersPart().resetListStatus();
							getScanLdapPart().resetListStatus();
							scanButton.setEnabled(false);
							cancelScanButton.setEnabled(true);
							scanAnimation.setImages(analyzeUsersImages);
							scanAnimation.start();
						}
					});
					
					// scan LDAP?
					boolean performLdapScan = false;
					
					// not only scanning directories using command line param?
					if (!ProgramParameters.isScanDirectories()) {
						// check if any of the OU's have been checked
						for (LdapOUInfo ouInfo : preferences.getLdapOUList()) {
							boolean scanLdapOU = ouInfo.isEnabled() || ProgramParameters.isForceScanAllLdapContainers() || ProgramParameters.isLdapOUInList(ouInfo.getBaseOU());
							performLdapScan = performLdapScan || scanLdapOU;
							
							if (!scanLdapOU) {
								logger.warn(MessageFormat.format("Skipping LDAP container because it has not been checked: connection {0} and baseDN [{1}]", ouInfo.getConnection().getConnectionName(), ouInfo.getBaseOU()));
							}
						}
					}
					
					// scan directories?
					boolean performDirectoryScan = false;
					
					// not only scanning LDAP using command line param?
					if (!ProgramParameters.isScanLdap()) {
						// check if any of the directoies have been checked
						for (DirectoryInfo directoryInfo : preferences.getDirectoryList()) {
							boolean scanDirectory = directoryInfo.isEnabled() || ProgramParameters.isForceScanAllDirectoryContainers() || ProgramParameters.isDirectoryInList(directoryInfo.getPath());
							performDirectoryScan = performDirectoryScan || scanDirectory;
							
							if (!scanDirectory) {
								logger.warn("Skipping directory because it has not been checked: " + directoryInfo.getPath());
							}
						}
					}
					
					// truncate the right data
					boolean truncateLdap = (preferences.isTruncateData() || ProgramParameters.isClearData()) && (performLdapScan || preferences.getLdapOUList().size() == 0);
					boolean truncateDirectories = ((preferences.isTruncateData() || ProgramParameters.isClearData()) && (performDirectoryScan || preferences.getDirectoryList().size() == 0)) || LicenseHelper.getEdition().equals(Edition.TRIAL);
					if (truncateLdap && truncateDirectories) {
						DBUtil.truncateAllData(statusListener);
						
					} else if (truncateLdap) {
						DBUtil.truncateLdapData(statusListener);
						
					} else if (truncateDirectories) {
						DBUtil.truncateDirectoryData(statusListener);
					}
					
					// clear caches
					DomainCache.clearCache();
					MemberCache.clearCache();
					TableCountCache.clearCache();
					FileCache.clearCache();
					RdnCache.clearCache();
					MembershipCache.clearCache();
					
					DBObject.setAutoCommit(false);
					
					if (performLdapScan) {
						// scan directories
						logger.info("Start scanning LDAP");
						statusListener.nextStep();
						
						// change the animation
						Display.getDefault().syncExec(new Runnable() {
							public void run() {
								scanAnimation.setImages(analyzeUsersImages);
							}
						});
						
						// scan ldap
						LdapConnector connector = LdapFactory.getLdapConnector();
						connector.scanOUs(preferences.getLdapOUList(), statusListener);
					
						if (statusListener.isCancelled()) {
							if (SyncMessageDialog.showMessage(MessageDialog.QUESTION, "Continue with file scan?", "The LDAP scan has been cancelled, do you want to continue with the directory scan and the calculation of nested group memberships?")) {
								statusListener.proceed();
								Display.getDefault().asyncExec(new Runnable(){
									public void run(){
										cancelScanButton.setEnabled(true);
									}
								});
							}
						} else {
							logger.info("Done scanning LDAP");
						}
						
						// TODO bij user limiet, toon hoeveel gebruikers uiteindelijk zijn gevonden (dus doorgaan zonder opslaan...)
						
						// Determine nested group memberships
						if (!statusListener.isCancelled()) {
							logger.info("Start determining nested group memberships");
							statusListener.nextStep();
							try {
								NestedGroupHelper.processNestedGroupMemberships(statusListener);
							} catch (DBException ex) {
								logger.error("Database error while determining nested group membership: " + ex.getMessage(), ex);
							}
						}
						
					} else {
						statusListener.nextStep(); // ldap scan skipped
						statusListener.nextStep(); // calculation of nested group memberships skipped
					}
					
					// continue with directory scan?
					if (!statusListener.isCancelled()) {
						
						if (performDirectoryScan) {
							logger.info("Start scanning directories");
	
							// change the animation
							Display.getDefault().syncExec(new Runnable() {
								public void run() {
									scanAnimation.setImages(analyzeFoldersImages);
								}
							});
							
							// scan directories
							statusListener.nextStep();
							AclScanner scanner = new AclScanner(statusListener);
							scanner.scan(preferences.getDirectoryList());
	
							if (statusListener.isCancelled()) {
								if (SyncMessageDialog.showMessage(MessageDialog.QUESTION, "Finish the scan?", "The directory scan has been cancelled, do you want to finish the scan (rematch existing filter values and update database statistics)?")) {
									statusListener.proceed();
									Display.getDefault().asyncExec(new Runnable(){
										public void run(){
											cancelScanButton.setEnabled(true);
										}
									});
								}
							} else {
								logger.info("Done scanning directories");
							}
							
						} else {
							statusListener.nextStep(); // local groups scan skipped
							statusListener.nextStep(); // directory scan skipped
						}
						
						if (!statusListener.isCancelled()) {
							
							// Run database statistics
							if (!statusListener.isCancelled()) {
								logger.info("Start running database statistics");
								statusListener.nextStep();
								try {
									DBUtil.updateStatistics();
								} catch (DBException ex) {
									logger.error("Error while updating database statistics: " + ex.getMessage(), ex);
								}
							}
							
							// Rematch filter values (like account name or file path), because they might have another db ID
							if (!statusListener.isCancelled()) {
								try {
									FilterHelper.matchFilterValuesAfterScan();
								} catch (DBException ex) {
									logger.error("Error while re-matching existing filter values: " + ex.getMessage(), ex);
								}
							}
						}
					}
					
				} catch (Exception ex) {
					logger.error("Error while performing the scan: " + ex.getMessage(), ex);
				} finally {
					try {
						DBObject.setAutoCommit(true);
					} catch (Exception e) {
						logger.error("Error while setting autoCommit on true: " + e.getMessage(), e);
					}
				}
								
				// clear the Member cache to free memory
				MemberCache.clearCache();
				
				// clear the table count cache again, because it might be that the user ran some overviews during the scan
				TableCountCache.clearCache();
				
				// restore the GUI
				Display.getDefault().syncExec(new Runnable() {
					public void run() {
						scanButton.setEnabled(true);
						cancelScanButton.setEnabled(false);
						scanAnimation.setImages(new Image[]{ImageFactory.getInstance().getImage(AppImage.ANALYZE_DONE_48X48)});
						scanAnimation.stop();
					}
				});
													
				// display the processing time
				long totalTime = System.currentTimeMillis() - curTime;
				long hours = totalTime / 3600000;
				long minutes = (totalTime - hours * 3600000) / 60000;
				long seconds = (totalTime - hours * 3600000 - minutes * 60000) / 1000;
				logger.info(MessageFormat.format("The scan has run {0} hours, {1} minutes and {2} seconds", hours, minutes, seconds));
				
				if (!statusListener.isCancelled()) {
					statusListener.updateStatus("Scan completed");
				} else {
					logger.warn("Scan process cancelled by the user");
					logger.warn("The database statistics have not been updated, this might impact the performance of the overviews");
					statusListener.updateStatus("Scan process cancelled by the user");
				}
				
				statusListener.nextStep();
										
				GUILogHandler.removeListener(statusListener);
				ApplicationUtil.setScanInProgress(false);
				
				LicenseHelper.updateUsageStatistics(MessageFormat.format("{0} hours, {1} minutes and {2} seconds", hours, minutes, seconds), 
						statusListener.getNumberOfUsers(), statusListener.getNumberOfGroups(), statusListener.getNumberOfFiles(), statusListener.getNumberOfDirectories());
				
				// was this an scheduled scan?
				if (ProgramParameters.isScan()) {
					
					// are there any reports scheduled?
					if (ProgramParameters.isRunReports()) {
						ExportHelper.exportScheduledReports(workbench);
						
					} else {
						// done, close the application
						Display.getDefault().syncExec(new Runnable() {
							public void run() {
								workbench.close();
							}
						});
					}
				} else {
					// refresh the Report View when it is opened by the user
					FilterHelper.getInstance().setReloadReportView(true);
				}
			}
		});

		myTread.start();
	}
}
