package com.perdemia.permissionanalyzer.application.preferences;

import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.KeySpec;
import java.util.Arrays;

import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.SecretKeySpec;

import com.unboundid.util.Base64;

/**
 * <p>
 * Provides methodes to encrypt / decrypt messages and to hash a password. The encryption is done using AES (Advanced Encryption Standard) in combination
 * with Cypher Block Chaining (CBC). Cypher Block Chaining splits the encryption algorithm up into blocks of data. It then encrypts one block and 
 * passes the running result into the encryption algorithm for the next block. As such, each block encryption depends on the aggregate result 
 * of all previously encrypted blocks in the same message.
 * </p><p>
 * The password for the encryption is first mixed with a salt value and then hashed using PBKDF2WithHmacSHA1. This hash algorithm iterates several times
 * (65536 times) to create the hash. The purpose of this is to delay a possible attack to retrieve the password value. The salt value makes sure
 * that the encrypted result differs from other AES implementations using the same password (to prevent rainbow table attacks).
 * </p>
 * 
 * @author Reinier
 *
 */
public class AESEncrypter {
	/** 16 bytes used as initialization vector for the encryption ciphers
	 * 
	 * Created using:
	 * 		byte[] iv = new byte[ecipher.getBlockSize()];
	 * 		new SecureRandom().nextBytes(iv);
	 */
	private static final byte[] IV = {-25, -94, -107, 99, -31, -10, -53, 13, -75, -42, 126, 99, 49, 99, 120, 46};
	
	/** Salt value used for encryption */
	private static final byte[] SALT_ENCRYPT = {-4, 87, -1, 100, 63, -20, -43, 49, -117, -27, -24, 126, 123, -41, 38, 93};
	
	/** Salt value used to hash the master password set in the preferences */
	private static final byte[] SALT_PASS_HASH = {74, 42, -110, -125, 126, 56, 28, -10, 43, -24, -75, -65, -34, -9, -53, -117};
			
	/** Iterations and key length for the PBKDF2WithHmacSHA1 algorithm */
	private static final int ITERATION_COUNT = 1000;
	private static final int KEY_LENGTH = 128;
	
	/** Encrypt and decrypt ciphers */
	private Cipher ecipher;
	private Cipher dcipher;
	
	/**
	 * Initialized the encryption and decryption ciphers using the given password. The password is mixed with a salt value and then hashed
	 * using the PBKDF2WithHmacSHA1 algorithm. The outcome is used as password for the AES encryption.
	 * 
	 * @param passPhrase password
	 * @throws Exception
	 */
	public AESEncrypter(String passPhrase) throws Exception {
		// mix the password and salt into a hash code using PBKDF2WithHmacSHA1. The hash code will be used as password for AES
		SecretKey key = hashPassword(passPhrase, SALT_ENCRYPT);
		SecretKey secret = new SecretKeySpec(key.getEncoded(), "AES");

		// use a hard-coded initialization vector, otherwise a random value is used which means the encrypted result is different every time
		// the program runs (and that it cannot be decrypted anymore)
		IvParameterSpec ivSpec = new IvParameterSpec(IV);
		
		ecipher = Cipher.getInstance("AES/CBC/PKCS5Padding");
		ecipher.init(Cipher.ENCRYPT_MODE, secret, ivSpec);

		dcipher = Cipher.getInstance("AES/CBC/PKCS5Padding");
		//byte[] iv = ecipher.getParameters().getParameterSpec(IvParameterSpec.class).getIV();
		dcipher.init(Cipher.DECRYPT_MODE, secret, ivSpec);
	}

	/**
	 * Creates a hash of the given password using the PBKDF2WithHmacSHA1 algorithm. The given password is mixed with the salt value into the hash. The PBKDF2 mechanism iterates 65536 times to calculate the secure hash.
	 * 
	 * @param password
	 * @return
	 * @throws NoSuchAlgorithmException
	 * @throws InvalidKeySpecException
	 * 
	 * @see http://adambard.com/blog/3-wrong-ways-to-store-a-password/
	 */
	private static SecretKey hashPassword(String password, byte[] salt) throws NoSuchAlgorithmException, InvalidKeySpecException {
		// mix the password and salt into a hash code using PBKDF2WithHmacSHA1. The hash code will be used as password for AES
		// see http://adambard.com/blog/3-wrong-ways-to-store-a-password/
		SecretKeyFactory factory = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA1");
		KeySpec spec = new PBEKeySpec(password.toCharArray(), salt, ITERATION_COUNT, KEY_LENGTH);
		return factory.generateSecret(spec);
	}

	/**
	 * Returns the hash of the password using the PBKDF2WithHmacSHA1 algorithm and converts it to a Base64 String.
	 * 
	 * @param password
	 * @return Base64 string representing the hash of the given password
	 * 
	 * @throws NoSuchAlgorithmException
	 * @throws InvalidKeySpecException
	 * 
	 * @see {@link #hashPassword(String)}
	 */
	public static String hashPasswordToString(String password) throws NoSuchAlgorithmException, InvalidKeySpecException {
		SecretKey key = hashPassword(password, SALT_PASS_HASH);
		return Base64.encode(key.getEncoded());
	}

	/**
	 * Encrypts a plain text using the cipher created in the constructor of this class.
	 * 
	 * @param plainText plain text
	 * @return cipher text
	 * 
	 * @throws Exception
	 */
	public String encrypt(String plainText) throws Exception {
		if (plainText == null) {
			plainText = "";
		}
		byte[] bytes = plainText.getBytes("UTF8");
		byte[] encrypted = ecipher.doFinal(bytes);
		return Base64.encode(encrypted);
	}
	
	/**
	 * Decrypts a text using the cipher created in the constructor of this class.
	 * 
	 * @param encryptedText plain text
	 * @return plain text
	 * 
	 * @throws Exception
	 */
	public String decrypt(String encryptedText) throws Exception {
		if (encryptedText == null) {
			throw new NullPointerException();
		}
		byte[] bytes = Base64.decode(encryptedText);
		byte[] decrypted = dcipher.doFinal(bytes);
		return new String(decrypted, "UTF8");
	}

	public static void main(String[] args) throws Exception {

		System.out.println("test!" + System.currentTimeMillis());
		
		String message = null;
		String password = "PASSWORD";

		AESEncrypter encrypter = new AESEncrypter(password);
		String encrypted = encrypter.encrypt(message);
		String decrypted = encrypter.decrypt(encrypted);
		
		System.out.println("Encrypt(\"" + message + "\", \"" + password + "\") = \"" + encrypted + "\"");
		System.out.println("Decrypt(\"" + encrypted + "\", \"" + password + "\") = \"" + decrypted + "\"");
		System.out.println("Hash: " + encrypter.hashPasswordToString(password));
		
		
		byte[] iv = new byte[16];
		new SecureRandom().nextBytes(iv);
		System.out.println("Bytes: " + Arrays.toString(iv));
		
	}
}