package com.perdemia.permissionanalyzer.application.preferences;

import java.io.IOException;
import java.net.Authenticator;
import java.net.InetSocketAddress;
import java.net.PasswordAuthentication;
import java.net.Proxy;
import java.net.Proxy.Type;
import java.net.ProxySelector;
import java.net.SocketAddress;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class CustomProxySelector extends ProxySelector {
	private static final Logger logger = LoggerFactory.getLogger(CustomProxySelector.class.getName());
	private static ProxySelector defaultSystemProxy;
	
	/** remember the default proxy */
	static {
		defaultSystemProxy = ProxySelector.getDefault();
		logSystemProxy();
	}
	
	@Override
	public void connectFailed(final URI uri, final SocketAddress sa, final IOException ioe) {
		logger.error("connectFailed for uri " + uri.toString(), ioe);
	}

	@Override
	public List<Proxy> select(final URI uri) {
		logger.debug("Get proxy for URL: " + uri.toString());
		String proxyHost = PreferenceHelper.getInstance().getProxyHost();
		int proxyPort = PreferenceHelper.getInstance().getProxyPort();
		
		if (proxyPort == 0) {
			logger.info("Using default proxy port 80.");
			proxyPort = 80;
		}
		
		List<Proxy> result = new ArrayList<Proxy>();
		Proxy proxy = new Proxy(Type.SOCKS, new InetSocketAddress(proxyHost, proxyPort));
		result.add(proxy);
		return result;
	}
	
	/**
	 * Read the preferences using {@link PreferenceHelper} and set the proxy using <code>ProxySelector.setDefault</code> and
	 * (if set) the authentication info using <code>Authenticator.setDefault</code>.
	 */
	public static void setProxy() {
		PreferenceHelper prefs = PreferenceHelper.getInstance();
		
		// proxy applicable?
		if (prefs.getProxyHost() != null && prefs.getProxyHost().length() > 0) {
			logger.info("Setting proxy: " + prefs.getProxyHost() + ", port: " + prefs.getProxyPort());
			
			ProxySelector.setDefault(new CustomProxySelector());
			
			final String proxyUsername = prefs.getProxyUsername();
			final String proxyPassword = prefs.getProxyPassword();
			
			if (proxyUsername != null && proxyUsername.length() > 0) {
				Authenticator.setDefault(new Authenticator() {
				    @Override
				    protected PasswordAuthentication getPasswordAuthentication() {
		               return new PasswordAuthentication(proxyUsername, proxyPassword.toCharArray());
				    }
				});
			} else {
				Authenticator.setDefault(null);
			}
		} else {
			if (prefs.isUseSystemProxy()) {
				// use default system proxy
				logger.info("Using default system proxy");
				logSystemProxy();
				ProxySelector.setDefault(defaultSystemProxy);
			} else {
				// use no proxy
				logger.info("Using no proxy");
				ProxySelector.setDefault(null);
			}
			Authenticator.setDefault(null);
		}
	}
	
	private static void logSystemProxy() {
		try {
			if (defaultSystemProxy != null) {
				List<Proxy> proxies = defaultSystemProxy.select(new URI("http://www.google.com"));
				if (proxies != null && proxies.size() > 0) {
					logger.info("System proxy address: " + proxies.get(0).address());	
					logger.info("System proxy type: " + proxies.get(0).type());
				}
			} else {
				logger.info("System proxy: null");
			}
		} catch (URISyntaxException e) {
			logger.error("Error while logging system proxy.", e);
		}
	}
}
