package com.perdemia.permissionanalyzer.application.preferences;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.preferences.IEclipsePreferences;
import org.eclipse.core.runtime.preferences.InstanceScope;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.preference.PreferenceDialog;
import org.eclipse.swt.widgets.Display;
import org.h2.tools.ChangeFileEncryption;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.AlertDialog;
import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.model.DirectoryInfo;
import com.perdemia.permissionanalyzer.application.model.DirectoryInfo.SourceType;
import com.perdemia.permissionanalyzer.application.model.LdapConnectionInfo;
import com.perdemia.permissionanalyzer.application.model.LdapConnectionInfo.Authentication;
import com.perdemia.permissionanalyzer.application.model.LdapConnectionInfo.Encryption;
import com.perdemia.permissionanalyzer.application.model.LdapOUInfo;
import com.perdemia.permissionanalyzer.application.model.LdapOUInfo.SearchObjectType;
import com.perdemia.permissionanalyzer.application.model.LdapOUInfo.SearchScope;
import com.perdemia.permissionanalyzer.application.swt.ExplainLabel;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.util.DBUtil;
import com.perdemia.permissionanalyzer.scanner.ldap.LdapConnectorImpl;
import com.unboundid.util.Base64;

public class PreferenceHelper {
	private static PreferenceHelper instance;
	private final static String PREF_ID = "com.perdemia.permissionanalyzer.preferences";
	private static Logger logger = LoggerFactory.getLogger(PreferenceHelper.class.getName());
	
	/** Preferences set in the Preference Dialog */
	private IEclipsePreferences prefs;
	private List<LdapConnectionInfo> ldapConnectionsList;
	private String proxyHost;
	private int proxyPort;
	private String proxyUsername;
	private String proxyPassword;
	private boolean useSystemProxy;
	private String smtpHost;
	private int smtpPort;
	private int smtpType;
	private String smtpUsername;
	private String smtpPassword;
	private String smtpFrom;
	private boolean smtpAttachReports;
	private String reportEmailTemplate;
	private String policyEmailTemplate;
	private String reportEmailSubject;
	private String policyEmailSubject;
	private boolean checkForUpdates;
	private boolean encryptDatabase;
	private boolean currentEncryptDatabase;
	private String logLevels;
	private String ldapAccountNameAttribute; //"sAMAccountName"
	private String ldapUserDisplayNameAttribute; //"displayName"
	private String ldapGroupDisplayNameAttribute; //"description"
	private String ldapGroupTypeAttribute; //"groupType"
	private String ldapDNAttribute; //"distinguishedName"
	private String ldapPrimaryGroupIdAttribute; //"primaryGroupID"
	private String ldapGroupMemberAttribute; //"member"
	private String ldapMemberOfAttribute; //"memberOf"
	private boolean debugLdap;
	private boolean followReferrals;
	private int dbConnectionType;
	private String dbConnectionString;
	private String dbConnectionDriver;
	private String dbConnectionUsername;
	private String dbConnectionPassword;
	private int batchSizeFiles;
	private int batchSizeAces;
	private int batchSizeMembers;
	private int batchSizeMemberships;
	private boolean calculateOwnershipRatioInOverview;
	private boolean calculateOwnershipRatioInEffectivePermissionsTab;
	private String logoFileName;
	private String logoBase64;
	private int logoWidth;
	private int logoHeight;
	private int logoLocationTop;
	private int logoLocationRight;
	private String csvSeparator;
	private int maxFileExportMegaBytes;
	private boolean useUserDisplayNameInReport;
	
	/** Preferences set in the Scan View */
	private List<DirectoryInfo> directoryList;
	private List<LdapOUInfo> ldapOUList;
	
	/** The Scan View adds a listener to react on changes in the Preference Dialog */
	private List<PreferenceListener> listeners = new ArrayList<PreferenceListener>();
	
	/** password given by the user in the login screen. In case this parameter is null or empty, a default password will be used. */
	private String masterPassword = "";
	private String currentMasterPassword = "";
	
	/** Preference name of the master password hash. All other key names are encrypted, but this preference must be available without encryption. It is used to check
	 *  the (encryption) password entered by the user. The preference value is no secret because it is a hash, but still obfuscate the name ;)
	 */
	private final static String MASTER_PASSWORD_HASH_KEY = "KjeUEH*3@763y2UAkdj3";
	
	/** Base64 encoded byte array used as default password for encrypting the preferences 
	 *  The user can override this password using the preferences dialog {@link PreferenceDialog}.
	 */
	private String defaultEncryptionPassword;
	
	/**
	 * Flag that indicates if the first use wizard should be displayed after the application starts up
	 */
	private boolean showFirstUseWizard;
	
	/**
	 * Flag that indicates if the license agreement should be displayed after the application starts up
	 */
	private boolean showLicenseAgreement;
	
	/**
	 * Flag that indicates what {@link #AlertDialog} dialogs should be displayed (based on the ordinal of {@link AlertDialog#DIALOG_TYPE})
	 */
	private List<Boolean> showAlertDialogs;
	
	/**
	 * Flag that indicates what {@link #ExplainLabel} dialogs should be displayed (based on the ordinal of {@link ExplainLabel#ExplainType})
	 */
	private List<Boolean> showExplainLabels;
	
	
	/**
	 * Checkbox on the Scan View that indicates the data needs to be cleared before the scan starts
	 */
	private boolean truncateData;
	
	/** Hidden preference to hide to Modify ACE button in the tabs */
	private boolean hideModifyAce;
	
	private AESEncrypter encrypter;
	private final static String DEFAULT_DB_PASSWORD = "permissionanalyzer";
	public final static String DEFAULT_CONNECTION_STRING = "jdbc:h2:[app_dir]H2DB;MAX_COMPACT_TIME=2000;ALIAS_COLUMN_NAME=TRUE;CACHE_SIZE=524288;MV_STORE=FALSE;MVCC=FALSE;"; //512 MB
	
	/**
	 * Constructor, retrieves the preferences from the Eclipse preference store.
	 */
	private PreferenceHelper(){
		prefs = InstanceScope.INSTANCE.getNode(PREF_ID);
		defaultEncryptionPassword = Base64.encode(new byte[]{71, 120, -1, 126, 18, 27, -101, -119, 94, 59, 108, 76, 64, 64, 92, 113});
	}
	
	/**
	 * Retrieves the master password hash from the preferences. This master password can be set by the user in the dialog {@link PreferenceDialog}.
	 * 
	 * @return the password hash from the preferences or <code>null</code> is case no password has been set.
	 */
	public String getMasterPasswordHash(){
		return prefs.get(MASTER_PASSWORD_HASH_KEY, null);
	}
	
	/**
	 * Loads the preferences from the <code>IEclipsePreferences</code> into the class variables. All preferences (names and values) are encrypted
	 * using the given password.
	 * 
	 */
	public void loadPreferences() {
		try {
			currentMasterPassword = masterPassword;
			String encryptionPassword = masterPassword;
			if (nullOrEmpty(encryptionPassword)) {
				encryptionPassword = defaultEncryptionPassword;
			}		
			
			encrypter = new AESEncrypter(encryptionPassword);
			
			proxyHost = loadPref("proxyHost", "");
			proxyPort = Integer.parseInt(loadPref("proxyPort", "0"));
			proxyUsername = loadPref("proxyUsername", "");
			proxyPassword = loadPref("proxyPassword", "");
			useSystemProxy = Boolean.parseBoolean(loadPref("useSystemProxy", "true"));
			smtpHost = loadPref("smtpHost", "");
			smtpPort = Integer.parseInt(loadPref("smtpPort", "0"));
			smtpType = Integer.parseInt(loadPref("smtpType", "0"));
			smtpUsername = loadPref("smtpUsername", "");
			smtpPassword = loadPref("smtpPassword", "");
			smtpFrom = loadPref("smtpFrom", "");
			smtpAttachReports = Boolean.parseBoolean(loadPref("smtpAttachReports", "false"));
			reportEmailTemplate = loadPref("reportEmailTemplate", "Dear Sir/Madam,\r\n\r\nPlease be informed that the report [report_name] has been renewed. The report can be found at [report_path] and has the following description:\r\n\r\n[report_description]");
			policyEmailTemplate = loadPref("policyEmailTemplate", "Dear Sir/Madam,\r\n\r\nYou receive this e-mail because the report [report_name] contains [report_threshold] or more files or folders. The report can be found at [report_path] and has the following description:\r\n\r\n[report_description]");
			reportEmailSubject = loadPref("reportEmailSubject", "Permission Analyzer report: [report_name]");
			policyEmailSubject = loadPref("policyEmailSubject", "Permission Analyzer alert: [report_name]");
			checkForUpdates = Boolean.parseBoolean(loadPref("checkForUpdates", "true"));
			encryptDatabase = Boolean.parseBoolean(loadPref("encryptDatabase", "false"));
			logLevels = loadPref("logLevels", "");
			currentEncryptDatabase = encryptDatabase;
			ldapAccountNameAttribute = loadPref("ldapAccountNameAttribute", "sAMAccountName");
			ldapUserDisplayNameAttribute = loadPref("ldapUserDisplayNameAttribute", "displayName");
			ldapGroupDisplayNameAttribute = loadPref("ldapGroupDisplayNameAttribute", "description");
			ldapGroupTypeAttribute = loadPref("ldapGroupTypeAttribute", "groupType");
			ldapDNAttribute = loadPref("ldapDNAttribute", "distinguishedName");
			ldapPrimaryGroupIdAttribute = loadPref("ldapPrimaryGroupIdAttribute", "primaryGroupID");
			ldapGroupMemberAttribute = loadPref("ldapGroupMemberAttribute", "member");
			ldapMemberOfAttribute = loadPref("ldapMemberOfAttribute", "memberOf");
			debugLdap = Boolean.parseBoolean(loadPref("debugLdap", "false"));
			calculateOwnershipRatioInOverview = Boolean.parseBoolean(loadPref("calculateOwnershipRatioInOverview", "false"));
			calculateOwnershipRatioInEffectivePermissionsTab = Boolean.parseBoolean(loadPref("calculateOwnershipRatioInEffectivePermissionsTab", "false"));
			logoFileName = loadPref("logoFileName", "");
			logoBase64 = loadPref("logoBase64", "");
			logoWidth = Integer.parseInt(loadPref("logoWidth", "90"));
			logoHeight = Integer.parseInt(loadPref("logoHeight", "90"));
			logoLocationTop = Integer.parseInt(loadPref("logoLocationTop", "40"));
			logoLocationRight = Integer.parseInt(loadPref("logoLocationRight", "40"));
			csvSeparator = loadPref("csvSeparator", ",");
			maxFileExportMegaBytes = Integer.parseInt(loadPref("maxFileExportMegaBytes", "0"));
			useUserDisplayNameInReport = Boolean.parseBoolean(loadPref("useUserDisplayNameInReport", "true"));
			
			followReferrals = Boolean.parseBoolean(loadPref("followReferrals", "true"));
			showFirstUseWizard = Boolean.parseBoolean(loadPref("showFirstUseWizard", "true"));
			showLicenseAgreement = Boolean.parseBoolean(loadPref("showLicenseAgreement", "true"));
			truncateData = Boolean.parseBoolean(loadPref("truncateData", "true"));
			dbConnectionType = Integer.parseInt(loadPref("dbConnectionType", "0"));
			dbConnectionString = loadPref("dbConnectionString", DEFAULT_CONNECTION_STRING);
			dbConnectionDriver = loadPref("dbConnectionDriver", "org.h2.Driver");
			dbConnectionUsername = loadPref("dbConnectionUsername", "");
			dbConnectionPassword = loadPref("dbConnectionPassword", "");
			batchSizeFiles = Integer.parseInt(loadPref("batchSizeFiles", "1000"));
			batchSizeAces = Integer.parseInt(loadPref("batchSizeAces", "5000"));
			batchSizeMembers = Integer.parseInt(loadPref("batchSizeMembers", "1000"));
			batchSizeMemberships = Integer.parseInt(loadPref("batchSizeMemberships", "10000"));
			hideModifyAce = Boolean.parseBoolean(loadPref("hideModifyAce", "false"));
			
			// get all alert dialog display flags
			showAlertDialogs = new ArrayList<Boolean>();
			for (int counter = 0; counter < AlertDialog.AlertType.values().length; counter++) {
				showAlertDialogs.add(Boolean.parseBoolean(loadPref("showAlertDialog" + counter, "true")));
			}

			// get all explain label display flags
			showExplainLabels = new ArrayList<Boolean>();
			for (int counter = 0; counter < ExplainLabel.ExplainType.values().length; counter++) {
				showExplainLabels.add(Boolean.parseBoolean(loadPref("showExplainLabel" + counter, "true")));
			}
			
			// get all ldap connections
			ldapConnectionsList = new ArrayList<LdapConnectionInfo>();
			for (int counter = 0; counter < 1000; counter++) {
				String connectionName = loadPref("ldapConnectionName" + counter, "");
				if (!nullOrEmpty(connectionName)) {
					LdapConnectionInfo connectionInfo = new LdapConnectionInfo();
					connectionInfo.setConnectionId(Integer.parseInt(loadPref("ldapConnectionId" + counter, "0"))); 
					connectionInfo.setConnectionName(loadPref("ldapConnectionName" + counter, ""));
					connectionInfo.setHostName(loadPref("ldapHostName" + counter, ""));
					connectionInfo.setHostPort(Integer.parseInt(loadPref("ldapHostPort" + counter, "0"))); 
					connectionInfo.setUsername(loadPref("ldapUsername" + counter, ""));
					connectionInfo.setPassword(loadPref("ldapPassword" + counter, ""));
					connectionInfo.setEncryption(Encryption.values()[((Integer.parseInt(loadPref("ldapEncryption" + counter, "0"))))]);
					connectionInfo.setAuthentication(Authentication.values()[((Integer.parseInt(loadPref("ldapAuthentication" + counter, "0"))))]);
					
					// SASL / GSSAPI settings
					connectionInfo.setQualityOfProtectionAuthAllowed(Boolean.parseBoolean(loadPref("ldapQualityOfProtectionAuthAllowed" + counter, "true")));
					connectionInfo.setQualityOfProtectionAuthIntAllowed(Boolean.parseBoolean(loadPref("ldapQualityOfProtectionAuthIntAllowed" + counter, "true")));
					connectionInfo.setQualityOfProtectionAuthConfAllowed(Boolean.parseBoolean(loadPref("ldapQualityOfProtectionAuthConfAllowed" + counter, "true")));
					connectionInfo.setRealm(loadPref("ldapRealm" + counter, ""));
					
					// GSSAPI settings
					connectionInfo.setConfigPath(loadPref("ConfigPath" + counter, ""));
					connectionInfo.setUseTicketCache(Boolean.parseBoolean(loadPref("ldapUseTicketCache" + counter, "true")));
					connectionInfo.setKDCAddress(loadPref("KDCAddress" + counter, ""));
					
					ldapConnectionsList.add(connectionInfo);
				} else {
					break;
				}
			}
			
			// get all directories configured in the Scan View
			directoryList = new ArrayList<DirectoryInfo>();
			for (int counter = 0; counter < 1000; counter++) {
				String directoryPath = loadPref("directoryPath" + counter, "");
				if (!nullOrEmpty(directoryPath)) {
					DirectoryInfo directoryInfo = new DirectoryInfo();
					directoryInfo.setDirectoryId(Integer.parseInt(loadPref("directoryId" + counter, "0")));
					directoryInfo.setPath(directoryPath);
					directoryInfo.setNumberOfSubdirectories(Integer.parseInt(loadPref("numberOfSubdirectories" + counter, "0")));
					directoryInfo.setIncludeFiles(Boolean.parseBoolean(loadPref("includeFiles" + counter, "false")));
					directoryInfo.setIncludeInheritedPermissions(Boolean.parseBoolean(loadPref("includeInheritedPermissions" + counter, "true")));
					directoryInfo.setSourceType(SourceType.values()[Integer.parseInt(loadPref("sourceType" + counter, "0"))]);
					directoryInfo.setFollowSymbolicLinks(Boolean.parseBoolean(loadPref("followSymbolicLinks" + counter, "true")));
					directoryInfo.setScanLocalGroups(Boolean.parseBoolean(loadPref("scanLocalGroups" + counter, "true")));
					directoryInfo.setIncludeHiddenShares(Boolean.parseBoolean(loadPref("includeHiddenShares" + counter, "true")));
					directoryInfo.setIncludeHiddenSharedDrives(Boolean.parseBoolean(loadPref("includeHiddenSharedDrives" + counter, "false")));
					directoryInfo.setExcludeShares(loadPref("excludeShares" + counter, "IPC$, PRINT$, ADMIN$"));
					directoryInfo.setServerLoginUsername(loadPref("serverLoginUsername" + counter, ""));
					directoryInfo.setServerLoginPassword(loadPref("serverLoginPassword" + counter, ""));
					directoryInfo.setServerLoginDomain(loadPref("serverLoginDomain" + counter, ""));
					directoryInfo.setEnabled(Boolean.parseBoolean(loadPref("directoryEnabled" + counter, "true")));
					directoryList.add(directoryInfo);
				} else {
					break;
				}
			}
			
			// get all ldap containers configured in the Scan View
			ldapOUList = new ArrayList<LdapOUInfo>();
			for (int counter = 0; counter < 1000; counter++) {
				String baseOUPath = loadPref("baseOUPath" + counter, "");
				if (!nullOrEmpty(baseOUPath)) {
					LdapOUInfo ldapOUInfo = new LdapOUInfo();
					ldapOUInfo.setLdapOUId(Integer.parseInt(loadPref("ldapOUId" + counter, "0")));
					ldapOUInfo.setBaseOU(baseOUPath);
					int connectionId = Integer.parseInt(loadPref("ldapOUConnectionId" + counter, "0"));
					//lookup the ldap connection
					for (LdapConnectionInfo connectionInfo : ldapConnectionsList) {
						if (connectionInfo.getConnectionId() == connectionId) {
							ldapOUInfo.setConnection(connectionInfo);
							break;
						}
					}
					
					String searchScope = loadPref("searchScope" + counter, "");
					ldapOUInfo.setSearchScope(SearchScope.valueOf(searchScope));
					
					String searchObjectType = loadPref("searchObjectType" + counter, "");
					ldapOUInfo.setSearchObjectType(SearchObjectType.valueOf(searchObjectType));
					
					ldapOUInfo.setEnabled(Boolean.parseBoolean(loadPref("ldapContainerEnabled" + counter, "true")));
					
					ldapOUList.add(ldapOUInfo);
				} else {
					break;
				}
			}
			
			// set proxy settings
			CustomProxySelector.setProxy();
			
			// apply log levels
			if (logLevels.length() > 0) {
				String[] logPackages = logLevels.split(";");
				for (String logPackage : logPackages) {
					if (logPackage.contains("=")) {
						String packageName = logPackage.split("=")[0];
						String logLevel = logPackage.split("=")[1];
						logger.info("Setting log level for package " + packageName + " to level " + logLevel);
						java.util.logging.Logger packageLogger = java.util.logging.Logger.getLogger(packageName);
						packageLogger.setLevel(java.util.logging.Level.parse(logLevel));
					}
				}
			}
			
			// check if we need to add a default LDAP connection
			if (ldapConnectionsList.isEmpty()) {
				try {
				LdapConnectionInfo connectionInfo = new LdapConnectionInfo();
				connectionInfo.setConnectionId(0); 
				connectionInfo.setConnectionName(LdapConnectorImpl.getDomainControllerName());
				connectionInfo.setHostName(LdapConnectorImpl.getDomainControllerName());
				connectionInfo.setHostPort(389); 
				connectionInfo.setUsername(LdapConnectorImpl.getDomainName() + "\\" + LdapConnectorImpl.getCurrentUsername());
				connectionInfo.setPassword("");
				connectionInfo.setEncryption(Encryption.PLAIN);
				connectionInfo.setAuthentication(Authentication.SIMPLE);
					ldapConnectionsList.add(connectionInfo);
				} catch (Exception ex) {
					logger.error("Error while adding default domain controller: " + ex.getMessage(), ex);
				}
				
			}
		
		} catch (Exception ex) {
			SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error retrieving preferences", "Error while retrieving the preferences: " + ex.getMessage());
        	logger.error("Error while retrieving the preferences: " + ex.getMessage(), ex);
		}
	}
	
	public String getDefaultDatabasePassword() {
		try {
			if (!nullOrEmpty(masterPassword)) {
				return encrypter.encrypt(DEFAULT_DB_PASSWORD);
			} else {
				return DEFAULT_DB_PASSWORD;
			}
		} catch (Exception e) {
			logger.error("Error while getting database password: " + e.getMessage(), e);
			return DEFAULT_DB_PASSWORD;
		}
	}
	
	private String loadPref(String name, String def) {
		try {
			return encrypter.decrypt(prefs.get(encrypter.encrypt(name), null));
		} catch (NullPointerException ex) {
			return def;
		} catch (Exception e) {
			logger.error("Error while loading preference " + name + ": " + e.getMessage(), e);
			return def;
		}
	}
	
	/**
	 * Creates a instance in case it doesn't exist, otherwise it returns the already existing instance (Singleton)
	 * 
	 * @return instance
	 */
	public static PreferenceHelper getInstance() {
		if (instance == null) {
			instance = new PreferenceHelper();
		}
		return instance;
	}
	
	/**
	 * Stores all the variables in this class as preferences in the <code><code>IEclipsePreferences</code></code>.
	 * All preferences (names and values) are encrypted using the given password.
	 * 
	 * @param password password given by the user in the login screen. In case this parameter is null or empty, a default password will be used.
	 */
	public void storePreferences(final boolean notifyListeners) {
		Display.getDefault().syncExec(new Runnable(){
			public void run() {
				ProgressMonitorDialog progressDialog = new ProgressMonitorDialog(Display.getDefault().getActiveShell());
				try {
					progressDialog.run(true, false, new IRunnableWithProgress(){
						@Override
						public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
					
							try {
								String encryptionPassword = masterPassword;
								if (nullOrEmpty(encryptionPassword)) {
									encryptionPassword = defaultEncryptionPassword;
								}
								
								encrypter = new AESEncrypter(encryptionPassword);
								
								// clear all the current settings (ldap connections may have been removed)
								prefs.clear();
								
								storePref("proxyHost", proxyHost);
								storePref("proxyPort", Integer.toString(proxyPort));
								storePref("proxyUsername", proxyUsername);
								storePref("proxyPassword", proxyPassword);
								storePref("useSystemProxy", Boolean.toString(useSystemProxy));
								storePref("smtpHost", smtpHost);
								storePref("smtpPort", Integer.toString(smtpPort));
								storePref("smtpType", Integer.toString(smtpType));
								storePref("smtpUsername", smtpUsername);
								storePref("smtpPassword", smtpPassword);
								storePref("smtpAttachReports", Boolean.toString(smtpAttachReports));			
								storePref("smtpFrom", smtpFrom);
								storePref("reportEmailTemplate", reportEmailTemplate);
								storePref("policyEmailTemplate", policyEmailTemplate);
								storePref("reportEmailSubject", reportEmailSubject);
								storePref("policyEmailSubject", policyEmailSubject);
								storePref("checkForUpdates", Boolean.toString(checkForUpdates));
								storePref("encryptDatabase", Boolean.toString(encryptDatabase));
								storePref("logLevels", logLevels);
								storePref("ldapAccountNameAttribute", ldapAccountNameAttribute);
								storePref("ldapUserDisplayNameAttribute", ldapUserDisplayNameAttribute);
								storePref("ldapGroupDisplayNameAttribute", ldapGroupDisplayNameAttribute);
								storePref("ldapGroupTypeAttribute", ldapGroupTypeAttribute);
								storePref("ldapDNAttribute", ldapDNAttribute);
								storePref("ldapPrimaryGroupIdAttribute", ldapPrimaryGroupIdAttribute);
								storePref("ldapGroupMemberAttribute", ldapGroupMemberAttribute);
								storePref("ldapMemberOfAttribute", ldapMemberOfAttribute);
								storePref("debugLdap", Boolean.toString(debugLdap));	
								storePref("followReferrals", Boolean.toString(followReferrals));
								storePref("showFirstUseWizard", Boolean.toString(showFirstUseWizard));
								storePref("showLicenseAgreement", Boolean.toString(showLicenseAgreement));
								storePref("truncateData", Boolean.toString(truncateData));
								storePref("dbConnectionType", Integer.toString(dbConnectionType));
								storePref("dbConnectionString", dbConnectionString);
								storePref("dbConnectionDriver", dbConnectionDriver);
								storePref("dbConnectionUsername", dbConnectionUsername);
								storePref("dbConnectionPassword", dbConnectionPassword);
								storePref("batchSizeFiles", Integer.toString(batchSizeFiles));
								storePref("batchSizeAces", Integer.toString(batchSizeAces));
								storePref("batchSizeMembers", Integer.toString(batchSizeMembers));
								storePref("batchSizeMemberships", Integer.toString(batchSizeMemberships));
								storePref("calculateOwnershipRatioInOverview", Boolean.toString(calculateOwnershipRatioInOverview));
								storePref("calculateOwnershipRatioInEffectivePermissionsTab", Boolean.toString(calculateOwnershipRatioInEffectivePermissionsTab));
								storePref("hideModifyAce", Boolean.toString(hideModifyAce));
								storePref("logoFileName", logoFileName);
								storePref("logoBase64", logoBase64);								
								storePref("logoWidth", Integer.toString(logoWidth));
								storePref("logoHeight", Integer.toString(logoHeight));
								storePref("logoLocationTop", Integer.toString(logoLocationTop));
								storePref("logoLocationRight", Integer.toString(logoLocationRight));
								storePref("csvSeparator", csvSeparator);
								storePref("maxFileExportMegaBytes", Integer.toString(maxFileExportMegaBytes));
								storePref("useUserDisplayNameInReport", Boolean.toString(useUserDisplayNameInReport));
																
								// store all alert dialog display flags
								for (int counter = 0; counter < showAlertDialogs.size(); counter++) {
									storePref("showAlertDialog" + counter, Boolean.toString(showAlertDialogs.get(counter)));
								}

								// store all explain label display flags
								for (int counter = 0; counter < showExplainLabels.size(); counter++) {
									storePref("showExplainLabel" + counter, Boolean.toString(showExplainLabels.get(counter)));
								}
																
								//store all ldap connections
								int counter = 0;
								for (LdapConnectionInfo connectionInfo : ldapConnectionsList) {
									storePref("ldapConnectionId" + counter, Integer.toString(connectionInfo.getConnectionId()));
									storePref("ldapConnectionName" + counter, connectionInfo.getConnectionName());
									storePref("ldapHostName" + counter, connectionInfo.getHostName());
									storePref("ldapHostPort" + counter, Integer.toString(connectionInfo.getHostPort()));
									storePref("ldapUsername" + counter, connectionInfo.getUsername());
									if (!connectionInfo.isOnlyStorePasswordInMemory()) {
										storePref("ldapPassword" + counter, connectionInfo.getPassword());
									}
									storePref("ldapEncryption" + counter, Integer.toString(connectionInfo.getEncryption().ordinal()));
									storePref("ldapAuthentication" + counter, Integer.toString(connectionInfo.getAuthentication().ordinal()));
									
									// SASL / GSSAPI settings
									storePref("ldapQualityOfProtectionAuthAllowed" + counter, Boolean.toString(connectionInfo.isQualityOfProtectionAuthAllowed()));
									storePref("ldapQualityOfProtectionAuthIntAllowed" + counter, Boolean.toString(connectionInfo.isQualityOfProtectionAuthIntAllowed()));
									storePref("ldapQualityOfProtectionAuthConfAllowed" + counter, Boolean.toString(connectionInfo.isQualityOfProtectionAuthConfAllowed()));
									storePref("ldapRealm" + counter, connectionInfo.getRealm());
									
									// GSSAPI settings
									storePref("ConfigPath" + counter, connectionInfo.getConfigPath());
									storePref("ldapUseTicketCache" + counter, Boolean.toString(connectionInfo.isUseTicketCache()));
									storePref("KDCAddress" + counter, connectionInfo.getKDCAddress());
									
									counter++;
								}
								
								//store all directories to scan
								counter = 0;
								for (DirectoryInfo directoryInfo : directoryList) {
									storePref("directoryId" + counter, Integer.toString(directoryInfo.getDirectoryId()));
									storePref("directoryPath" + counter, directoryInfo.getPath());
									storePref("numberOfSubdirectories" + counter, Integer.toString(directoryInfo.getNumberOfSubdirectories()));
									storePref("includeFiles" + counter, Boolean.toString(directoryInfo.isIncludeFiles()));
									storePref("includeInheritedPermissions" + counter, Boolean.toString(directoryInfo.isIncludeInheritedPermissions()));
									storePref("followSymbolicLinks" + counter, Boolean.toString(directoryInfo.isFollowSymbolicLinks()));
									storePref("scanLocalGroups" + counter, Boolean.toString(directoryInfo.isScanLocalGroups()));
									storePref("includeHiddenShares" + counter, Boolean.toString(directoryInfo.isIncludeHiddenShares()));
									storePref("includeHiddenSharedDrives" + counter, Boolean.toString(directoryInfo.isIncludeHiddenSharedDrives()));
									storePref("excludeShares" + counter, directoryInfo.getExcludeShares());
									storePref("directoryEnabled" + counter, Boolean.toString(directoryInfo.isEnabled()));
									storePref("sourceType" + counter, Integer.toString(directoryInfo.getSourceType().ordinal()));
									storePref("serverLoginUsername" + counter, directoryInfo.getServerLoginUsername());
									storePref("serverLoginPassword" + counter, directoryInfo.getServerLoginPassword());
									storePref("serverLoginDomain" + counter, directoryInfo.getServerLoginDomain());
									counter++;
								}
								
								//store all ldap containers to scan
								counter = 0;
								for (LdapOUInfo ouInfo : ldapOUList) {
									storePref("ldapOUId" + counter, Integer.toString(ouInfo.getLdapOUId()));
									storePref("baseOUPath" + counter, ouInfo.getBaseOU());
									storePref("ldapOUConnectionId" + counter, Integer.toString(ouInfo.getConnection().getConnectionId()));
									storePref("searchScope" + counter, ouInfo.getSearchScope().name());
									storePref("searchObjectType" + counter, ouInfo.getSearchObjectType().name());
									storePref("ldapContainerEnabled" + counter, Boolean.toString(ouInfo.isEnabled()));
									counter++;
								}
								
								//save the preferences
						        prefs.flush();
								
						        // notify the listeners
						        if (notifyListeners) {
							        for (PreferenceListener listener : listeners) {
							        	listener.preferencesChanged();
							        }
						        }
						        
						        // store the password hash
						        if (!nullOrEmpty(masterPassword)) {
									prefs.put(MASTER_PASSWORD_HASH_KEY, AESEncrypter.hashPasswordToString(masterPassword));
								}
						        
						        if (DBUtil.isLocalDB()) {
						        	String dbFolder;
						        	String path = dbConnectionString.substring(8, dbConnectionString.indexOf(";"));
									path = path.replace("[app_dir]", Platform.getLocation().toFile().getParent() + File.separator);
						        	int lastSlash = path.lastIndexOf("\\");
						        	if (lastSlash == -1) {
						        		lastSlash = path.lastIndexOf("/");
						        	}
						        	if (lastSlash > 0) {
						        		dbFolder = path.substring(0, lastSlash);
						        	} else {
						        		throw new Exception("Could not determine database folder: " + path);
						        	}
						        	
						        	//------------------------------------------------------
							        // STEP 1: application password cleared?
							        // OR the option "encrypt database" has been unchecked
							        //------------------------------------------------------
							        if ((!nullOrEmpty(currentMasterPassword) && nullOrEmpty(masterPassword))
							        	|| (currentEncryptDatabase && !encryptDatabase)) {
										
							        	if (nullOrEmpty(masterPassword)) {
							        		//reset the database user password
											DBObject.executeStatement("set password '" + DEFAULT_DB_PASSWORD + "'");
							        	}
							        	
							        	// decrypt the database		        	
							        	if (currentEncryptDatabase) {
											
											monitor.beginTask("Decrypting database...", IProgressMonitor.UNKNOWN);
			
							        		String[] args = new String[]{"-dir", dbFolder, "-db", "H2DB", "-cipher", "AES", "-decrypt", 
													AESEncrypter.hashPasswordToString(currentMasterPassword)};
											
											// close connection before decrypting
											DBUtil.closeDBConnection();
											long startTime = System.currentTimeMillis();
											new ChangeFileEncryption().runTool(args);
											logger.info("Done decrypting DB, duration: " + (System.currentTimeMillis() - startTime) + "ms");
											
											//success, update the preferences
											storePref("databaseEncrypted", "false");
											encryptDatabase = false;
									        prefs.flush();
									        
											DBUtil.getDBConnection();
							        	}
									}
							        
							        //------------------------------------------------------------------------------
									// STEP 2: if the master password has changed then change the database password as well
							        //------------------------------------------------------------------------------
									if (!currentMasterPassword.equals(masterPassword) && !nullOrEmpty(masterPassword)) {
										
										//change the database user password to offer access protection
										String newDatabasePassword = encrypter.encrypt(DEFAULT_DB_PASSWORD);
										DBObject.executeStatement("set password '" + newDatabasePassword + "'");
										
										//encrypt the database (or change encryption key) if required
										if (encryptDatabase) {
											
											monitor.beginTask("Encrypting database...", IProgressMonitor.UNKNOWN);
											
											//java -cp h2*.jar org.h2.tools.ChangeFileEncryption -dir ~ -db test -cipher AES -encrypt filepwd
											List<String> args = new ArrayList<String>(Arrays.asList("-dir", dbFolder, "-db", "H2DB", "-cipher", "AES", "-encrypt", 
													AESEncrypter.hashPasswordToString(masterPassword)));
											
											// decrypt first?
											if (!nullOrEmpty(currentMasterPassword)) {
												args.add("-decrypt");
												args.add(AESEncrypter.hashPasswordToString(currentMasterPassword));
											}
											
											// close connection before encrypting
											DBUtil.closeDBConnection();
											long startTime = System.currentTimeMillis();
											new ChangeFileEncryption().runTool(args.toArray(new String[0]));
											logger.info("Done encrypting DB, duration: " + (System.currentTimeMillis() - startTime) + "ms");
																
											//success, update the preferences
											storePref("databaseEncrypted", "true");
											encryptDatabase = true;
									        prefs.flush();
									        
									        DBUtil.getDBConnection();
										}
									}
									
									//-----------------------------------------------------------------------------------
									// STEP 3: is the database encryption preference enabled and is the password unchanged?
									// (decryption of the database is handled in step 1 and password changes in step 2)
									//-----------------------------------------------------------------------------------
									if (!currentEncryptDatabase && encryptDatabase && currentMasterPassword.equals(masterPassword)) {
										//only encrypt/decrypt DB is there is a password set (should always be the case). Password changes are handled in the IF statement above!!
										if (!nullOrEmpty(masterPassword)) {
											
											monitor.beginTask("Encrypting database...", IProgressMonitor.UNKNOWN);
											
											//encrypt DB
											String[] args = new String[]{"-dir", dbFolder, "-db", "H2DB", "-cipher", "AES", "-encrypt", 
													AESEncrypter.hashPasswordToString(masterPassword)};
															
											// close connection before changing the encryption
											DBUtil.closeDBConnection();
											long startTime = System.currentTimeMillis();
											new ChangeFileEncryption().runTool(args);
											logger.info("Done ecnrypting DB, duration: " + (System.currentTimeMillis() - startTime) + "ms");
																
											storePref("databaseEncrypted", "true");
											encryptDatabase = true;
									        prefs.flush();
									        
											DBUtil.getDBConnection();
										}
									}
						        }
								
								//save the preferences
						        prefs.flush();
						        
						        currentMasterPassword = masterPassword;
						        currentEncryptDatabase = encryptDatabase;
						        
						        // set proxy settings
						     	CustomProxySelector.setProxy();
						     	
						     	// apply log levels
								if (logLevels.length() > 0) {
									String[] logPackages = logLevels.split(";");
									for (String logPackage : logPackages) {
										if (logPackage.contains("=")) {
											String packageName = logPackage.split("=")[0];
											String logLevel = logPackage.split("=")[1];
											logger.info("Setting log level for package " + packageName + " to level " + logLevel);
											java.util.logging.Logger packageLogger = java.util.logging.Logger.getLogger(packageName);
											packageLogger.setLevel(java.util.logging.Level.parse(logLevel.toUpperCase()));
										}
									}
								}	
						
						    } catch (Exception ex) {
						    	//------------------------------------------------------------------------------------------------
						    	// restore the previous application password, because encryption or decryption might have failed.
						    	//------------------------------------------------------------------------------------------------
						    	encryptDatabase = currentEncryptDatabase;
						    	masterPassword = currentMasterPassword;
						    	        	
						    	try {
						        	storePref("isDatabaseEncrypted", Boolean.toString(currentEncryptDatabase));
						        	if (nullOrEmpty(currentMasterPassword)) {
										prefs.remove(MASTER_PASSWORD_HASH_KEY);
									} else {
										prefs.put(MASTER_PASSWORD_HASH_KEY, AESEncrypter.hashPasswordToString(currentMasterPassword));
									}
						    	
									prefs.flush();
						    	} catch (Exception e) {
									//bad luck
									logger.error("Error while restoring original password because saving preferences failed: " + ex.getMessage(), e);
								}
						    				
						    	SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error saving preferences", "Error while saving the preferences: " + ex.getMessage());
						    	logger.error("Error while saving the preferences: " + ex.getMessage(), ex);
						    	
						    }
						}	
					});
				} catch (InvocationTargetException | InterruptedException ex) {
					SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error saving preferences", "Error while saving the preferences: " + ex.getMessage());
			    	logger.error("Error while saving the preferences: " + ex.getMessage(), ex);
				}
			}
		});
	}
	
	public void addListener(PreferenceListener listener) {
		listeners.add(listener);
	}
	
	public void removeListener(PreferenceListener listener) {
		listeners.remove(listener);
	}
	
	private boolean nullOrEmpty(String value) {
		return value == null || value.length() == 0;
	}
	
	private void storePref(String name, String value) throws Exception {
		prefs.put(encrypter.encrypt(name), encrypter.encrypt(value));
	}

	public List<LdapConnectionInfo> getLdapConnectionsList() {
		return ldapConnectionsList;
	}

	public void setLdapConnectionsList(List<LdapConnectionInfo> ldapConnectionsList) {
		this.ldapConnectionsList = ldapConnectionsList;
	}

	public String getProxyHost() {
		return proxyHost;
	}

	public void setProxyHost(String proxyHost) {
		this.proxyHost = proxyHost;
	}

	public int getProxyPort() {
		return proxyPort;
	}

	public void setProxyPort(int proxyPort) {
		this.proxyPort = proxyPort;
	}

	public boolean isCheckForUpdates() {
		return checkForUpdates;
	}

	public void setCheckForUpdates(boolean checkForUpdates) {
		this.checkForUpdates = checkForUpdates;
	}

	public String getMasterPassword() {
		return masterPassword;
	}

	public void setMasterPassword(String masterPassword) {
		this.masterPassword = masterPassword;
	}

	public boolean isEncryptDatabase() {
		return encryptDatabase;
	}

	public void setEncryptDatabase(boolean encryptDatabase) {
		this.encryptDatabase = encryptDatabase;
	}

	public List<DirectoryInfo> getDirectoryList() {
		return directoryList;
	}

	public void setDirectoryList(List<DirectoryInfo> directoryList) {
		this.directoryList = directoryList;
	}

	public List<LdapOUInfo> getLdapOUList() {
		return ldapOUList;
	}

	public void setLdapOUList(List<LdapOUInfo> ldapOUList) {
		this.ldapOUList = ldapOUList;
	}

	public String getProxyUsername() {
		return proxyUsername;
	}

	public void setProxyUsername(String proxyUsername) {
		this.proxyUsername = proxyUsername;
	}

	public String getProxyPassword() {
		return proxyPassword;
	}

	public void setProxyPassword(String proxyPassword) {
		this.proxyPassword = proxyPassword;
	}

	public String getLogLevels() {
		return logLevels;
	}

	public void setLogLevels(String logLevels) {
		this.logLevels = logLevels;
	}

	public String getLdapAccountNameAttribute() {
		return ldapAccountNameAttribute;
	}

	public void setLdapAccountNameAttribute(String ldapAccountNameAttribute) {
		this.ldapAccountNameAttribute = ldapAccountNameAttribute;
	}

	public String getLdapUserDisplayNameAttribute() {
		return ldapUserDisplayNameAttribute;
	}

	public void setLdapUserDisplayNameAttribute(String ldapUserDisplayNameAttribute) {
		this.ldapUserDisplayNameAttribute = ldapUserDisplayNameAttribute;
	}

	public String getLdapGroupDisplayNameAttribute() {
		return ldapGroupDisplayNameAttribute;
	}

	public void setLdapGroupDisplayNameAttribute(
			String ldapGroupDisplayNameAttribute) {
		this.ldapGroupDisplayNameAttribute = ldapGroupDisplayNameAttribute;
	}

	public String getLdapGroupTypeAttribute() {
		return ldapGroupTypeAttribute;
	}

	public void setLdapGroupTypeAttribute(String ldapGroupTypeAttribute) {
		this.ldapGroupTypeAttribute = ldapGroupTypeAttribute;
	}

	public String getLdapDNAttribute() {
		return ldapDNAttribute;
	}

	public void setLdapDNAttribute(String ldapDNAttribute) {
		this.ldapDNAttribute = ldapDNAttribute;
	}

	public String getLdapPrimaryGroupIdAttribute() {
		return ldapPrimaryGroupIdAttribute;
	}

	public void setLdapPrimaryGroupIdAttribute(String ldapPrimaryGroupIdAttribute) {
		this.ldapPrimaryGroupIdAttribute = ldapPrimaryGroupIdAttribute;
	}

	public String getLdapGroupMemberAttribute() {
		return ldapGroupMemberAttribute;
	}

	public void setLdapGroupMemberAttribute(String ldapGroupMemberAttribute) {
		this.ldapGroupMemberAttribute = ldapGroupMemberAttribute;
	}

	public boolean isDebugLdap() {
		return debugLdap;
	}

	public void setDebugLdap(boolean debugLdap) {
		this.debugLdap = debugLdap;
	}

	public boolean isFollowReferrals() {
		return followReferrals;
	}

	public void setFollowReferrals(boolean followReferrals) {
		this.followReferrals = followReferrals;
	}

	public boolean isShowFirstUseWizard() {
		return showFirstUseWizard;
	}

	public void setShowFirstUseWizard(boolean showFirstUseWizard) {
		this.showFirstUseWizard = showFirstUseWizard;
	}

	public String getDbConnectionString() {
		return dbConnectionString;
	}

	public void setDbConnectionString(String dbConnectionString) {
		this.dbConnectionString = dbConnectionString;
	}

	public String getDbConnectionUsername() {
		return dbConnectionUsername;
	}

	public void setDbConnectionUsername(String dbConnectionUsername) {
		this.dbConnectionUsername = dbConnectionUsername;
	}

	public String getDbConnectionPassword() {
		return dbConnectionPassword;
	}

	public void setDbConnectionPassword(String dbConnectionPassword) {
		this.dbConnectionPassword = dbConnectionPassword;
	}

	public String getDbConnectionDriver() {
		return dbConnectionDriver;
	}

	public void setDbConnectionDriver(String dbConnectionDriver) {
		this.dbConnectionDriver = dbConnectionDriver;
	}

	public int getDbConnectionType() {
		return dbConnectionType;
	}

	public void setDbConnectionType(int dbConnectionType) {
		this.dbConnectionType = dbConnectionType;
	}

	public String getSmtpHost() {
		return smtpHost;
	}

	public void setSmtpHost(String smtpHost) {
		this.smtpHost = smtpHost;
	}

	public int getSmtpPort() {
		return smtpPort;
	}

	public void setSmtpPort(int smtpPort) {
		this.smtpPort = smtpPort;
	}

	public int getSmtpType() {
		return smtpType;
	}

	public void setSmtpType(int smtpType) {
		this.smtpType = smtpType;
	}

	public String getSmtpUsername() {
		return smtpUsername;
	}

	public void setSmtpUsername(String smtpUsername) {
		this.smtpUsername = smtpUsername;
	}

	public String getSmtpPassword() {
		return smtpPassword;
	}

	public void setSmtpPassword(String smtpPassword) {
		this.smtpPassword = smtpPassword;
	}

	public String getReportEmailTemplate() {
		return reportEmailTemplate;
	}

	public void setReportEmailTemplate(String reportEmailTemplate) {
		this.reportEmailTemplate = reportEmailTemplate;
	}

	public String getPolicyEmailTemplate() {
		return policyEmailTemplate;
	}

	public void setPolicyEmailTemplate(String policyEmailTemplate) {
		this.policyEmailTemplate = policyEmailTemplate;
	}

	public String getSmtpFrom() {
		return smtpFrom;
	}

	public void setSmtpFrom(String smtpFrom) {
		this.smtpFrom = smtpFrom;
	}

	public boolean isSmtpAttachReports() {
		return smtpAttachReports;
	}

	public void setSmtpAttachReports(boolean smtpAttachReports) {
		this.smtpAttachReports = smtpAttachReports;
	}

	public String getReportEmailSubject() {
		return reportEmailSubject;
	}

	public void setReportEmailSubject(String reportEmailSubject) {
		this.reportEmailSubject = reportEmailSubject;
	}

	public String getPolicyEmailSubject() {
		return policyEmailSubject;
	}

	public void setPolicyEmailSubject(String policyEmailSubject) {
		this.policyEmailSubject = policyEmailSubject;
	}

	public String getLdapMemberOfAttribute() {
		return ldapMemberOfAttribute;
	}

	public void setLdapMemberOfAttribute(String ldapMemberOfAttribute) {
		this.ldapMemberOfAttribute = ldapMemberOfAttribute;
	}

	public boolean isShowLicenseAgreement() {
		return showLicenseAgreement;
	}

	public void setShowLicenseAgreement(boolean showLicenseAgreement) {
		this.showLicenseAgreement = showLicenseAgreement;
	}

	public boolean isTruncateData() {
		return truncateData;
	}

	public void setTruncateData(boolean truncateData) {
		this.truncateData = truncateData;
	}

	public List<Boolean> getShowAlertDialogs() {
		return showAlertDialogs;
	}

	public List<Boolean> getShowExplainLabels() {
		return showExplainLabels;
	}

	public int getBatchSizeFiles() {
		return batchSizeFiles;
	}

	public void setBatchSizeFiles(int batchSizeFiles) {
		this.batchSizeFiles = batchSizeFiles;
	}

	public int getBatchSizeAces() {
		return batchSizeAces;
	}

	public void setBatchSizeAces(int batchSizeAces) {
		this.batchSizeAces = batchSizeAces;
	}

	public int getBatchSizeMembers() {
		return batchSizeMembers;
	}

	public void setBatchSizeMembers(int batchSizeMembers) {
		this.batchSizeMembers = batchSizeMembers;
	}

	public int getBatchSizeMemberships() {
		return batchSizeMemberships;
	}

	public void setBatchSizeMemberships(int batchSizeMemberships) {
		this.batchSizeMemberships = batchSizeMemberships;
	}

	public boolean isUseSystemProxy() {
		return useSystemProxy;
	}

	public void setUseSystemProxy(boolean useSystemProxy) {
		this.useSystemProxy = useSystemProxy;
	}

	public boolean isCalculateOwnershipRatioInOverview() {
		return calculateOwnershipRatioInOverview;
	}

	public void setCalculateOwnershipRatioInOverview(boolean calculateOwnershipRatioInOverview) {
		this.calculateOwnershipRatioInOverview = calculateOwnershipRatioInOverview;
	}

	public boolean isCalculateOwnershipRatioInEffectivePermissionsTab() {
		return calculateOwnershipRatioInEffectivePermissionsTab;
	}

	public void setCalculateOwnershipRatioInEffectivePermissionsTab(boolean calculateOwnershipRatioInEffectivePermissionsTab) {
		this.calculateOwnershipRatioInEffectivePermissionsTab = calculateOwnershipRatioInEffectivePermissionsTab;
	}

	public boolean isHideModifyAce() {
		return hideModifyAce;
	}

	public void setHideModifyAce(boolean hideModifyAce) {
		this.hideModifyAce = hideModifyAce;
	}

	public String getLogoFileName() {
		return logoFileName;
	}

	public void setLogoFileName(String logoFileName) {
		this.logoFileName = logoFileName;
	}

	public String getLogoBase64() {
		return logoBase64;
	}

	public void setLogoBase64(String logoBase64) {
		this.logoBase64 = logoBase64;
	}

	public int getLogoWidth() {
		return logoWidth;
	}

	public void setLogoWidth(int logoWidth) {
		this.logoWidth = logoWidth;
	}

	public int getLogoHeight() {
		return logoHeight;
	}

	public void setLogoHeight(int logoHeight) {
		this.logoHeight = logoHeight;
	}

	public int getLogoLocationTop() {
		return logoLocationTop;
	}

	public void setLogoLocationTop(int logoLocationTop) {
		this.logoLocationTop = logoLocationTop;
	}

	public int getLogoLocationRight() {
		return logoLocationRight;
	}

	public void setLogoLocationRight(int logoLocationRight) {
		this.logoLocationRight = logoLocationRight;
	}

	public String getCsvSeparator() {
		return csvSeparator;
	}

	public void setCsvSeparator(String csvSeparator) {
		this.csvSeparator = csvSeparator;
	}

	public int getMaxFileExportMegaBytes() {
		return maxFileExportMegaBytes;
	}

	public void setMaxFileExportMegaBytes(int maxFileExportMegaBytes) {
		this.maxFileExportMegaBytes = maxFileExportMegaBytes;
	}

	public boolean isUseUserDisplayNameInReport() {
		return useUserDisplayNameInReport;
	}

	public void setUseUserDisplayNameInReport(boolean useUserDisplayNameInReport) {
		this.useUserDisplayNameInReport = useUserDisplayNameInReport;
	}
}
 