package com.perdemia.permissionanalyzer.application.swt;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Text;

/**
 * This class builds a SWT Text component with a KeyListener that is executed after a given delay. Whenever a new key has been pressed within the delay, the key event is cancelled.
 * This component is useful when each keypress trigger a time consuming process, like a query in a search dialog. So, instead of executing the query after each keypress, the query is only executed
 * once after the user hasn't pressed a key for x ms (default is 400 ms).
 *   
 * @author Reinier
 */
public final class DelayedKeyListenerText {
	/** Not thread safe, but that's ok */
	private static long lastKeyPress = 0;
	
	public static Text create (Composite composite, int swtStyle, final KeyListener keyListener) {
		return create(composite, swtStyle, keyListener, 400);
	}
	
	/**
	 * Creates the SWT Text component and attaches the given keyListener which is exected after the user didn't press a key for x milliseconds.
	 * 
	 * @param composite
	 * @param swtStyle
	 * @param keyListener
	 * @param delay
	 * @return
	 */
	public static Text create (Composite composite, int swtStyle, final KeyListener keyListener, final int delay) {
		Text text = new Text(composite, swtStyle);
		text.addKeyListener(new KeyAdapter() {
			@Override
			public void keyReleased(final KeyEvent e) {
				if (e.keyCode != 13 && e.keyCode != SWT.ARROW_RIGHT && e.keyCode != SWT.ARROW_LEFT && e.keyCode != SWT.HOME && e.keyCode != SWT.END) {
					lastKeyPress = System.currentTimeMillis();
					Display.getDefault().timerExec(delay + 100, new Runnable() {
						@Override
						public void run() {
							if (System.currentTimeMillis() - lastKeyPress > delay) {
								keyListener.keyReleased(e);
								lastKeyPress = 0;
							}
						}
					});					
				}
			}
		});
		
		return text;
	}
}
