package com.perdemia.permissionanalyzer.application.util;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.export.ExportHelper;
import com.perdemia.permissionanalyzer.data.Filter;

/**
 * Helper class that represents the checkboxes / radio buttons on the Filter composites. Those settings are stored as one Filter object in
 * the memberFilterList, permissionFilterList or fileFilterList.
 * 
 * @author Reinier
 *
 */
public class FilterConfig {
	private static Logger logger = LoggerFactory.getLogger(FilterConfig.class.getName());
	
	/** Member filter settings */
	private boolean includeGroupMembership;
	private boolean includeGroups;
	private boolean includeUsers;
	private boolean includeBuiltIn;
	private boolean includeADGroupsAndUsers;
	private boolean includeDisabledUsers;
	
	/** Permission filter settings */
	private boolean includePermissions; // false is exclude
	private boolean aceNeedAllPermissions; // false is an ACE can have any of the selected permissions
	private boolean hideInheritedPermissions;
	
	/** Folder filter settings */
	private boolean includeFolders;
	private boolean includeFiles;
	private boolean includeFilesThatBlockInheritance;
	private boolean invertFolders;
	
	/** Actual Filter objects that hold the config information from the database */
	private Filter memberConfigFilter;
	private Filter permissionConfigFilter;
	private Filter folderConfigFilter;
	
	/**
	 * Constructor only visible for this package (the FilterHelper). This constructor will initialize the
	 * filter configurations based on the actual Filter database objects. 
	 * 
	 * @param memberConfigFilter
	 * @param permissionConfigFilter
	 * @param folderConfigFilter
	 */
	public FilterConfig(Filter memberConfigFilter, Filter permissionConfigFilter, Filter folderConfigFilter) {
		this.memberConfigFilter = memberConfigFilter;
		this.permissionConfigFilter = permissionConfigFilter;
		this.folderConfigFilter = folderConfigFilter;
		
		// load the filter values
		// member filter configs
		if (memberConfigFilter != null) {
			includeGroupMembership = memberConfigFilter.getValue().charAt(0) == '1';
			includeUsers = memberConfigFilter.getValue().charAt(1) == '1';
			includeGroups = memberConfigFilter.getValue().charAt(2) == '1';
			includeBuiltIn = memberConfigFilter.getValue().length() < 4 || memberConfigFilter.getValue().charAt(3) == '1';
			includeADGroupsAndUsers = memberConfigFilter.getValue().length() < 5 || memberConfigFilter.getValue().charAt(4) == '1';
			includeDisabledUsers = memberConfigFilter.getValue().length() < 6 || memberConfigFilter.getValue().charAt(5) == '1';
		} else {
			logger.warn("memberConfigFilter is null");
		}
		
		// permission filter configs
		if (permissionConfigFilter != null) {
			includePermissions = permissionConfigFilter.getValue().charAt(0) == '1';
			aceNeedAllPermissions = permissionConfigFilter.getValue().charAt(1) == '1';
			hideInheritedPermissions = permissionConfigFilter.getValue().charAt(2) == '1';
		} else {
			logger.warn("permissionConfigFilter is null");
		}
		
		// folder filter configs
		if (folderConfigFilter != null) {
			includeFolders = folderConfigFilter.getValue().charAt(0) == '1';
			includeFiles = folderConfigFilter.getValue().charAt(1) == '1';
			includeFilesThatBlockInheritance = folderConfigFilter.getValue().length() > 2 && folderConfigFilter.getValue().charAt(2) == '1';
			invertFolders = folderConfigFilter.getValue().length() > 3 && folderConfigFilter.getValue().charAt(3) == '1';
		} else {
			logger.warn("folderConfigFilter is null");
		}
	}
	
	/**
	 * Updates the actual Filter objects (which might be stored in the database by the GUI)
	 */
	private void updateConfigFilters() {
		// member filter config values
		StringBuilder memberValues = new StringBuilder();
		memberValues.append(includeGroupMembership ? '1' : '0');
		memberValues.append(includeUsers ? '1' : '0');
		memberValues.append(includeGroups ? '1' : '0');
		memberValues.append(includeBuiltIn ? '1' : '0');
		memberValues.append(includeADGroupsAndUsers ? '1' : '0');
		memberValues.append(includeDisabledUsers ? '1' : '0');
		memberConfigFilter.setValue(memberValues.toString());
		
		// permission filter config values
		StringBuilder permissionValues = new StringBuilder();
		permissionValues.append(includePermissions ? '1' : '0');
		permissionValues.append(aceNeedAllPermissions ? '1' : '0');
		permissionValues.append(hideInheritedPermissions ? '1' : '0');
		permissionConfigFilter.setValue(permissionValues.toString());
		
		// folder file config values
		StringBuilder folderValues = new StringBuilder();
		folderValues.append(includeFolders ? '1' : '0');
		folderValues.append(includeFiles ? '1' : '0');
		folderValues.append(includeFilesThatBlockInheritance ? '1' : '0');
		folderValues.append(invertFolders ? '1' : '0');
		folderConfigFilter.setValue(folderValues.toString());
	}
	
	public boolean isIncludeGroupMembership() {
		return includeGroupMembership;
	}
	public void setIncludeGroupMembership(boolean includeGroupMembership) {
		this.includeGroupMembership = includeGroupMembership;
		updateConfigFilters();
	}
	public boolean isIncludeGroups() {
		return includeGroups;
	}
	public void setIncludeGroups(boolean includeGroups) {
		this.includeGroups = includeGroups;
		updateConfigFilters();
	}
	public boolean isIncludeUsers() {
		return includeUsers;
	}
	public void setIncludeUsers(boolean includeUsers) {
		this.includeUsers = includeUsers;
		updateConfigFilters();
	}
	public boolean isIncludePermissions() {
		return includePermissions;
	}
	public void setIncludePermissions(boolean includePermissions) {
		this.includePermissions = includePermissions;
		updateConfigFilters();
	}
	public boolean isAceNeedAllPermissions() {
		return aceNeedAllPermissions;
	}
	public void setAceNeedAllPermissions(boolean aceNeedAllPermissions) {
		this.aceNeedAllPermissions = aceNeedAllPermissions;
		updateConfigFilters();
	}

	public boolean isHideInheritedPermissions() {
		return hideInheritedPermissions;
	}

	public void setHideInheritedPermissions(boolean hideInheritedPermissions) {
		this.hideInheritedPermissions = hideInheritedPermissions;
		updateConfigFilters();
	}

	public boolean isIncludeFolders() {
		return includeFolders;
	}

	public void setIncludeFolders(boolean includeFolders) {
		this.includeFolders = includeFolders;
		updateConfigFilters();
	}

	public boolean isIncludeFiles() {
		return includeFiles;
	}

	public void setIncludeFiles(boolean includeFiles) {
		this.includeFiles = includeFiles;
		updateConfigFilters();
	}

	public boolean isIncludeBuiltIn() {
		return includeBuiltIn;
	}

	public void setIncludeBuiltIn(boolean includeBuiltIn) {
		this.includeBuiltIn = includeBuiltIn;
		updateConfigFilters();
	}

	public Filter getMemberConfigFilter() {
		return memberConfigFilter;
	}

	public Filter getPermissionConfigFilter() {
		return permissionConfigFilter;
	}

	public Filter getFolderConfigFilter() {
		return folderConfigFilter;
	}

	public boolean isIncludeFilesThatBlockInheritance() {
		return includeFilesThatBlockInheritance;
	}

	public void setIncludeFilesThatBlockInheritance(boolean includeFilesThatBlockInheritance) {
		this.includeFilesThatBlockInheritance = includeFilesThatBlockInheritance;
		updateConfigFilters();
	}

	public boolean isIncludeADGroupsAndUsers() {
		return includeADGroupsAndUsers;
	}

	public void setIncludeADGroupsAndUsers(boolean includeADGroupsAndUsers) {
		this.includeADGroupsAndUsers = includeADGroupsAndUsers;
		updateConfigFilters();
	}
	
	public boolean isIncludeDisabledUsers() {
		return includeDisabledUsers;
	}

	public void setIncludeDisabledUsers(boolean includeDisabledUsers) {
		this.includeDisabledUsers = includeDisabledUsers;
		updateConfigFilters();
	}

	public boolean isInvertFolders() {
		return invertFolders;
	}

	public void setInvertFolders(boolean invertFolders) {
		this.invertFolders = invertFolders;
		updateConfigFilters();
	}
}

