package com.perdemia.permissionanalyzer.application.util;

import gnu.trove.list.array.TIntArrayList;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Pattern;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.widgets.Text;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.handlers.toolbar.ApplyFilterHandler;
import com.perdemia.permissionanalyzer.application.handlers.toolbar.QuickFileSearchHandler;
import com.perdemia.permissionanalyzer.application.handlers.toolbar.QuickFileSearchToolControl;
import com.perdemia.permissionanalyzer.application.jface.contentproviders.MainFileTreeContentProvider;
import com.perdemia.permissionanalyzer.application.jface.dialogs.SearchMembersDialog;
import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.logging.GUILogHandler;
import com.perdemia.permissionanalyzer.application.parts.FilterPermissionPart;
import com.perdemia.permissionanalyzer.application.parts.MemberScopePart;
import com.perdemia.permissionanalyzer.application.parts.TracePart;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper.Edition;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Filter;
import com.perdemia.permissionanalyzer.data.Filter.FilterType;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.PermissionLookup;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SimplePermission;
import com.perdemia.permissionanalyzer.data.Rdn;
import com.perdemia.permissionanalyzer.data.Report;
import com.perdemia.permissionanalyzer.data.util.CustomProcessor;
import com.perdemia.permissionanalyzer.data.util.DomainCache;
import com.perdemia.permissionanalyzer.data.util.FileCache;
import com.perdemia.permissionanalyzer.data.util.MembershipCache;

/**
 * Singleton class that hold all the filter information.
 * 
 * @author Reinier
 * 
 */
public class FilterHelper {
	private static Logger logger = LoggerFactory.getLogger(FilterHelper.class.getName());

	/** instance of this singleton */
	private static FilterHelper instance;

	/** A Filters collection for each filter Composite (these values are immediately changed when the UI is changed) **/
	private java.util.List<Filter> memberFilterList = new ArrayList<Filter>();
	private java.util.List<Filter> permissionFilterList = new ArrayList<Filter>();
	private java.util.List<Filter> folderFilterList = new ArrayList<Filter>();
	
	/** A Filters collection for each filter Composite (these values represent the state when the filter was applied by the user, not the actual values in the UI **/
	private java.util.List<Filter> memberFilterListApplied = new ArrayList<Filter>();
	private java.util.List<Filter> permissionFilterListApplied = new ArrayList<Filter>();
	private java.util.List<Filter> folderFilterListApplied = new ArrayList<Filter>();
	
	/** Represents the FilterConfig at the point the filter was applied in the GUI */
	private FilterConfig filterConfigApplied;
	
	/** This field is filled when a user selects "Load report" in the ReportDialog. */
	private Report selectedReport;
		
	/** Objects (table) to return in the query with the filter 
	 *	<ul>	
	 *	<li>FILE: Returns all file IDs from the file_info table that match the filters selected by the users. This type is used to fill the 
	 *			lists {@link #applicableFileIds} and {@link applicableFileIdsIncludingParents}</li>
	 *	<li>ACL_ITEM: Returns all records from the AclItem table that match the filters of the user and the selected folder id!</li>
	 *	<li>MEMBER: Returns all distinct records from the member table that match the filters selected by the user and that are either on the
	 *			selected file (in the ACE's) or in a group that is attached to one of the ACE's.  When the option 'include group memberships' is selected 
	 *			and the user wants to include all the members from the group 'Freelancers' then we want to display the members from the Freelancers group, not
	 *			the direct members on the directory. For example, the group Administrators may appear on a directory because on of the members of Freelancers is
	 *			an administrator. This type of query gets the members of the selected file in the <code>additionalFilter</code> parameters of the method
	 *			{@link FilterHelper#buildFilterQuery(ObjectType, boolean, int, String, String, Object...)}, like <code>MEMBERSHIP.GROUP_ID IN (2,43,509)</code>. See {@link MemberScopePart#getMembersThatMatchFilterPS(int)}</li>
	 *	<li>MEMBERS_OF_A_FILE: Instead of returning the file ids of all the files that match the filter criteria, this type of query returns the distinct members
	 *			that belong to all the files that have been found. It is used by the {@link MemberScopePart} to build the list of all (indirect) members found in the directory tree.</li>
	 *	<ul>
	 */
	public enum ObjectType {
		FILE, ACL_ITEM, MEMBER, MEMBERS_IN_FILE_TREE, MEMBER_COUNT, GROUP_MEMBERS
	};
		
	/** Base queries used to retrieve Files or AclItems */
	public static final String FILE_BASE_QUERY_WITH_PARENTS = "SELECT DISTINCT FILEINFO.PARENT_FILE_IDS FROM {0}.FILEINFO ";
	
	// don't use DISTINCT in this query because it is used in the reports and table viewer which adds a ORDER BY name. 
	// (and all order by fields must appear in the select when using a DISTINCT)
	public static final String FILE_BASE_QUERY_NO_PARENTS = "SELECT FILEINFO.ID FROM {0}.FILEINFO ";
	public static final String ACL_BASE_QUERY = "SELECT DISTINCT ACLITEM.* FROM {0}.ACLITEM ";
	//public static final String MEMBER_BASE_QUERY = "SELECT DISTINCT FILTER_MEMBER.* FROM {0}.MEMBER FILTER_MEMBER INNER JOIN {0}.MEMBERSHIP ON MEMBERSHIP.MEMBER_ID = FILTER_MEMBER.ID ";
	//public static final String ALL_MEMBERS_BASE_QUERY = "SELECT DISTINCT FILTER_MEMBER.* FROM {0}.MEMBER FILTER_MEMBER ";
	public static final String MEMBER_BASE_QUERY = "SELECT DISTINCT FILTER_MEMBER.* FROM {0}.MEMBER FILTER_MEMBER ";
	public static final String GROUP_MEMBERS_BASE_QUERY = "SELECT DISTINCT FILTER_MEMBER.*, VIA_GROUP.NAME VIA_GROUP_NAME FROM {0}.MEMBER FILTER_MEMBER INNER JOIN {0}.MEMBERSHIP FILTER_MEMBERSHIP ON FILTER_MEMBERSHIP.MEMBER_ID = FILTER_MEMBER.ID " + 
																"LEFT OUTER JOIN {0}.MEMBER VIA_GROUP ON VIA_GROUP.ID = FILTER_MEMBERSHIP.VIA_GROUP_ID ";
	public static final String MEMBER_COUNT_BASE_QUERY = "SELECT COUNT(DISTINCT FILTER_MEMBER.ID) FROM {0}.MEMBER FILTER_MEMBER INNER JOIN {0}.MEMBERSHIP ON MEMBERSHIP.MEMBER_ID = FILTER_MEMBER.ID ";
	
	public static final String INVERT_FOLDERS_QUERY_SUFFIX = " WHERE ID NOT IN (SELECT FILEINFO.ID FROM {0}.FILEINFO ";
	
	/** Cache of the queries built */
	private final Map<String, PreparedStatement> queryCache = new HashMap<String, PreparedStatement>();
	private final Map<String, List<Object>> queryParameterCache = new HashMap<String, List<Object>>();

	/** PreparedStatement parameters to build the query */
	List<Object> queryParameters = new ArrayList<Object>();
	
	/** List of file id's (including their parents) that have been found after applying the filter.
	 * This list is used when a Tree leaf is expanded and to determine which children to display */
	private TIntArrayList applicableFileIdsIncludingParents = new TIntArrayList(); 
	
	/** List of file id's (NOT including their parents) that have been found after applying the filter.
	 * This list is used in the TableViewer when the user scrolls in the list of results or when a report is generated */
	private TIntArrayList applicableFileIds = new TIntArrayList(); 
	
	/**
	 * The Tree is lazy loaded and when more files are loaded (by expanding the tree) this int value is increased.
	 * This value is used to display a progress bar when the user clicks the Expand All button.
	 */
	private int filesDisplayedInTree;
	
	/** Boolean to indicate if the Cancel button has been pressed while expanding all tree items */
	private boolean stopExpandingAll;
	
	/** Value entered in the quick file search text box in the top right corner to be used in the filter
	 * @see {@link QuickFileSearchToolControl}
	 * @see {@link QuickFileSearchHandler} 
	 * */
	private String quickFileSearch = "";
	private Text quickFileSearchText;
	
	/** 
	 * At some places we need the number of parameter in the PreparedStatement. Using <code>ps.getParameterMetaData().getParameterCount()</code> did
	 * not work because it throws an <code>"SQLServerException: Incorrect syntax near ','"</code> for the PS with a query containing <code>WHERE (MEMBERSHIP.GROUP_ID IN (?,?,?,?,?,?,?)</code>.
	 * That is why we fill the variable after generating a new query (by counting the question marks)
	 * @see {@link SearchMembersDialog#SearchMembersDialog(org.eclipse.swt.widgets.Shell, PreparedStatement, int)}
	 */
	private int paramCountOfLastFilterQuery;
	
	/**
	 * This flag triggers a refresh when the report view is displayed. The flag is set after a new scan has been finished.
	 */
	private boolean reloadReportView;
	
	/** Flag that indicates whether or not a filter has been applied */
	private boolean filterHasBeenApplied;
	
	/** Number of items as result of the filtering 
	 * @see {@link ApplyFilterHandler}
	 */
	private int itemCount;
	
	/**
	 * Number of matching members that are found. Used to calculated the ownership ratio in the overview and reports.
	 */
	private int allFilteredMembers;
	
	/**
	 * Maximum number of member filters allowed in the trial version
	 */
	private final static String MAX_MEMBER_FILTERS_FOR_TRIAL = "3";
	
	/**
	 * The value used as prefix in case a wildcard filter on the folder name must be applied on top of existing folder filters.
	 * In other words, the filter will be applied as AND operator, instead of OR.
	 */
	public final static String PREFIX_FOLDER_WILDCARD_FILTER_AFTER_OTHER_FILTERS = "%%%AFTER%%%";
	
	/**
	 * The value used as prefix in case a wildcard filter on the folder name must include all subfolders.
	 */
	public final static String PREFIX_FOLDER_WILDCARD_FILTER_INCLUDE_SUBFOLDERS = "%%%%INCLUDESUBFOLDERS%%%%";
	
	/**
	 * Default constructor
	 */
	public FilterHelper(){
		// make sure the filter configs are initialized
		getMemberConfigFilter();
		getPermissionConfigFilter();
		getFolderConfigFilter();
	}
	
	/**
	 * Instance method for the Singleton pattern
	 * 
	 * @return current static instance. A new one is created if it doesn't exists
	 */
	public static FilterHelper getInstance() {
		if (instance == null) {
			instance = new FilterHelper();
		}
		return instance;
	}
	
	/**
	 * Build a new query based on the current Filters. The <code>baseQuery</code> is used to determine the data to return and this method adds the right INNER JOINS and WHERE clauses.
	 * 
	 * <p>
	 * <b>Example to retrieve all AclItems for a FileInfo:</b>
	 * </p>
	 * <code>objectType: ObjectType.ACL_ITEM</code><br/>
	 * <code>additionalFilter: ACLITEM.FILEINFO_ID = ?</code> <code>additionalFilterValue: 6</code>
	 * 
	 * @param objectType
	 *            object type (table) used in the SELECT statement of the query
	 * @param additionalFilter
	 *            a filter that will be appended at the end of the query, like ACLITEM.FILEINFO_ID = ?
	 * @param additionalFilterValue
	 *            value for the ? character in the additionalFilter
	 * 
	 * @return query with the right joins and where clauses based on the current Filters
	 */
	@SuppressWarnings("resource")
	public PreparedStatement buildFilterQuery(ObjectType objectType, boolean useInTreeViewer, boolean onlyUseExcludes, int limit, String orderBy, String additionalFilter, Object ... additionalFilterValue) throws DBException {
		PreparedStatement filterQuery = null;
		StringBuilder baseQuery = new StringBuilder();
		String cacheKey = objectType + additionalFilter + limit + useInTreeViewer;
		
		// First check if the set of filters contains any filters that are not valid anymore (because a new scan changed the database content)
		// see matchFilterValuesAfterScan()
		for (Filter filter: memberFilterListApplied) {
			if (filter.getValue().equals("-1")) {
				SyncMessageDialog.showMessage(MessageDialog.ERROR, "Invalid filter", "The member filter with value " + filter.getName() + " is not valid anymore because the member does no longer exists in the database. \r\n\r\nModify the filter or re-scan the network.");
				throw new DBException("The member filter with value " + filter.getName() + " is not valid anymore because the member does no longer exists in the database.");
			}
		}
		for (Filter filter : folderFilterListApplied) {
			if (filter.getValue().equals("-1")) {
				SyncMessageDialog.showMessage(MessageDialog.ERROR, "Invalid filter", "The folder filter with value " + filter.getName() + " is not valid anymore because the folder does no longer exists in the database. \r\n\r\nModify the filter or re-scan the network.");
				throw new DBException("The folder filter with value " + filter.getName() + " is not valid anymore because the folder does no longer exists in the database.");
			}
		}
		
		try {
			// ----------------------------------------------------------
			// determine the applicable filters
			// ----------------------------------------------------------
			boolean hasFilterOnMemberId = false;
			boolean hasFilterOnOwnerId = false;
			boolean hasFilterOnMemberName = false;
			boolean hasFilterOnGroupId = false;
			boolean hasFilterOnGroupName = false;
			boolean hasFilterOnLdapOU = false;
			boolean hasFilterOnPermission = false;
			boolean hasFilterOnFile = false;
						
			for (Filter filter: memberFilterListApplied) {
				hasFilterOnMemberId = hasFilterOnMemberId || filter.getType().equals(FilterType.MEMBER_ID) || filter.getType().equals(FilterType.MEMBER_REPORT);
				hasFilterOnOwnerId = hasFilterOnOwnerId || filter.getType().equals(FilterType.OWNER_ID);
				hasFilterOnMemberName = hasFilterOnMemberName || filter.getType().equals(FilterType.MEMBER_NAME);
				hasFilterOnGroupId = hasFilterOnGroupId || filter.getType().equals(FilterType.MEMBER_GROUP_ID);
				hasFilterOnGroupName = hasFilterOnGroupName || filter.getType().equals(FilterType.MEMBER_GROUP_NAME);
				hasFilterOnLdapOU = hasFilterOnLdapOU || filter.getType().equals(FilterType.MEMBER_LDAP_OU);
			}
			
			// 6 simple permissions, 13 special permissions, * 2 because the list contains allow and deny filters and + 1 for the configuration filter
			hasFilterOnPermission = permissionFilterListApplied.size() > 1 && // not only the FILTER_CONFIG type 
										(permissionFilterListApplied.size() != (6 * 2 + 13 * 2 + 1) || filterConfigApplied.isAceNeedAllPermissions());
			
			hasFilterOnFile = hasFilterOnFile() || hasFilterOnOwnerId;
							
			boolean filterRequired = hasFilterOnMemberId || hasFilterOnOwnerId || hasFilterOnMemberName || hasFilterOnGroupId || hasFilterOnGroupName || hasFilterOnLdapOU || 
					!filterConfigApplied.isIncludeUsers() || !filterConfigApplied.isIncludeGroups() || !filterConfigApplied.isIncludeBuiltIn() || !filterConfigApplied.isIncludeADGroupsAndUsers() || 
					!filterConfigApplied.isIncludeDisabledUsers() || hasFilterOnPermission || hasFilterOnFile ||	filterConfigApplied.isHideInheritedPermissions() || objectType.equals(ObjectType.MEMBER) ||
					quickFileSearch.length() > 0 || (additionalFilter != null && additionalFilter.length() > 0);
			
			filterRequired = filterRequired && checkNumberOfMemberFiltersForTrial(memberFilterListApplied, false);
						
			filterHasBeenApplied = filterRequired;

			// not in cache yet?
			if (!queryCache.containsKey(cacheKey)) {
												
				queryParameters = new ArrayList<Object>();
				
				// determine base query
				switch (objectType) {
					case FILE:
						if (useInTreeViewer) {
							baseQuery.append(FILE_BASE_QUERY_WITH_PARENTS);
						} else {
							baseQuery.append(FILE_BASE_QUERY_NO_PARENTS);
						}
						
						// invert matching folders?
						if (filterConfigApplied.isInvertFolders()) {
							baseQuery.append(INVERT_FOLDERS_QUERY_SUFFIX);
						}
						break;
					case ACL_ITEM:
						baseQuery.append(ACL_BASE_QUERY);
						break;
					case MEMBER:
						baseQuery.append(MEMBER_BASE_QUERY);
						break;
					case MEMBERS_IN_FILE_TREE:
						baseQuery.append(MEMBER_BASE_QUERY);
						break;
					case MEMBER_COUNT:
						baseQuery.append(MEMBER_COUNT_BASE_QUERY);
						break;
					case GROUP_MEMBERS:
						baseQuery.append(GROUP_MEMBERS_BASE_QUERY);
						break;
				}
								
				if (filterRequired) {
										
					// ----------------------------------------------------------
					// determine the right JOINS
					// ----------------------------------------------------------
					if (objectType.equals(ObjectType.FILE) && (hasFilterOnMemberId || hasFilterOnMemberName || hasFilterOnGroupId || hasFilterOnGroupName || hasFilterOnLdapOU || hasFilterOnPermission
							|| filterConfigApplied.isHideInheritedPermissions()
							|| !filterConfigApplied.isIncludeUsers() || !filterConfigApplied.isIncludeGroups() || !filterConfigApplied.isIncludeBuiltIn() || !filterConfigApplied.isIncludeADGroupsAndUsers()
							|| !filterConfigApplied.isIncludeDisabledUsers())) {
						// acl info is required
						baseQuery.append("INNER JOIN {0}.ACLITEM ON FILEINFO.ID = ACLITEM.FILEINFO_ID "); 
					}
					
					// A fully joined query looks as follow:
					//	[ACLITEM.MEMBER_ID] 
					//			>> [NESTED_MEMBERSHIP.GROUP_ID] 		  		all groups of the members found (including the member itself)
					//			   [NESTED_MEMBERSHIP.MEMBER_ID] 
					//					>> [FILTER_MEMBER.ID] 				  	members of the 'include members of group ABC' filter
					//						>> [FILTER_MEMBERSHIP.MEMBER_ID]  
					//					   	   [FILTER_MEMBERSHIP.GROUP_ID]
					//								>> [FILTER_GROUP.ID]		filter for 'include members of group ABC'
					if (objectType.equals(ObjectType.FILE) || objectType.equals(ObjectType.ACL_ITEM)) {
						// join on folders for ACE's of expanded members in the All Members tab!
						if (objectType.equals(ObjectType.ACL_ITEM) && additionalFilterValue.length == 0 && hasFilterOnFile) {
							baseQuery.append("INNER JOIN {0}.FILEINFO ON ACLITEM.FILEINFO_ID = FILEINFO.ID ");
						}
						
						if (filterConfigApplied.isIncludeGroupMembership()) {
							if (hasFilterOnGroupName || hasFilterOnGroupId) {
								//baseQuery.append("INNER JOIN {0}.MEMBER ON ACLITEM.MEMBER_ID = MEMBER.ID ");
								//baseQuery.append("INNER JOIN {0}.MEMBERSHIP NESTED_MEMBERSHIP ON MEMBER.ID = NESTED_MEMBERSHIP.GROUP_ID ");
								baseQuery.append("INNER JOIN {0}.MEMBERSHIP NESTED_MEMBERSHIP ON ACLITEM.MEMBER_ID = NESTED_MEMBERSHIP.GROUP_ID ");
								baseQuery.append("INNER JOIN {0}.MEMBER FILTER_MEMBER ON NESTED_MEMBERSHIP.MEMBER_ID = FILTER_MEMBER.ID ");
								baseQuery.append("INNER JOIN {0}.MEMBERSHIP FILTER_MEMBERSHIP ON FILTER_MEMBER.ID = FILTER_MEMBERSHIP.MEMBER_ID ");
								
								if (hasFilterOnGroupName) {
									baseQuery.append("INNER JOIN {0}.MEMBER FILTER_GROUP ON FILTER_MEMBERSHIP.GROUP_ID = FILTER_GROUP.ID ");
								}
								
							// only filter on member name or id, with nested group membership
							} else if (hasFilterOnMemberName || hasFilterOnMemberId || hasFilterOnLdapOU) {
								baseQuery.append("INNER JOIN {0}.MEMBERSHIP ON ACLITEM.MEMBER_ID = MEMBERSHIP.GROUP_ID INNER JOIN {0}.MEMBER FILTER_MEMBER ON MEMBERSHIP.MEMBER_ID = FILTER_MEMBER.ID ");
							
							// filter on member.is_group or member.is_builtin?
							} else if (!filterConfigApplied.isIncludeUsers() || !filterConfigApplied.isIncludeGroups() || !filterConfigApplied.isIncludeBuiltIn() || !filterConfigApplied.isIncludeADGroupsAndUsers()
									|| !filterConfigApplied.isIncludeDisabledUsers()) {
								baseQuery.append("INNER JOIN {0}.MEMBER FILTER_MEMBER ON ACLITEM.MEMBER_ID = FILTER_MEMBER.ID ");
							}
						} else {
							if (hasFilterOnGroupName || hasFilterOnGroupId) {
								baseQuery.append("INNER JOIN {0}.MEMBERSHIP FILTER_MEMBERSHIP ON ACLITEM.MEMBER_ID = FILTER_MEMBERSHIP.MEMBER_ID ");
								
								if (hasFilterOnGroupName) {
									baseQuery.append("INNER JOIN {0}.MEMBER FILTER_GROUP ON FILTER_MEMBERSHIP.GROUP_ID = FILTER_GROUP.ID ");
								}
							}
							
							// filter on member name, is_group, is_builtin or RDN?
							if (hasFilterOnMemberName || hasFilterOnLdapOU || !filterConfigApplied.isIncludeUsers() || !filterConfigApplied.isIncludeGroups() || !filterConfigApplied.isIncludeBuiltIn() || !filterConfigApplied.isIncludeADGroupsAndUsers()
									|| !filterConfigApplied.isIncludeDisabledUsers()) {
								 baseQuery.append("INNER JOIN {0}.MEMBER FILTER_MEMBER ON ACLITEM.MEMBER_ID = FILTER_MEMBER.ID "); 
							}
						}
												
						// join the PermissionLookup table?
						List<Integer> allowSpecialPermissionList = new ArrayList<Integer>();
						List<Integer> denySpecialPermissionList = new ArrayList<Integer>();
						for (Filter filter : permissionFilterListApplied) {
							int filterValue = Integer.parseInt(filter.getValue());
							// allow?
							if (filter.getType().equals(FilterType.PERMISSION_ALLOW)) {
								if (filterValue < FilterPermissionPart.SIMPLE_PERMISSIONS_INDEX_OFFSET) {
									allowSpecialPermissionList.add(filterValue);
								}
							
							// deny?
							} else if (filter.getType().equals(FilterType.PERMISSION_DENY)) {
								if (filterValue < FilterPermissionPart.SIMPLE_PERMISSIONS_INDEX_OFFSET) {
									denySpecialPermissionList.add(filterValue);
								}
							}
						}
						
						SimplePermission allowSimplePermission = PermissionUtil.getSimplePermission(allowSpecialPermissionList, false, false); 
						SimplePermission denySimplePermission = PermissionUtil.getSimplePermission(denySpecialPermissionList, false, false);
						
						boolean hasSpecialAllowPermissions = !allowSpecialPermissionList.isEmpty() && (allowSimplePermission == null || !allowSimplePermission.equals(SimplePermission.FULL));
						boolean hasSpecialDenyPermissions = !denySpecialPermissionList.isEmpty() && (denySimplePermission == null || !denySimplePermission.equals(SimplePermission.FULL));
						
						/*
						// in case the ACE must have ALL permissions, then we need at least one permission filter and a special permission (a simple permission would be matched on permissionId)
						boolean aceMustHaveAllAndFilterHasSpecialAllowPermissions = filterConfigApplied.isAceNeedAllPermissions() && !allowSimplePermission.equals(SimplePermission.FULL) && allowPermissionCount > 0;
						boolean aceMustHaveAllAndFilterHasSpecialDenyPermissions = filterConfigApplied.isAceNeedAllPermissions() && !denySimplePermission.equals(SimplePermission.FULL) && denyPermissionCount > 0;

						// incase the ACE must have ANY permission then we need at least one special permission filter and not the FULL permission because then every ACE would match
						boolean aceCanHaveAnyAndFilterHasSpecialAllowPermissions = !filterConfigApplied.isAceNeedAllPermissions() && !allowSpecialPermissionList.isEmpty() && !allowSimplePermission.equals(SimplePermission.FULL);
						boolean aceCanHaveAnyAndFilterHasSpecialDenyPermissions = !filterConfigApplied.isAceNeedAllPermissions() && !denySpecialPermissionList.isEmpty() && !denySimplePermission.equals(SimplePermission.FULL);
						*/
						
						/*
						boolean aceMustHaveAllAndFilterHasSpecialAllowPermissions = filterConfigApplied.isAceNeedAllPermissions() && (allowPermissionCount != 19 && allowSpecialPermissionCount > 0);
						boolean aceMustHaveAllAndFilterHasSpecialDenyPermissions = filterConfigApplied.isAceNeedAllPermissions() && (denyPermissionCount != 19 && denySpecialPermissionCount > 0);
						boolean aceCanHaveAnyAndFilterHasSpecialAllowPermissions = !filterConfigApplied.isAceNeedAllPermissions() && allowSpecialPermissionCount > 0 && allowSpecialPermissionCount != 13;
						boolean aceCanHaveAnyAndFilterHasSpecialDenyPermissions = !filterConfigApplied.isAceNeedAllPermissions() && denySpecialPermissionCount > 0 && denySpecialPermissionCount != 13;
						*/
						
						if (hasSpecialAllowPermissions || hasSpecialDenyPermissions) {
							baseQuery.append("INNER JOIN {0}.PERMISSIONLOOKUP ON ACLITEM.PERMISSION_ID = PERMISSIONLOOKUP.ID ");
						}
						
					} else if (objectType.equals(ObjectType.MEMBER) || objectType.equals(ObjectType.MEMBER_COUNT)) {
						// the base filter is:
						// "SELECT DISTINCT FILTER_MEMBER.* FROM {0}.MEMBER FILTER_MEMBER "
						// now if we have a filter on group ID or group name and nested group memberships enabled, we will need an additional JOIN table 
						if (filterConfigApplied.isIncludeGroupMembership() && (hasFilterOnGroupId || hasFilterOnGroupName)) {
							baseQuery.append("INNER JOIN {0}.MEMBERSHIP NESTED_MEMBERSHIP ON FILTER_MEMBER.ID = NESTED_MEMBERSHIP.MEMBER_ID ");
							baseQuery.append("INNER JOIN {0}.MEMBERSHIP FILTER_MEMBERSHIP ON FILTER_MEMBER.ID = FILTER_MEMBERSHIP.MEMBER_ID ");
							
							if (additionalFilter != null) {
								additionalFilter = additionalFilter.toUpperCase().replace("FILTER_MEMBERSHIP", "NESTED_MEMBERSHIP");
							}
						} else {
							// do we need the membership table?
							if (hasFilterOnGroupId || hasFilterOnGroupName || (additionalFilter != null && additionalFilter.toUpperCase().contains("MEMBERSHIP"))) {
								baseQuery.append(" INNER JOIN {0}.MEMBERSHIP FILTER_MEMBERSHIP ON FILTER_MEMBER.ID = FILTER_MEMBERSHIP.MEMBER_ID ");
							}
						}					
						// member table is required if we want to filter on group name
						if (hasFilterOnGroupName) {
							baseQuery.append("INNER JOIN {0}.MEMBER FILTER_GROUP ON FILTER_MEMBERSHIP.GROUP_ID = FILTER_GROUP.ID ");
						}
						
						
					} else if (objectType.equals(ObjectType.GROUP_MEMBERS)) {
						// member table is required if we want to filter on group name
						if (hasFilterOnGroupName) {
							baseQuery.append("INNER JOIN {0}.MEMBER FILTER_GROUP ON FILTER_MEMBERSHIP.GROUP_ID = FILTER_GROUP.ID ");
						}
						
					} else if (objectType.equals(ObjectType.MEMBERS_IN_FILE_TREE)) {
						
						// do we need the membership table?
						if (filterConfigApplied.isIncludeGroupMembership()) {
							baseQuery.append("INNER JOIN {0}.MEMBERSHIP ON MEMBERSHIP.MEMBER_ID = FILTER_MEMBER.ID INNER JOIN {0}.ACLITEM ON ACLITEM.MEMBER_ID = MEMBERSHIP.GROUP_ID ");

							if (hasFilterOnGroupName || hasFilterOnGroupId) {
								baseQuery.append("INNER JOIN {0}.MEMBERSHIP FILTER_MEMBERSHIP ON FILTER_MEMBER.ID = FILTER_MEMBERSHIP.MEMBER_ID ");
								
								if (hasFilterOnGroupName) {
									baseQuery.append("INNER JOIN {0}.MEMBER FILTER_GROUP ON FILTER_MEMBERSHIP.GROUP_ID = FILTER_GROUP.ID ");
								}
							}
						} else {
							baseQuery.append("INNER JOIN {0}.ACLITEM ON ACLITEM.MEMBER_ID = FILTER_MEMBER.ID ");
							if (hasFilterOnGroupName || hasFilterOnGroupId) {
								baseQuery.append("INNER JOIN {0}.MEMBERSHIP FILTER_MEMBERSHIP ON ACLITEM.MEMBER_ID = FILTER_MEMBERSHIP.MEMBER_ID ");
								
								if (hasFilterOnGroupName) {
									baseQuery.append("INNER JOIN {0}.MEMBER FILTER_GROUP ON FILTER_MEMBERSHIP.GROUP_ID = FILTER_GROUP.ID ");
								}
							}
						}
						
						if (hasFilterOnFile) {
							baseQuery.append("INNER JOIN {0}.FILEINFO ON ACLITEM.FILEINFO_ID = FILEINFO.ID ");
						}
					}
					
					baseQuery.append("WHERE");
					
					// when the option invert folders has been checked, we only display the folders that don't match the filter.
					// in that case we don't apply filters for ACE's or members (because those matches have been excluded from the overview using NOT IN...
					boolean applyNoFilterBecauseInvertFolders = filterConfigApplied.isInvertFolders() && !objectType.equals(ObjectType.FILE);
					
					if (!applyNoFilterBecauseInvertFolders) {
						
						baseQuery.append(" ((");
						
						// ----------------------------------------------------------
						// first apply all the INCLUDE filters
						// ----------------------------------------------------------
						boolean includeQueryFound = false;
												
						// apply include filters on members
						if (!onlyUseExcludes && !applyNoFilterBecauseInvertFolders) {
							includeQueryFound = includeQueryFound | addFilterOnMemberName(baseQuery, queryParameters, true);
							includeQueryFound = includeQueryFound | addFilterOnMemberId(baseQuery, true, objectType);
							includeQueryFound = includeQueryFound | addFilterOnGroupName(baseQuery, queryParameters, true, objectType);
							includeQueryFound = includeQueryFound | addFilterOnGroupId(baseQuery, true, objectType);
							includeQueryFound = includeQueryFound | addFilterOnMemberRDN(baseQuery, true);
						
							// owner filter
							if ((objectType.equals(ObjectType.FILE) || objectType.equals(ObjectType.MEMBERS_IN_FILE_TREE)) && hasFilterOnFile) {
								includeQueryFound = includeQueryFound | addFilterOnOwnerId(baseQuery, true);
							}
						}
						
						if (includeQueryFound) {
							// close the filter on member
							baseQuery.append(")");
						} else {
							// remove the parentheses because no member filters have been found
							baseQuery.deleteCharAt(baseQuery.length() - 1);
						}
						
						// filter on is_group?
						if (filterConfigApplied.isIncludeUsers() != filterConfigApplied.isIncludeGroups()) {
							if (!baseQuery.toString().trim().endsWith("(")) {
								baseQuery.append(" AND ");
							}
							
							baseQuery.append("FILTER_MEMBER.IS_GROUP = " + Boolean.toString(filterConfigApplied.isIncludeGroups()));
							includeQueryFound = true;
						}
						
						// exclude is_builtin?
						if (!filterConfigApplied.isIncludeBuiltIn()) {
							if (!baseQuery.toString().trim().endsWith("(")) {
								baseQuery.append(" AND ");
							}
							
							baseQuery.append("FILTER_MEMBER.IS_BUILTIN = FALSE");
							includeQueryFound = true;
						}
						
						// exclude AD groups and users?
						if (!filterConfigApplied.isIncludeADGroupsAndUsers()) {
							if (!baseQuery.toString().trim().endsWith("(")) {
								baseQuery.append(" AND ");
							}
							
							baseQuery.append("FILTER_MEMBER.GROUP_TYPE = " + Member.GroupType.COMPUTER_LOCAL.ordinal());
							includeQueryFound = true;
						}
						
						// exclude disabled users?
						if (!filterConfigApplied.isIncludeDisabledUsers()) {
							if (!baseQuery.toString().trim().endsWith("(")) {
								baseQuery.append(" AND ");
							}
							
							baseQuery.append("(FILTER_MEMBER.IS_ENABLED = TRUE OR FILTER_MEMBER.IS_ENABLED IS NULL)");
							includeQueryFound = true;
						}
						
						// apply include filters on permissions
						if (!objectType.equals(ObjectType.MEMBER) && !objectType.equals(ObjectType.GROUP_MEMBERS) && !objectType.equals(ObjectType.MEMBER_COUNT) && hasFilterOnPermission && filterConfigApplied.isIncludePermissions()) {
							// append AND(....)
							if (includeQueryFound) {
								baseQuery.append(" AND ");
							}
							baseQuery.append(" (");
							includeQueryFound = includeQueryFound | addFilterOnPermissions(baseQuery);
							baseQuery.append(") ");
						}
						
						// filter out inherited permissions?
						if (!objectType.equals(ObjectType.MEMBER) && !objectType.equals(ObjectType.GROUP_MEMBERS) && !objectType.equals(ObjectType.MEMBER_COUNT) && filterConfigApplied.isHideInheritedPermissions()) {
							if (includeQueryFound) {
								baseQuery.append(" AND ");
							}
							
							baseQuery.append("ACLITEM.FLAG_INHERITED_ACE = false");
													
							includeQueryFound = true;
						}
						
						// is the quick file search filter filled out? (the search box in the top right corner of the application)
						if (quickFileSearch.length() > 0 && objectType.equals(ObjectType.FILE)) {
							if (includeQueryFound) {
								baseQuery.append(" AND ");
							}
							quickFileSearch = quickFileSearch.replace('*', '%');
							queryParameters.add(quickFileSearch); // folder name
							
							if (quickFileSearch.contains("%")) {
								baseQuery.append("FILEINFO.NAME_LOWERCASE LIKE ? ");
							} else {
								baseQuery.append("FILEINFO.NAME_LOWERCASE = ? ");
							}
							includeQueryFound = true;
						}
						
						// apply exclude filters on folders (for file queries, member_in_file_tree and ACE's of expanded members in the All Members tab!
						//if (!objectType.equals(ObjectType.MEMBER) && hasFilterOnFile) {
						if ((objectType.equals(ObjectType.FILE) || objectType.equals(ObjectType.MEMBERS_IN_FILE_TREE) || (objectType.equals(ObjectType.ACL_ITEM) && additionalFilterValue.length == 0)) && hasFilterOnFile) {
							includeQueryFound = includeQueryFound | addFilterOnFolders(baseQuery, true);						
						}
						
						if (includeQueryFound) {
							// close the parentheses 
							baseQuery.append(") "); 
						} else {
							// remove the parentheses, because there is no where condition added
							baseQuery.deleteCharAt(baseQuery.length() - 1);
						}
						
						// ----------------------------------------------------------
						// apply EXCLUDE filters
						// ----------------------------------------------------------
						if (!baseQuery.toString().trim().endsWith("WHERE")) {
							baseQuery.append(" AND (");
						} else {
							baseQuery.append(" (");
						}
						boolean excludeQueryFound = false;
						
						excludeQueryFound = excludeQueryFound | addFilterOnMemberName(baseQuery, queryParameters, false);
						excludeQueryFound = excludeQueryFound | addFilterOnMemberId(baseQuery, false, objectType);
						excludeQueryFound = excludeQueryFound | addFilterOnGroupName(baseQuery, queryParameters, false, objectType);
						excludeQueryFound = excludeQueryFound | addFilterOnGroupId(baseQuery, false, objectType);
						excludeQueryFound = excludeQueryFound | addFilterOnMemberRDN(baseQuery, false);
											
						// apply exclude filters for permissions
						if (!objectType.equals(ObjectType.MEMBER) && !objectType.equals(ObjectType.GROUP_MEMBERS) && !objectType.equals(ObjectType.MEMBER_COUNT) && hasFilterOnPermission && !filterConfigApplied.isIncludePermissions()) {
							// append AND(....)
							if (excludeQueryFound) {
								baseQuery.append(" AND ");
							}
							baseQuery.append(" (");
							excludeQueryFound = excludeQueryFound | addFilterOnPermissions(baseQuery);
							baseQuery.append(") ");	
						}
						
						// apply exclude filters on folders (for file queries, member_in_file_tree and ACE's of expanded members in the All Members tab!
						if ((objectType.equals(ObjectType.FILE) || objectType.equals(ObjectType.MEMBERS_IN_FILE_TREE) || (objectType.equals(ObjectType.ACL_ITEM) && additionalFilterValue.length == 0)) && hasFilterOnFile) {
							excludeQueryFound = excludeQueryFound | addFilterOnFolders(baseQuery, false);
							excludeQueryFound = excludeQueryFound | addFilterOnOwnerId(baseQuery, false);
						}
						
						if (excludeQueryFound) {
							// close the parentheses 
							baseQuery.append(") "); 
						} else {
							// remove the "AND (", because there is no where condition added
							if (baseQuery.toString().endsWith("AND (")) {
								baseQuery.delete(baseQuery.length() - 5, baseQuery.length());
							} else {
								// only remove the "(" character
								baseQuery.deleteCharAt(baseQuery.length() - 1);
							}
						}
					}
					
					// ----------------------------------------------------------
					// add additional query
					// ----------------------------------------------------------
					if (additionalFilter != null) {
						if (!baseQuery.toString().trim().endsWith("WHERE")) {	
							baseQuery.append(" AND");
						}
						baseQuery.append(" (");
						baseQuery.append(additionalFilter);
						baseQuery.append(")");
					}
					
					// close the subselect (NOT IN ...) which was added by invertFolders
					if (filterConfigApplied.isInvertFolders() && objectType.equals(ObjectType.FILE)) {
						baseQuery.append(")");
					}
					
					// apply schema
					String queryString = MessageFormat.format(baseQuery.toString(), DBObject.SCHEMA);
							
					// apply the LIMIT if applicable
					queryString = DBObject.addLimitAndOrderBy(queryString, orderBy, null, limit);
					
					// build the prepared statement
					filterQuery = DBObject.getPreparedStatement(queryString);
			
					// cache the query as PreparedStatement
					queryCache.put(cacheKey, filterQuery);
					queryParameterCache.put(cacheKey, queryParameters);
					
					logger.info("Cached filter query: " + queryString  + queryParameters.toString());
				} else {
					String query;
					if (objectType.equals(ObjectType.FILE)) {
						if (useInTreeViewer) {
							// no filter required
							paramCountOfLastFilterQuery = 0;
							return null;
							
						} else {
							// no filter has been applied, but we need to query anyway because we need to sort the list with file names
							// (since the scan agents have been introduced the file names are not sorted anymore in the database)
							
							// apply schema and order by
							query = MessageFormat.format(baseQuery.toString(), DBObject.SCHEMA);
							query = DBObject.addLimitAndOrderBy(query, "FILEINFO.PARENT_FILE_IDS ASC, FILEINFO.NAME_LOWERCASE ASC", null, 0);
							filterQuery = DBObject.getPreparedStatement(query);							
						}
					} else if (objectType.equals(ObjectType.MEMBERS_IN_FILE_TREE)) {
						
						if (filterConfigApplied.isIncludeGroupMembership()) {
							baseQuery.append("INNER JOIN {0}.MEMBERSHIP ON MEMBERSHIP.MEMBER_ID = FILTER_MEMBER.ID INNER JOIN {0}.ACLITEM ON ACLITEM.MEMBER_ID = MEMBERSHIP.GROUP_ID ");
						} else {
							baseQuery.append("INNER JOIN {0}.ACLITEM ON ACLITEM.MEMBER_ID = FILTER_MEMBER.ID ");
						}
						
						// apply schema and order by
						query = MessageFormat.format(baseQuery.toString(), DBObject.SCHEMA);
						query = DBObject.addLimitAndOrderBy(query, orderBy, null, 0);
						filterQuery = DBObject.getPreparedStatement(query);
					} else {
						logger.warn("Invalid object type given while no filter was required: " + objectType);
						return null;
					}
					
					// cache the query as PreparedStatement
					queryCache.put(cacheKey, filterQuery);
					queryParameterCache.put(cacheKey, queryParameters);
					logger.info("Cached filter query: " + query  + queryParameters.toString());
				}
			}

			filterQuery = queryCache.get(cacheKey);
			queryParameters = queryParameterCache.get(cacheKey);
			
			// ----------------------------------------------------------
			// parameters for the query
			// ----------------------------------------------------------
			paramCountOfLastFilterQuery = queryParameters.size();
			int index = 1;
			for (int counter = 0; counter < queryParameters.size(); counter++) {
				
				Object param = queryParameters.get(counter);
				if (param instanceof String) {
					param = param.toString().replace("\\", "\\\\");
				}
				filterQuery.setObject(index, param);
				index++;
			}
			
			if (additionalFilterValue != null) {
				paramCountOfLastFilterQuery += additionalFilterValue.length;
				for (Object value : additionalFilterValue) {
					filterQuery.setObject(index, value);
					index++;
				}					
			}

			if (logger.isDebugEnabled()) {
				logger.debug("Returning query for filtering: " + (filterQuery != null ? filterQuery.toString() : "null"));
			}

			return filterQuery;

		} catch (Exception ex) {
			logger.error("Error while applying filter: " + ex.getMessage() + ", Query: " + baseQuery.toString() + ", PreparedStatement: " + (filterQuery != null ? filterQuery.toString() : "null"), ex);
			SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while applying filter: " + GUILogHandler.getNestedMessage(ex));
			throw new DBException("Error while applying filter: " + GUILogHandler.getNestedMessage(ex), ex);
		}
	}
	
	/**
	 * Determines if there is a filter on folder name, file/folder includes, etc.
	 * 
	 * @return
	 */
	public boolean hasFilterOnFile() {
		return folderFilterListApplied.size() > 1 || filterConfigApplied.isIncludeFiles() != filterConfigApplied.isIncludeFolders() || filterConfigApplied.isIncludeFilesThatBlockInheritance();
	}
			
	/**
	 * Loops through the filters with type {@link FilterType#MEMBER_NAME}. Each MEMBER_NAME filter is applied on the column NAME_LOWERCASE.
	 * 
	 * @return true in case a filter has been added
	 */
	private boolean addFilterOnMemberName(StringBuilder baseQuery, List<Object> queryParameters, boolean processIncludes) {	
		int count = 0;
		for (Filter filter : memberFilterListApplied) {
			if (filter.getType().equals(FilterType.MEMBER_NAME) && filter.isInclude() == processIncludes) {
				queryParameters.add(filter.getValue().toLowerCase().replace('*', '%')); // member name
				count++;
			}
		}

		if (count > 0) {
			String concatenation = processIncludes ? " OR " : " AND ";
			
			if (!baseQuery.toString().trim().endsWith("(")) {
				baseQuery.append(concatenation);
			}
			
			for (int counter = 1; counter <= count; counter++) {
				baseQuery.append("FILTER_MEMBER.NAME_LOWERCASE ");
				if (processIncludes) {
					baseQuery.append("LIKE ?");
				} else {
					baseQuery.append("NOT LIKE ?");
				}
				
				if (counter < count) {
					baseQuery.append(concatenation);
				}
			}
			
			return true;
		}
		
		return false;
	}
	
	/**
	 * Loops through the filters with type {@link FilterType#MEMBER_LDAP_OU}. Each MEMBER_LDAP_OU filter is applied on the column RDN_ID.
	 * 
	 * @return true in case a filter has been added
	 * @throws DBException 
	 */
	private boolean addFilterOnMemberRDN(StringBuilder baseQuery, boolean processIncludes) throws DBException {	
		List<String> rdns = new ArrayList<String>();
		List<Integer> rdnIds = new ArrayList<Integer>();

	
		// first collect all the RDNs from the filters
		for (Filter filter : memberFilterListApplied) {
			if (filter.getType().equals(FilterType.MEMBER_LDAP_OU) && filter.isInclude() == processIncludes) {
				rdns.add(filter.getValue());
			}
		}
		
		// get all RDN id's (including child RDN's) that belong to the selected RDN names
		if (rdns.size() > 0) {
			StringBuilder rdnQuery = new StringBuilder();
			Object[] params = new Object[rdns.size() * 2];
			int index = 0;
			
			for (String rdn : rdns) {
				if (rdnQuery.length() > 0) {
					rdnQuery.append(" OR ");
				}
				rdnQuery.append("RDN = ? OR RDN LIKE ?");
				params[index++] = rdn;
				params[index++] = "%," + rdn; //child RDN's
			}
			
			List<Rdn> allRdns = Rdn.where(rdnQuery.toString(), params);
			
			// get all id's
			for (Rdn rdn : allRdns) {
				rdnIds.add(rdn.getId());
			}
			
		} else {
			// no filters found
			return false;
		}

		// if we don't add anything to the query, we will get an exception because the query uses a WHERE (there is a filter on
		// OU selected, so the WHERE is already added)
		if (rdnIds.size() == 0) {
			rdnIds.add(Integer.MAX_VALUE);
		}
		
		// determine the SQL operator (invert using NOT, equals for one member id and IN for multiple member ids)
		String operator;
		if (rdnIds.size() == 1) {
			operator = processIncludes ? " = " : " <> ";
		} else {
			operator = processIncludes ? " IN " : " NOT IN ";
		}
								
		if (!baseQuery.toString().trim().endsWith("(")) {
			if (processIncludes) {
				baseQuery.append(" OR ");
			} else {
				baseQuery.append(" AND ");
			}
		}
		
		baseQuery.append(" FILTER_MEMBER.RDN_ID").append(operator);
		
		if (rdnIds.size() == 1) {
			baseQuery.append(rdnIds.get(0));
		} else {
			baseQuery.append(rdnIds.toString().replace("[", "(").replace("]", ")"));
		}
		
		return true;
	}

	/**
	 * Loops through the filters with type {@link FilterType#MEMBER_ID} and type {@link FilterType#MEMBER_REPORT}. For nested group membership each MEMBER_ID filter is applied 
	 * on the column FILTER_MEMBER.ID otherwise it is applied directly on ACLITEM.MEMBER_ID
	 * 
	 * @return true in case a filter has been added
	 * @throws DBException 
	 * @throws NumberFormatException 
	 */
	private boolean addFilterOnMemberId(StringBuilder baseQuery, boolean processIncludes, ObjectType objectType) throws NumberFormatException, DBException {
		List<Integer> memberIds = new ArrayList<Integer>();

		// first collect all the id's
		for (Filter filter : memberFilterListApplied) {
			if (filter.getType().equals(FilterType.MEMBER_ID) && filter.isInclude() == processIncludes) {
				memberIds.add(Integer.valueOf(filter.getValue()));
			}
			
			// members from a report?
			if (filter.getType().equals(FilterType.MEMBER_REPORT) && filter.isInclude() == processIncludes) {
				Report report = Report.getById(Integer.parseInt(filter.getValue()));
				List<Member> reportMembers = report.getMembersInFolderTree();
				for (Member member : reportMembers) {
					memberIds.add(member.getId());
				}
			}
		}

		if (memberIds.size() > 0) {
			// determine the SQL operator (invert using NOT, equals for one member id and IN for multiple member ids)
			String operator;
			if (memberIds.size() == 1) {
				operator = processIncludes ? " = " : " <> ";
			} else {
				operator = processIncludes ? " IN " : " NOT IN ";
			}
			
			// determine column
			String column = !filterConfigApplied.isIncludeGroupMembership() && (objectType.equals(ObjectType.FILE) || objectType.equals(ObjectType.ACL_ITEM)) ? " ACLITEM.MEMBER_ID" : " FILTER_MEMBER.ID";
			
			// when a exclude filter is used with group memberships, then we also want to exclude the ACE's that are inherited through group memberships
			if (!processIncludes && filterConfigApplied.isIncludeGroupMembership() && (objectType.equals(ObjectType.FILE) || objectType.equals(ObjectType.ACL_ITEM))) {
				if (!baseQuery.toString().trim().endsWith("(")) {
					baseQuery.append(" AND ");
				}
				if (memberIds.size() == 1) {
					baseQuery.append("ACLITEM.MEMBER_ID <> " + memberIds.get(0));
				} else {
					baseQuery.append("ACLITEM.MEMBER_ID NOT IN " + memberIds.toString().replace("[", "(").replace("]", ")"));
				}
			}
		
			if (!baseQuery.toString().trim().endsWith("(")) {
				if (processIncludes) {
					baseQuery.append(" OR ");
				} else {
					baseQuery.append(" AND ");
				}
			}
			
			baseQuery.append(column).append(operator);
			
			if (memberIds.size() == 1) {
				baseQuery.append(memberIds.get(0));
			} else {
				baseQuery.append(memberIds.toString().replace("[", "(").replace("]", ")"));
			}
			
			return true;
		}
		
		return false;
	}

	/**
	 * Loops through the filters with type {@link FilterType#OWNER_ID}.
	 * 
	 * @return true in case a filter has been added
	 */
	private boolean addFilterOnOwnerId(StringBuilder baseQuery, boolean processIncludes) {
		List<Integer> ownerIds = new ArrayList<Integer>();

		// first collect all the id's
		for (Filter filter : memberFilterListApplied) {
			if (filter.getType().equals(FilterType.OWNER_ID) && filter.isInclude() == processIncludes) {
				ownerIds.add(Integer.valueOf(filter.getValue()));
			}
		}

		if (ownerIds.size() > 0) {
			// determine the SQL operator (invert using NOT, equals for one member id and IN for multiple member ids)
			String operator;
			if (ownerIds.size() == 1) {
				operator = processIncludes ? " = " : " <> ";
			} else {
				operator = processIncludes ? " IN " : " NOT IN ";
			}
			
			if (!baseQuery.toString().trim().endsWith("(")) {
				if (processIncludes) {
					baseQuery.append(" OR ");
				} else {
					baseQuery.append(" AND ");
				}
			}
			
			baseQuery.append(" FILEINFO.OWNER_ID ").append(operator);
			
			if (ownerIds.size() == 1) {
				baseQuery.append(ownerIds.get(0));
			} else {
				baseQuery.append(ownerIds.toString().replace("[", "(").replace("]", ")"));
			}
			
			return true;
		}
		
		return false;
	}
	
	/**
	 * Loops through the filters with type {@link FilterType#MEMBER_GROUP_NAME}. When nested group membership is enabled then this filter 
	 * requires a fully joined query as determined in the buildFilterQuery method:
	 * 
	 * <pre>
	 * [ACL.MEMBER_ID] 
	 * 			>> [NESTED_MEMBERSHIP.GROUP_ID] 		  		all groups of the members found (including the member itself)
	 * 			   [NESTED_MEMBERSHIP.MEMBER_ID] 
	 * 					>> [FILTER_MEMBER.ID] 				  	members of the 'include members of group ABC' filter
	 * 						>> [FILTER_MEMBERSHIP.MEMBER_ID]  
	 * 						   [FILTER_MEMBERSHIP.GROUP_ID]
	 * 								>> [FILTER_GROUP.ID]		filter for 'include members of group ABC'
	 * </pre>
	 * 
	 * When nested group membership is <b>not</b> used the following inner joins are used:
	 * 
	 * <pre>
	 * [ACL.MEMBER_ID]   	
	 * 		>> [FILTER_MEMBERSHIP.MEMBER_ID]  	members of the 'include members of group ABC' filter
	 * 		   [FILTER_MEMBERSHIP.GROUP_ID]
	 * 				>> [FILTER_GROUP.ID]		filter for 'include members of group ABC'
	 * </pre>
	 * 
	 * <pre>
	 * 
	 * When the objectType {@link ObjectType#MEMBER} is set the following inner joins are used:
	 * 
	 * <pre>
	 * [filter_member.id] > [filter_membership.member_id]
	 *		[filter_membership.group_id] > [filter_group.id] 
	 * </pre>
	 * 
	 * @return true in case a filter has been added
	 */
	private boolean addFilterOnGroupName(StringBuilder baseQuery, List<Object> queryParameters, boolean processIncludes, ObjectType objectType) {
		int count = 0;
		for (Filter filter : memberFilterListApplied) {
			if (filter.getType().equals(FilterType.MEMBER_GROUP_NAME) && filter.isInclude() == processIncludes) {
				queryParameters.add(filter.getValue().toLowerCase().replace('*', '%')); // group name (member name)

				// add the parameter a second time for excludes of the nested group memberships (see query below)
				if (!processIncludes && filterConfigApplied.isIncludeGroupMembership() && (objectType.equals(ObjectType.FILE) || objectType.equals(ObjectType.ACL_ITEM))) {
					queryParameters.add(filter.getValue().toLowerCase().replace('*', '%')); // group name (member name)
				}
				count++;
			}
		}	
		
		if (count > 0) {
			String concatenation = processIncludes ? " OR " : " AND ";
			
			if (!baseQuery.toString().trim().endsWith("(")) {
				baseQuery.append(concatenation);
			}
			
			for (int counter = 1; counter <= count; counter++) {
				if (processIncludes) {
					baseQuery.append(" FILTER_GROUP.NAME_LOWERCASE LIKE ?");
				} else {
					baseQuery.append(" FILTER_MEMBERSHIP.MEMBER_ID NOT IN (SELECT MEMBER_ID FROM PERMISSIONANALYZER.MEMBERSHIP INNER JOIN PERMISSIONANALYZER.MEMBER EXCLUDE_GROUP ON MEMBERSHIP.GROUP_ID = EXCLUDE_GROUP.ID WHERE EXCLUDE_GROUP.NAME_LOWERCASE LIKE ?)");
				}
				
				// when a exclude filter is used with group memberships, then we also want to exclude the ACE's that are inherited through group memberships
				if (!processIncludes && filterConfigApplied.isIncludeGroupMembership() && (objectType.equals(ObjectType.FILE) || objectType.equals(ObjectType.ACL_ITEM))) {
						baseQuery.append(" AND ACLITEM.MEMBER_ID NOT IN (SELECT MEMBER_ID FROM PERMISSIONANALYZER.MEMBERSHIP INNER JOIN PERMISSIONANALYZER.MEMBER EXCLUDE_GROUP ON MEMBERSHIP.GROUP_ID = EXCLUDE_GROUP.ID WHERE EXCLUDE_GROUP.NAME_LOWERCASE LIKE ?)");
				}

				if (counter < count) {
					baseQuery.append(concatenation);
				}
			}
			
			return true;
		}
		return false;
	}

	/**
	 * Loops through the filters with type {@link FilterType#MEMBER_GROUP_ID}.
	 * 
	 * @return true in case a filter has been added
	 * @see {@link #addFilterOnGroupName(StringBuilder, List, boolean)}
	 */
	private boolean addFilterOnGroupId(StringBuilder baseQuery, boolean processIncludes, ObjectType objectType) {
		List<Integer> groupIds = new ArrayList<Integer>();

		// first collect all the id's
		for (Filter filter : memberFilterListApplied) {
			if (filter.getType().equals(FilterType.MEMBER_GROUP_ID) && filter.isInclude() == processIncludes) {
				groupIds.add(Integer.valueOf(filter.getValue()));
			}
		}

		if (groupIds.size() > 0) {
					
			// determine the SQL operator (invert using NOT, equals for one member id and IN for multiple member ids)
			String condition = "";
			if (processIncludes) {
				if (groupIds.size() == 1) {
					condition = "(( FILTER_MEMBERSHIP.GROUP_ID = ";
				} else {
					condition = "(( FILTER_MEMBERSHIP.GROUP_ID IN ";
				}
			} else {
				// when a exclude filter is used with group memberships, then we also want to exclude the ACE's that are inherited through group memberships
				if (!processIncludes && filterConfigApplied.isIncludeGroupMembership() && (objectType.equals(ObjectType.FILE) || objectType.equals(ObjectType.ACL_ITEM))) {
					if (groupIds.size() == 1) {
						condition = "(ACLITEM.MEMBER_ID NOT IN (SELECT MEMBER_ID FROM PERMISSIONANALYZER.MEMBERSHIP WHERE GROUP_ID = " + groupIds.get(0) + ") OR ACLITEM.MEMBER_ID = " + groupIds.get(0) + ") AND ";
					} else {
						condition = "(ACLITEM.MEMBER_ID NOT IN (SELECT MEMBER_ID FROM PERMISSIONANALYZER.MEMBERSHIP WHERE GROUP_ID IN " + groupIds.toString().replace("[", "(").replace("]", ")") + ") OR ACLITEM.MEMBER_ID IN " + groupIds.toString().replace("[", "(").replace("]", ")") + ") AND ";
					}
				}

				if (groupIds.size() == 1) {
					condition += "(( FILTER_MEMBERSHIP.MEMBER_ID NOT IN (SELECT MEMBER_ID FROM PERMISSIONANALYZER.MEMBERSHIP WHERE GROUP_ID = ";
				} else {
					condition += "(( FILTER_MEMBERSHIP.MEMBER_ID NOT IN (SELECT MEMBER_ID FROM PERMISSIONANALYZER.MEMBERSHIP WHERE GROUP_ID IN ";
				}
			}
			
			if (!baseQuery.toString().trim().endsWith("(")) {
				if (processIncludes) {
					baseQuery.append(" OR ");
				} else {
					baseQuery.append(" AND ");
				}
			}
						
			baseQuery.append(condition);
			
			if (groupIds.size() == 1) {
				baseQuery.append(groupIds.get(0));
			} else {
				baseQuery.append(groupIds.toString().replace("[", "(").replace("]", ")"));
			}
			
			if (processIncludes) {
				// don't include the group itself, only its members
				if (groupIds.size() == 1) {
					baseQuery.append(") AND FILTER_MEMBERSHIP.MEMBER_ID <> ");
					baseQuery.append(groupIds.get(0));
					baseQuery.append(")");
				} else {
					baseQuery.append(") AND FILTER_MEMBERSHIP.MEMBER_ID NOT IN ");
					baseQuery.append(groupIds.toString().replace("[", "(").replace("]", ")"));
					baseQuery.append(")");
				}
			} else {
				// don't exclude the group itself, only its members
				if (groupIds.size() == 1) {
					baseQuery.append(")) OR FILTER_MEMBERSHIP.MEMBER_ID = ");
					baseQuery.append(groupIds.get(0));
					baseQuery.append(")");
				} else {
					baseQuery.append(")) OR FILTER_MEMBERSHIP.MEMBER_ID IN ");
					baseQuery.append(groupIds.toString().replace("[", "(").replace("]", ")"));
					baseQuery.append(")");
				}
			}
			
			return true;
		}
		
		return false;
	}
	
	/**
	 * Loops through the filters with type {@link FilterType#PERMISSION_ALLOW}, {@link FilterType#PERMISSION_DENY}.
	 * <p>
	 * When {@link FilterConfig#isAceNeedAllPermissions()} is true then a JOIN with the table PERMISSION_LOOKUP is created and for each
	 * special permission a where condition is added, like P2=1 AND P3=1 AND P8=1. The simple permissions are ignored, because the corresponding
	 * special permissions are automatically selected by the GUI when a simple permission is checked. Only when all the special permissions are selected (FULL)
	 * then the filter is applied on the PERMISSION_ID column in the ACLITEM table, because there can only be one match.
	 * </p>
	 * <p>
	 * When {@link FilterConfig#isAceNeedAllPermissions()} is false then simple permissions are directly compared with the PERMISSION_ID column and special permissions
	 * are compared with their corresponding column in the PERMISSIONLOOKUP table.
	 * </p>
	 * @return true in case a filter has been added
	 */
	private boolean addFilterOnPermissions(StringBuilder baseQuery) {
		boolean aceNeedAllPermissions = getFilterConfig(false).isAceNeedAllPermissions();
		boolean include = getFilterConfig(false).isIncludePermissions();
		boolean filterAdded = false;
		
		String operator = include ? " OR " : " AND ";
		String compareOperator = include ? " = " : " <> ";
		
		// first determine how many ALLOW and DENY permissions we have
		List<Integer> allowPermissionValues = new ArrayList<Integer>();
		List<Integer> denyPermissionValues = new ArrayList<Integer>();
		for (Filter filter : permissionFilterListApplied) {
			if (filter.getType().equals(FilterType.PERMISSION_ALLOW)) {
				allowPermissionValues.add(Integer.parseInt(filter.getValue()));
			} else if (filter.getType().equals(FilterType.PERMISSION_DENY)) {
				denyPermissionValues.add(Integer.parseInt(filter.getValue()));
			}
		}
		
		// if the allow permissions and deny permissions are equal then don't filter on the permission type
		boolean processAllow = (filterConfigApplied.isAceNeedAllPermissions() && allowPermissionValues.size() > 0)
									|| (!filterConfigApplied.isAceNeedAllPermissions() && allowPermissionValues.size() != 0);
		boolean processDeny = (filterConfigApplied.isAceNeedAllPermissions() && denyPermissionValues.size() > 0)
									|| (!filterConfigApplied.isAceNeedAllPermissions() && denyPermissionValues.size() != 0);
		boolean allowAndDenyEqual = allowPermissionValues.containsAll(denyPermissionValues) && allowPermissionValues.size() == denyPermissionValues.size() && processAllow == processDeny;
		 
		// loop through this process two times, first for the Allow permissions and then for the Deny permissions
		// only when allowAndDenyEqual = true this loop is only done once, because there is no need to filter on a specific permission type
		// (the filterType field is reset at the end of the loop)
		FilterType filterType = processAllow ? FilterType.PERMISSION_ALLOW : FilterType.PERMISSION_DENY;
		for (int round = 1; round <= 2; round++) {
			
			if (round == 2) {
				baseQuery.append(operator);
			}
			
			// determine the special permissions selected by the permission filters
			List<Integer> specialPermissionsInFilter = new ArrayList<Integer>();
			for (Filter filter : permissionFilterListApplied) {
				int filterValue = Integer.parseInt(filter.getValue());
				if (filter.getType().equals(filterType) && filterValue < FilterPermissionPart.SIMPLE_PERMISSIONS_INDEX_OFFSET) {
					// special permission
					specialPermissionsInFilter.add(filterValue);
				}
			}
			
			if (aceNeedAllPermissions) {			
				// found any filters for Allow or Deny?
				if (specialPermissionsInFilter.size() > 0) {
					filterAdded = true;
					
					// full permissions selected? Then just filter on that specific permissionId
					// This doesn't work for other simple permissions, because a ACE might have READ + Delete files and still match the READ filter
					SimplePermission simplePermission = PermissionUtil.getSimplePermission(specialPermissionsInFilter, false, false);
					if (SimplePermission.FULL.equals(simplePermission)) {
						baseQuery.append(" (ACLITEM.PERMISSION_ID ");
						baseQuery.append(compareOperator);
						baseQuery.append(PermissionLookup.getIdBySpecialPermissions(specialPermissionsInFilter));
						
					} else {
						// use the lookup table to select ACE's that have all the selected special permissions
						// for Include this query looks like <previous conditions> OR (P0=1 AND P8=1 AND P12=1 AND ACLITEM.IS_DENY = <deny>)
						// for Exclude this query looks like <previous conditions> AND (P0<>1 OR P8<>1 OR P12<>1 OR ACLITEM.IS_DENY <> <deny>)
						baseQuery.append("(");
						
						for (int counter = 0; counter < specialPermissionsInFilter.size(); counter++) {
							baseQuery.append(" P" + specialPermissionsInFilter.get(counter)); // column name like P1
							baseQuery.append(compareOperator);
							baseQuery.append("1");
							
							if (counter < specialPermissionsInFilter.size() - 1) {
								baseQuery.append(include ? " AND " : " OR "); //JUST THE OPPOSITE OF THE operator DECLARED ABOVE!
							}
						}
					}
					
					// include ALL has only allow or deny
					baseQuery.append(include ? " AND " : " OR "); //JUST THE OPPOSITE OF THE operator DECLARED ABOVE!
					baseQuery.append("ACLITEM.IS_DENY");
					baseQuery.append(compareOperator);
					baseQuery.append(filterType == FilterType.PERMISSION_DENY);
					baseQuery.append(")");
				}
			} else {
				// query ACE's that have ANY of the selected permissions
				// for Include this query look like <previous conditions> OR ((P0=1 OR P3=1 OR ACLITEM.PERMISSION_ID=128 OR ACLITEM.PERMISSION_ID=256) AND ACLITEM.IS_DENY = <deny>)
				// for Exclude this query look like <previous conditions> AND ((P0<>1 AND P3<>1 AND ACLITEM.PERMISSION_ID<>128 AND ACLITEM.PERMISSION_ID<>256) OR ACLITEM.IS_DENY <> <deny>)
				
				// if all allow or deny permissions are selected, then only filter on ACLITEM.IS_DENY = true (or ACLITEM.IS_DENY = false)
				// 19 = all permissions (13 special and 6 simple)
				if ((filterType.equals(FilterType.PERMISSION_ALLOW) && allowPermissionValues.size() == 19) ||
						(filterType.equals(FilterType.PERMISSION_DENY) && denyPermissionValues.size() == 19)) {
					
					baseQuery.append("ACLITEM.IS_DENY");
					baseQuery.append(compareOperator);
					baseQuery.append(filterType == FilterType.PERMISSION_DENY);
					filterAdded = true;
					
				} else {
				
					baseQuery.append("((");
					boolean foundFilter = false;
					
					for (Filter filter : permissionFilterListApplied) {
						if (!filter.getType().equals(FilterType.PERMISSION_FILTER_CONFIG)) {
							int filterValue = Integer.parseInt(filter.getValue());
							if (filter.getType().equals(filterType)) {
								
								foundFilter = true;
								filterAdded = true;
								
								if (!baseQuery.toString().trim().endsWith("((")) {
									baseQuery.append(operator);
								}
								
								if (filterValue < FilterPermissionPart.SIMPLE_PERMISSIONS_INDEX_OFFSET) {
									// special permission
									baseQuery.append(" P" + filterValue); // column name like P1
									baseQuery.append(compareOperator);
									baseQuery.append("1");
									
								} else {
									// simple permission
									List<Integer> specialPermissions = Arrays.asList(PermissionLookup.permissionMappings.get(SimplePermission.values()[filterValue - FilterPermissionPart.SIMPLE_PERMISSIONS_INDEX_OFFSET]));
									baseQuery.append(" ACLITEM.PERMISSION_ID ");
									baseQuery.append(compareOperator);
									baseQuery.append(PermissionLookup.getIdBySpecialPermissions(specialPermissions));
								}
							}
						}
					}
					
					if (foundFilter) {
						baseQuery.append(") ");
						
						if (!allowAndDenyEqual) {
							baseQuery.append(include ? " AND " : " OR "); //JUST THE OPPOSITE OF THE operator DECLARED ABOVE!
							baseQuery.append("ACLITEM.IS_DENY");
							baseQuery.append(compareOperator);
							baseQuery.append(filterType == FilterType.PERMISSION_DENY);
						} 
						
						baseQuery.append(")");
						
					} else {
						// no query applied, remove the (( characters
						baseQuery.deleteCharAt(baseQuery.length() - 1);
						baseQuery.deleteCharAt(baseQuery.length() - 1);
					}
				}
			}
			
			//next round process the DENY permissions, if applicable
			if (!allowAndDenyEqual && processDeny && processAllow) {
				filterType = FilterType.PERMISSION_DENY;
			} else {
				break;
			}
		}
				
		return filterAdded;
	}
	
	/**
	 * Loops through the filters with type {@link FilterType#FOLDER_PATH} and {@link FilterType#FOLDER_NAME}. Each FOLDER_NAME filter is applied on the column NAME_LOWERCASE and
	 * each FOLDER_PATH filter is applied on the column PARENT_FILE_IDS (which contains values like "1,45,234,2342")
	 * 
	 * @return true in case a filter has been added
	 * @throws DBException 
	 */
	private boolean addFilterOnFolders(StringBuilder baseQuery, boolean processIncludes) throws DBException {	
		boolean addedDirectoryCheck = false;
		int count = 0;
		String concatenation = processIncludes ? " OR " : " AND ";
		
		if (filterConfigApplied.isIncludeFiles() != filterConfigApplied.isIncludeFolders() && processIncludes) {
			// start the whole WHERE clause with either ( or AND (
			if (!baseQuery.toString().trim().endsWith("(")) {
				baseQuery.append(" AND (");  
			} else {
				baseQuery.append(" (");
			}
			
			baseQuery.append("FILEINFO.IS_DIRECTORY = ?) ");
			queryParameters.add(filterConfigApplied.isIncludeFolders());
			
			addedDirectoryCheck = true;
		}
		
		if (filterConfigApplied.isIncludeFilesThatBlockInheritance() && processIncludes) {
			// start the whole WHERE clause with either ( or AND (
			if (!baseQuery.toString().trim().endsWith("(")) {
				baseQuery.append(" AND (");  
			} else {
				baseQuery.append(" (");
			}
			
			baseQuery.append("FILEINFO.IS_DACL_PROTECTED = true) ");			
			addedDirectoryCheck = true;
		}
		
//		FIRST LOOP -  NOT ON TOP OF OTHER FILTERS:
//		======================================
//		Without subfolders and not on top of other filters:
//		(FileInfo.ParentIDs like "2,45,63,%" OR FileInfo.ParentIDs like "4,87,98,%" OR FileInfo.Name LIKE "%test1%")
//
//		With subfolders and not on top other other filters:
//		<list ids> = Name like "%test1%"
//		(FileInfo.ParentIDs like "2,45,63,%" OR FileInfo.ParentIDs like "4,87,98,%" OR FileInfo.ParentIDs LIKE <list ids>)				
//
//		SECOND LOOP - ON TOP OF OTHER FILTERS:
//		======================================
//		Without subfolders and on top of other filters:
//		((FileInfo.ParentIDs like "2,45,63,%" OR FileInfo.ParentIDs like "4,87,98,%") AND FileInfo.Name LIKE "%test1%")
//		
//		With subfolders and on top other other filters:
//		<list ids> = (FileInfo.ParentIDs like "2,45,63,%" OR FileInfo.ParentIDs like "4,87,98,%") AND Name LIKE "%test1%"
//		(FileInfo.ParentIDs like <list ids>)
		
		boolean processTheFiltersOnTopOfOthers = false;
		// loop two times through the filters, once for the normal filters and once for the filters with the checkbox "Apply after other filters"
		do {
			// we have two different filters, the regular folder filters, and the filters that must be applied after the regular filters (on the result, instead of the entire directory tree)
			// we loop through the filters twice to process both the filter types. This sub-counter variable indicates the number of filters per type 
			int subcount = 0;
			for (Filter filter : folderFilterListApplied) {
				if ((filter.getType().equals(FilterType.FOLDER_NAME) || filter.getType().equals(FilterType.FOLDER_PATH)) && filter.isInclude() == processIncludes) {
					
					String filterValue = filter.getValue();
					
					// process the possible filter prefixes for wild card searches 
					// (the two flags "Include subfolders" and "Use wildcard filter on the result of other filters, instead of the entire directory tree")
					boolean isFilterAfterOthers = false;
					boolean includeSubfolders = false;
					if (filterValue.contains(FilterHelper.PREFIX_FOLDER_WILDCARD_FILTER_AFTER_OTHER_FILTERS)) {
						isFilterAfterOthers = true;
					}
					if (filterValue.contains(FilterHelper.PREFIX_FOLDER_WILDCARD_FILTER_INCLUDE_SUBFOLDERS)) {
						includeSubfolders = true;
					}
					
					// remove the filter prefixes and apply to lowercase (same as the database)
					filterValue = filterValue.replace(FilterHelper.PREFIX_FOLDER_WILDCARD_FILTER_AFTER_OTHER_FILTERS, "").replace(FilterHelper.PREFIX_FOLDER_WILDCARD_FILTER_INCLUDE_SUBFOLDERS, "").toLowerCase();
					
					if (processTheFiltersOnTopOfOthers == isFilterAfterOthers) {
					
						// start the whole WHERE clause with either ( or AND (
						if (!baseQuery.toString().trim().endsWith("(") && subcount == 0) {
							baseQuery.append(" AND (");
						} else if (subcount == 0){
							baseQuery.append(" (");
						}
										
						if (!baseQuery.toString().trim().endsWith("(")) {
							baseQuery.append(concatenation);
						}
						
						if (filter.getType().equals(FilterType.FOLDER_PATH)) {
							queryParameters.add(filterValue); // exactly this folder (based on parent_file_ids)
							queryParameters.add(filterValue + ",%"); // all folders that start with this comma separated parent_file_ids
							
							if (processIncludes) {
								baseQuery.append("FILEINFO.PARENT_FILE_IDS = ? OR FILEINFO.PARENT_FILE_IDS LIKE ? ");
							} else {
								baseQuery.append("FILEINFO.PARENT_FILE_IDS <> ? AND FILEINFO.PARENT_FILE_IDS NOT LIKE ? ");
							}
						} else {
							
							// replace the * wildcard to a database wildcard (%)
							filterValue = filterValue.replace('*', '%');
							
							// don't include subfolders
							if (!includeSubfolders) {
								// and use a wildcard search
								if (filterValue.contains("%")) {
									queryParameters.add(filterValue); // folder name
									baseQuery.append(processIncludes ? "FILEINFO.NAME_LOWERCASE LIKE ? " : "FILEINFO.NAME_LOWERCASE NOT LIKE ? ");
								} else {
									 // use an exact search
									 queryParameters.add(filterValue); // folder name
									baseQuery.append(processIncludes ? "FILEINFO.NAME_LOWERCASE = ? " : "FILEINFO.NAME_LOWERCASE <> ? ");
								 }
							} else {
								// include subfolders, meaning first retrieve all matching folders, then add the folder Id's to the main query
								// We cann't add a query clause like baseQuery.append("FILEINFO.NAME_LOWERCASE NOT LIKE ? ");
								// because that will only include or exclude the parent folders, but not the underlying folders/files!
								// we need all the matching files and then filter on their PARENT_FILE_IDS attribute.
								// one more thing, if this filter is applied on top of other filters, then we need to include those filters
								// as well when retrieving all folders with a matching wildcard name (not just search the entire table!)
								StringBuilder queryPrefix = new StringBuilder("");
								
								// not present yet, get the list from the database
								if (isFilterAfterOthers) {
									// we need to include the other filters as well
									for (Filter scopeFilter : folderFilterListApplied) {
										if (scopeFilter.getType().equals(FilterType.FOLDER_PATH) && scopeFilter.isInclude()) {
											if (queryPrefix.length() == 0) {
												queryPrefix.append("(");
											} else {
												queryPrefix.append("OR ");
											}
											
											queryPrefix.append("FILEINFO.PARENT_FILE_IDS = '" + scopeFilter.getValue() + "' OR FILEINFO.PARENT_FILE_IDS LIKE '" + scopeFilter.getValue() + ",%' ");
										}
									}
									
									if (queryPrefix.length() > 0) {
										queryPrefix.append(") AND ");
									}
								}
								
								List<FileInfo> matchingFiles = FileCache.getWildcardMatches(queryPrefix.toString() + filterValue);
								if (matchingFiles == null) {
									if (filterValue.contains("%")) {
										matchingFiles = FileInfo.where(queryPrefix.toString() + "NAME_LOWERCASE LIKE ?", new Object[]{filterValue});
									} else {
										matchingFiles = FileInfo.where(queryPrefix.toString() + "NAME_LOWERCASE = ?", new Object[]{filterValue});
									}
									
									// cache the result
									FileCache.cacheWildcardSearch(queryPrefix.toString() + filterValue, matchingFiles);
								}
								
								// now build the right filter for the main query
								for (int counter = 0; counter < matchingFiles.size(); counter++) {
									FileInfo file = matchingFiles.get(counter);
									
									if (processIncludes) {
										baseQuery.append("FILEINFO.PARENT_FILE_IDS = ''" + file.getParentFileIds() + "'' OR FILEINFO.PARENT_FILE_IDS LIKE ''" + file.getParentFileIds() + ",%'' ");
									} else {
										baseQuery.append("FILEINFO.PARENT_FILE_IDS <> ''" + file.getParentFileIds() + "'' AND FILEINFO.PARENT_FILE_IDS NOT LIKE ''" + file.getParentFileIds() + ",%'' ");
									}
															
									if (counter < matchingFiles.size() - 1) {
										baseQuery.append(processIncludes ? "OR " : "AND ");
									}
								}
																		
								// no filters apply? Make sure the WHERE clause results in a valid syntax
								if (matchingFiles.isEmpty()) {
									if (processIncludes) {
										baseQuery.append("1 = 0 ");
									} else {
										baseQuery.append("1 = 1");
									}
								}
							}
						}
						
						count++;
						subcount++;
					}	
				}
			}
			
			if (subcount > 0) {
				baseQuery.append(")");
			}
			
			// toggle flag (this do-while loop will run twice)
			processTheFiltersOnTopOfOthers = !processTheFiltersOnTopOfOthers;
		} while (processTheFiltersOnTopOfOthers);
		
		return addedDirectoryCheck || count > 0;
	}
	
	/**
	 * Clears the query cache and enforces a rebuild of the queries. New filters are not applied until this cache is cleared.
	 */
	public void applyNewFilter(boolean clearTraceView) {
		queryCache.clear();
		queryParameterCache.clear();
		
		if (clearTraceView) {
			TracePart.clearSelectedMember();
		}
		
		filterConfigApplied = new FilterConfig(getMemberConfigFilter(), getPermissionConfigFilter(), getFolderConfigFilter());
		
		// copy the current GUI values to the 'applied' lists which are used to build the queries
		memberFilterListApplied.clear();
		memberFilterListApplied.addAll(memberFilterList);
		permissionFilterListApplied.clear();
		permissionFilterListApplied.addAll(permissionFilterList);
		folderFilterListApplied.clear();
		folderFilterListApplied.addAll(folderFilterList);
		
		allFilteredMembers = -1;
		MembershipCache.clearFilteredMembershipCountCache();
	}
	
	/**
	 * Checks if the applied filters differ from the current filters (like memberFilterListApplied.equals(memberFilterList))
	 *  
	 * @return true in case the current filters don't match the filters applied in the overview
	 */
	public boolean hasUnappliedChanges() {
		boolean hasUnappliedChanges = !memberFilterListApplied.equals(memberFilterList) ||
									  !permissionFilterListApplied.equals(permissionFilterList) ||
									  !folderFilterListApplied.equals(folderFilterList);
		
		return hasUnappliedChanges;
	}
		
	/**
	 * Returns an object containing filter configurations. There is a difference between 'applied' values and 'current' values.
	 * The Filter composites in the GUI are interested in the current values, but the Trace view for example needs to know the
	 * values when the button "Apply filter" was pressed.
	 * 
	 * @param useCurrentValues true to use values set by the GUI or false to use values applied in the file Tree
	 * 
	 * @return
	 */
	public FilterConfig getFilterConfig(boolean useCurrentValues) {
		if (useCurrentValues) {
			return new FilterConfig(getMemberConfigFilter(), getPermissionConfigFilter(), getFolderConfigFilter());
		} else {
			return filterConfigApplied;
		}
	}
	
	/**
	 * Returns the filter of type {@link FilterType#FILTER_CONFIG} in the list {@link #memberFilterList}
	 * 
	 * @return Filter with all the member configs (checkbox values etc)
	 */
	private Filter getMemberConfigFilter() {
		// get the member filter config
		for (Filter filter : memberFilterList) {
			if (filter.getType().equals(FilterType.MEMBER_FILTER_CONFIG)) {
				return filter;
			}
		}
		
		// filter not found yet, create a new one
		Filter filterConfig = new Filter(FilterType.MEMBER_FILTER_CONFIG, true, "00000", "Configuration for the member filters");
		memberFilterList.add(filterConfig);
		return filterConfig;
	}
	
	/**
	 * Returns the filter of type {@link FilterType#FILTER_CONFIG} in the list {@link #permissionFilterList}
	 * 
	 * @return Filter with all the permission configs (checkbox values etc)
	 */
	private Filter getPermissionConfigFilter() {
		// get the permission filter config
		for (Filter filter : permissionFilterList) {
			if (filter.getType().equals(FilterType.PERMISSION_FILTER_CONFIG)) {
				return filter;
			}
		}

		// filter not found yet, create a new one
		Filter filterConfig = new Filter(FilterType.PERMISSION_FILTER_CONFIG, true, "00000", "Configuration for the permission filters");
		permissionFilterList.add(filterConfig);
		return filterConfig;		
	}
	
	/**
	 * Returns the filter of type {@link FilterType#FILTER_CONFIG} in the list {@link #folderFilterList}
	 * 
	 * @return Filter with all the file configs (checkbox values etc)
	 */
	private Filter getFolderConfigFilter() {
		// get the file filter config
		for (Filter filter : folderFilterList) {
			if (filter.getType().equals(FilterType.FOLDER_FILTER_CONFIG)) {
				return filter;
			}
		}

		// filter not found yet, create a new one
		Filter filterConfig = new Filter(FilterType.FOLDER_FILTER_CONFIG, true, "00000", "Configuration for the file filters");
		folderFilterList.add(filterConfig);
		return filterConfig;	
	}
	
	public java.util.List<Filter> getMemberFilterList() {
		return memberFilterList;
	}

	public void setMemberFilterList(java.util.List<Filter> memberFilterList) {
		this.memberFilterList = memberFilterList;
	}

	public java.util.List<Filter> getPermissionFilterList() {
		return permissionFilterList;
	}

	public void setPermissionFilterList(java.util.List<Filter> permissionFilterList) {
		this.permissionFilterList = permissionFilterList;
	}

	public java.util.List<Filter> getFolderFilterList() {
		return folderFilterList;
	}

	public void setFolderFilterList(java.util.List<Filter> fileFilterList) {
		this.folderFilterList = fileFilterList;
	}
	
	public static void test(String[] args) {
		/*
		//TIntArrayList testList = new TIntArrayList(10000);
		ArrayList<Integer> testList = new ArrayList(10000);
		//int counter2 = 0;
		
		for (int counter = 0; counter < 10000000; counter++) {
			testList.add(counter);
			
			if (counter % 100000 == 0) {
				double currentMemory = ( (double)((double)(Runtime.getRuntime().totalMemory()/1024)/1024))- ((double)((double)(Runtime.getRuntime().freeMemory()/1024)/1024));
				logger.debug("counter: " + counter);
				logger.debug("currentMemory: " + currentMemory);
			}
		}
		
		logger.debug("DONE, start contains");
		
		boolean contains = testList.contains(7500000);
		
		logger.debug("contains: " + contains);
		
		//int search = testList.binarySearch(7500000);
		//int search = Arrays.binarySearch(, arg1) testList.binarySearch(7500000);
		
		//logger.debug("search: " + search);
		*/
		/*
		//Map<Integer, List<Integer>> files = new HashMap<Integer, List<Integer>>();
		double currentMemory = ( (double)((double)(Runtime.getRuntime().totalMemory()/1024)/1024))- ((double)((double)(Runtime.getRuntime().freeMemory()/1024)/1024));
		
		logger.debug("INIT: " + currentMemory);
		
		TIntArrayList fileList = new TIntArrayList(10000);
		//ArrayList<Integer> fileList = new ArrayList<Integer>(10000);
		Set<Integer> fileSet = new HashSet<Integer>(10000);
		
		ArrayList<String> testItems = new ArrayList<String>(500000);
		
		Random ran = new Random();
		
		for (int counter = 0; counter < 500000; counter++) {
			testItems.add(ran.nextInt(500000) + "," + 
					ran.nextInt(500000) + "," + 
					ran.nextInt(500000) + "," + 
					ran.nextInt(500000) + "," + 
					ran.nextInt(500000) + "," + 
					ran.nextInt(500000) + "," + 
					ran.nextInt(500000) + "," + 
					ran.nextInt(500000) + "," + 
					ran.nextInt(500000) + "," + 
					ran.nextInt(500000));
		}
		
		currentMemory = ( (double)((double)(Runtime.getRuntime().totalMemory()/1024)/1024))- ((double)((double)(Runtime.getRuntime().freeMemory()/1024)/1024));
		logger.debug("Initialized array: " + currentMemory);
		
		for (int counter = 0; counter < 500000; counter++) {
			StringTokenizer tokens = new StringTokenizer(testItems.get(counter), ",");
			while (tokens.hasMoreTokens()) {
				fileList.add(Integer.parseInt(tokens.nextToken()));
				
				//if (fileList.binarySearch(id) < 0) {
				//	fileList.add(id);
				//	fileList.sort();
				//}		
				
				//if (!fileMap.contains(id)) {
					//fileMap.add(id);
				//fileSet.add(id);
				//}
			}
			if (counter % 10000 == 0) {
				logger.debug("processed items: " + counter);
			}
		}
		
		currentMemory = ( (double)((double)(Runtime.getRuntime().totalMemory()/1024)/1024))- ((double)((double)(Runtime.getRuntime().freeMemory()/1024)/1024));
		logger.debug("Done with loop: " + currentMemory);
		
		fileList.sort();
		//fileSet.addAll(fileList);
		
		logger.debug("files set size: " + fileSet.size());
		logger.debug("files list size: " + fileList.size());

		TIntArrayList newFileList = new TIntArrayList(10000);
		for (int counter = 0; counter < fileList.size(); counter++) {
			if (counter == 0 || fileList.getQuick(counter-1) != fileList.getQuick(counter)) {
				newFileList.add(fileList.getQuick(counter));
			}
		}
		
		logger.debug("new file list size: " + newFileList.size());*/
		/*
		PreparedStatement ps;
		try {
			//ps = DBObject.getPreparedStatement("select parentfile.* from PERMISSIONANALYZER.parentfile inner join permissionanalyzer.aclitem on parentfile.file_id = aclitem.fileinfo_id where aclitem.member_id IN (9, 10, 2, 1)");
			ps = DBObject.getPreparedStatement("select distinct parentfile.parent_id from PERMISSIONANALYZER.parentfile inner join permissionanalyzer.aclitem on parentfile.fileinfo_id = aclitem.fileinfo_id where aclitem.member_id IN (9, 10, 2, 1)");
			
			//and parentfile.parent_id <> parentfile.child_id 
			ParentFile.fillListUsingPreparedStatement(ps, files);
		} catch (SQLException e) {
			e.printStackTrace();
		} catch (DBException e) {
			e.printStackTrace();
		}
		
		logger.debug("DONE!");
		logger.debug("MAP SIZE: " + files.size());
		
		logger.debug("CONTAINS 273653: " + files.contains(273653));
		logger.debug("CONTAINS 12876: " + files.contains(12876));
		logger.debug("CONTAINS 123987: " + files.contains(123987));
		logger.debug("CONTAINS 6000: " + files.contains(6000));
		
		logger.debug("DONE! START CLEAR");
		files.clear();
		logger.debug("DONE!");
		*/
		/*Filter filter = new Filter(0, FilterType.MEMBER_GROUP_NAME, true, "gr%1", "group1");
		
		Filter filter2 = new Filter(0, FilterType.MEMBER_NAME, false, "User12", "User12");
		Filter filter3 = new Filter(0, FilterType.MEMBER_NAME, false, "User13", "User13");
		Filter filter0 = new Filter(0, FilterType.MEMBER_INCLUDE_FILTER_MEMBERSHIP, true, "", "");
		
		getInstance().memberFilterList.clear();
		getInstance().memberFilterList.add(filter);
		getInstance().memberFilterList.add(filter0);
		getInstance().memberFilterList.add(filter2);
		getInstance().memberFilterList.add(filter3);
		getInstance().clearCache();
		
		PreparedStatement ps = getInstance().buildFilterQuery(ObjectType.MEMBER, "MEMBERSHIP.GROUP_ID = ?", 3);
		try {
			Membership mship = new Membership(5, 15);
			mship.save();
			
			MessageDialog.openInformation(null, "Test4", Member.getByPreparedStatement(ps).toString());
		} catch (DBException e) {
			e.printStackTrace();
		}*/
		
		/*
		for (int counter = 2; counter < 21; counter++) {
			Membership membership = new Membership(counter, counter);
			try {
				membership.save();
			} catch (DBException e) {
				e.printStackTrace();
			}
		}*/
	}
		
	public TIntArrayList getApplicableFileIds(PreparedStatement ps, final boolean sortAllowed, boolean addRootFolders) throws DBException {
		final TIntArrayList fileIds = new TIntArrayList(10000); 
				
		DBObject.getUsingCustomProcessor(ps, new CustomProcessor(){
			@Override
			public boolean processResultSet(ResultSet rs) throws SQLException {
				StringTokenizer tokens = new StringTokenizer(rs.getString(1), ",");
				while (tokens.hasMoreTokens()) {
					int nextToken = Integer.parseInt(tokens.nextToken());
					// don't allow duplicate folder ids (in case of sortAllowed duplicates are removed below!)
					if (sortAllowed || !fileIds.contains(nextToken)) {
						fileIds.add(nextToken);
					}
				}
				
				return true;
			}
		});
		
		// add the root folders for the overview of the report (customer requirement)
		if (addRootFolders && !hasFilterOnFile() && !filterConfigApplied.isInvertFolders()) {
			List<FileInfo> rootFiles = FileInfo.where("parent_id is null", null);
			for (int counter = 0; counter < rootFiles.size(); counter++) {
				// don't allow duplicate folder ids (in case of sortAllowed duplicates are removed below!)
				if (sortAllowed || !fileIds.contains(rootFiles.get(counter).getId())) {
					fileIds.add(rootFiles.get(counter).getId());
				}
			}
		}
				
		// sort and remove duplicates
		if (sortAllowed) {
			fileIds.sort();
		}
		
		TIntArrayList uniqueFileIds = new TIntArrayList();
				
		for (int counter = 0; counter < fileIds.size(); counter++) {
			if (counter == 0 || fileIds.getQuick(counter-1) != fileIds.getQuick(counter)) {
				uniqueFileIds.add(fileIds.getQuick(counter));
			}
		}
	
		return uniqueFileIds;
	}
	
	/**
	 * The filter types MEMBER_ID, MEMBER_GROUP_ID and FOLDER_PATH refer to id's in the database. These id's may change after a new scan. This method will lookup the 
	 * filter values (like account name, or file path) in the database and updates the filter with the right id. 
	 */
	public static void matchFilterValuesAfterScan() throws DBException {
		List<Filter> filters = Filter.getAll();
		List<Filter> filtersUsedInGUI = new ArrayList<Filter>();
		filtersUsedInGUI.addAll(getInstance().getMemberFilterList());
		filtersUsedInGUI.addAll(getInstance().getPermissionFilterList());
		filtersUsedInGUI.addAll(getInstance().getFolderFilterList());
				
		for (Filter filter : filters) {
			if (filter.getType().equals(FilterType.MEMBER_ID) || filter.getType().equals(FilterType.MEMBER_GROUP_ID) || filter.getType().equals(FilterType.OWNER_ID)) { 
				// lookup the account name
				// the filter name has a format of <domain>\\<account>[<display name>], so extract the account name
				String accountName = filter.getName();
				
				int endIndex = accountName.indexOf(" [");
				if (endIndex >= 0) {
					accountName = accountName.substring(0, endIndex);
				}
				
				Member member = null;
				int beginIndex = accountName.indexOf("\\");
				if (beginIndex >= 0) {
					// extract domain and account name
					String domain = accountName.substring(0, beginIndex);
					accountName = accountName.substring(beginIndex + 1);
					
					Integer domainId = DomainCache.getDomainId(domain);
					if (domainId != null) {
						member = Member.findFirst("NAME_LOWERCASE = ? AND DOMAIN_ID = ?", new Object[]{accountName.toLowerCase(), domainId});
					} else {
						member = Member.findFirst("NAME_LOWERCASE = ?", new Object[]{accountName.toLowerCase()});
					}
				} else {
					member = Member.findFirst("NAME_LOWERCASE = ?", new Object[]{accountName.toLowerCase()});
				}			
				
				if (member != null) {
					// update the filter
					if (filter.getValue().equals("-1")) {
						logger.info(MessageFormat.format("Filter enabled again: Account {0} has been found in the database.", accountName));
					}
					
					filter.setValue(String.valueOf(member.getId()));
					filter.save();
				} else {
					logger.warn(MessageFormat.format("Filter disabled: Account {0} is not found in the database anymore and the existing filter [{1}] has no usable value anymore.", accountName, filter.getValue()));
					filter.setValue("-1");
					filter.save();
				}
			}
			
			if (filter.getType().equals(FilterType.FOLDER_PATH)) {
				StringBuffer filterDisabledCause = new StringBuffer("");
				
				// lookup the file record that belongs to this file path
				String path = filter.getName(); 
				List<FileInfo> rootFiles = FileInfo.where("PARENT_ID IS NULL", null);
				FileInfo applicableRootFile = null;
				
				for (FileInfo file : rootFiles) { 
					if ((path.startsWith(file.getName()) && applicableRootFile == null) || path.equals(file.getName())) {
						applicableRootFile = file;
					}
				}
				
				FileInfo filterFile = null;
				
				if (applicableRootFile != null) {	
					if (!path.equals(applicableRootFile.getName())) {
						filterDisabledCause.append("Filter path \"").append(path).append("\" does not exactly match the root folder \"").append(applicableRootFile.getName()).append("\". ");
						
						// strip the root folder and lookup the database FileInfo record
						path = path.substring(applicableRootFile.getName().length() + 1);
					
						// split the path into folders
						// Replace forward slash with backward slash, just to be sure
						path = path.replace("/", "\\");
						String[] folders = path.split(Pattern.quote("\\"));
						filterFile = applicableRootFile;
						for (String folder : folders) {
							FileInfo subfolder = FileInfo.findFirst("PARENT_ID = ? AND NAME_LOWERCASE = ?", new Object[]{filterFile.getId(), folder.toLowerCase()});
							
							if (subfolder != null) {
								filterFile = subfolder;
							} else {
								filterDisabledCause.append("Sub-folder \"").append(folder.toLowerCase()).append("\" not found using parent_id \"").append(filterFile.getId()).append("\".");
								filterFile = null;
								break;
							}
						}
					} else {
						filterFile = applicableRootFile;
					}
					
					if (filterFile != null) {
						// update the filter
						if (filter.getValue().equals("-1")) {
							logger.info(MessageFormat.format("Filter enabled again: File {0} has been found in the database.", filter.getName()));
						}
						filter.setValue(String.valueOf(filterFile.getParentFileIds()));
						filter.save();
						
					}
				} else {
					filterDisabledCause.append("Filter not found in the current root folders [");
					for (FileInfo file : rootFiles) {
						filterDisabledCause.append(file.getName());
						filterDisabledCause.append(", ");
					}
					filterDisabledCause.append("].");
				}
				
				if (filterFile == null) {
					// if we get here then the file has not been found in the DB
					logger.warn(MessageFormat.format("Filter disabled: File {0} has not been not found in the database anymore and the corresponding filter has no usable value anymore. " + filterDisabledCause.toString(), filter.getName()));
					filter.setValue("-1");
					filter.save();	
				}
			}
			
			// update the filter value in the GUI (if applicable)
			for (Filter filterInGUI : filtersUsedInGUI) {
				if (filter.getId().equals(filterInGUI.getId())) {
					filterInGUI.setValue(filter.getValue());
				}
			}
		}
	}
	
	/**
	 * Determines if the user has filted on a single member. 
	 * 
	 * @return a Filter object in case the user has filtered on a specific member, otherwise null
	 */
	public Filter isFilterOnSingleMember() {
		// filtered on one member?
		Filter singleMemberFilter = null;
		for (Filter memberFilter : getMemberFilterList()) {
			if (memberFilter.isInclude()) {
				if (memberFilter.getType().equals(FilterType.MEMBER_ID)) {
					if (singleMemberFilter != null) {
						// we found a second filter on a member
						singleMemberFilter = null;
						break;
					} else {
						singleMemberFilter = memberFilter;
					}
					
				} else if (!memberFilter.getType().equals(FilterType.MEMBER_FILTER_CONFIG)) {
					singleMemberFilter = null;
					break;
				}
			}
		}
		
		return singleMemberFilter;
	}
	
	/**
	 * Returns (and holds) the total number of members that match the filter criteria
	 * 
	 * @return member count
	 * @throws DBException 
	 * @see {@link MainFileTreeContentProvider#determineOwnershipRatio(FilterHelper, FileInfo)}
	 */
	public int getAllFilteredMembers() {
		if (allFilteredMembers == -1) {
			try {
				PreparedStatement membersThatMatchFilterPS = buildFilterQuery(ObjectType.MEMBER_COUNT, false, false, 0, null, null);
				ResultSet rs = null;
				
				try {
					rs = DBObject.executePreparedStatement(membersThatMatchFilterPS, false);
					if (rs.next()){
						allFilteredMembers = rs.getInt(1);
					}
				} finally {
					rs.close();
				}
			} catch (Exception ex) {
				logger.error("Error while counting all members that match the filter criteria", ex);
				SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while counting all members that match the filter criteria: " + ex.getMessage());
			}
		}
			
		return allFilteredMembers;
	}
	
	/**
	 * Checks if the number of member filters exceeds the trial limits
	 * 
	 * @param memberFilters this method can either be called from the FilterMemberPart of the New Policy Wizard!
	 * @param true in case this check is called before adding a new filter
	 */
	public static boolean checkNumberOfMemberFiltersForTrial(List<Filter> memberFilters, boolean checkBeforeAdd) {
		int treshhold = Integer.parseInt(MAX_MEMBER_FILTERS_FOR_TRIAL);
		
		// check for the existence of the MemberConfig filter (it might not be present in the New Policy Wizard!)
		boolean foundMemberConfigFilter = false;
		for (Filter filter : memberFilters) {
			if (filter.getType().equals(FilterType.MEMBER_FILTER_CONFIG)) {
				foundMemberConfigFilter = true;
				break;
			}
		}
		
		// add one for the MemberConfig filter, because that filter doesn't count
		if (foundMemberConfigFilter) {
			treshhold++;
		}
		
		if (!checkBeforeAdd) {
			// allow the maximum, since we're not adding a new filter
			treshhold++;
		}
		boolean allowed = !LicenseHelper.getEdition().equals(Edition.TRIAL) || memberFilters.size() < treshhold;
		
		if (!allowed) {
			logger.warn("Maximum number of member filters has been reached!");
			SyncMessageDialog.showMessage(MessageDialog.WARNING, "Reached maximum member filters", "The trial version is limited to 3 member filters, you have already added 3 member filters.");
		}
		
		return allowed;
	}
	
	public boolean isApplyMemberFilter() {
		// 1 is the MemberConfig filter object, and in case invertFolder is true we don't want to apply any member filters
		// because all the matching folders are excluded from the overview
		return getMemberFilterListApplied().size() > 1 && !filterConfigApplied.isInvertFolders();
	}

	public TIntArrayList getApplicableFileIdsIncludingParents() {
		return applicableFileIdsIncludingParents;
	}

	public void setApplicableFileIdsIncludingParents(TIntArrayList applicableFileIdsIncludingParents) {
		this.applicableFileIdsIncludingParents = applicableFileIdsIncludingParents;
	}

	public TIntArrayList getApplicableFileIds() {
		return applicableFileIds;
	}

	public void setApplicableFileIds(TIntArrayList applicableFileIds) {
		this.applicableFileIds = applicableFileIds;
	}

	public int getFilesDisplayedInTree() {
		return filesDisplayedInTree;
	}

	public void setFilesDisplayedInTree(int filesDisplayedInTree) {
		this.filesDisplayedInTree = filesDisplayedInTree;
	}

	public boolean isStopExpandingAll() {
		return stopExpandingAll;
	}

	public void setStopExpandingAll(boolean stopExpandingAll) {
		this.stopExpandingAll = stopExpandingAll;
	}

	public Report getSelectedReport() {
		return selectedReport;
	}

	public void setSelectedReport(Report selectedReport) {
		this.selectedReport = selectedReport;
	}

	public java.util.List<Filter> getMemberFilterListApplied() {
		return memberFilterListApplied;
	}

	public void setMemberFilterListApplied(
			java.util.List<Filter> memberFilterListApplied) {
		this.memberFilterListApplied = memberFilterListApplied;
	}

	public java.util.List<Filter> getPermissionFilterListApplied() {
		return permissionFilterListApplied;
	}

	public void setPermissionFilterListApplied(
			java.util.List<Filter> permissionFilterListApplied) {
		this.permissionFilterListApplied = permissionFilterListApplied;
	}

	public java.util.List<Filter> getFolderFilterListApplied() {
		return folderFilterListApplied;
	}

	public void setFolderFilterListApplied(
			java.util.List<Filter> folderFilterListApplied) {
		this.folderFilterListApplied = folderFilterListApplied;
	}

	public String getQuickFileSearch() {
		return quickFileSearch;
	}

	public void setQuickFileSearch(String quickFileSearch) {
		this.quickFileSearch = quickFileSearch;
	}

	public Text getQuickFileSearchText() {
		return quickFileSearchText;
	}

	public void setQuickFileSearchText(Text quickFileSearchText) {
		this.quickFileSearchText = quickFileSearchText;
	}

	public boolean isReloadReportView() {
		return reloadReportView;
	}

	public void setReloadReportView(boolean reloadReportView) {
		this.reloadReportView = reloadReportView;
	}

	public boolean isFilterHasBeenApplied() {
		return filterHasBeenApplied;
	}

	public int getParamCountOfLastFilterQuery() {
		return paramCountOfLastFilterQuery;
	}

	public int getItemCount() {
		return itemCount;
	}

	public void setItemCount(int itemCount) {
		this.itemCount = itemCount;
	}
}
