package com.perdemia.permissionanalyzer.application.util;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.zip.Adler32;
import java.util.zip.CheckedInputStream;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.license4j.License;
import com.license4j.LicenseText;
import com.license4j.LicenseValidator;
import com.license4j.ValidationStatus;
import com.license4j.util.FileUtils;
import com.perdemia.permissionanalyzer.application.jface.dialogs.LicenseDialog;
import com.perdemia.permissionanalyzer.data.Member;

public class LicenseHelper {
	private static Logger logger = LoggerFactory.getLogger(LicenseHelper.class.getName());
	
	/** REQUIRED. There is a unique public key for each product. */
	private final static String PUBLIC_KEY = "30819f300d06092a864886f70d010101050003818d00308189NOECG02818" +
									  "10087274120468e14e20f577b62dec060343048760c61edadfcac9a2936c" +
									  "46706e7d9e87548721d499c773707652a268aad0638b48134a6df24b9aa8" +
									  "519952d92fd108b9aa323eec6b8ae90d1195803e0daee4f76073d6b192a0" +
									  "ec6c929ce347e2b03RSA4102413SHA512withRSA845061e21e2100fe65e5" +
									  "20051b0ee842da4e290c55e43c22646837ccc904a4d30203010001";
	
	/** REQUIRED. The product ID is used to validate activated license text. */
    private final static String PRODUCT_ID = "permission-analyzer-id";

    /**
     * OPTIONAL. If you are using edition and version validation, you should
     * define this product's edition and version here.
     */
    private final static String PRODUCT_EDITION = null;
    private final static String PRODUCT_VERSION = null;

    /**
     * OPTINAL. If it is null, new Date() is called and used. If you want to use
     * another custom date/time source, you can define date here.
     */
    private final static Date CURRENT_DATE = null;

    /**
     * OPTIONAL. This variable is used to check for valid maintenance period. If
     * you defined a maintenance period in license then it will be valid for
     * versions before maintenance period expires.
     */
    private final static Date CURRENT_VERSION_RELEASE_DATE = null;
    
    /** Path where the imported license is saved and where the activated license is saved (which overwrites a license that has not been activated) */
    public final static String LICENSE_PATH = ApplicationUtil.getWorkspacePath() + "license.lic";
    
    /** Indicates the edition of the license, which is stored as attribute in the license */
    public static enum Edition {TRIAL, BASIC, STANDARD, ADVANCED, ENTERPRISE, CONSULTANT, SCAN_AGENT};
    
    /** The key under which the edition text is stored in the license text */
    private final static String EDITION_FIELD = "edition"; 
    
    /** License information retrieved from file and cached */
    private static License license;
    private static boolean licenseLoaded;
    
    /**
     * Reads the license file from the configured location and validates the license information with the hardware fingerprint, productId, etc.
     * 
     * @return the license object. Null in case the license file is not found or if the license is not valid or not activated.
     * @see {@link #refreshLicenseInfo}
     */
    public static License getLicense() {
        String licenseTextString = null;
		
        if (!licenseLoaded) {
	        try {
				licenseTextString = FileUtils.readFile(LICENSE_PATH);
			} catch (IOException e) {
				//logger.info("License file not found: " + LICENSE_PATH);
			}
	        
	        try {
	        	InputStream input = LicenseText.class.getResourceAsStream("/com/license4j/LicenseText.class");
	        	long licenseTextChecksum = getChecksum(input);
	        	input.close();
		        input = LicenseValidator.class.getResourceAsStream("/com/license4j/LicenseValidator.class");
		        long licenseValidatorChecksum = getChecksum(input);
		        input.close();
		        
		        if (!String.valueOf(licenseTextChecksum).equals("2150637563")) { // using String so it will be obfuscated
		        	///logger.error("License classes are not valid!");
		        	return null;
		        }
		        
		        if (!String.valueOf(licenseValidatorChecksum).equals("2904157826")) { // using String so it will be obfuscated
		        	//logger.error("License classes are not valid!");
		        	return null;
		        }
	        } catch (Exception e) {
				//logger.error("License classes could not be validated: " + e.getMessage(), e);
	        	return null;
			}
	        
			license = getLicense(licenseTextString);
	        
	        licenseLoaded = true;
        }
        
        return license;
    }
    
	public static long getChecksum(InputStream is) {
		CheckedInputStream cis = null;
		long checksum = 0;
		try {
			cis = new CheckedInputStream(is, new Adler32());
			byte[] tempBuf = new byte[128];
			while (cis.read(tempBuf) >= 0) {}
			checksum = cis.getChecksum().getValue();
		} catch (IOException e) {
			checksum = 0;
		} finally {
			if (cis != null) {
				try {
					cis.close();
				} catch (IOException ioe) {
				}
			}
		}
		return checksum;
	}
    
    /** 
     * Builds a license object based on the license file content (licenseText parameter)
     * 
     * @param licenseText
     * 
     * @return License object with ValidationStatus and ActivationStatus
     * @see {@link #getLicense()}
     * @see {@link LicenseDialog#importLicense()}
     */
    public static License getLicense(String licenseText) {
    	if (licenseText != null && licenseText.length() > 0) {
    		License license = LicenseValidator.validate(
	        		licenseText,
	                PUBLIC_KEY,
	                PRODUCT_ID,
	                PRODUCT_EDITION,
	                PRODUCT_VERSION,
	                CURRENT_DATE,
	                CURRENT_VERSION_RELEASE_DATE
	        );
	        
	        //logger.info("License status: " + license.getValidationStatus() + ", activationStatus: " + license.getActivationStatus());
	        return license;
	        
		} else {
			//logger.info("License text empty: " + LICENSE_PATH);
			return null;
		}
    }
    
    /**
     * Puts the flag {@link #licenseLoaded} to false so that the method {@link #getLicense()} will read the license file again.
     */
    public static void refreshLicenseInfo(boolean recalculateDatabaseUsage) {
    	licenseLoaded = false;
    	    	
    	if (recalculateDatabaseUsage) {
    		Member.clearStatistics();	
    	}
    }
    
    /**
     * Retrieves the license content from file ({@link #getLicense()} and validates the license status and activation status.
     * 
     * @return true in case the license is present and valid, false otherwise.
     */
    public static boolean hasValidLicense() {
    	License license = getLicense();
    	
    	if (license == null) {
    		return false;
    	}
    	
    	return license.getValidationStatus().equals(ValidationStatus.LICENSE_VALID) && (license.isActivationCompleted() || !license.isActivationRequired());
    }
    
    /**
     * Delete the license information from disk. This method is used for deactivation or when the current activation is not valid anymore (blacklisted).
     * 
     * @throws IOException
     */
    public static void deleteLicenseFile() throws IOException {
    	Files.delete(new File(LICENSE_PATH).toPath());
    }
    
    /**
     * Validates the license and activation on Online.license4j website. In case the activation and/or license has been delete or moved
     * it will delete the license file.
     * 
     * @return
     * @throws IOException
     */
    public static void validateLicenseOnline() {
		Thread checkOnlineThread = new Thread() {
			public void run() {
				// activation blacklisted?
				if (hasValidLicense()) {
					int result = LicenseValidator.checkOnlineAvailability(PUBLIC_KEY, getLicense(), 5000);
					
					// 1 = valid, -1 = time-out, 0 = invalid
					if (result == 0) {
						//logger.info("License activation no longer valid!");
						try {
							LicenseHelper.deleteLicenseFile();
						} catch (IOException e) {
							//logger.error("License activation no longer valid, but could not delete license file.", e);
						}
						LicenseHelper.refreshLicenseInfo(true);
					}
					
					// update usage statistics
					LicenseValidator.updateLicenseUseInfo(PUBLIC_KEY, getLicense(), 5000);
				}
			}
		};

		checkOnlineThread.start();
    }
    
    /**
     * Send statistics to the License server after a scan
     * 
     * @param scanDuration
     * @param usersGroupCount
     * @param fileCount
     */
    public static void updateUsageStatistics(String scanDuration, int usersCount, int groupCount, int fileCount, int directoryCount) {
    	// upload statistics
    	if (hasValidLicense()) {
			HashMap<String, String> statistics = new HashMap<String, String>();
			statistics.put("scanDate", new SimpleDateFormat("yyyy-MM-dd HH:mm:ss").format(new Date()));
			statistics.put("scanDuration", scanDuration);
			statistics.put("usersCount", String.valueOf(usersCount));
			statistics.put("groupCount", String.valueOf(groupCount));
			statistics.put("fileCount", String.valueOf(fileCount));
			statistics.put("directoryCount", String.valueOf(directoryCount));
			
			LicenseValidator.updateLicenseUseTrackingInfo(PUBLIC_KEY, getLicense(), statistics, 5000);
    	}
    }
    
    /**
     * Read the edition field from the license text.
     * 
     * @return edition: trial, basic, standard, advanced, enterprise, consultant or scan_agent 
     */
    public static Edition getEdition() {
		if (hasValidLicense()) {
	    	License license = getLicense();
			String edition = license.getLicenseText().getCustomSignedFeature(EDITION_FIELD);
			switch (edition) {
				case "basic": return Edition.BASIC;
				case "standard": return Edition.STANDARD;
				case "advanced": return Edition.ADVANCED;
				case "enterprise": return Edition.ENTERPRISE;
				case "consultant": return Edition.CONSULTANT;
				case "scan_agent": return Edition.SCAN_AGENT;
				default: return Edition.TRIAL;
			}			
		} else {
			return Edition.TRIAL;
		}
    }
    
    /**
     * Translates the Edition enumeration to a human readable String
     * 
     * @return edition
     */
    public static String getEditionString() {
    	Edition edition = LicenseHelper.getEdition();
		switch (edition) {
			case TRIAL: return "Trial version";
			case BASIC: return "Basic edition";
			case STANDARD: return "Standard edition";
			case ADVANCED: return "Advanced edition";
			case ENTERPRISE: return "Enterprise edition";
			case CONSULTANT: return "Consultant edition";
			case SCAN_AGENT: return "Scan Agent edition";
		}
		
		return "";
    }
    
    /**
     * Translates the ActivateStatus of the license to a human readable String
     * 
     * @return activation status text
     */
    public static String getActivateStatusString(License license) {
    	if (license != null && license.getActivationStatus() != null) {
			switch (license.getActivationStatus()) {
				case ACTIVATION_COMPLETED: return "Completed"; 
				case ACTIVATION_HARDWAREID_ERROR: return "Activation not valid for this computer"; 
				case ACTIVATION_NOT_FOUND_ON_SERVER: return "Activation not found on the server"; 
				case ACTIVATION_REQUIRED: return "Not activated"; 
				case DEACTIVATION_COMPLETED: return "Deactivated"; 
				case LICENSE_NOT_FOUND_ON_ACTIVATION_SERVER: return "License not found on the server"; 
				case ACTIVATION_SERVER_CONNECTION_ERROR: return "Could not connect to activation server"; 
				case ALREADY_ACTIVATED_ON_ANOTHER_COMPUTER: return "This license is already activated on another computer"; 
				case MULTIPLE_ACTIVATION_LIMIT_REACHED: return "Activation limit has been reached"; 
				case MULTIPLE_DEACTIVATION_LIMIT_REACHED: return "Deactivation limit has been reached";
				case ACTIVATION_NOT_REQUIRED: return "Activation not required";
				default: return "Invalid";
			}
		} else {
			return "Not activated";
		}
    }
    
    /**
     * Translates the ValidationStatus of the license to a human readable String
     * 
     * @return license status text
     */
    public static String getLicenseStatusString(License license) {
		if (license != null) {
			switch (license.getValidationStatus()) {
				case LICENSE_VALID: return "Valid";
				case LICENSE_INVALID: return "Invalid";
				case LICENSE_EXPIRED: return "Expired";
				case INCORRECT_SYSTEM_TIME: return "Invalid (no valid system time found)";
				case LICENSE_MAINTENANCE_EXPIRED: return "Expired";
				case MISMATCH_HARDWARE_ID: return "License not valid for this computer";
				default: return "Invalid";
			}
		} else {
			return "No license available";
		}
    }
}
