package com.perdemia.permissionanalyzer.application.util;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.widgets.Composite;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.contentproviders.MainFileTreeContentProvider;
import com.perdemia.permissionanalyzer.application.parts.AclPart;
import com.perdemia.permissionanalyzer.application.parts.FilePart;
import com.perdemia.permissionanalyzer.application.parts.TracePart;
import com.perdemia.permissionanalyzer.application.util.ImageFactory.AppImage;
import com.perdemia.permissionanalyzer.data.AclItem;
import com.perdemia.permissionanalyzer.data.EffectivePermissions;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.PermissionLookup;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SimplePermission;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SpecialPermission;

public final class PermissionUtil {
	private final static Logger logger = LoggerFactory.getLogger(PermissionUtil.class.getName());

	/**
	 * List of Integers indicating the order of special permissions to display in the GUI
	 */
	public static List<Integer> customPermissionOrderList;

	/**
	 * Determines a match between a list of permissions values (Integers) and a {@link SimplePermission}. Simple permissions are a combination of special permissions.
	 * 
	 * @param aclItem
	 * @return a {@link SimplePermission} value, which can be SPECIAL in case no match is found with one of the other simple permissions.
	 * 
	 */
	private static SimplePermission getSimplePermission(AclItem aclItem) {
		SimplePermission simplePermission = getSimplePermission(aclItem.getPermissionId(), aclItem.getFile() != null && aclItem.getFile().isDirectory(), aclItem.isFlagFileInherit());
		
		if (simplePermission != null) {
			return simplePermission;
		}

		if (aclItem.getPermissionId() != null && aclItem.getPermissionId() != 0) {
			return SimplePermission.SPECIAL;
		} else {
			return null; // no permissions
		}
	}

	/**
	 * Determines a match between a list of effective permissions values (Integers) and a {@link SimplePermission}. Simple permissions are a combination of special permissions.
	 * 
	 * @param file
	 *            file object where the effectve permissions are calculated using {@link #determineEffectivePermissions(List, FileInfo)}
	 * @return a {@link SimplePermission} value, which can be SPECIAL in case no match is found with one of the other simple permissions.
	 * 
	 */
	private static SimplePermission getSimplePermission(EffectivePermissions effectivePermissions, boolean isDirectory) {
		if (effectivePermissions.getDeny() == null && effectivePermissions.getPermissions().size() > 0) {
			// The list of effective permissions contains a combination of allow
			// and deny permissions, it's not one simple permission
			return SimplePermission.SPECIAL;
		}

		SimplePermission simplePermission = getSimplePermission(effectivePermissions.getPermissions().keySet(), isDirectory, effectivePermissions.isFlagFileInherit());
		
		if (simplePermission != null) {
			return simplePermission;
		}

		if (!effectivePermissions.getPermissions().keySet().isEmpty()) {
			return SimplePermission.SPECIAL;
		} else {
			return null; // no permissions
		}
	}
	
	/**
	 * Checks if the given list of special permissions corresponds with a simple permission
	 * 
	 * @param specialPermissions
	 * @param isDirectory used to diffentiate between READ_EXECUTE and LIST_FOLDER_CONTENT
	 * @param isFlagFileInherit used to diffentiate between READ_EXECUTE and LIST_FOLDER_CONTENT
	 * @return
	 */
	public static SimplePermission getSimplePermission(Collection<Integer> specialPermissions, boolean isDirectory, boolean isFlagFileInherit) {
		int permissionId = PermissionLookup.getIdBySpecialPermissions(specialPermissions);
		return getSimplePermission(permissionId, isDirectory, isFlagFileInherit);
	}
	
	/**
	 * Checks if the given permissionId corresponds with a simple permission
	 * 
	 * @param specialPermissions
	 * @param isDirectory used to diffentiate between READ_EXECUTE and LIST_FOLDER_CONTENT
	 * @param isFlagFileInherit used to diffentiate between READ_EXECUTE and LIST_FOLDER_CONTENT
	 * @return
	 */
	public static SimplePermission getSimplePermission(Integer permissionId, boolean isDirectory, boolean isFlagFileInherit) {
		
		SimplePermission simplePermission = null;
		Set<Entry<SimplePermission, Integer>> entries = PermissionLookup.simplePermissionValues.entrySet();

		for (Entry<SimplePermission, Integer> entry : entries) {
			if (entry.getValue().equals(permissionId)) {
				simplePermission = entry.getKey();
				break;
			}
		}

		if (simplePermission != null) {
			// READ_EXECUTE and LIST_FOLDER_CONTENT have the same permission set
			if (simplePermission.equals(SimplePermission.READ_EXECUTE) || simplePermission.equals(SimplePermission.LIST_FOLDER_CONTENT)) {
				if (isFlagFileInherit || !isDirectory) {
					return SimplePermission.READ_EXECUTE;
				} else {
					return SimplePermission.LIST_FOLDER_CONTENT;
				}
			}

			return simplePermission;
		}
		
		return null;
	}

	/**
	 * Converts the AclItem to a String based on the permission list it contains. It returns either a simple permission or a specific special permission otherwise the text 'Special'.
	 * 
	 * @param aclItem
	 *            aclItem
	 * 
	 * @return String representation of the permissions in the given AclItem
	 * 
	 */
	public static String permissionsToString(AclItem aclItem, boolean showInheritanceInfo) {
		return permissionsToString(aclItem.isGenericAll(), aclItem.isGenericExecute(), aclItem.isGenericWrite(), aclItem.isGenericRead(), aclItem.isDeny(), aclItem.getPermissionId(), getSimplePermission(aclItem),
				showInheritanceInfo ? !aclItem.isFlagInheritedAce() : null);
	}

	/**
	 * Converts the effective permissions of a FileInfo to a String based on the permission list it contains. It returns either a simple permission or a specific special permission otherwise the text
	 * 'Special'.
	 * 
	 * @param file
	 *            file where the effective permissions have been determined using {@link #determineEffectivePermissions(List, FileInfo)}
	 * 
	 * @return String representation of the permissions for the given FileInfo
	 * 
	 */
	public static String effectivePermissionsToString(EffectivePermissions effectivePermissions, boolean isDirectory) {
		return permissionsToString(effectivePermissions.isGenericAll(), effectivePermissions.isGenericExecute(), effectivePermissions.isGenericWrite(), effectivePermissions.isGenericRead(), effectivePermissions.getDeny(),
				PermissionLookup.getIdBySpecialPermissions(effectivePermissions.getPermissions().keySet()), getSimplePermission(effectivePermissions, isDirectory), null);
	}

	/**
	 * Converts the effective permissions to a String based on the permission list it contains. It returns either a simple permission or a specific special permission otherwise the text 'Special'.
	 * 
	 * @param isGenericAll
	 * @param isGenericExecute
	 * @param isGenericWrite
	 * @param isGenericRead
	 * @param isDeny
	 *            null means there is a mix of allow and deny permissions
	 * @param permissionId
	 * @param simplePermission
	 * 
	 * @return String representation of the permissions
	 * 
	 */
	private static String permissionsToString(boolean isGenericAll, boolean isGenericExecute, boolean isGenericWrite, boolean isGenericRead, Boolean isDeny, Integer permissionId, SimplePermission simplePermission, Boolean isExplicit) {
		String permissionString;

		// Generic permissions?
		if (isGenericAll || isGenericExecute || isGenericWrite || isGenericRead) {
			StringBuilder stringBuilder = new StringBuilder("Generic: ");
			if (isGenericAll) {
				stringBuilder.append("All/");
			}
			if (isGenericExecute) {
				stringBuilder.append("Execute/");
			}
			if (isGenericWrite) {
				stringBuilder.append("Write/");
			}
			if (isGenericRead) {
				stringBuilder.append("Read/");
			}
			stringBuilder.deleteCharAt(stringBuilder.length() - 1);

			permissionString = stringBuilder.toString();

		} else if (permissionId != null && permissionId != 0 && PermissionLookup.getSpecialPermissions(permissionId).size() == 1) {
			// Return a specific special permission if we only have one special
			// permission
			permissionString = specialPermissionToString(PermissionLookup.getSpecialPermissions(permissionId).get(0), isDeny, null);

		} else {
			// we have several special permissions, determine the 'simple'
			// permission name (or 'special' if no simple name matches)
			permissionString = simplePermissionToString(simplePermission);
		}

		if (isDeny != null && isDeny && !permissionString.startsWith("Deny:")) {
			permissionString = "Deny: " + permissionString;
		}

		if (isExplicit != null && !permissionString.equals("None")) {
			if (isExplicit) {
				permissionString += " (explicit)";
			} else {
				permissionString += " (inherited)";
			}
		}

		return permissionString;
	}

	/**
	 * Returns the name for a {@link SimplePermission}.
	 * 
	 * @param simplePermission
	 * 
	 * @return human readable name to be used in the GUI
	 */
	public static String simplePermissionToString(SimplePermission simplePermission) {
		if (simplePermission == null) {
			return "None";
		} else {
			switch (simplePermission) {
			case FULL:
				return "Full";
			case MODIFY:
				return "Modify";
			case READ_EXECUTE:
				return "Read and execute";
			case LIST_FOLDER_CONTENT:
				return "List folder content";
			case READ:
				return "Read";
			case WRITE:
				return "Write";
			case SPECIAL:
				return "Special";
			default:
				return "Unknown";
			}
		}
	}

	/**
	 * Returns the name for a {@link SimplePermission}.
	 * 
	 * @param simplePermission
	 * 
	 * @return human readable name to be used in the GUI
	 */
	public static Image getSimplePermissionImage(SimplePermission simplePermission) {
		if (simplePermission == null) {
			return null;
		} else {
			switch (simplePermission) {
			case FULL:
				return ImageFactory.getInstance().getImage(AppImage.FOLDER_FULL_16X16);
			case MODIFY:
				return ImageFactory.getInstance().getImage(AppImage.FOLDER_MODIFY_16X16);
			case READ_EXECUTE:
				return ImageFactory.getInstance().getImage(AppImage.FOLDER_EXECUTE_16X16);
			case LIST_FOLDER_CONTENT:
				return ImageFactory.getInstance().getImage(AppImage.FOLDER_LIST_FOLDER_CONTENT_16X16);
			case READ:
				return ImageFactory.getInstance().getImage(AppImage.FOLDER_READ_16X16);
			case WRITE:
				return ImageFactory.getInstance().getImage(AppImage.FOLDER_WRITE_16X16);
			case SPECIAL:
				return ImageFactory.getInstance().getImage(AppImage.PERM_READ_ACL_16X16);
			default:
				return null;
			}
		}
	}

	/**
	 * Converts a special permission to a String representation.
	 * 
	 * @param permission
	 *            ordinal of the permission
	 * @param deny
	 *            indicates if it is a deny permission
	 * 
	 * @return human readable String to be used in the GUI
	 */
	public static String specialPermissionToString(int permission, boolean deny, Boolean isExplicit) {
		String permissionString = null;

		if (permission == SpecialPermission.READ_DATA.ordinal()) {
			permissionString = "Read data";
		} else if (permission == SpecialPermission.WRITE_DATA.ordinal()) {
			permissionString = "Write data";
		} else if (permission == SpecialPermission.APPEND_DATA.ordinal()) {
			permissionString = "Append data";
		} else if (permission == SpecialPermission.READ_NAMED_ATTRS.ordinal()) {
			permissionString = "Read extended attributes";
		} else if (permission == SpecialPermission.WRITE_NAMED_ATTRS.ordinal()) {
			permissionString = "Write extended attributes";
		} else if (permission == SpecialPermission.EXECUTE.ordinal()) {
			permissionString = "Execute";
		} else if (permission == SpecialPermission.DELETE_CHILD.ordinal()) {
			permissionString = "Delete subfolders and files";
		} else if (permission == SpecialPermission.READ_ATTRIBUTES.ordinal()) {
			permissionString = "Read attributes";
		} else if (permission == SpecialPermission.WRITE_ATTRIBUTES.ordinal()) {
			permissionString = "Write attributes";
		} else if (permission == SpecialPermission.DELETE.ordinal()) {
			permissionString = "Delete";
		} else if (permission == SpecialPermission.READ_ACL.ordinal()) {
			permissionString = "Read permissions";
		} else if (permission == SpecialPermission.WRITE_ACL.ordinal()) {
			permissionString = "Change permissions";
		} else if (permission == SpecialPermission.WRITE_OWNER.ordinal()) {
			permissionString = "Take ownership";
		} else if (permission == SpecialPermission.SYNCHRONIZE.ordinal()) {
			permissionString = "Sync";
		} else {
			permissionString = "Unknown";
		}

		if (deny) {
			permissionString = "Deny: " + permissionString;
		}

		if (isExplicit != null) {
			if (isExplicit) {
				permissionString += " (explicit)";
			} else {
				permissionString += " (inherited)";
			}
		}

		return permissionString;
	}

	/**
	 * Determines the Image that belongs to a special permission.
	 * 
	 * @param permission
	 *            the special permission
	 * @param isDeny
	 *            if true the deny image is used
	 * @param size16
	 *            if true the 16x16 image if returns, otherwise the size 20x20 is returned
	 * 
	 * @return Image object representing the special permission
	 */
	public static Image getSpecialPermissionImage(int permission, boolean isDeny, boolean size16) {
		AppImage image = getSpecialPermissionImageEnum(permission, isDeny, size16);
		if (image != null) {
			return ImageFactory.getInstance().getImage(image);
		} else {
			return null;
		}
	}

	/**
	 * Determines the Image that belongs to a special permission.
	 * 
	 * @param permission
	 *            the special permission
	 * @param isDeny
	 *            if true the deny image is used
	 * @param size16
	 *            if true the 16x16 image if returns, otherwise the size 20x20 is returned
	 * 
	 * @return Image object representing the special permission
	 */
	public static AppImage getSpecialPermissionImageEnum(int permission, boolean isDeny, boolean size16) {
		AppImage image = null;

		if (permission == SpecialPermission.READ_DATA.ordinal()) {
			image = size16 ? AppImage.PERM_READ_DATA_16X16 : AppImage.PERM_READ_DATA_20X20;
		}
		if (permission == SpecialPermission.WRITE_DATA.ordinal()) {
			image = size16 ? AppImage.PERM_WRITE_DATA_16X16 : AppImage.PERM_WRITE_DATA_20X20;
		}
		if (permission == SpecialPermission.APPEND_DATA.ordinal()) {
			image = size16 ? AppImage.PERM_APPEND_DATA_16X16 : AppImage.PERM_APPEND_DATA_20X20;
		}
		if (permission == SpecialPermission.READ_NAMED_ATTRS.ordinal()) {
			image = size16 ? AppImage.PERM_READ_EXT_ATTR_16X16 : AppImage.PERM_READ_EXT_ATTR_20X20;
		}
		if (permission == SpecialPermission.WRITE_NAMED_ATTRS.ordinal()) {
			image = size16 ? AppImage.PERM_WRITE_EXT_ATTR_16X16 : AppImage.PERM_WRITE_EXT_ATTR_20X20;
		}
		if (permission == SpecialPermission.EXECUTE.ordinal()) {
			image = size16 ? AppImage.PERM_EXECUTE_16X16 : AppImage.PERM_EXECUTE_20X20;
		}
		if (permission == SpecialPermission.DELETE_CHILD.ordinal()) {
			image = size16 ? AppImage.PERM_DELETE_CHILD_16X16 : AppImage.PERM_DELETE_CHILD_20X20;
		}
		if (permission == SpecialPermission.READ_ATTRIBUTES.ordinal()) {
			image = size16 ? AppImage.PERM_READ_ATTR_16X16 : AppImage.PERM_READ_ATTR_20X20;
		}
		if (permission == SpecialPermission.WRITE_ATTRIBUTES.ordinal()) {
			image = size16 ? AppImage.PERM_WRITE_ATTR_16X16 : AppImage.PERM_WRITE_ATTR_20X20;
		}
		if (permission == SpecialPermission.DELETE.ordinal()) {
			image = size16 ? AppImage.PERM_DELETE_16X16 : AppImage.PERM_DELETE_20X20;
		}
		if (permission == SpecialPermission.READ_ACL.ordinal()) {
			image = size16 ? AppImage.PERM_READ_ACL_16X16 : AppImage.PERM_READ_ACL_20X20;
		}
		if (permission == SpecialPermission.WRITE_ACL.ordinal()) {
			image = size16 ? AppImage.PERM_WRITE_ACL_16X16 : AppImage.PERM_WRITE_ACL_20X20;
		}
		if (permission == SpecialPermission.WRITE_OWNER.ordinal()) {
			image = size16 ? AppImage.PERM_TAKE_OWNERSHIP_16X16 : AppImage.PERM_TAKE_OWNERSHIP_20X20;
		}

		return image;
	}

	/**
	 * Create a Image based on the given permissions. The Image contains a folder or file icon including a graphical indication about the 'highest' permission. Used to determine the main image for a
	 * row in the Main FileInfo View tree and Trace View tree
	 * 
	 * @param aclItem
	 *            the aclItem of a row in the Trace View table. When null the method will use {@link FileInfo#getEffectivePermissions()}.
	 * @param file
	 *            the corresponding file object of the AclItem to determine the file icon in case the file is not a directory (based on extension)
	 * 
	 * @return FileInfo or folder image of 20x20
	 */
	public static AppImage getFileImage(AclItem aclItem, FileInfo file) {
		AppImage image = null;
		boolean isDeny = false;
		String permissionText = null; // permission name used for file icons
		SimplePermission simplePermission = null;
		Integer permissionId = null;

		if (file.getEffectivePermissions() != null) {
			// use the effective permissions in the file object
			isDeny = file.getEffectivePermissions().getDeny() != null && file.getEffectivePermissions().getDeny();
			simplePermission = getSimplePermission(file.getEffectivePermissions(), file.isDirectory());
			permissionId = PermissionLookup.getIdBySpecialPermissions(file.getEffectivePermissions().getPermissions().keySet());
		} else if (aclItem != null) {			
			isDeny = aclItem.isDeny();
			simplePermission = getSimplePermission(aclItem);
			permissionId = aclItem.getPermissionId();
		} else {
			simplePermission = SimplePermission.READ;
			permissionId = -1;
		}

		if (simplePermission != null && permissionId != null && permissionId != 0) {
			switch (simplePermission) {
			case READ:
				image = (isDeny ? AppImage.FOLDER_READ_20X20 : AppImage.FOLDER_READ_20X20);
				permissionText = "read";
				break;
			case READ_EXECUTE:
				image = (isDeny ? AppImage.FOLDER_EXECUTE_20X20 : AppImage.FOLDER_EXECUTE_20X20);
				permissionText = "execute";
				break;
			case LIST_FOLDER_CONTENT:
				image = (isDeny ? AppImage.FOLDER_LIST_FOLDER_CONTENT_20X20 : AppImage.FOLDER_LIST_FOLDER_CONTENT_20X20);
				permissionText = "execute";
				break;
			case WRITE:
				image = (isDeny ? AppImage.FOLDER_WRITE_20X20 : AppImage.FOLDER_WRITE_20X20);
				permissionText = "write";
				break;
			case MODIFY:
				image = (isDeny ? AppImage.FOLDER_MODIFY_20X20 : AppImage.FOLDER_MODIFY_20X20);
				permissionText = "modify";
				break;
			case FULL:
				image = (isDeny ? AppImage.FOLDER_FULL_20X20 : AppImage.FOLDER_FULL_20X20);
				permissionText = "full";
				break;
			case SPECIAL:

				// the determine the best matching icon
				if (PermissionLookup.containsSpecialPermission(permissionId, 11) || PermissionLookup.containsSpecialPermission(permissionId, 12)) {
					image = (isDeny ? AppImage.FOLDER_FULL_20X20 : AppImage.FOLDER_FULL_20X20);
					permissionText = "full";

				} else if (PermissionLookup.containsSpecialPermission(permissionId, 6) || PermissionLookup.containsSpecialPermission(permissionId, 9)) {
					image = (isDeny ? AppImage.FOLDER_DELETE_20X20 : AppImage.FOLDER_DELETE_20X20);
					permissionText = "delete";

				} else if (PermissionLookup.containsSpecialPermission(permissionId, 5)) {
					image = (isDeny ? AppImage.FOLDER_EXECUTE_20X20 : AppImage.FOLDER_EXECUTE_20X20);
					permissionText = "execute";

				} else if (PermissionLookup.containsSpecialPermission(permissionId, 1) || PermissionLookup.containsSpecialPermission(permissionId, 2) || PermissionLookup.containsSpecialPermission(permissionId, 4)
						|| PermissionLookup.containsSpecialPermission(permissionId, 8)) {
					image = (isDeny ? AppImage.FOLDER_WRITE_20X20 : AppImage.FOLDER_WRITE_20X20);
					permissionText = "write";

				} else if (PermissionLookup.containsSpecialPermission(permissionId, 0) || PermissionLookup.containsSpecialPermission(permissionId, 3) || PermissionLookup.containsSpecialPermission(permissionId, 7)
						|| PermissionLookup.containsSpecialPermission(permissionId, 10)) {
					image = (isDeny ? AppImage.FOLDER_READ_20X20 : AppImage.FOLDER_READ_20X20);
					permissionText = "read";
				}
			}
		} else {
			image = AppImage.FOLDER_NONE_20X20;
			permissionText = "none";
		}

		// file icon?
		if (!file.isDirectory()) {
			image = getImageForFile(permissionText, file);
		}

		return image;
	}

	/**
	 * Retrieves a {@link AppImage} based on the file extension and the permission on the file.
	 * 
	 * @param permission
	 *            permission text like 'read', 'write', etc...
	 * @param file
	 *            FileInfo object
	 * 
	 * @return AppImage enum value
	 */
	private static AppImage getImageForFile(String permission, FileInfo file) {
		int extensionIndex = file.getName().lastIndexOf(".");
		String extension = "";

		if (extensionIndex != -1) {
			extension = file.getName().substring(extensionIndex + 1).toLowerCase();
		}

		String fileType;
		if (ImageFactory.FILE_EXTENSIONS.containsKey(extension)) {
			fileType = ImageFactory.FILE_EXTENSIONS.get(extension);
		} else {
			fileType = "file";
		}

		return AppImage.valueOf("FILE_" + fileType.toUpperCase() + "_" + permission.toUpperCase() + "_20X20");
	}

	/**
	 * Determines a list of unique effective permissions based on a total list of permissions. The following order rules apply:
	 * <ul>
	 * <li>Explicit Deny</li>
	 * <li>Explicit Allow</li>
	 * <li>Inherited Deny</li>
	 * <li>Inherited Allow</li>
	 * </ul>
	 * 
	 * <b>BUT</b>, note the following:
	 * 
	 * <pre>
	 * Consider the scenario below:
	 *  
	 *                         Type        Deny       Allow
	 * Folder1            Inherit
	 *                        Explicit      X
	 *  
	 *      Folder2       Inherit           X
	 *                        Explicit                  X
	 *  
	 *          Folder3   Inherit           X           X
	 *                        Explicit
	 *  
	 * Now what? What happen if an user tries to access Folder3 directly?
	 *  
	 * The access will be allowed!! How? Simple, the user is inheriting both permissions (Deny and Allow). OK, but everyone knows on NTFS Deny has precedence over Allow Not always!
	 *  
	 * Extracted from Microsoft Technet:
	 *  
	 * "Inherited Deny permissions do not prevent access to an object if the object has an explicit Allow permission entry."
	 *  
	 * "Explicit permissions take precedence over inherited permissions, even inherited Deny permissions."
	 *  
	 * However, none of this apply to our scenario above.
	 *  
	 * NTFS has a logic for checking permissions:
	 * 
	 *     Check if exist explicit Deny on Folder3 >> No.
	 *     Check if exist explicit Allow on Folder3 >> No.
	 *     Check if inherit is enabled on Folder3 >> Yes, It is. Check inherited permissions on parent folder (Folder2).
	 *     Check if exist explicit Deny on Folder2 >> No.
	 *     Check if exist explicit Allow on Folder2 >> YES!!
	 *     Allow access to object!!
	 * 
	 * Theres no explicit permissions on Folder3, the last inherited permission will take precedence.
	 * 
	 * Thats why the user has access to Folder3!
	 * </pre>
	 * 
	 * The list of Permissions is returned as a new {@link AclItem}
	 * 
	 * @param aclItems
	 *            list of {@link AclItem} objects
	 * @return List with unique effective permissions in a new {@link AclItem} object. Each <code>Permission</code> contains its original <code>AclItem</code>.
	 * 
	 * @see {@link MainFileTreeContentProvider#updateElement(Object, int)}
	 */
	public static EffectivePermissions determineEffectivePermissions(List<AclItem> aclItems) {
		EffectivePermissions effectivePermissions = new EffectivePermissions();
		Map<Integer, AclItem> permissions = new HashMap<Integer, AclItem>();

		for (AclItem aclItem : aclItems) {

			// skip aclItems with flag INHERIT_ONLY, because they only apply to the sub files or folders
			if (!aclItem.isFlagInheritOnly()) {
				List<Integer> specialPermissions = PermissionLookup.getSpecialPermissions(aclItem.getPermissionId());
				for (Integer permission : specialPermissions) {
					AclItem currentAclItem = permissions.get(permission);

					if (currentAclItem == null) {
						// permission is not present yet, so always use this new permission
						permissions.put(permission, aclItem);
					} else {

						if (!aclItem.isFlagInheritedAce()) {
							if (currentAclItem.isFlagInheritedAce() || aclItem.isDeny()) {
								// explicit wins from non-explicit and explicit deny always wins
								permissions.put(permission, aclItem);
							}

						} else if (currentAclItem.isFlagInheritedAce()) {
							int currentInheritedFrom = currentAclItem.getInheritedFromFolderId() != null ? currentAclItem.getInheritedFromFolderId() : -1;
							int newInheritedFrom = aclItem.getInheritedFromFolderId() != null ? aclItem.getInheritedFromFolderId() : -1;

							// current permission and this new one are both inherited, the most 'direct' inheritedFrom folder wins. When equal deny wins.
							if ((currentInheritedFrom < newInheritedFrom) || (currentInheritedFrom == newInheritedFrom && aclItem.isDeny())) {
								// TODO test effective permissions, direct parent should win from indirect parent
								permissions.put(permission, aclItem);
							}
						}
					}
				}
			}
		}

		// check if all permissions have the same isDeny() value (used to determine the text representation, mixed deny/allow rights have always the label "special")
		Boolean isDeny = null;
		Collection<AclItem> effectiveAclItems = permissions.values();
		for (AclItem aclItem : effectiveAclItems) {
			if (isDeny == null) {
				isDeny = aclItem.isDeny();
			} else {
				if (isDeny != aclItem.isDeny()) {
					// deny rights combined with allow rights
					effectivePermissions.setDeny(null);
					break;
				}
			}
			effectivePermissions.setDeny(isDeny);
		}

		// check if the flag FILE_INHERIT is set on all the permissions. This flag is used to determine the difference between READ_EXECUTE and LIST_FOLDER_CONTENT. They both have the same
		// set of special permissions, only READ_EXECUTE has also the flag AclEntryFlag.FILE_INHERIT.
		// also check if all AclItems share the same generic permission(s).
		boolean isFileInherit = aclItems.size() > 0;
		boolean genericAll = aclItems.size() > 0;
		boolean genericExecute = aclItems.size() > 0;
		boolean genericWrite = aclItems.size() > 0;
		boolean genericRead = aclItems.size() > 0;

		for (AclItem aclItem : aclItems) {
			isFileInherit = isFileInherit && aclItem.isFlagFileInherit();
			genericAll = genericAll && aclItem.isGenericAll();
			genericExecute = genericExecute && aclItem.isGenericExecute();
			genericWrite = genericWrite && aclItem.isGenericWrite();
			genericRead = genericRead && aclItem.isGenericRead();
		}

		effectivePermissions.setFlagFileInherit(isFileInherit);
		effectivePermissions.setGenericAll(genericAll);
		effectivePermissions.setGenericExecute(genericExecute);
		effectivePermissions.setGenericRead(genericRead);
		effectivePermissions.setGenericWrite(genericWrite);
		effectivePermissions.setPermissions(permissions);

		if (logger.isTraceEnabled()) {
			logger.trace("Determined effective permissions: " + permissions.keySet().toString());
		}

		return effectivePermissions;
	}

	/**
	 * Returns a List of Integers indicating the (better) order of special permissions to display in the GUI (permission columns for example)
	 * 
	 * @return list of special permission ordinals
	 * @see {@link FilePart#createPartControl(Composite)}
	 * @see {@link AclPart#createPartControl(Composite)}
	 * @see {@link TracePart#createPartControl(Composite)}
	 */
	public static List<Integer> getCustomPermissionOrderList() {
		if (customPermissionOrderList == null) {
			customPermissionOrderList = Arrays.asList(new Integer[] { 0, 1, 2, 5, 7, 8, 3, 4, 6, 9, 10, 11, 12 });

		}
		return customPermissionOrderList;
	}

	/**
	 * Builds a String representing a list of inheritance flags of the given {@link AclItem} object.
	 * 
	 * @param aclItem
	 *            AclItem containing the inheritance flags
	 * 
	 * @return Human readable string representing the inheritance strategy of the AclItem
	 */
	public static String aclEntryFlagsToString(AclItem aclItem, boolean isDirectory) {
		String flags;

		if (!aclItem.isFlagDirectoryInherit() && !aclItem.isFlagFileInherit() && !aclItem.isFlagInheritOnly()) {
			if (isDirectory) {
				flags = "This folder only";
			} else {
				flags = "This file only";
			}
		} else if (aclItem.isFlagDirectoryInherit() && !aclItem.isFlagFileInherit() && !aclItem.isFlagInheritOnly()) {
			flags = "This folder and subfolders";
		} else if (!aclItem.isFlagDirectoryInherit() && aclItem.isFlagFileInherit() && !aclItem.isFlagInheritOnly()) {
			flags = "This folder and files";
		} else if (aclItem.isFlagDirectoryInherit() && aclItem.isFlagFileInherit() && aclItem.isFlagInheritOnly()) {
			flags = "Subfolders and files only (not this folder)";
		} else if (aclItem.isFlagDirectoryInherit() && aclItem.isFlagFileInherit() && !aclItem.isFlagInheritOnly()) {
			flags = "This folder, subfolders and files";
		} else if (aclItem.isFlagDirectoryInherit() && !aclItem.isFlagFileInherit() && aclItem.isFlagInheritOnly()) {
			flags = "Subfolders only (not this folder)";
		} else if (!aclItem.isFlagDirectoryInherit() && aclItem.isFlagFileInherit() && aclItem.isFlagInheritOnly()) {
			flags = "Files only (not this folder)";
		} else {
			logger.error("Could not determine ACL flag text for the acl " + aclItem.toString());
			return "Unknown";
		}

		if (aclItem.isFlagNoPropagateInherit()) {
			flags += " (no propagate)";
		}

		return flags;
	}
}
