package com.perdemia.permissionanalyzer.data;

import java.nio.file.attribute.AclEntryFlag;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.data.util.FileCache;
import com.perdemia.permissionanalyzer.data.util.MemberCache;
import com.perdemia.permissionanalyzer.data.util.TableCountCache;

/**
 * Represents a ACL entry for one FileInfo (a set of Permissions belonging to one Member, one FileInfo and one set of inheritance flags). 
 * It contains the information for one row in the Trace view or ACL view.
 * 
 * @author Reinier
 *
 */
public class AclItem extends DBObject {
private final static Logger logger = LoggerFactory.getLogger(AclItem.class.getName());
	
	/** Table attributes */
	private final static String[] tableAttributes = new String[]{"MEMBER_ID", "FILEINFO_ID", "PERMISSION_ID", "INHERITED_FROM_FOLDER_ID", "FLAG_DIRECTORY_INHERIT", "FLAG_FILE_INHERIT",
			"FLAG_INHERIT_ONLY", "FLAG_NO_PROPAGATE_INHERIT", "FLAG_INHERITED_ACE", "IS_DENY", "GENERIC_ALL", "GENERIC_EXECUTE", "GENERIC_WRITE", "GENERIC_READ", "IS_DIRECTORY"};
	
	/** Table fields */
	private Integer memberId;
	private Integer fileId;
	private Integer permissionId;
	private Integer inheritedFromFolderId;
	private boolean flagDirectoryInherit;
	private boolean flagFileInherit;
	private boolean flagInheritOnly;
	private boolean flagNoPropagateInherit;
	private boolean flagInheritedAce;
	private boolean deny;
	private boolean genericAll;
	private boolean genericExecute;
	private boolean genericWrite;
	private boolean genericRead;
	private boolean directory;
	
	/**helper fields */
	private Member member;
	private List<AclEntryFlag> aclEntryFlags;
	
	/** Only used in Trace view and the All Members report */
	private FileInfo file;	
	private Member viaMember;
	
	/** Only used in ACL View */
	private String inheritedFromFolder;
	
	/** Helper field, used in the EffectivePermissionsPartContentProvider. */
	private boolean applicableOnSelectedFolder = true;
	
	/**
	 * The last ID that was used for an INSERT of this class in the database. If 0 is returned then the
	 * {@link DBObject#insert()} method will retrieve the max ID from the database.
	 */
	private static int lastId;
	
	//========================================
	// GENERAL DB ACTIONS
	//========================================
	
	public static AclItem getById(Integer id) throws DBException {
		return (AclItem) getById(AclItem.class, id);
	}
		
	@SuppressWarnings("unchecked")
	public static List<AclItem> where(String whereClause, Object[] parameters) throws DBException {
		return where(AclItem.class, whereClause, parameters, null, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<AclItem> where(String whereClause, Object[] parameters, String orderBy) throws DBException {
		return where(AclItem.class, whereClause, parameters, orderBy, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<AclItem> where(String whereClause, Object[] parameters, String orderBy, int limit) throws DBException {
		return where(AclItem.class, whereClause, parameters, orderBy, limit);
	}
	
	public static AclItem findFirst(String whereClause, Object... parameters) throws DBException {
		return (AclItem) findFirst(AclItem.class, whereClause, parameters);
	}
	
	@SuppressWarnings("unchecked")
	public static List<AclItem> getByPreparedStatement(PreparedStatement ps) throws DBException {
		return getByPreparedStatement(AclItem.class, ps);
	}
	
	public static int delete(Integer id) throws DBException {
		return delete(AclItem.class, id);
	}
	
	public static int delete(String whereClause, Object...parameters) throws DBException {
		return delete(AclItem.class, whereClause, parameters);
	}
	
	@Override
	protected String[] getTableAttributes(){
		return tableAttributes;
	}
	
	@Override
	protected void buildObject(ResultSet rs) throws SQLException {
		id = (rs.getInt("id"));
		memberId = rs.getInt(tableAttributes[0]);
		fileId = rs.getInt(tableAttributes[1]);
		permissionId = rs.getInt(tableAttributes[2]);
		inheritedFromFolderId = getNullableInteger(rs, tableAttributes[3]);
		flagDirectoryInherit = rs.getBoolean(tableAttributes[4]);
		flagFileInherit = rs.getBoolean(tableAttributes[5]);
		flagInheritOnly = rs.getBoolean(tableAttributes[6]);
		flagNoPropagateInherit = rs.getBoolean(tableAttributes[7]);
		flagInheritedAce = rs.getBoolean(tableAttributes[8]);
		deny = rs.getBoolean(tableAttributes[9]);
		genericAll = rs.getBoolean(tableAttributes[10]);
		genericExecute = rs.getBoolean(tableAttributes[11]);
		genericWrite = rs.getBoolean(tableAttributes[12]);
		genericRead = rs.getBoolean(tableAttributes[13]);
		directory = rs.getBoolean(tableAttributes[14]);
	}
	
	@Override
	protected void setQueryParameters(PreparedStatement ps) throws SQLException {
		ps.setInt(1, memberId);
		ps.setInt(2, fileId);
		ps.setInt(3, permissionId);
		setNullableInteger(ps, 4, inheritedFromFolderId);
		ps.setBoolean(5, flagDirectoryInherit);
		ps.setBoolean(6, flagFileInherit);
		ps.setBoolean(7, flagInheritOnly);
		ps.setBoolean(8, flagNoPropagateInherit);
		ps.setBoolean(9, flagInheritedAce);
		ps.setBoolean(10, deny);
		ps.setBoolean(11, genericAll);
		ps.setBoolean(12, genericExecute);
		ps.setBoolean(13, genericWrite);
		ps.setBoolean(14, genericRead);
		ps.setBoolean(15, directory);
	}
	
	public static int getCount() throws DBException {
		if (TableCountCache.getCount(AclItem.class) == null) {
			TableCountCache.cacheCount(AclItem.class, getCount(AclItem.class));
		}
		return TableCountCache.getCount(AclItem.class);
	}
	
	//========================================
	// SPECIFIC DB ACTIONS
	//========================================
	@SuppressWarnings("unchecked")
	public static List<AclItem> getByFileId(Integer fileId) throws DBException {
		if (fileId != null) {
			return where(AclItem.class, "fileinfo_id = ?", new Object[]{fileId}, null, 0);
		} else {
			return where(AclItem.class, "fileinfo_id is null", null, null, 0);
		}
	}
	
	/**
	 * Returns a clone of the current object, but linking to the same Member and Permission List
	 * 
	 * @return AclItem, clone of this object
	 */
	public AclItem clone() {
		AclItem clone = new AclItem();
		clone.setMemberId(getMemberId());
		clone.setMember(getMember(false));
		clone.setFile(getFile());
		clone.setPermissionId(getPermissionId());
		clone.setDeny(isDeny());
		clone.setFileId(getFileId());
		clone.setFlagDirectoryInherit(isFlagDirectoryInherit());
		clone.setFlagFileInherit(isFlagFileInherit());
		clone.setFlagInheritOnly(isFlagInheritOnly());
		clone.setFlagInheritedAce(isFlagInheritedAce());
		clone.setFlagNoPropagateInherit(isFlagNoPropagateInherit());
		clone.setInheritedFromFolderId(getInheritedFromFolderId());
		clone.setGenericAll(isGenericAll());
		clone.setGenericExecute(isGenericExecute());
		clone.setGenericWrite(isGenericWrite());
		clone.setGenericRead(isGenericRead());
		clone.setDirectory(isDirectory());
		return clone;
	}

	//========================================
	// GETTERS AND SETTERS
	//========================================

	public Integer getMemberId() {
		return memberId;
	}

	public void setMemberId(Integer memberId) {
		this.memberId = memberId;
	}

	public Integer getFileId() {
		return fileId;
	}

	public void setFileId(Integer fileId) {
		this.fileId = fileId;
	}

	public boolean isFlagDirectoryInherit() {
		return flagDirectoryInherit;
	}

	public void setFlagDirectoryInherit(boolean flagDirectoryInherit) {
		this.flagDirectoryInherit = flagDirectoryInherit;
	}

	public boolean isFlagFileInherit() {
		return flagFileInherit;
	}

	public void setFlagFileInherit(boolean flagFileInherit) {
		this.flagFileInherit = flagFileInherit;
	}

	public boolean isFlagInheritOnly() {
		return flagInheritOnly;
	}

	public void setFlagInheritOnly(boolean flagInheritOnly) {
		this.flagInheritOnly = flagInheritOnly;
	}

	public boolean isFlagNoPropagateInherit() {
		return flagNoPropagateInherit;
	}

	public void setFlagNoPropagateInherit(boolean flagNoPropagateInherit) {
		this.flagNoPropagateInherit = flagNoPropagateInherit;
	}

	public boolean isDeny() {
		return deny;
	}

	public void setDeny(boolean deny) {
		this.deny = deny;
	}

	public Member getMember(boolean loadIfNull) {
		if (member == null && loadIfNull) {
			try {
				member = MemberCache.getMember(memberId);
			} catch (DBException e) {
				logger.error("Error while retrieving additional member information in AclItem " + getId() + ", memberId: " + memberId, e);
			}
		}
		return member;
	}

	public void setMember(Member member) {
		if (member != null) {
			this.memberId = member.getId();
		}
		this.member = member;
	}

	public Integer getPermissionId() {
		return permissionId;
	}

	public void setPermissionId(Integer permissionId) {
		this.permissionId = permissionId;
	}

	public FileInfo getFile() {
		return file;
	}
	
	public FileInfo getFile(boolean loadIfNull) {
		if (file == null && loadIfNull) {
			try {
				if (fileId != null) {
					file = FileCache.getFile(fileId);
				}
			} catch (DBException e) {
				logger.error("Error while retrieving additional file information in AclItem ID " + getId() + ", fileId: " + fileId + ", memberId: " + memberId + (member != null ? ", member: " + member.toString() : ""), e);
			}
		}
		return file;
	}

	public void setFile(FileInfo file) {
		if (file != null) {
			this.fileId = file.getId();
		}
		this.file = file;
	}

	public Member getViaMember() {
		return viaMember;
	}

	public void setViaMember(Member viaGroup) {
		this.viaMember = viaGroup;
	}

	public Integer getInheritedFromFolderId() {
		return inheritedFromFolderId;
	}

	public void setInheritedFromFolderId(Integer inheritedFromFolderId) {
		this.inheritedFromFolderId = inheritedFromFolderId;
	}

	public String getInheritedFromFolder() {
		return inheritedFromFolder;
	}

	public void setInheritedFromFolder(String inheritedFromFolder) {
		this.inheritedFromFolder = inheritedFromFolder;
	}

	public void setAclEntryFlags(List<AclEntryFlag> aclEntryFlags) {
		this.aclEntryFlags = aclEntryFlags;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("AclItem [memberId=");
		builder.append(memberId);
		builder.append(", fileId=");
		builder.append(fileId);
		builder.append(", flagDirectoryInherit=");
		builder.append(flagDirectoryInherit);
		builder.append(", flagFileInherit=");
		builder.append(flagFileInherit);
		builder.append(", flagInheritOnly=");
		builder.append(flagInheritOnly);
		builder.append(", flagNoPropagateInherit=");
		builder.append(flagNoPropagateInherit);
		builder.append(", flagInheritedAce=");
		builder.append(flagInheritedAce);
		builder.append(", deny=");
		builder.append(deny);
		builder.append(", permissionId=");
		builder.append(permissionId);
		builder.append(", aclEntryFlags=");
		builder.append(aclEntryFlags);
		builder.append(", viaGroup=");
		builder.append(viaMember);
		builder.append(", inheritedFromFolderId=");
		builder.append(inheritedFromFolderId);
		builder.append(", inheritedFromFolder=");
		builder.append(inheritedFromFolder);
		builder.append(", genericAll=");
		builder.append(genericAll);
		builder.append(", genericExecute=");
		builder.append(genericExecute);
		builder.append(", genericWrite=");
		builder.append(genericWrite);
		builder.append(", genericRead=");
		builder.append(genericRead);
		builder.append(", directory=");
		builder.append(directory);
		builder.append(", id=");
		builder.append(id);
		builder.append("]");
		return builder.toString();
	}

	public boolean isGenericAll() {
		return genericAll;
	}

	public void setGenericAll(boolean genericAll) {
		this.genericAll = genericAll;
	}

	public boolean isGenericExecute() {
		return genericExecute;
	}

	public void setGenericExecute(boolean genericExecute) {
		this.genericExecute = genericExecute;
	}

	public boolean isGenericWrite() {
		return genericWrite;
	}

	public void setGenericWrite(boolean genericWrite) {
		this.genericWrite = genericWrite;
	}

	public boolean isGenericRead() {
		return genericRead;
	}

	public void setGenericRead(boolean genericRead) {
		this.genericRead = genericRead;
	}

	public boolean isFlagInheritedAce() {
		return flagInheritedAce;
	}

	public void setFlagInheritedAce(boolean flagInheritedAce) {
		this.flagInheritedAce = flagInheritedAce;
	}

	public boolean isDirectory() {
		return directory;
	}

	public void setDirectory(boolean directory) {
		this.directory = directory;
	}

	public boolean isApplicableOnSelectedFolder() {
		return applicableOnSelectedFolder;
	}

	public void setApplicableOnSelectedFolder(boolean applicableOnSelectedFolder) {
		this.applicableOnSelectedFolder = applicableOnSelectedFolder;
	}
	
	@Override
	public int getLastId() {
		return lastId;
	}

	@Override
	public void setLastId(int lastId) {
		AclItem.lastId = lastId;
	}	
}