package com.perdemia.permissionanalyzer.data;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.SQLIntegrityConstraintViolationException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.preferences.PreferenceHelper;
import com.perdemia.permissionanalyzer.application.util.ApplicationUtil;
import com.perdemia.permissionanalyzer.data.util.CustomProcessor;
import com.perdemia.permissionanalyzer.data.util.DBUtil;
import com.perdemia.permissionanalyzer.data.util.DBUtil.DbType;

public abstract class DBObject {
	protected Integer id;
	private static Logger logger = LoggerFactory.getLogger(DBObject.class.getName());
	public final static String SCHEMA = "PERMISSIONANALYZER";
	private final static Map<String, PreparedStatement> preparedStatementCache = new HashMap<String, PreparedStatement>();

	/** Queries cached per subclass */
	private final static Map<Class<? extends DBObject>, String> selectQueryCache = new HashMap<Class<? extends DBObject>, String>();
	private final static Map<Class<? extends DBObject>, String> updateQueryCache = new HashMap<Class<? extends DBObject>, String>();
	private final static Map<Class<? extends DBObject>, String> insertQueryCache = new HashMap<Class<? extends DBObject>, String>();
	private final static Map<String, String> deleteQueryCache = new HashMap<String, String>();
	private final static Map<String, String> whereQueryCache = new HashMap<String, String>();
	
	private static List<FileInfo> fileInfoBatchQueue = new ArrayList<FileInfo>();
	private static List<AclItem> aclItemBatchQueue = new ArrayList<AclItem>();
	private static List<Member> memberBatchQueue = new ArrayList<Member>();
	private static List<Membership> membershipBatchQueue = new ArrayList<Membership>();
	private static List<ReportComparator1> reportComparator1BatchQueue = new ArrayList<ReportComparator1>();
	private static List<ReportComparator2> reportComparator2BatchQueue = new ArrayList<ReportComparator2>();
	
	private static boolean isAutoCommit = true;
	
	/**
	 * Fills the properties of this (subclass) instance with values from the <code>ResultSet</code>
	 * 
	 * @param rs
	 *            Row from the database
	 * 
	 * @throws SQLException
	 */
	protected abstract void buildObject(ResultSet rs) throws SQLException;

	/**
	 * Returns a list of table attribute names for the subclass. This does <b>not</b> include the id column.
	 * 
	 * @return table attribute names
	 */
	protected abstract String[] getTableAttributes();

	/**
	 * Fills the <code>PreparedStatement</code> with the right parameters. The order of parameters matches the order of {@link #getTableAttributes()}
	 * 
	 * @param ps
	 *            PreparedStatement
	 */
	protected abstract void setQueryParameters(PreparedStatement ps) throws SQLException;

	/**
	 * Returns the last ID that was used for an INSERT of this DBObject subclass in the database. If 0 is returned then the
	 * {@link #insert()} method will retrieve the max ID from the database.
	 *  
	 * @return last Id
	 */
	public abstract int getLastId();
	
	/**
	 * Sets the last ID that was being inserted in the database for the subclass of this DBObject.
	 * 
	 * @param id id
	 */
	public abstract void setLastId(int id);
	
	/**
	 * Gets the database connection from {@link ApplicationUtil}. This class will use the application preferences to build a connection and it hold one instance of a open connection.
	 * 
	 * @return open connection
	 */
	public static Connection getConnection() throws DBException {
		return DBUtil.getDBConnection();
	}

	private static void executeFileInfoBatch() throws DBException, SQLException {
//		// TEMP REINIER
//		if (fileInfoBatchQueue != null && fileInfoBatchQueue.size() > 8) {
//			logger.warn("EXECUTE FILE BATCH: firstID=" + fileInfoBatchQueue.get(0).getId() + ", " + fileInfoBatchQueue.get(1).getId()+ fileInfoBatchQueue.get(2).getId() + ", "+ fileInfoBatchQueue.get(3).getId() + ", "+ fileInfoBatchQueue.get(4).getId() + ", "+ fileInfoBatchQueue.get(5).getId() + ", "+ fileInfoBatchQueue.get(6).getId() + ", "+ fileInfoBatchQueue.get(7).getId());
//		} else {
//			logger.warn("EXECUTE FILE BATCH with batch size: " + (fileInfoBatchQueue != null ? fileInfoBatchQueue.size() : "null"));
//		}
		
		executeBatch(fileInfoBatchQueue, new FileInfo(), new DuplicateKeyHandler(){
			@Override
			public void handleDuplicateKey(int increaseValue, int firstFailingId) {
				
				// TEMP REINIER
				//logger.warn("DUPLICATE KEY: firstFailingId=" + firstFailingId + ", increaseValue=" + increaseValue);
				
				
				// update the AclItems that are still in the queue
				for (AclItem aclItem : aclItemBatchQueue) {
					if (aclItem.getFileId() >= firstFailingId) {
						aclItem.setFileId(aclItem.getFileId() + increaseValue);
					}
					
					if (aclItem.getInheritedFromFolderId() != null && aclItem.getInheritedFromFolderId() >= firstFailingId) {
						aclItem.setInheritedFromFolderId(aclItem.getInheritedFromFolderId() + increaseValue);
					}
				}
				
				// update the Parent ID's in the FileInfo objects that are now in the queue
				for (FileInfo fileInfo : fileInfoBatchQueue) {
					if (fileInfo.getParentId() != null && fileInfo.getParentId() >= firstFailingId) {
						fileInfo.setParentId(fileInfo.getParentId() + increaseValue);
					}
					
					if (fileInfo.getParentFileIds() != null && !fileInfo.getParentFileIds().isEmpty()) {
												
						// convert the comma separated list of ids to a list of Integers, and increase the id's
						String[] parentIds = fileInfo.getParentFileIds().split(",");
						List<Integer> newParentIds = new ArrayList<Integer>();
						for (String nextParentId : parentIds) {
							Integer nextParentIdInt = Integer.parseInt(nextParentId);
							if (nextParentIdInt >= firstFailingId) {
								nextParentIdInt += increaseValue;
							}
							newParentIds.add(nextParentIdInt);
						}
						
						// build the new comma separated list of parent id's again
						StringBuilder parentFileIds = new StringBuilder();
						for (int counter = 0; counter < newParentIds.size(); counter++) {
							if (counter > 0) {
								parentFileIds.append(",");
							}
							parentFileIds.append(newParentIds.get(counter));
						}
												
						fileInfo.setParentFileIds(parentFileIds.toString());
					}
				}
			}			
		});		
	}
	
	private static void executeMemberBatch() throws DBException, SQLException {
		executeBatch(memberBatchQueue, new Member(), new DuplicateKeyHandler(){
			@Override
			public void handleDuplicateKey(int increaseValue, int firstFailingId) {
				// update the Memberships that are still in the queue
				for (Membership membership : membershipBatchQueue) {
					if (membership.getMemberId() >= firstFailingId) {
						membership.setMemberId(membership.getMemberId() + increaseValue);
					}
					if (membership.getGroupId() >= firstFailingId) {
						membership.setGroupId(membership.getGroupId() + increaseValue);
					}
				}
				
				// update the AclItems that are still in the queue
				for (AclItem aclItem : aclItemBatchQueue) {
					if (aclItem.getMemberId() >= firstFailingId) {
						aclItem.setMemberId(aclItem.getMemberId() + increaseValue);
					}
				}
			}			
		});		
	}
		
	private static void executeAclItemBatch() throws DBException, SQLException {
		executeBatch(aclItemBatchQueue, new AclItem(), null);	
	}
	
	private static void executeMembershipBatch() throws DBException, SQLException {
		executeBatch(membershipBatchQueue, new Membership(), null);	
	}
	
	private static void executeReportComparator1Batch() throws DBException, SQLException {
		executeBatch(reportComparator1BatchQueue, new ReportComparator1(), null);
	}
	
	private static void executeReportComparator2Batch() throws DBException, SQLException {
		executeBatch(reportComparator2BatchQueue, new ReportComparator2(), null);
	}
	
	private static void executeBatch(List<? extends DBObject> batchList, DBObject dbObject, DuplicateKeyHandler duplicateKeyHandler) throws SQLException, DBException {	
		if (batchList.size() > 0) {
			int increaseId = 0;
			
			String query = getInsertQuery(dbObject);
			PreparedStatement ps = getPreparedStatement(query);
			boolean objectsNotInserted = true;
			int savetyCounter = 0;
		
			while (objectsNotInserted) {
				// is the file Id increased?
				if (increaseId > 0) {
					// update related items, like AclItems or Members that are still in the queue
					if (duplicateKeyHandler != null) {
						// TEMP REINIER
						//logger.warn(dbObject.getClass().getSimpleName());
						if (batchList != null && batchList.size() > 8) {
							logger.warn("DUPLICATE KEY IN EXECUTE BATCH of " + dbObject.getClass().getSimpleName() + ": batchList ID's=" + batchList.get(0).getId() + ", " + batchList.get(1).getId() + ", " +  batchList.get(2).getId() + ", "+ batchList.get(3).getId() + ", "+ batchList.get(4).getId() + ", "+ batchList.get(5).getId() + ", "+ batchList.get(6).getId() + ", "+ batchList.get(7).getId());
						} else {
							logger.warn("DUPLICATE KEY IN EXECUTE BATCH of " + dbObject.getClass().getSimpleName() + ": batchList size: " + (batchList != null ? batchList.size() : "null"));
						}
						duplicateKeyHandler.handleDuplicateKey(increaseId, batchList.get(0).getId());
					}
				}
				
				for (DBObject nextDBObject : batchList) {
					if (increaseId > 0) {
						nextDBObject.setId(nextDBObject.getId() + increaseId);
					}
					nextDBObject.setQueryParameters(ps);
					ps.setInt(nextDBObject.getTableAttributes().length + 1, nextDBObject.getId());
					ps.addBatch();
				}
				
				try {
					ps.executeBatch();
					getConnection().commit();
					objectsNotInserted = false;
					batchList.clear();
					
				} catch (SQLException ex) {
					
					getConnection().rollback();
					ps.clearBatch();
					
					// 23505 = H2, Postgresql
					// 23000 = MSSQL, MySQL, Oracle
					if ((ex instanceof SQLIntegrityConstraintViolationException || ex.getSQLState().equals("23000") || ex.getSQLState().equals("23505")) && savetyCounter < 10) {
						// the ID of the object is probably taken by other scan agent, increase the ID's and try again
						increaseId = 500000;
						
						// TEMP REINIER
						//logger.warn("dbObject.getLastId() " + dbObject.getLastId() + " + " + increaseId);
						
						dbObject.setLastId(dbObject.getLastId() + increaseId);
						
						if (logger.isDebugEnabled()) {
							logger.debug("Found duplicate key, increase the id's");
						}
						
						// TEMP REINIER
						//logger.warn("SQLException in insert of " + dbObject.getClass().getSimpleName() + ": " + ex.getMessage());
								
						savetyCounter++;
						
					} else {
						// something else went wrong
						logger.error("Error while inserting batch of records of type " + dbObject.getClass() + ": " + ex.getMessage(), ex);
						
						// bad luck, the batch failed, exit the loop
						objectsNotInserted = false;
						batchList.clear();
						
						throw ex;					
					}
				}
			}
		}
	}
	
	private static void checkForOutgoingBatch(boolean executeAll) throws SQLException, DBException {
		if (executeAll) {
			// the order is important! For example, the FileInfo records must be present in the database when the AclItems are persisted
			executeMemberBatch();
			executeFileInfoBatch();
			executeAclItemBatch();			
			executeMembershipBatch();
			executeReportComparator1Batch();
			executeReportComparator2Batch();
			
		} else {
			// check the threshold values
			if (fileInfoBatchQueue.size() > PreferenceHelper.getInstance().getBatchSizeFiles()) {
				executeMemberBatch(); // for the file owner records
				executeFileInfoBatch();
			}
			
			if (memberBatchQueue.size() > PreferenceHelper.getInstance().getBatchSizeMembers()) {
				executeMemberBatch();
				//executeFileInfoBatch(); ?? why files? Tevens in aclItemBatchQueue de memberBatch en fileBatch omgedraaid
			}
			
			if (aclItemBatchQueue.size() > PreferenceHelper.getInstance().getBatchSizeAces()) {
				executeMemberBatch();
				executeFileInfoBatch();
				executeAclItemBatch();
			}

			if (membershipBatchQueue.size() > PreferenceHelper.getInstance().getBatchSizeMemberships()) {
				executeMemberBatch();
				executeMembershipBatch();
			}
			
			if (reportComparator1BatchQueue.size() > 1000) {
				executeReportComparator1Batch();
			}
			
			if (reportComparator2BatchQueue.size() > 1000) {
				executeReportComparator2Batch();
			}			
		}
	}
	
	public static void setAutoCommit(boolean autoCommit) throws SQLException, DBException {
		logger.debug("Setting auto-commit: " + autoCommit);
		if (!isAutoCommit) {
			checkForOutgoingBatch(true);
			getConnection().commit();
		}
		getConnection().setAutoCommit(autoCommit);
		isAutoCommit = autoCommit;
		
		// The PreparedStatement object caches the connection's AutoCommit property when the statement is prepared. 
		// Later changes to the AutoCommit property have no effect on the prepared statement. So clear the cache!
		preparedStatementCache.clear();
	}

	/**
	 * Checks if there is already a <code>PreparedStatement</code> for the given query in the cache {@link #preparedStatementCache}. If it is not present it will add the new <code>PreparedStatement</code> in the cache.
	 * 
	 * @param query
	 * @return <code>PreparedStatement</code> object from cache
	 */
	public static PreparedStatement getPreparedStatement(String query) throws SQLException, DBException {
		return getPreparedStatement(query, "");
	}

	/**
	 * Checks if there is already a <code>PreparedStatement</code> for the given query and auto-increment column in the cache {@link #preparedStatementCache}. If it is not present it will add the new <code>PreparedStatement</code> in the cache.
	 * 
	 * @param query
	 * @return <code>PreparedStatement</code> object from cache
	 */
	private static PreparedStatement getPreparedStatement(String query, String incrementColumn) throws SQLException, DBException {
		PreparedStatement ps;
		query = checkQueryForDialect(query);
		if (preparedStatementCache.containsKey(query + incrementColumn)) {
			ps = preparedStatementCache.get(query + incrementColumn);
		} else {
			logger.info("Caching query: " + query);
			
			if (incrementColumn.length() > 0) {
				ps = getConnection().prepareStatement(query, new String[] { incrementColumn });
			} else {
				if ("mysql".equalsIgnoreCase(getConnection().getMetaData().getDatabaseProductName())) {
					ps = getConnection().prepareStatement(query, ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);
					ps.setFetchSize(Integer.MIN_VALUE);
				} else {
					ps = getConnection().prepareStatement(query);
				}
			}
			preparedStatementCache.put(query + incrementColumn, ps);
		}
		ps.clearParameters();
		return ps;
	}
	
	/**
	 * Clears the preparedStatementCache so that existing (cached) PreparedStatement don't fail in case the DB connection has been closed!
	 */
	 public static void clearPreparedStatementCache() {
		preparedStatementCache.clear();
	}

	/**
	 * Returns a list of object with the type of the given <code>Class</code> that will match the <code>whereClause</code>.
	 * 
	 * @param clazz
	 *            type of object to return and to query
	 * @param orderBy
	 *            an optional ORDER BY statement, like "NAME DESC"
	 * @param limit
	 *            a limit on the number of rows to return, 0 means no limit
	 * 
	 * @return List containing objects of type <code>Class</code>
	 * @throws DBException
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected static List getAll(Class<? extends DBObject> clazz, String orderBy, int limit) throws DBException {
		ResultSet rs = null;
		PreparedStatement ps = null;
		List result = new ArrayList<DBObject>();
		String query = null;

		try {
			query = getAllQuery(clazz, orderBy, limit);
			ps = getPreparedStatement(query);
			
			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + query);
			}

			rs = ps.executeQuery();

			while (rs.next()) {
				try {
					DBObject dbObject = clazz.newInstance();
					dbObject.buildObject(rs);
					result.add(dbObject);
				} catch (InstantiationException | IllegalAccessException e) {
					logger.error("Error while creating new instance of " + clazz.getName(), e);
				}
			}

		} catch (SQLException ex) {
			logger.error("Error while getting records from database", ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while getting records from database: " + ex.getMessage(), ex);
			}
			
			throw new DBException(query, ex);

		} finally {
			closeResultSet(rs);
		}

		return result;
	}
	
	/**
	 * Retrieves an instance of the given <code>Class</code> and id from the database. It will use the class name as table name and assumes there is a column named id.
	 * 
	 * @param clazz
	 *            object type to retrieve
	 * @param id
	 *            database id
	 * 
	 * @return an instance of the given <code>Class</code>
	 * @throws DBException
	 * 
	 * @see {@link #buildObject(ResultSet)

	 */
	public static DBObject getById(Class<? extends DBObject> clazz, Integer id) throws DBException {
		ResultSet rs = null;
		PreparedStatement ps = null;
		String query = null;
		DBObject dbObject = null;
				
		try {
			dbObject = clazz.newInstance();
		} catch (InstantiationException | IllegalAccessException e) {
			logger.error("Error while creating new instance of " + clazz.getName(), e);
		}

		try {
			checkForOutgoingBatch(true);
			
			if (id != null) {
				query = getSelectQuery(clazz);
				ps = getPreparedStatement(query);
				ps.setInt(1, id);
			}

			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + query);
			}

			rs = ps.executeQuery();

			if (rs.next()) {
				dbObject.buildObject(rs);
				return dbObject;
			}

		} catch (SQLException ex) {
			logger.error("Error while getting record by id from database using query [" + query + "]", ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while getting record by id from database using query [" + query + "]", ex);
				
			}
			
			throw new DBException(getSelectQuery(clazz), new Object[] { id }, ex);

		} finally {
			closeResultSet(rs);
		}

		return null;
	}

	/**
	 * Returns a list of object with the type of the given <code>Class</code> that will match the <code>whereClause</code>.
	 * 
	 * @param clazz
	 *            type of object to return and to query
	 * @param whereClause
	 *            sql where clause (for example: "name = ? or address = ?")
	 * @param orderBy
	 *            an optional ORDER BY statement, like "NAME DESC"
	 * @param limit
	 *            a limit on the number of rows to return, 0 means no limit
	 * @param parameters
	 *            parameters of the where clause
	 * 
	 * @return List containing objects of type <code>Class</code>
	 * @throws DBException
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected static List where(Class<? extends DBObject> clazz, String whereClause, Object[] parameters, String orderBy, int limit) throws DBException {
		ResultSet rs = null;
		PreparedStatement ps = null;
		List result = new ArrayList<DBObject>();
		String query = null;

		try {
			checkForOutgoingBatch(true);
			
			query = getWhereQuery(clazz, whereClause, orderBy, limit);
						
			ps = getPreparedStatement(query);
			if (parameters != null) {
				int counter = 1;
				for (Object param : parameters) {
					if (param instanceof String) {
						param = param.toString().replace("\\", "\\\\");
					}
					ps.setObject(counter++, param);
				}
			}

			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + query);
			}

			rs = ps.executeQuery();

			while (rs.next()) {
				try {
					DBObject dbObject = clazz.newInstance();
					dbObject.buildObject(rs);
					result.add(dbObject);
				} catch (InstantiationException | IllegalAccessException e) {
					logger.error("Error while creating new instance of " + clazz.getName(), e);
				}
			}

		} catch (SQLException ex) {
			logger.error("Error while getting records from database using query [" + query + "]", ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) wwhile getting records from database using query [" + query + "]", ex);
			}
			
			throw new DBException(query, parameters, ex);

		} finally {
			closeResultSet(rs);
		}

		return result;
	}

	/**
	 * Returns a list of object with the type of the given <code>Class</code> that will match the <code>whereClause</code>.
	 * 
	 * @param clazz
	 *            type of object to return and to query
	 * @param whereClause
	 *            sql where clause (for example: "name = ? or address = ?")
	 * @param limit
	 *            a limit on the number of rows to return, 0 means no limit
	 * @param parameters
	 *            parameters of the where clause
	 * 
	 * @return List containing objects of type <code>Class</code>
	 * @throws DBException
	 */
	@SuppressWarnings({ "rawtypes", "unchecked" })
	protected static List getByPreparedStatement(Class<? extends DBObject> clazz, PreparedStatement ps) throws DBException {
		ResultSet rs = null;
		List result = new ArrayList<DBObject>();

		try {
			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + ps.toString());
			}

			rs = ps.executeQuery();

			while (rs.next()) {
				try {
					DBObject dbObject = clazz.newInstance();
					dbObject.buildObject(rs);
					result.add(dbObject);
				} catch (InstantiationException | IllegalAccessException e) {
					logger.error("Error while creating new instance of " + clazz.getName(), e);
				}
			}

		} catch (SQLException ex) {
			logger.error("Error while getting records from database", ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while getting records from database" + ex.getMessage(), ex);
			}
			
			throw new DBException(ps != null ? ps.toString() : "null", null, ex);

		} finally {
			closeResultSet(rs);
		}

		return result;
	}

	/**
	 * Executes a PreparedStatement and lets a custom processor handle all the record sets
	 * 
	 * @param ps
	 *            preparedstatement to execute
	 * @param customProcessor
	 *            implementation class that will handle the record sets
	 * 
	 * @throws DBException
	 */
	public static void getUsingCustomProcessor(PreparedStatement ps, CustomProcessor customProcessor) throws DBException {
		ResultSet rs = null;

		try {
			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + ps.toString());
			}

			rs = ps.executeQuery();
			
			boolean processNext = true;
			while (rs.next() && processNext) {
				processNext = customProcessor.processResultSet(rs);
			}

		} catch (SQLException ex) {
			logger.error("Error while getting records from database", ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while getting records from database: " + ex.getMessage(), ex);
			}
			
			throw new DBException(ps != null ? ps.toString() : "null", null, ex);

		} finally {
			closeResultSet(rs);
		}
	}

	
	/**
	 * Returns the first result of a query or <code>null</code> if none is found.
	 * 
	 * @param clazz
	 *            type of object to return and to query
	 * @param whereClause
	 *            sql where clause (for example: "name = ? or address = ?")
	 * @param parameters
	 *            parameters of the where clause
	 * 
	 * @return Object of type <code>Clazz</code>
	 * @throws DBException
	 */
	@SuppressWarnings("rawtypes")
	protected static DBObject findFirst(Class<? extends DBObject> clazz, String whereClause, Object... parameters) throws DBException {
		List result = where(clazz, whereClause, parameters, "", 1);
		if (result != null && result.size() > 0) {
			return (DBObject) result.get(0);
		} else {
			return null;
		}
	}

	/**
	 * Deletes a record from the database based on the given <code>Class</code> (table name) and Id.
	 * 
	 * @param clazz
	 *            type of object to delete. This parameter is used to determine the table name
	 * @param id
	 *            id of the object to delete
	 * 
	 * @return number of affected rows
	 * @throws DBException
	 */
	protected static int delete(Class<? extends DBObject> clazz, Integer id) throws DBException {
		PreparedStatement ps = null;
		String query = null;
		
		try {
			if (id != null) {
				query = getDeleteQuery(clazz, "id = ?");
				ps = getPreparedStatement(query);
				ps.setInt(1, id);
			}

			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + query);
			}

			int result = ps.executeUpdate();
			
			return result;
			
		} catch (SQLException ex) {
			logger.error("Error while deleting record by id from database using query [" + query + "]", ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while deleting record by id from database using query [" + query + "]", ex);
			}
			
			throw new DBException(getDeleteQuery(clazz, "id = ?"), new Object[] { id }, ex);
		}
	}

	/**
	 * Deletes records from the database based on the given <code>whereClause</code>.
	 * 
	 * @param clazz
	 *            type of object to delete. This parameter is used to determine the table name
	 * @param whereClause
	 *            criteria for the delete statement
	 * 
	 * @return number of affected rows
	 * @throws DBException
	 */
	protected static int delete(Class<? extends DBObject> clazz, String whereClause, Object... parameters) throws DBException {
		PreparedStatement ps = null;

		String query = getDeleteQuery(clazz, whereClause);
		
		try {
			ps = getPreparedStatement(query);
			if (parameters != null) {
				int counter = 1;
				for (Object param : parameters) {
					ps.setObject(counter++, param);
				}
			}

			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + query);
			}

			checkForOutgoingBatch(true);
			
			return ps.executeUpdate();
			
		} catch (SQLException ex) {
			logger.error("Error while deleting records from database using query " + query, ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while deleting records from database using query " + query, ex);
			}
			
			throw new DBException(query, parameters, ex);
		}
	}

	/**
	 * Deletes the current object instance from the database.
	 * 
	 * @return number of affected rows
	 * @throws DBException
	 */
	public int delete() throws DBException {
		return delete(getClass(), getId());
	}

	/**
	 * Updates the current object instance in the database.
	 * 
	 * @return number of affected rows
	 * @throws DBException
	 * 
	 * @see {@link #getTableAttributes()}
	 * @see {@link #setQueryParameters(PreparedStatement)}
	 */
	protected int update() throws DBException {
		PreparedStatement ps = null;
		String query = null;

		if (logger.isTraceEnabled()) {
			logger.trace("Executing update query: " + getUpdateQuery(this));
		}

		try {
			query = getUpdateQuery(this);
			ps = getPreparedStatement(getUpdateQuery(this));
			setQueryParameters(ps);

			ps.setInt(getTableAttributes().length + 1, id);

			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + query);
			}

			checkForOutgoingBatch(true);
			
			int result = ps.executeUpdate();
			
			// always commit updates
			if (!isAutoCommit) {
				getConnection().commit();
			}
			
			return result;
			
		} catch (SQLException ex) {
			logger.error("Error while updating record in database using query [" + query + "]: " + ex.getMessage(), ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while updating record in database using query [" + query + "]: " + ex.getMessage(), ex);
				
			}
			
			throw new DBException(getUpdateQuery(this), new Object[] { id }, ex);
		}
	}
	
	
	/**
	 * Gets the max id.
	 *
	 * @param objectClass the object class
	 * @return the max id
	 */
	public static int getMaxId(Class<? extends DBObject> objectClass) {
		ResultSet rs = null;
		
		try {
			// get the max id from the database
			rs = executeQuery("select max(id) from " + SCHEMA + "." + objectClass.getSimpleName().toUpperCase());
			if (rs.next() && rs.getInt(1) > 0) {
				return rs.getInt(1);
			} else {
				return 1;
			}
		} catch (Exception e) {
			logger.error("Error while retrieving the max id: " + e.getMessage(), e);
			return 0;
		}
		finally {
			DBObject.closeResultSet(rs);
		}
	}
	
	/**
	 * Inserts the current object instance in the database.
	 * 
	 * @throws DBException
	 * 
	 * @see {@link #getTableAttributes()}
	 * @see {@link #setQueryParameters(PreparedStatement)}
	 */
	protected void insert() throws DBException {
		PreparedStatement ps = null;
		String query = null;
		
		if (logger.isTraceEnabled()) {
			logger.trace("Executing insert query: " + getInsertQuery(this));
		}

		try {
			// determine the next id
			// when isAutoCommit is true, we are not in scanning mode, and don't use ExecuteBatch and don't check on DuplicateKeyException. In this case we always want to retrieve the latest ID from the db,
			// because other instances may have insert new records
			if (getLastId() == 0 || isAutoCommit) {
				setLastId(getMaxId(this.getClass()));																
				
				// TEMP REINIER
				//logger.warn("Retrieving max id from the database for table " + this.getClass().getSimpleName() + ": " + getLastId());
				
				logger.debug("Retrieving max id from the database for table " + this.getClass().getSimpleName() + ": " + getLastId());
			}
			
			if (!isAutoCommit && (this instanceof FileInfo || this instanceof AclItem || this instanceof Member || this instanceof Membership || this instanceof ReportComparator1 || this instanceof ReportComparator2)) {
				if (logger.isTraceEnabled()) {
					logger.trace("Batching object: " + this.getClass().getSimpleName() + ", " + getLastId());
				}
				
				// generate a new id
				id = getLastId() + 1;
				setLastId(id);
								
				// add this object to the queue
				if (this instanceof FileInfo) {
					fileInfoBatchQueue.add((FileInfo) this);
				} else if (this instanceof AclItem) {
					aclItemBatchQueue.add((AclItem) this);
				} else if (this instanceof Member) {
					memberBatchQueue.add((Member) this);
				} else if (this instanceof Membership) {
					membershipBatchQueue.add((Membership) this);
				} else if (this instanceof ReportComparator1) {
					reportComparator1BatchQueue.add((ReportComparator1) this);
				} else if (this instanceof ReportComparator2) {
					reportComparator2BatchQueue.add((ReportComparator2) this);
				}
				
				checkForOutgoingBatch(false);
				
			} else {
				// no batching
				query = getInsertQuery(this);
				
				/*
				// the second argument (ID) determine the auto-generate id column. It's uppercase because Derby is case sensitive for this value and uses uppercase
				// column names. We don't want to quote column names in de create table statements.
				if (DBUtil.getDbType().equals(DbType.POSTGRESQL)) {
					// column names are folded to lower case in PostgreSQL :(
					ps = getPreparedStatement(query, "id");
				} else {
					ps = getPreparedStatement(query, "ID");
				}*/
				
				ps = getPreparedStatement(query);
				setQueryParameters(ps);
				
				id = getLastId() + 1;
				ps.setInt(getTableAttributes().length + 1, id);
				setLastId(id); 
				
				if (logger.isTraceEnabled()) {
					logger.trace("Executing query: " + query);
				}
								
				ps.executeUpdate();
				
				if (!isAutoCommit) {
					getConnection().commit();
				}
			}
			

			// get the new (auto generated) id
			/*
			ResultSet rs = null;
			try {
				rs = ps.getGeneratedKeys();
				if (rs.next()) {
					id = rs.getInt(1);
				} else {
					logger.error("Failed to find out the auto-incremented value for the new id. object: " + this.toString());
				}
			} catch (Exception e) {
				logger.error("Failed to find out the auto-incremented value for the new id. object: " + this.toString(), e);
			} finally {
				closeResultSet(rs);
			}*/

		} catch (SQLException ex) {
			logger.error("Error while inserting record in database using query [" + query + "]: " + ex.getMessage(), ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while inserting record in database using query [" + query + "]: " + ex.getMessage(), ex);
			}
			
			throw new DBException(getInsertQuery(this), new Object[] { id }, ex);
		}
	}

	/**
	 * Either inserts or updates the object in the database, based on the current id. In case the id is <code>null</code> an insert will take place, otherwise the update method is called.
	 * 
	 * @throws DBException
	 * 
	 * @see {@link #update()}
	 * @see {@link #insert()}
	 */
	public void save() throws DBException {
		if (id == null) {
			insert();
		} else {
			update();
		}
	}
	
	/**
	 * Execute a SELECT COUNT for the given class and returns the int result
	 * 
	 * @param clazz table to select
	 * @return number of rows in the table
	 * @throws DBException
	 * 
	 */
	protected static int getCount(Class<? extends DBObject> clazz) throws DBException {
		Statement statement = null;
		String countQuery = "select count(*) from " + SCHEMA + "." + clazz.getSimpleName().toUpperCase();
		ResultSet rs = null;
		
		try {
			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + countQuery);
			}

			statement = getConnection().createStatement();
			rs = statement.executeQuery(countQuery);

			if (rs.next()) {
				return rs.getInt(1);
			} else {
				logger.warn("Table count did not return any results: " + countQuery);
				return -1;
			}
		} catch (SQLException ex) {
			logger.error("Error while executing query " + countQuery, ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while executing query " + countQuery, ex);
			}
			
			throw new DBException(countQuery, ex);
		} finally {
			closeResultSet(rs);
			closeStatement(statement);
		}
	}
	
	/**
	 * Execute a SELECT COUNT for the given class and returns the int result
	 * 
	 * @param clazz table to select
	 * @param where where clause to add to the select count query (may include ? placeholders)
	 * @param params parameters for the prepared statement
	 * 
	 * @return number of rows
	 * @throws DBException
	 * 
	 */
	protected static int getCount(Class<? extends DBObject> clazz, String where, Object ... params) throws DBException {
		PreparedStatement statement = null;
		ResultSet rs = null;
		String countQuery = "select count(*) from " + SCHEMA + "." + clazz.getSimpleName().toUpperCase() + " where " + where;
		countQuery = checkQueryForDialect(countQuery);
		
		try {
			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + countQuery);
			}
						
			statement = getConnection().prepareStatement(countQuery);
			
			// apply parameters
			for (int counter = 0; counter < params.length; counter++) {
				statement.setObject(counter, params[counter]);
			}
			
			rs = statement.executeQuery();

			if (rs.next()) {
				return rs.getInt(1);
			} else {
				logger.warn("Table count did not return any results: " + countQuery);
				return -1;
			}
		} catch (SQLException ex) {
			logger.error("Error while executing query " + countQuery, ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while executing query " + countQuery, ex);
			}
			
			throw new DBException(countQuery, ex);
		} finally {
			closeResultSet(rs);
			closeStatement(statement);
		}
	}

	/**
	 * Executes a query directly on the DB connection.
	 * 
	 * @param query
	 *            query to execute
	 * @throws DBException
	 * 
	 */
	public static void executeStatement(String query) throws DBException {
		Statement statement = null;

		try {
			query = checkQueryForDialect(query);
			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + query);
			}

			checkForOutgoingBatch(true);
			
			statement = getConnection().createStatement();
			statement.execute(query);
			
		} catch (SQLException ex) {
			logger.error("Error while executing query " + query, ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while executing query " + query, ex);
			}
			
			throw new DBException(query, ex);
		} finally {
			closeStatement(statement);
		}
	}

	/**
	 * Executes a <code>PreparedStatement</code> on the database with the given parameters.
	 * 
	 * @param ps
	 *            PreparedStatement to execute
	 * @param isUpdate
	 *            Indicates whether or not to use executeUpdate or executeQuery
	 * 
	 * @return ResultSet
	 * @throws DBException
	 */
	public static ResultSet executePreparedStatement(PreparedStatement ps, boolean isUpdate) throws DBException {
		try {
			if (logger.isTraceEnabled()) {
				logger.trace("Executing query: " + (ps != null ? ps.toString() : "null"));
			}
			
			checkForOutgoingBatch(true);
			
			ResultSet result = null;
			if (isUpdate) {
				ps.executeUpdate();
			} else {
				result = ps.executeQuery();
			}
			
			return result;

		} catch (SQLException ex) {
			logger.error("Error while executing query: " + ps, ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while executing query: " + ps, ex);
			}
			
			throw new DBException(ps != null ? ps.toString() : "null", null, ex);
		}
	}

	/**
	 * Executes a <code>PreparedStatement</code> on the database with the given parameters.
	 * 
	 * @param query
	 *            query to be executed (use ? for parameters)
	 * @param parameters
	 *            list of parameters
	 * 
	 * @return ResultSet
	 * @throws DBException
	 */
	public static ResultSet executeQuery(String query, Object... parameters) throws DBException {
		PreparedStatement ps = null;

		try {
			checkForOutgoingBatch(true);
			
			ps = getPreparedStatement(query);
			for (int counter = 0; counter < parameters.length; counter++) {
				ps.setObject(counter + 1, parameters[counter]);
			}
			
			return executePreparedStatement(ps, false);
			
		} catch (SQLException ex) {
			logger.error("Error while executing query: " + query + "[" + Arrays.asList(parameters).toString() + "]", ex);
			
			if (ex.getNextException() != null) {
				ex = ex.getNextException();
				logger.error("Error (nested exception) while executing query: " + query + "[" + Arrays.asList(parameters).toString() + "]", ex);
			}
			
			throw new DBException(query, parameters, ex);
		}
	}

	/**
	 * Retrieves a select query for the given <code>Class</code> from the cache. If it doesn't exists it will build the select query and adds it to the cache.
	 * 
	 * @param clazz
	 * @return
	 */
	private static String getSelectQuery(Class<? extends DBObject> clazz) {
		if (selectQueryCache.containsKey(clazz)) {
			return selectQueryCache.get(clazz);
		} else {
			String selectQuery = "select * from " + SCHEMA + "." + clazz.getSimpleName().toUpperCase() + " where id = ?";
			selectQueryCache.put(clazz, selectQuery);
			return selectQuery;
		}
	}

	/**
	 * Retrieves an update query for the given <code>Class</code> from the cache. If it doesn't exists it will build the update query and adds it to the cache.
	 * 
	 * @param clazz
	 * @return
	 */
	private static String getUpdateQuery(DBObject dbObject) {
		if (updateQueryCache.containsKey(dbObject.getClass())) {
			return updateQueryCache.get(dbObject.getClass());
		} else {
			// build the update query for this class
			StringBuilder builder = new StringBuilder("update ");
			builder.append(SCHEMA);
			builder.append(".");
			builder.append(dbObject.getClass().getSimpleName().toUpperCase());
			builder.append(" set ");

			boolean first = true;
			for (String attribute : dbObject.getTableAttributes()) {
				if (!first) {
					builder.append(", ");
				}

				builder.append(attribute);
				builder.append(" = ? ");
				first = false;
			}
			builder.append("where id = ?");

			String updateQuery = builder.toString();
			updateQueryCache.put(dbObject.getClass(), updateQuery);
			return updateQuery;
		}
	}

	/**
	 * Retrieves an insert query for the given <code>Class</code> from the cache. If it doesn't exists it will build the insert query and adds it to the cache.
	 * 
	 * @param clazz
	 * @return
	 */
	private static String getInsertQuery(DBObject dbObject) throws DBException {
		if (insertQueryCache.containsKey(dbObject.getClass())) {
			return insertQueryCache.get(dbObject.getClass());
		} else {
			// build the query
			StringBuilder builder = new StringBuilder("insert into ");
			builder.append(SCHEMA);
			builder.append(".");
			builder.append(dbObject.getClass().getSimpleName().toUpperCase());
			builder.append(" (");

			// attribute names, comma separated
			boolean first = true;
			
//			if (DBUtil.getDbType().equals(DbType.ORACLE)) {
//				builder.append("ID");
//				first = false;
//			}
			
			for (String attribute : dbObject.getTableAttributes()) {
				if (!first) {
					builder.append(", ");
				}

				builder.append(attribute);
				first = false;
			}

			builder.append(",id) values (");

			first = true;
			
//			if (DBUtil.getDbType().equals(DbType.ORACLE)) {
//				builder.append(dbObject.getClass().getSimpleName().toUpperCase() + "_SEQ.NEXTVAL");
//				first = false;
//			}
			
			// (include the 'id' value, which is not in the TableAttributes)
			for (int counter = 0; counter <= dbObject.getTableAttributes().length; counter++) {
				if (!first) {
					builder.append(", ");
				}
				builder.append("?");
				first = false;
			}
			builder.append(")");

			String insertQuery = builder.toString();
			insertQueryCache.put(dbObject.getClass(), insertQuery);
			return insertQuery;
		}
	}

	/**
	 * Retrieves a delete query for the given <code>Class</code> from the cache. If it doesn't exists it will build the delete query and adds it to the cache.
	 * 
	 * @param clazz
	 * @return
	 */
	private static String getDeleteQuery(Class<? extends DBObject> clazz, String whereClause) {
		if (deleteQueryCache.containsKey(clazz.getSimpleName() + whereClause)) {
			return deleteQueryCache.get(clazz.getSimpleName() + whereClause);
		} else {
			String deleteQuery = "delete from " + SCHEMA + "." + clazz.getSimpleName().toUpperCase() + " where " + whereClause;
			deleteQueryCache.put(clazz.getSimpleName() + whereClause, deleteQuery);
			return deleteQuery;
		}
	}

	/**
	 * Retrieves a where query for the given <code>Class</code> from the cache. If it doesn't exists it will build the where query and adds it to the cache.
	 * 
	 * @param clazz
	 *            object type to return (and to build based on the ResultSet)
	 * @param whereClause
	 *            WHERE statement to add to the base SELECT statement, like "NAME LIKE ?" * @param orderBy an optional ORDER BY statement, like "NAME DESC"
	 * @param orderBy
	 *            an optional ORDER BY statement, like "NAME DESC"
	 * @param limit
	 *            optional the maximum number of rows
	 * 
	 * @return
	 * @throws DBException 
	 */
	protected static String getWhereQuery(Class<? extends DBObject> clazz, String whereClause, String orderBy, int limit) throws DBException {
		if (whereQueryCache.containsKey(clazz.getSimpleName() + whereClause + limit + orderBy)) {
			return whereQueryCache.get(clazz.getSimpleName() + whereClause + limit + orderBy);
		} else {
			String whereQuery = "select * from " + SCHEMA + "." + clazz.getSimpleName().toUpperCase() + " where " + whereClause;
			whereQuery = addLimitAndOrderBy(whereQuery, orderBy, null, limit);
			whereQuery = checkQueryForDialect(whereQuery);
			
			whereQueryCache.put(clazz.getSimpleName() + whereClause + limit + orderBy, whereQuery);
			return whereQuery;
		}		
	}
	
	/**
	 * Retrieves a select all query for the given <code>Class</code> from the cache. If it doesn't exists it will build the select query and adds it to the cache.
	 * 
	 * @param clazz
	 *            object type to return (and to build based on the ResultSet)
	 * @param orderBy
	 *            an optional ORDER BY statement, like "NAME DESC"
	 * @param limit
	 *            optional the maximum number of rows
	 * 
	 * @return
	 */
	protected static String getAllQuery(Class<? extends DBObject> clazz, String orderBy, int limit) throws DBException {
		if (whereQueryCache.containsKey(clazz.getSimpleName() + limit + orderBy)) {
			return whereQueryCache.get(clazz.getSimpleName() + limit + orderBy);
		} else {
			String allQuery = "select * from " + SCHEMA + "." + clazz.getSimpleName().toUpperCase();
			if (limit > 0) {
				allQuery = addLimitAndOrderBy(allQuery, orderBy, null, limit);
			}
			whereQueryCache.put(clazz.getSimpleName() + limit + orderBy, allQuery);
			return allQuery;
		}		
	}

	/**
	 * Adds an ORDER BY and result limit to the query depending on the database type.
	 * 
	 * @param query
	 *            query or statement to add the order by and limit
	 * @param orderBy
	 *            an optional ORDER BY statement, like "NAME DESC"
	 * @param limit
	 *            optional the maximum number of rows
	 * 
	 * @return updated sql statement containing a limit statement and/or the ORDER BY statement
	 * 
	 * @throws DBException
	 */
	public static String addLimitAndOrderBy(String query, String orderBy, String groupBy, int limit) {
		if (orderBy != null && orderBy.length() > 0) {
			orderBy = " ORDER BY " + orderBy;
		} else {
			orderBy = "";
		}

		if (groupBy != null && groupBy.length() > 0) {
			groupBy = " GROUP BY " + groupBy;
		} else {
			groupBy = "";
		}
		
		if (limit > 0) {
			try {
				if (DBUtil.getDbType().equals(DbType.ORACLE)) {
					query += (" AND ROWNUM <= " + limit + groupBy + orderBy);
	
				} else if (DBUtil.getDbType().equals(DbType.DB2) || DBUtil.getDbType().equals(DbType.DERBY)) {
					query += (groupBy + orderBy + " FETCH FIRST " + limit + " ROWS ONLY");
	
				} else if (DBUtil.getDbType().equals(DbType.MSSQL)) {
					//query += (" AND ROWNUMBER <= " + limit + orderBy);
					if (query.toLowerCase().startsWith("select distinct")) {
						query = "SELECT DISTINCT TOP " + limit + query.substring(15) + groupBy + orderBy;
					} else {
						query = "SELECT TOP " + limit + query.substring(6) + groupBy + orderBy;
					}
			
				} else {
					// MySQL, PostGreSql, H2
					query += (groupBy + orderBy + " LIMIT " + limit);
				}
			} catch (DBException ex) {
				logger.error("Error while determining database type and could not add ORDER BY or LIMIT.", ex);
			}
		} else {
			query += groupBy + orderBy;
		}

		return query;
	}

	public static void closeStatement(final Statement stmt) {
		if (stmt != null) {
			try {
				stmt.close();
			} catch (SQLException ex) {
				logger.error("Error while closing SQL statement", ex);
			}
		}
	}

	public static void closeResultSet(final ResultSet rs) {
		if (rs != null) {
			try {
				rs.close();
			} catch (SQLException ex) {
				logger.error("Error while closing SQL ResultSet", ex);
			}
		}
	}

	/**
	 * Reads an Integer value from the RecordSet and checks if the value is null. If so it returns null instead of 0 like the default JDBC implementation :(
	 * 
	 * @param rs
	 * @param columnName
	 * @return <code>null</code> in case the column value is null otherwise the Integer value
	 * 
	 * @throws SQLException
	 */
	public static Integer getNullableInteger(ResultSet rs, String columnName) throws SQLException {
		return rs.getObject(columnName) != null ? rs.getInt(columnName) : null;
	}

	/**
	 * Sets an Integer value as parameter in the given PreparedStatement. In case the value is null it uses <code>ps.setNull(paramIndex, Types.INTEGER)</code> to process the null value.
	 * 
	 * @param ps
	 * @param paramIndex
	 * @param value
	 * 
	 * @throws SQLException
	 */
	public static void setNullableInteger(PreparedStatement ps, int paramIndex, Integer value) throws SQLException {
		if (value != null) {
			ps.setInt(paramIndex, value);
		} else {
			ps.setNull(paramIndex, Types.INTEGER);
		}
	}
	
	/**
	 * Sets an Timestamp value as parameter in the given PreparedStatement. In case the value is null it uses <code>ps.setNull(paramIndex, Types.TIMESTAMP)</code> to process the null value.
	 * 
	 * @param ps
	 * @param paramIndex
	 * @param value java.util.date which is converted to a Timestamp
	 * 
	 * @throws SQLException
	 */
	public static void setNullableTimestamp(PreparedStatement ps, int paramIndex, Date value) throws SQLException {
		if (value != null) {
			ps.setTimestamp(paramIndex, new Timestamp(value.getTime()));
		} else {
			ps.setNull(paramIndex, Types.TIMESTAMP);
		}
	}
	
	/**
	 * Reads an Integer value from the RecordSet and checks if the value is null. If so it returns null instead of 0 like the default JDBC implementation :(
	 * 
	 * @param rs
	 * @param columnName
	 * @return <code>null</code> in case the column value is null otherwise the Integer value
	 * 
	 * @throws SQLException
	 */
	public static String getNullAsEmptyString(ResultSet rs, String columnName) throws SQLException {
		return rs.getObject(columnName) != null ? rs.getString(columnName) : "";
	}
		
	/**
	 * Applies a specific dialect to a query. Things like LIMIT syntax is already handled in other methods. This method offers the possibilty to
	 * check every query before is it executed. For now only one check if performed, if we're dealing with MS SQL, ORACLE, DB2 or Derby we replace 'true' for 1 and 'false' for 1,
	 * because those DB's have no boolean column types, only bit, number(1) or smallint.
	 * @param query
	 * @return
	 */
	public static String checkQueryForDialect(String query) throws DBException {
		if (DBUtil.getDbType().equals(DbType.MSSQL) || DBUtil.getDbType().equals(DbType.ORACLE) || DBUtil.getDbType().equals(DbType.DB2)
				|| DBUtil.getDbType().equals(DbType.DERBY)) {
			query = query.replace("false", "0");
			query = query.replace("FALSE", "0");
			query = query.replace("true", "1");
			query = query.replace("TRUE", "1");
		}
		
		return query;
	}
	
	public Integer getId() {
		return id;
	}

	public void setId(Integer id) {
		this.id = id;
	}
}

interface DuplicateKeyHandler {
	public void handleDuplicateKey(int increaseValue, int firstFailingId);
}