package com.perdemia.permissionanalyzer.data;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class Filter extends DBObject {
private final static Logger logger = LoggerFactory.getLogger(Filter.class.getName());
	
	/** Table attributes */
	private final static String[] tableAttributes = new String[]{"NAME", "FILTERSET_ID", "REPORT_ID", "FILTER_TYPE", "INCLUDE", "FILTER_VALUE"};
	
	/** Table fields */
	private String name;
	private Integer filterSetId;
	private Integer reportId;
	private FilterType type;
	private boolean include;
	private String value;
	
	/**
	 * Defines the possible filters. 
	 * @author Reinier
	 *
	 */
	public static enum FilterType {
		MEMBER_FILTER_CONFIG, MEMBER_NAME, MEMBER_ID, MEMBER_GROUP_ID, MEMBER_LDAP_OU, MEMBER_GROUP_NAME,
		PERMISSION_FILTER_CONFIG, PERMISSION_ALLOW, PERMISSION_DENY,
		FOLDER_FILTER_CONFIG, FOLDER_NAME, FOLDER_PATH, OWNER_ID, MEMBER_REPORT
	}
		
	/** Helper field, indicates if a MEMBER_ID filter is a group */
	private boolean isMemberGroup;
	
	/**
	 * The last ID that was used for an INSERT of this class in the database. If 0 is returned then the
	 * {@link DBObject#insert()} method will retrieve the max ID from the database.
	 */
	private static int lastId;
	
	public Filter() {}	

	public Filter(FilterType type, boolean include, String value, String name) {
		this.type = type;
		this.include = include;
		this.value = value;
		this.name = name;
	}
	
	public Filter clone() {
		return new Filter(type, include, value, name);
	}
	
	//========================================
	// GENERAL DB ACTIONS
	//========================================
	
	@SuppressWarnings("unchecked")
	public static List<Filter> getAll() throws DBException {
		return (List<Filter>) getAll(Filter.class, null, 0);
	}
	
	public static Filter getById(Integer id) throws DBException {
		return (Filter) getById(Filter.class, id);
	}
		
	@SuppressWarnings("unchecked")
	public static List<Filter> where(String whereClause, Object[] parameters) throws DBException {
		return where(Filter.class, whereClause, parameters, null, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<Filter> where(String whereClause, Object[] parameters, String orderBy) throws DBException {
		return where(Filter.class, whereClause, parameters, orderBy, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<Filter> where(String whereClause, Object[] parameters, String orderBy, int limit) throws DBException {
		return where(Filter.class, whereClause, parameters, orderBy, limit);
	}
	
	public static Filter findFirst(String whereClause, Object... parameters) throws DBException {
		return (Filter) findFirst(Filter.class, whereClause, parameters);
	}
	
	public static int delete(Integer id) throws DBException {
		return delete(Filter.class, id);
	}
	
	public static int delete(String whereClause, Object...parameters) throws DBException {
		return delete(Filter.class, whereClause, parameters);
	}
	
	@Override
	protected String[] getTableAttributes(){
		return tableAttributes;
	}
	
	@Override
	protected void buildObject(ResultSet rs) throws SQLException {
		id = (rs.getInt("id"));
		name = rs.getString(tableAttributes[0]);
		filterSetId = getNullableInteger(rs, tableAttributes[1]);
		reportId = getNullableInteger(rs, tableAttributes[2]);
		type = FilterType.values()[rs.getInt(tableAttributes[3])];
		include = rs.getBoolean(tableAttributes[4]);
		value = rs.getString(tableAttributes[5]);
	}
	
	@Override
	protected void setQueryParameters(PreparedStatement ps) throws SQLException {
		ps.setString(1, name);
		setNullableInteger(ps, 2, filterSetId);
		setNullableInteger(ps, 3, reportId);
		ps.setInt(4, type != null ? type.ordinal() : null);
		ps.setBoolean(5, include);
		ps.setString(6, value);
	}

	//========================================
	// HELPER DB METHODS
	//========================================
	public static List<Filter> getMemberFiltersForReport(int reportId) throws DBException {
		return where("REPORT_ID = ? AND (FILTER_TYPE = ? OR FILTER_TYPE = ? OR FILTER_TYPE = ? OR FILTER_TYPE = ? OR FILTER_TYPE = ? OR FILTER_TYPE = ? OR FILTER_TYPE = ? OR FILTER_TYPE = ?)", 
			new Object[]{ reportId,
			FilterType.MEMBER_FILTER_CONFIG.ordinal(),
			FilterType.MEMBER_NAME.ordinal(),
			FilterType.MEMBER_ID.ordinal(),
			FilterType.MEMBER_GROUP_ID.ordinal(),
			FilterType.MEMBER_LDAP_OU.ordinal(),
			FilterType.MEMBER_GROUP_NAME.ordinal(),
			FilterType.OWNER_ID.ordinal(),
			FilterType.MEMBER_REPORT.ordinal()});				
	}
	
	public static List<Filter> getPermissionFiltersForReport(int reportId) throws DBException {
		return where("REPORT_ID = ? AND (FILTER_TYPE = ? OR FILTER_TYPE = ? OR FILTER_TYPE = ?)", 
			new Object[]{ reportId,
			FilterType.PERMISSION_FILTER_CONFIG.ordinal(),
			FilterType.PERMISSION_ALLOW.ordinal(),
			FilterType.PERMISSION_DENY.ordinal()});				
	}
	
	public static List<Filter> getFolderFiltersForReport(int reportId) throws DBException {
		return where("REPORT_ID = ? AND (FILTER_TYPE = ? OR FILTER_TYPE = ? OR FILTER_TYPE = ?)", 
			new Object[]{ reportId,
			FilterType.FOLDER_FILTER_CONFIG.ordinal(),
			FilterType.FOLDER_NAME.ordinal(),
			FilterType.FOLDER_PATH.ordinal()});				
	}
	
	//========================================
	// GETTERS AND SETTERS
	//========================================
	
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public Integer getFilterSetId() {
		return filterSetId;
	}

	public void setFilterSetId(Integer filterSetId) {
		this.filterSetId = filterSetId;
	}

	public FilterType getType() {
		return type;
	}

	public void setType(FilterType type) {
		this.type = type;
	}

	public boolean isInclude() {
		return include;
	}

	public void setInclude(boolean include) {
		this.include = include;
	}

	public String getValue() {
		return value;
	}

	public void setValue(String value) {
		this.value = value;
	}

	public boolean isMemberGroup() {
		return isMemberGroup;
	}

	public void setMemberGroup(boolean isMemberGroup) {
		this.isMemberGroup = isMemberGroup;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("Filter [name=");
		builder.append(name);
		builder.append(", filterSetId=");
		builder.append(filterSetId);
		builder.append(", reportId=");
		builder.append(reportId);
		builder.append(", type=");
		builder.append(type);
		builder.append(", include=");
		builder.append(include);
		builder.append(", value=");
		builder.append(value);
		builder.append(", isMemberGroup=");
		builder.append(isMemberGroup);
		builder.append(", id=");
		builder.append(id);
		builder.append("]");
		return builder.toString();
	}

	public Integer getReportId() {
		return reportId;
	}

	public void setReportId(Integer reportId) {
		this.reportId = reportId;
	}
	
	@Override
	public int getLastId() {
		return lastId;
	}

	@Override
	public void setLastId(int lastId) {
		Filter.lastId = lastId;
	}	
}
