package com.perdemia.permissionanalyzer.data;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.text.MessageFormat;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.jface.dialogs.MessageDialog;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.MemberInfoDialog;
import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.model.StatusListener;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper;
import com.perdemia.permissionanalyzer.data.util.DomainCache;
import com.perdemia.permissionanalyzer.data.util.MemberCache;
import com.perdemia.permissionanalyzer.data.util.MembershipCache;
import com.perdemia.permissionanalyzer.scanner.acl.AceReader;
import com.perdemia.permissionanalyzer.scanner.groups.NestedGroupHelper;

/**
 * Represents a user or group.
 * 
 * @author Reinier
 * 
 */
public class Member extends DBObject {
	private final static Logger logger = LoggerFactory.getLogger(Member.class.getName());
	
	/** Cache map with domain/computer names and database id's. This map is used when the username is set and to determine the corresponding domain/computer for this user */
	private static final Map<String, Integer> domains = new HashMap<String, Integer>();
	
	/** Table attributes */
	private final static String[] tableAttributes = new String[] { "NAME", "NAME_LOWERCASE", "DISPLAY_NAME", "DISPLAY_NAME_LOWERCASE", "IS_GROUP", "GROUP_TYPE", "CN", "RDN_ID", "DOMAIN_ID", "IS_BUILTIN", "IS_ENABLED" };

	/** Table fields */
	private String name;
	private String displayName;
	private boolean group;
	private GroupType groupType = GroupType.NONE;
	private String cn;
	private Integer rdnId;
	private Integer domainId;
	private boolean builtIn;
	private boolean enabled = true;

	public static enum GroupType {
		NONE, COMPUTER_LOCAL, DOMAIN_LOCAL, GLOBAL, UNIVERSAL, DOMAIN_BUILTIN
	};

	/** Helper fields */
	private List<Integer> memberOfList;
	private Domain domain;
	private Rdn rdn;
	
	/** Helper fields for the {@link MemberInfoDialog} 
	 * @see {@link #getMembers(Member, String, boolean, String, int)
	 * @see {@link #getMemberOfList(Member, String, boolean, String, int)
	 * */
	private final static String VIA_GROUP_COLUMN = "VIA_GROUP_NAME";
	private String viaGroupName = "";
	
	/** statistics */
	private static int totalNumberOfUsersSaved;
	private static boolean totalNumberOfUsersReached;
	private static int totalNumberOfGroupsSaved;
	private static boolean totalNumberOfGroupsReached;
	private static int maxNumberOfGroups;
	private static int maxNumberOfUsers;
	private static Boolean userGroupCountAllowed;
	
	/** license limits. The maximum limits are stored as String so that they are obfuscated by Zelix Klassmaster  */
	private static Map<String, String> licenseLimits = new HashMap<String, String>();
	
	/** Effective permissions used in the Effective permissions tab */
	private EffectivePermissions effectivePermissions;
	
	/** List of applicable AclItem for this user, includes parent folders and group memberships. Used in the Effective permissions tab */
	private List<AclItem> aclItemList;
	
	/** List of members for a group, used in the Effective permissions tab to drill down in the Tree */
	private List<Member> members;
	
	/**
	 * The last ID that was used for an INSERT of this class in the database. If 0 is returned then the
	 * {@link DBObject#insert()} method will retrieve the max ID from the database.
	 */
	private static int lastId;
	
	static {
		licenseLimits.put("TRIAL_LIMIT_USERS", "2147483647");
		licenseLimits.put("TRIAL_LIMIT_GROUPS", "2147483647");
		licenseLimits.put("BASIC_LIMIT_USERS", "500");
		licenseLimits.put("BASIC_LIMIT_GROUPS", "100");
		licenseLimits.put("STANDARD_LIMIT_USERS", "3000");
		licenseLimits.put("STANDARD_LIMIT_GROUPS", "1000");
		licenseLimits.put("ADVANCED_LIMIT_USERS", "15000");
		licenseLimits.put("ADVANCED_LIMIT_GROUPS", "5000");
		licenseLimits.put("ENTERPRISE_LIMIT_USERS", "2147483647");
		licenseLimits.put("ENTERPRISE_LIMIT_GROUPS", "2147483647");
		licenseLimits.put("CONSULTANT_LIMIT_USERS", "2147483647");
		licenseLimits.put("CONSULTANT_LIMIT_GROUPS", "2147483647");
		licenseLimits.put("SCAN_AGENT_LIMIT_USERS", "2147483647");
		licenseLimits.put("SCAN_AGENT_LIMIT_GROUPS", "2147483647");
		clearStatistics();
	}
	
	/** Constructors */
	public Member() {}

	public Member(String name) {
		setName(name);
	}

	public Member(String name, Integer id) {
		this.id = id;
		setName(name);
	}
	
	/**
	 * Clear the number of users and groups that have been saved and the boolean flags that indicate if the limits have been reached.
	 */
	public static void clearStatistics() {
		totalNumberOfUsersSaved = 0;
		totalNumberOfUsersReached = false;
		totalNumberOfGroupsSaved = 0;
		totalNumberOfGroupsReached = false;
		maxNumberOfGroups = Integer.parseInt(licenseLimits.get(LicenseHelper.getEdition().toString() + "_LIMIT_GROUPS"));
		maxNumberOfUsers = Integer.parseInt(licenseLimits.get(LicenseHelper.getEdition().toString() + "_LIMIT_USERS"));
		userGroupCountAllowed = null;
	}

	// ========================================
	// GENERAL DB ACTIONS
	// ========================================

	public static Member getById(Integer id) throws DBException {
		return (Member) getById(Member.class, id);
	}

	@SuppressWarnings("unchecked")
	public static List<Member> where(String whereClause, Object[] parameters) throws DBException {
		return where(Member.class, whereClause, parameters, "name_lowercase", 0);
	}

	@SuppressWarnings("unchecked")
	public static List<Member> where(String whereClause, Object[] parameters, String orderBy) throws DBException {
		return where(Member.class, whereClause, parameters, orderBy, 0);
	}

	@SuppressWarnings("unchecked")
	public static List<Member> where(String whereClause, Object[] parameters, String orderBy, int limit) throws DBException {
		return where(Member.class, whereClause, parameters, orderBy, limit);
	}

	public static Member findFirst(String whereClause, Object... parameters) throws DBException {
		return (Member) findFirst(Member.class, whereClause, parameters);
	}

	public static int delete(Integer id) throws DBException {
		MemberCache.clearMember(id);
		return delete(Member.class, id);
	}

	public static int delete(String whereClause, Object... parameters) throws DBException {
		MemberCache.clearCache();
		MembershipCache.clearCache();
		return delete(Member.class, whereClause, parameters);
	}

	@Override
	protected String[] getTableAttributes() {
		return tableAttributes;
	}

	@Override
	protected void buildObject(ResultSet rs) throws SQLException {
		id = (rs.getInt("id"));
		name = rs.getString(tableAttributes[0]);
		// don't load name_lowercase in memory because it's only used for filtering in the database!
		displayName = rs.getString(tableAttributes[2]);
		
		// a report will crash of the displayName contains new lines. Prevent this.
		if (displayName != null) {
			displayName = displayName.replace("\r\n", " ").replace("\n", " ");
		}
		
		// don't load display_name_lowercase in memory because it's only used for filtering in the database!
		group = rs.getBoolean(tableAttributes[4]);
		groupType = GroupType.values()[rs.getInt(tableAttributes[5])];
		cn = rs.getString(tableAttributes[6]);
		rdnId = getNullableInteger(rs, tableAttributes[7]);
		domainId = getNullableInteger(rs, tableAttributes[8]);
		builtIn = rs.getBoolean(tableAttributes[9]);
		enabled = rs.getBoolean(tableAttributes[10]);
				
		// via group name available?
		try {
			int index = rs.findColumn(VIA_GROUP_COLUMN);
			viaGroupName = rs.getString(index);
		} catch (SQLException ex) {
			//no action required, column not in this query
		}
	}

	@Override
	protected void setQueryParameters(PreparedStatement ps) throws SQLException {
		ps.setString(1, name);
		ps.setString(2, name != null ? name.toLowerCase() : null);
		ps.setString(3, displayName);
		ps.setString(4, displayName != null ? displayName.toLowerCase() : null);
		ps.setBoolean(5, group);
		ps.setInt(6, groupType.ordinal());
		ps.setString(7, cn);
		setNullableInteger(ps, 8, rdnId);
		setNullableInteger(ps, 9, domainId);
		ps.setBoolean(10, builtIn);
		ps.setBoolean(11, enabled);
		
		if (id != null && id > 0) {
			MemberCache.clearMember(id);
		}
	}

	// ========================================
	// HELPER METHODS
	// ========================================

	/**
	 * Builds a display name based on the member name (username) and displayName
	 * 
	 * @param includeName
	 *            if <code>true</code> then the account name of the member is used including the displayName. Otherwise only the displayName is returned.
	 * @return display name
	 */
	public String getDisplayName(boolean includeName, boolean includeGroupDescription) {
		String result;
		String domain = "";
		if (includeName && getDomain() != null) {
			domain = getDomain().getDomain() + "\\";
		}
		
		if (includeName) {
			if ((displayName != null && displayName.length() > 0) &&
					(!group || includeGroupDescription)) {
				result = domain + name + " [" + displayName + "]";
			} else {
				result = domain + name;
			}
		} else {
			result = (displayName != null ? displayName : "");
		}
		
		if (group) {
			switch (groupType) {
			case COMPUTER_LOCAL:
				result += " [local group]"; break;
			case DOMAIN_LOCAL:
				result += " [domain local group]"; break;
			case GLOBAL:
				result += " [global group]"; break;
			case UNIVERSAL:
				result += " [universal group]"; break;
			case DOMAIN_BUILTIN:
				result += " [domain built-in group]"; break;
			default:
				result += " [group]"; break;
			}
		} 
		
		return result;
	}

	/**
	 * After a new <code>Member</code> is inserted in the database, a new record in the Membership table is created as well. This is because this table is joined when nested group membership is required in the overview. In that case we also want to
	 * include the user himself, that's why we create a record in the Membership table where the member_id and group_id are the same.
	 */
	@Override
	protected void insert() throws DBException {
		if (isGroup()) {			
			//license limit of maximum number of groups reached?
			if (totalNumberOfGroupsSaved >= maxNumberOfGroups) {
				if (!totalNumberOfGroupsReached) {
					logger.error(MessageFormat.format("The {0} edition of Permission Analyzer has a limit of {1} groups to scan. This limit has been reached.", LicenseHelper.getEdition().toString(), maxNumberOfGroups));
					totalNumberOfGroupsReached = true;
				}
				return;
			}
			totalNumberOfGroupsSaved++;
		} else {
			
			//license limit of maximum number of users reached?
			if (totalNumberOfUsersSaved >= maxNumberOfUsers) {
				if (!totalNumberOfUsersReached) {
					logger.error(MessageFormat.format("The {0} edition of Permission Analyzer has a limit of {1} users to scan. This limit has been reached.", LicenseHelper.getEdition().toString(), maxNumberOfUsers));
					totalNumberOfUsersReached = true;
				}
				return;
			}
			totalNumberOfUsersSaved++;
		}
		
		super.insert();
		Membership membership = new Membership(getId(), getId());
		membership.save();
	}
	
	@Override
	public int delete() throws DBException {
		if (isGroup()) {
			totalNumberOfGroupsSaved--;
		} else {
			totalNumberOfUsersSaved--;
		}
		return super.delete();
	}
	
	/**
	 * Checks if the number of users and group in the database does not exceed the maximum defined by the license. This may happen when the network is scanned with
	 * a ENTERPRISE license and other installation connects to the same database using a BASIC license.
	 * 
	 * @return true in case the number of users and groups in the database does <b>not</b> exceed the maximum defined by the license
	 */
	public static boolean isUserGroupCountAllowed(){
		if (userGroupCountAllowed == null) {
			try {
				final int groupCount = getCount(Member.class, "IS_GROUP = TRUE");
				final int userCount = getCount(Member.class, "IS_GROUP = FALSE OR IS_GROUP IS NULL");
				
				logger.debug("groupCount: " + groupCount);
				logger.debug("userCount: " + userCount);
				logger.debug("maxNumberOfGroups: " + maxNumberOfGroups);
				logger.debug("maxNumberOfUsers: " + maxNumberOfUsers);
				
				userGroupCountAllowed = (groupCount <= maxNumberOfGroups && userCount <= maxNumberOfUsers);
				
				if (!userGroupCountAllowed) {
					String msg = MessageFormat.format("The number of users and groups in the database exceeds the maximum number of users and groups that are allowed for your license. \r\n\r\n You have {0} users and {1} groups in the database and your license allows {2} users and {3} groups. The Report View is therefore disabled."
							, userCount, groupCount, maxNumberOfUsers, maxNumberOfGroups);
					SyncMessageDialog.showMessage(MessageDialog.ERROR, "License mismatch", msg);
					logger.error("License mismatch - " + msg);
				}
			} catch (DBException e) {
				logger.error("Error while getting user and group count", e);
				userGroupCountAllowed = false;
			}
		}
		
		return userGroupCountAllowed;
	}

	/**
	 * Filters on member name or display name by composing two separate queries and executing them with a UNION statement. This makes sure that both the index on Member.Name and Member.Display_name is used.
	 * 
	 * @param name
	 *            search string, may contain % as wildcard. The method will figure out whether to use LIKE or = in the query
	 * @param returnUsers
	 *            <code>false</code> to exclude users from the result list
	 * @param returnGroups
	 *            <code>false</code> to exclude groups from the result list
	 * @param domainId optional domain id to filter on. 0 or -1 means don't apply a domain filter.
	 * @return list with members that match the search criteria
	 * 
	 * @throws DBException
	 */
	@SuppressWarnings("unchecked")
	public static List<Member> findByName(String name, boolean returnUsers, boolean returnGroups, int domainId, String orderBy, int limit, String baseQuery, Object baseQueryParameter) throws DBException {
		StringBuilder nameQuery;
		StringBuilder displayNameQuery;
		String wherePrefix;
		String additionalWhereClause = "";
		boolean useTwoSeparateQueries = true;
		boolean filterOnName;
		
		// replace wild card * with %
		if (name != null) {
			name = name.replace('*', '%');
			
			// remove the double quotes for an exact search
			if (name.startsWith("\"")) {
				name = name.substring(1);
			}
			if (name.endsWith("\"")) {
				name = name.substring(0, name.length() - 1);
			}
		}
		
		if (baseQuery != null && baseQuery.length() > 0) {
			nameQuery = new StringBuilder(baseQuery);
			displayNameQuery = new StringBuilder(baseQuery);
			wherePrefix = " AND ";
		} else {
			nameQuery = new StringBuilder("select * from " + SCHEMA + ".MEMBER ");
			displayNameQuery = new StringBuilder("select * from " + SCHEMA + ".MEMBER ");
			wherePrefix = "";
		}
		
		filterOnName = name != null && name.length() > 0 && !name.equals("%"); 
		if (filterOnName) {
			if (wherePrefix.length() == 0) {
				nameQuery.append("where ");
				displayNameQuery.append("where ");
			}
			
			if (name.contains("%")) {
				if (name.startsWith("%")) {
					// starts with wild card, DB index will not be used in that case
					// in this case there is no point in using two separate queries on the index columns, we just use OR
					nameQuery.append(wherePrefix);
					nameQuery.append("(MEMBER.name_lowercase like ? OR MEMBER.display_name_lowercase like ?)");
					displayNameQuery = null; // don't use, because we can't use any index
					useTwoSeparateQueries = false;
				} else {
					// use two separate queries both on a indexed column using the LIKE operator
					nameQuery.append(wherePrefix);
					nameQuery.append("MEMBER.name_lowercase like ?");
					displayNameQuery.append(wherePrefix);
					displayNameQuery.append("MEMBER.display_name_lowercase like ?");
				}
			} else {
				// use two separate queries both on a indexed column using the equals operator
				nameQuery.append(wherePrefix);
				nameQuery.append("MEMBER.name_lowercase = ?");
				displayNameQuery.append(wherePrefix);
				displayNameQuery.append("MEMBER.display_name_lowercase = ?");
			}
		} else {
			// no filter on name, just use one query
			useTwoSeparateQueries = false;
		}
		
		if (returnUsers && !returnGroups) {
			// only show users
			additionalWhereClause += " and (MEMBER.is_group = false or MEMBER.is_group is null)";
			
		} else if (!returnUsers && returnGroups) {
			// only show groups
			additionalWhereClause += " and (MEMBER.is_group = true or MEMBER.is_group is null)";
		}
		
		if (domainId > 0) {
			// filter on domain
			additionalWhereClause += " and (MEMBER.domain_id = ?)";
		}
		
		// additional where clause found?
		if (additionalWhereClause.length() > 0 && nameQuery.indexOf(" where ") == -1) {
			nameQuery.append("where ");
			displayNameQuery.append("where ");
			
			// remove the first " and" from the additional where clause
			additionalWhereClause = additionalWhereClause.substring(4);
		}

		String finalQuery;
		if (useTwoSeparateQueries) {
			finalQuery = "(" + nameQuery.toString() + additionalWhereClause + ") UNION (" + displayNameQuery.toString() + additionalWhereClause + ")";
		} else {
			finalQuery = nameQuery.toString() + additionalWhereClause;
		}
		
		finalQuery = addLimitAndOrderBy(finalQuery, orderBy, null, limit);

		PreparedStatement ps = null;
		try {
			ps = getPreparedStatement(finalQuery);
			
			// apply the PreparedStatement parameters for the first query
			int index = 1;
			if (baseQueryParameter != null) {
				ps.setObject(index++, baseQueryParameter);
			}
			// filter on name?
			if (filterOnName) {
				ps.setString(index++, name.toLowerCase());
				
				if (!useTwoSeparateQueries) {
					// both name and display name are in the same query
					ps.setString(index++, name.toLowerCase());
				}
			}
			// filter on domain?
			if (domainId > 0) {
				ps.setInt(index++, domainId);
			}
			
			if (useTwoSeparateQueries) {
				// apply the same parameters for the second query
				if (baseQueryParameter != null) {
					ps.setObject(index++, baseQueryParameter);
				}
				// filter on name?
				if (filterOnName) {
					ps.setString(index++, name.toLowerCase());
				}
				// filter on domain?
				if (domainId > 0) {
					ps.setInt(index++, domainId);
				}						
			}
			
		} catch (SQLException ex) {
			logger.error("Error while getting members from database on name", ex);
			throw new DBException(ps != null ? ps.toString() : "null", null, ex);
		}

		return (List<Member>) getByPreparedStatement(Member.class, ps);
	}
	
	@SuppressWarnings("unchecked")
	public static List<Member> getByPreparedStatement(PreparedStatement ps) throws DBException {
		return (List<Member>) getByPreparedStatement(Member.class, ps);
	}
	
	/**
	 * Returns a query that can be used to loop through all the members. The query selects all members and order by name_lowercase.
	 * 
	 * @return query String
	 * @throws SQLException
	 * @throws DBException
	 * 
	 * @see {@link NestedGroupHelper#processNestedGroupMemberships(StatusListener)}
	 */
	public static String getAllQuery() throws DBException {
		return getAllQuery(Member.class, "NAME_LOWERCASE", 0);
	}
	
	/**
	 * Gets a list of members of the given group
	 * 
	 * @param group
	 *            the group to return the members of
	 * @param filterOnName
	 *            name filter, may include the wild character %
	 * @param includeNestedGroups
	 *            if <code>true</code> all members of all the groups in the given group are returned
	 * @param orderBy
	 *            column to sort on, like "name desc". Null means no order by.
	 * @param limit
	 *            limit on the results, 0 means no limit
	 * @return
	 * @throws DBException
	 */
	public static List<Member> getMembers(Member group, String filterOnName, boolean returnUsers, boolean returnGroups, boolean includeNestedGroups, int domainId, String orderBy, int limit) throws DBException {
		String baseQuery;
		
		if (includeNestedGroups) {
			baseQuery = "select member.*, via_group.name VIA_GROUP_NAME from permissionanalyzer.member inner join permissionanalyzer.membership on member.id = membership.member_id"
				+ " left outer join permissionanalyzer.member via_group on via_group.id = membership.via_group_id"
				+ " where membership.group_id = ? and membership.group_id <> membership.member_id";
		} else {
			baseQuery = "select member.* from permissionanalyzer.member inner join permissionanalyzer.membership on member.id = membership.member_id"
				+ " where membership.group_id = ? and membership.group_id <> membership.member_id and membership.via_group_id is null";
		}

		return findByName(filterOnName, returnUsers, returnGroups, domainId, orderBy, limit, baseQuery, group.getId());
	}

	/**
	 * Gets a list of groups where the given user/group is member of
	 * 
	 * @param group
	 *            user or group
	 * @param filterOnName
	 *            name filter, may include the wild character %
	 * @param includeNestedGroups
	 *            if <code>true</code> all members of all the groups in the given group are returned
	 * @param orderBy
	 *            column to sort on, like "name desc". Null means no order by.
	 * @param limit
	 *            limit on the results, 0 means no limit
	 * @return
	 * @throws DBException
	 */
	public static List<Member> getMemberOfList(Member member, String filterOnName, boolean includeNestedGroups, int domainId, String orderBy, int limit) throws DBException {
		String baseQuery = "select member.*, via_group.name VIA_GROUP_NAME from permissionanalyzer.member inner join permissionanalyzer.membership on member.id = membership.group_id"
				+ " left outer join permissionanalyzer.member via_group on via_group.id = membership.via_group_id"
				+ " where membership.member_id = ? and membership.group_id <> membership.member_id";

		if (!includeNestedGroups) {
			baseQuery += " AND membership.via_group_id is null";
		}
		return findByName(filterOnName, true, true, domainId, orderBy, limit, baseQuery, member.getId());
	}

	// ========================================
	// GETTERS AND SETTERS
	// ========================================

	public String getName(boolean includeDomain) {
		String domain = "";
		if (includeDomain && getDomain() != null) {
			domain = getDomain().getDomain() + "\\";
		}
		
		return domain + (name != null ? name : "");
	}

	public void setName(String name) {
		// upper case the computer/domain name to be uniform with other parts of the scan
		if (name.contains("\\")) {
			String accountName = name.substring(name.indexOf("\\") + 1);
			String domainName = name.substring(0, name.indexOf("\\")).toUpperCase();
			
			this.name = accountName;
			this.domainId = DomainCache.getDomainId(domainName);
			
		} else {
			this.name = name;
		}
	}
	
	/**
	 * Helper method that check if the given account is a BUILT-IN account, if so it replaces the "BUILT-IN" prefix 
	 * with the given server name
	 * 
	 * @param name
	 * @param sid
	 * @param domain
	 * @param server
	 */
	public void setName(String name, String sid, String domain, String server) {
		// we now check for BUILT-IN accounts. We convert those account names to <server>\\Administrators instead of BUILT-IN\Administrators.
		if (AceReader.isLocalBuiltinAccount(sid.toUpperCase())) {
										
			// don't use the account.domain because it says "BUILT-IN", use the server name instead
			// remove the BUILT-IN prefix and replace it with the machine name
			int firstBackslash = name.indexOf("\\");
			if (firstBackslash >= 0) {
				setName(server + name.substring(firstBackslash));
			} else {
				setName(name = server + "\\" + name);
			}
			
			setBuiltIn(true);
		}
		
		// no built-in group, or failed determining if it was a built-in group?
		if (getName(false) == null || getName(false).length() == 0) {
			if (domain != null) {
				setName(domain + "\\" + name);
			} else {
				setName(name);
			}
		}
	}

	public String getDisplayName() {
		return displayName;
	}
	
	public void setDisplayName(String displayName) {
		if (displayName != null) {
			displayName = displayName.replace("\r\n", " ").replace("\n", " ");
		}
		
		this.displayName = displayName;
	}

	public boolean isGroup() {
		return group;
	}

	public void setGroup(boolean group) {
		this.group = group;
	}

	public GroupType getGroupType() {
		return groupType;
	}

	public void setGroupType(GroupType groupType) {
		this.groupType = groupType;
	}

	public String getCn() {
		return cn;
	}

	public void setCn(String cn) {
		this.cn = cn;
	}

	public Integer getRdnId() {
		return rdnId;
	}

	public void setRdnId(Integer rdnId) {
		this.rdnId = rdnId;
	}

	public Integer getDomainId() {
		return domainId;
	}

	public void setDomainId(Integer domainId) {
		this.domainId = domainId;
	}

	public List<Integer> getMemberOfList() {
		return memberOfList;
	}

	public void setMemberOfList(List<Integer> memberOfList) {
		this.memberOfList = memberOfList;
	}

	public Domain getDomain() {
		if (domain == null) {
			try {
				domain = DomainCache.getDomain(domainId);
			} catch (DBException e) {
				logger.error("Error while determining domain for user " + name, e);
			}
		}
		return domain;
	}

	public void setDomain(Domain domain) {
		this.domain = domain;
	}

	public Rdn getRdn() {
		if (rdn == null) {
			try {
				rdn = Rdn.getById(rdnId);
			} catch (DBException e) {
				logger.error("Error while determining RDN for user " + name, e);
			}
		}
		
		return rdn;
	}

	public void setRdn(Rdn rdn) {
		this.rdn = rdn;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("Member [name=");
		builder.append(name);
		builder.append(", displayName=");
		builder.append(displayName);
		builder.append(", group=");
		builder.append(group);
		builder.append(", builtIn=");
		builder.append(builtIn);
		builder.append(", groupType=");
		builder.append(groupType);
		builder.append(", cn=");
		builder.append(cn);
		builder.append(", rdnId=");
		builder.append(rdnId);
		builder.append(", domainId=");
		builder.append(domainId);
		builder.append(", domain=");
		builder.append(domain);
		builder.append(", rdn=");
		builder.append(rdn);
		builder.append(", id=");
		builder.append(id);
		builder.append("]");
		return builder.toString();
	}

	public String getViaGroupName() {
		return viaGroupName;
	}

	public void setViaGroupName(String viaGroupName) {
		this.viaGroupName = viaGroupName;
	}

	public static int getTotalNumberOfUsersSaved() {
		return totalNumberOfUsersSaved;
	}

	public static boolean isTotalNumberOfUsersReached() {
		return totalNumberOfUsersReached;
	}

	public static int getTotalNumberOfGroupsSaved() {
		return totalNumberOfGroupsSaved;
	}

	public static boolean isTotalNumberOfGroupsReached() {
		return totalNumberOfGroupsReached;
	}

	public boolean isBuiltIn() {
		return builtIn;
	}

	public void setBuiltIn(boolean builtIn) {
		this.builtIn = builtIn;
	}

	public EffectivePermissions getEffectivePermissions() {
		return effectivePermissions;
	}

	public void setEffectivePermissions(EffectivePermissions effectivePermissions) {
		this.effectivePermissions = effectivePermissions;
	}

	public List<AclItem> getAclItemList() {
		return aclItemList;
	}

	public void setAclItemList(List<AclItem> aclItemList) {
		this.aclItemList = aclItemList;
	}

	public List<Member> getMembers() {
		return members;
	}

	public void setMembers(List<Member> members) {
		this.members = members;
	}
	
	@Override
	public int getLastId() {
		return lastId;
	}

	@Override
	public void setLastId(int lastId) {
		Member.lastId = lastId;
	}

	public boolean isEnabled() {
		return enabled;
	}

	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
	}	
}
