package com.perdemia.permissionanalyzer.data;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


public class PermissionLookup extends DBObject {
	private final static Logger logger = LoggerFactory.getLogger(PermissionLookup.class.getName());
	
	/** Table attributes */
	private final static String[] tableAttributes = new String[]{"P0","P1","P2","P3","P4","P5","P6","P7","P8","P9","P10","P11","P12"};
	
	/** Table fields */
	private int p0;
	private int p1;
	private int p2;
	private int p3;
	private int p4;
	private int p5;
	private int p6;
	private int p7;
	private int p8;
	private int p9;
	private int p10;
	private int p11;
	private int p12;
	
	/**
	 * The last ID that was used for an INSERT of this class in the database. If 0 is returned then the
	 * {@link DBObject#insert()} method will retrieve the max ID from the database.
	 */
	private static int lastId;
	
	/** The enum that defines all the simple permissions (combination of special permissions). */
	public static enum SimplePermission {
		FULL, MODIFY, WRITE, READ_EXECUTE, LIST_FOLDER_CONTENT, READ, SPECIAL
	}
	
	public static enum SpecialPermission
	{
	  READ_DATA(0x1), WRITE_DATA(0x2), APPEND_DATA(0x4), READ_NAMED_ATTRS(0x8), WRITE_NAMED_ATTRS(0x10), EXECUTE(0x20), DELETE_CHILD(0x40), READ_ATTRIBUTES(0x80), WRITE_ATTRIBUTES(0x100), DELETE(0x10000), READ_ACL(0x20000), WRITE_ACL(0x40000), WRITE_OWNER(0x80000), SYNCHRONIZE(0x100000);
	  
	  private int bitValue;
	  
	  private SpecialPermission(int bitValue) {
		  this.bitValue = bitValue;
	  }

		public int getBitValue() {
			return bitValue;
		}
	}
	
	/*
	 * READ_DATA = 0 WRITE_DATA = 1 APPEND_DATA = 2 READ_NAMED_ATTRS = 3 WRITE_NAMED_ATTRS = 4 EXECUTE = 5 DELETE_CHILD = 6 READ_ATTRIBUTES = 7 WRITE_ATTRIBUTES = 8 DELETE = 9 READ_ACL = 10 WRITE_ACL = 11 WRITE_OWNER = 12 SYNCHRONIZE = 13
	 */

	/**
	 * Mapping between special permissions and 'simple' permissions
	 */
	public static Map<SimplePermission, Integer[]> permissionMappings;
	public static Map<SimplePermission, Integer> simplePermissionValues;
	
	static {
		// synchronize is not included in the arrays because its not used
		permissionMappings = new EnumMap<SimplePermission, Integer[]>(SimplePermission.class);
		permissionMappings.put(SimplePermission.FULL, new Integer[] { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12}); // 11, 12 extra (write acl and owner)
		permissionMappings.put(SimplePermission.MODIFY, new Integer[] { 0, 1, 2, 3, 4, 5, 7, 8, 9, 10}); // 6, 9 extra (deletes)
		permissionMappings.put(SimplePermission.WRITE, new Integer[] { 1, 2, 4, 8});
		permissionMappings.put(SimplePermission.READ_EXECUTE, new Integer[] { 0, 3, 5, 7, 10}); // same as FOLDER_CONTENT, but with FILE_INHERIT flag
		permissionMappings.put(SimplePermission.LIST_FOLDER_CONTENT, new Integer[] { 0, 3, 5, 7, 10}); // 5 (execute)
		permissionMappings.put(SimplePermission.READ, new Integer[] { 0, 3, 7, 10});
		
		// calculate the simple permission values based on their mapping with special permissions
		simplePermissionValues = new HashMap<SimplePermission, Integer>();
		// skip SimplePermission.SPECIAL
		for (int counter = 0; counter < SimplePermission.values().length - 1; counter++) {
			Integer[] specialPermissions = permissionMappings.get(SimplePermission.values()[counter]);
			simplePermissionValues.put(SimplePermission.values()[counter], getIdBySpecialPermissions(Arrays.asList(specialPermissions)));
		}
	}
	
	//========================================
	// GENERAL DB ACTIONS
	//========================================
	
	public static PermissionLookup getById(Integer id) throws DBException {
		return (PermissionLookup) getById(PermissionLookup.class, id);
	}
		
	@SuppressWarnings("unchecked")
	public static List<PermissionLookup> where(String whereClause, Object[] parameters) throws DBException {
		return where(PermissionLookup.class, whereClause, parameters, null, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<PermissionLookup> where(String whereClause, Object[] parameters, String orderBy) throws DBException {
		return where(PermissionLookup.class, whereClause, parameters, orderBy, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<PermissionLookup> where(String whereClause, Object[] parameters, String orderBy, int limit) throws DBException {
		return where(PermissionLookup.class, whereClause, parameters, orderBy, limit);
	}
	
	public static PermissionLookup findFirst(String whereClause, Object... parameters) throws DBException {
		return (PermissionLookup) findFirst(PermissionLookup.class, whereClause, parameters);
	}
	
	public static int delete(Integer id) throws DBException {
		return delete(PermissionLookup.class, id);
	}
	
	public static int delete(String whereClause, Object...parameters) throws DBException {
		return delete(PermissionLookup.class, whereClause, parameters);
	}
	
	@Override
	protected String[] getTableAttributes(){
		return tableAttributes;
	}
	
	@Override
	protected void buildObject(ResultSet rs) throws SQLException {
		id = (rs.getInt("id"));
		p0 = rs.getInt(tableAttributes[0]);
		p1 = rs.getInt(tableAttributes[1]);
		p2 = rs.getInt(tableAttributes[2]);
		p3 = rs.getInt(tableAttributes[3]);
		p4 = rs.getInt(tableAttributes[4]);
		p5 = rs.getInt(tableAttributes[5]);
		p6 = rs.getInt(tableAttributes[6]);
		p7 = rs.getInt(tableAttributes[7]);
		p8 = rs.getInt(tableAttributes[8]);
		p9 = rs.getInt(tableAttributes[9]);
		p10 = rs.getInt(tableAttributes[10]);
		p11 = rs.getInt(tableAttributes[11]);
		p12 = rs.getInt(tableAttributes[12]);
	}
	
	@Override
	protected void setQueryParameters(PreparedStatement ps) throws SQLException {
		ps.setInt(1, p0);
		ps.setInt(2, p1);
		ps.setInt(3, p2);
		ps.setInt(4, p3);
		ps.setInt(5, p4);
		ps.setInt(6, p5);
		ps.setInt(7, p6);
		ps.setInt(8, p7);
		ps.setInt(9, p8);
		ps.setInt(10, p9);
		ps.setInt(11, p10);
		ps.setInt(12, p11);
		ps.setInt(13, p12);
	}
	
	//========================================
	// HELPER METHODS
	//========================================
	public static int getIdBySpecialPermissions(Collection<Integer> specialPermissions) {
		int result = 0;
		for (Integer specialPermission : specialPermissions) {
			//result += Math.pow(2, specialPermission + 1); //specialPermission is zero-based, but the bit isn't
			result += Math.pow(2, specialPermission);
		}
		return result;
	}
	
	public static boolean containsSpecialPermission(Integer permissionId, int specialPermission) {
		if (permissionId == null) {
			return false;
		}
				
		//int specialPermissionBit = (int) Math.pow(2, specialPermission + 1); //specialPermission is zero-based, but the bit isn't
		int specialPermissionBit = (int) Math.pow(2, specialPermission);
		return (permissionId & specialPermissionBit) == specialPermissionBit;
	}
	
	public static List<Integer> getSpecialPermissions(Integer permissionId) {
		List<Integer> specialPermissions = new ArrayList<Integer>();
		for (int counter = 0; counter < 13; counter++) {
			if (containsSpecialPermission(permissionId, counter)) {
				specialPermissions.add(counter);
			}
		}
		return specialPermissions;
	}
	
	public static boolean tableIsFilled() throws DBException {
		return getById(1) != null;
	}
	
	public static void main (String[] args) {
		System.out.println(getSpecialPermissions(1983));
		System.out.println("-------------");
		System.out.println(getSpecialPermissions(1193));
	}
	
	public static void fillTable() throws DBException {
		PreparedStatement ps = null;
		
		logger.info("Start filling permission lookup table");
		
		try {
			ps = getPreparedStatement("INSERT INTO " + SCHEMA + ".PERMISSIONLOOKUP VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?)");
			
			//int recordCount = (int) Math.pow(2, 14);
			int recordCount = (int) Math.pow(2, 13);
			
			for (int counter = 0; counter < recordCount; counter++) {
				ps.setInt(1, counter); // id
				
				//for (int specialPermission = 1; specialPermission <= 13; specialPermission++) { //specialPermission is zero-based, but the bit isn't
				for (int specialPermission = 0; specialPermission <= 12; specialPermission++) {
					//ps.setInt(specialPermission + 1, (specialPermission & counter) == specialPermission ? 1 : 0);
					ps.setInt(specialPermission + 2, containsSpecialPermission(counter, specialPermission) ? 1 : 0);
				}
				executePreparedStatement(ps, true);
			}
		} catch (SQLException ex) {
			logger.error("Error while filling permission lookup table: " + ps, ex);
			throw new DBException(ps != null ? ps.toString() : "null", null, ex);
		}
		
		logger.info("Permission lookup table filled");
	}
	
	//========================================
	// GETTERS AND SETTERS
	//========================================

	public int getP0() {
		return p0;
	}

	public void setP0(int p0) {
		this.p0 = p0;
	}

	public int getP1() {
		return p1;
	}

	public void setP1(int p1) {
		this.p1 = p1;
	}

	public int getP2() {
		return p2;
	}

	public void setP2(int p2) {
		this.p2 = p2;
	}

	public int getP3() {
		return p3;
	}

	public void setP3(int p3) {
		this.p3 = p3;
	}

	public int getP4() {
		return p4;
	}

	public void setP4(int p4) {
		this.p4 = p4;
	}

	public int getP5() {
		return p5;
	}

	public void setP5(int p5) {
		this.p5 = p5;
	}

	public int getP6() {
		return p6;
	}

	public void setP6(int p6) {
		this.p6 = p6;
	}

	public int getP7() {
		return p7;
	}

	public void setP7(int p7) {
		this.p7 = p7;
	}

	public int getP8() {
		return p8;
	}

	public void setP8(int p8) {
		this.p8 = p8;
	}

	public int getP9() {
		return p9;
	}

	public void setP9(int p9) {
		this.p9 = p9;
	}

	public int getP10() {
		return p10;
	}

	public void setP10(int p10) {
		this.p10 = p10;
	}

	public int getP11() {
		return p11;
	}

	public void setP11(int p11) {
		this.p11 = p11;
	}

	public int getP12() {
		return p12;
	}

	public void setP12(int p12) {
		this.p12 = p12;
	}
	
	@Override
	public int getLastId() {
		return lastId;
	}

	@Override
	public void setLastId(int lastId) {
		PermissionLookup.lastId = lastId;
	}	
}
