package com.perdemia.permissionanalyzer.data;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.FilterHelper.ObjectType;

public class Report extends DBObject {
	private final static Logger logger = LoggerFactory.getLogger(Report.class.getName());
	public enum ReportType {
		ACL, ACL_NESTED_GROUP_MEMBERS, EFFECTIVE, ALL_MEMBERS, MATCHING_GROUPS_AND_DIRECT_MEMBERS, MATCHING_GROUPS_AND_NESTED_MEMBERS, ACL_DIRECT_GROUP_MEMBERS, 
		GROUPS_WITH_PERMISSIONS_AND_DIRECT_MEMBERS, GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS, MATCHING_USERS_AND_DIRECT_GROUP_MEMBERSHIPS, MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS,
		FOLDERS_AND_USERS_EFFECTIVE
	}

	public enum FileType {
		HTML, HTML_JQUERY, CSV
	}
	
	/** Table attributes */
	private final static String[] tableAttributes = new String[]{"NAME", "DESCRIPTION", "LAST_RUN_DATE", "LAST_COUNT", "LAST_FILE_PATH", "REPORT_TYPE", "FILE_TYPE", "FILE_PATH", "TEMPLATE_PATH", "EMAIL_ADDRESS", "EMAIL_THRESHOLD", "LAST_STATUS_MESSAGE", "USE_SIMPLE_PRESENTATION", "CREATE_COMPARISON_REPORT"};
	
	/** Table fields */
	private String name = "";
	private String description= "";
	private String filePath= "";
	private String templatePath= "";
	private Date lastRunDate;
	private Integer lastCount;
	private String lastFilePath;
	private ReportType reportType = ReportType.ACL;
	private FileType fileType = FileType.HTML;
	private String emailAddress;
	private int emailThreshold;
	private String lastStatusMessage;
	private boolean useSimplePresentation;
	private boolean createComparisonReport;
	
	/** References to the reports filter sets */
	private FilterSet memberFilterSet;
	private FilterSet permissionFilterSet;
	private FilterSet folderFilterSet;
	private boolean filterSetsLoaded;
	
	/** Cached helper fields */
	private List<Filter> memberFilters;
	private List<Filter> permissionFilters;
	private List<Filter> folderFilters;
	
	/**
	 * The last ID that was used for an INSERT of this class in the database. If 0 is returned then the
	 * {@link DBObject#insert()} method will retrieve the max ID from the database.
	 */
	private static int lastId;
	
	//========================================
	// GENERAL DB ACTIONS
	//========================================
	
	@SuppressWarnings("unchecked")
	public static List<Report> getAll() throws DBException {
		return (List<Report>) getAll(Report.class, "NAME", 0);
	}
	
	public static Report getById(Integer id) throws DBException {
		return (Report) getById(Report.class, id);
	}
		
	@SuppressWarnings("unchecked")
	public static List<Report> where(String whereClause, Object[] parameters) throws DBException {
		return where(Report.class, whereClause, parameters, null, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<Report> where(String whereClause, Object[] parameters, String orderBy) throws DBException {
		return where(Report.class, whereClause, parameters, orderBy, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<Report> where(String whereClause, Object[] parameters, String orderBy, int limit) throws DBException {
		return where(Report.class, whereClause, parameters, orderBy, limit);
	}
	
	public static Report findFirst(String whereClause, Object... parameters) throws DBException {
		return (Report) findFirst(Report.class, whereClause, parameters);
	}
	
	public static int delete(Integer id) throws DBException {
		return delete(Report.class, id);
	}
	
	public static int delete(String whereClause, Object...parameters) throws DBException {
		return delete(Report.class, whereClause, parameters);
	}
	
	@Override
	protected String[] getTableAttributes(){
		return tableAttributes;
	}
	
	@Override
	protected void buildObject(ResultSet rs) throws SQLException {
		id = rs.getInt("id");
		name = rs.getString(tableAttributes[0]);
		description = rs.getString(tableAttributes[1]);
		lastRunDate = rs.getTimestamp(tableAttributes[2]);
		lastCount = rs.getInt(tableAttributes[3]);
		lastFilePath = rs.getString(tableAttributes[4]);
		reportType = ReportType.values()[rs.getInt(tableAttributes[5])];
		fileType = FileType.values()[rs.getInt(tableAttributes[6])];
		
		// we only have one HTML type now
		if (fileType.equals(FileType.HTML_JQUERY)) {
			fileType = FileType.HTML;
		}
					
		filePath = rs.getString(tableAttributes[7]);
		templatePath = rs.getString(tableAttributes[8]);
		emailAddress = rs.getString(tableAttributes[9]);
		emailThreshold = rs.getInt(tableAttributes[10]);
		lastStatusMessage = rs.getString(tableAttributes[11]);
		useSimplePresentation = rs.getBoolean(tableAttributes[12]);
		createComparisonReport = rs.getBoolean(tableAttributes[13]);		
	}
	
	@Override
	protected void setQueryParameters(PreparedStatement ps) throws SQLException {
		ps.setString(1, name);
		ps.setString(2, description);
		setNullableTimestamp(ps, 3, lastRunDate);
		setNullableInteger(ps, 4, lastCount);
		ps.setString(5, lastFilePath);
		ps.setInt(6, reportType != null ? reportType.ordinal() : null);
		ps.setInt(7, fileType != null ? fileType.ordinal() : null);
		ps.setString(8, filePath);
		ps.setString(9, templatePath);
		ps.setString(10, emailAddress);
		ps.setInt(11, emailThreshold);
		ps.setString(12, lastStatusMessage);
		ps.setBoolean(13, useSimplePresentation);
		ps.setBoolean(14, createComparisonReport);
	}
	
	//========================================
	// HELPER METHODS
	//========================================
	public static String reportTypeToString(ReportType reportType) {
		switch (reportType) {
			case EFFECTIVE: return "Folders/files and the sum of their permissions";
			case FOLDERS_AND_USERS_EFFECTIVE: return "Folders/files and the users with their effective permissions";
			case ACL: return "Folders/files and their Access Control List";
			case ACL_DIRECT_GROUP_MEMBERS: return "Folders/files and the ACL with expanded groups showing direct members and their effective permissions";
			case ACL_NESTED_GROUP_MEMBERS: return "Folders/files and the ACL with expanded groups showing nested members and their effective permissions";
			case ALL_MEMBERS: return "Users/groups and all their permissions";
			case MATCHING_GROUPS_AND_DIRECT_MEMBERS: return "Groups that match the filter criteria and their direct members";
			case MATCHING_GROUPS_AND_NESTED_MEMBERS: return "Groups that match the filter criteria and their nested members";
			case GROUPS_WITH_PERMISSIONS_AND_DIRECT_MEMBERS: return "Groups that have permissions in the folder tree and their direct members";
			case GROUPS_WITH_PERMISSIONS_AND_NESTED_MEMBERS: return "Groups that have permissions in the folder tree and their nested members";
			case MATCHING_USERS_AND_DIRECT_GROUP_MEMBERSHIPS: return "Users that match the filter criteria and their direct group memberships";
			case MATCHING_USERS_AND_NESTED_GROUP_MEMBERSHIPS: return "Users that match the filter criteria and their nested group memberships";
			
			default: return "";
		}
	}
	
	public static String fileTypeToString(FileType fileType) {
		switch (fileType) {
			case HTML: return "HTML";
			case HTML_JQUERY: return "HTML with interactive table";
			case CSV: return "CSV";
			default: return "";
		}
	}
	
	/**
	 * Loads the ReportFilterSet objects from the database that belong to this Report. For each ReportFilterSet the related FilterSet is loaded. Based on
	 * {@link FilterSet#getType()} the class fields memberFilterSet, permissionFilterSet and folderFilterSet are filled.
	 * @throws DBException
	 */
	public void loadFilterSets() throws DBException {
		if (!filterSetsLoaded) {
			List<ReportFilterSet> reportFilters = ReportFilterSet.where("report_id = ?", new Object[]{id});
			for (ReportFilterSet reportFilter : reportFilters) {
				FilterSet filterSet = FilterSet.getById(reportFilter.getFilterSetId());
				filterSet.loadFilters();
				switch (filterSet.getType()) {
					case MEMBER: memberFilterSet = filterSet; break;
					case PERMISSION: permissionFilterSet = filterSet; break;
					case FOLDER: folderFilterSet = filterSet; break;
				}
			}
		}
	}
	
	/**
	 * Checks if this report has a reference to a FilterSet of the type MEMBER. If so it returns the filters of that FilterSet, otherwise
	 * it returns all Filters with a report_id that corresponds to this report id.
	 * 
	 * @return list of member filters that belong to this report
	 * @throws DBException
	 */
	public List<Filter> getMemberFilters() throws DBException {
		if (memberFilters == null) {
			loadFilterSets();
			if (memberFilterSet != null) {
				memberFilters = memberFilterSet.getFilters();
			} else {
				memberFilters = Filter.getMemberFiltersForReport(id);
			}
		}
		
		return memberFilters;
	}
	
	/**
	 * Checks if this report has a reference to a FilterSet of the type PERMISSION. If so it returns the filters of that FilterSet, otherwise
	 * it returns all Filters with a report_id that corresponds to this report id.
	 * 
	 * @return list of permission filters that belong to this report
	 * @throws DBException
	 */
	public List<Filter> getPermissionFilters() throws DBException {
		if (permissionFilters == null) {
			loadFilterSets();
			if (permissionFilterSet != null) {
				permissionFilters = permissionFilterSet.getFilters();
			} else {
				permissionFilters =Filter.getPermissionFiltersForReport(id);
			}
		}
		
		return permissionFilters;
	}
	
	/**
	 * Checks if this report has a reference to a FilterSet of the type FOLDER. If so it returns the filters of that FilterSet, otherwise
	 * it returns all Filters with a report_id that corresponds to this report id.
	 * 
	 * @return list of folder filters that belong to this report
	 * @throws DBException
	 */
	public List<Filter> getFolderFilters() throws DBException {
		if (folderFilters == null) {		
			loadFilterSets();
			if (folderFilterSet != null) {
				folderFilters = folderFilterSet.getFilters();
			} else {
				folderFilters = Filter.getFolderFiltersForReport(id);
			}
		}
		
		return folderFilters;
	}
	
	/**
	 * Returns the members that are found in the folder tree of this report. This list corresponds with the "All members" tab when the report is loaded
	 * in the application.
	 * 
	 * @return a list of members that are found in the report folder tree
	 * @throws DBException 
	 * @see {@link AllMembersPartContentProvider}
	 */
	public List<Member> getMembersInFolderTree() throws DBException {
		// Create a new local filter helper without affecting the current UI
		FilterHelper filterHelper = new FilterHelper();
		filterHelper.setMemberFilterList(getMemberFilters());
		filterHelper.setPermissionFilterList(getPermissionFilters());
		filterHelper.setFolderFilterList(getFolderFilters());
		filterHelper.applyNewFilter(false);
		filterHelper.setSelectedReport(this);
		return Member.getByPreparedStatement(filterHelper.buildFilterQuery(ObjectType.MEMBERS_IN_FILE_TREE, false, false, 0, null, null));
	}
	
	//========================================
	// GETTERS AND SETTERS
	//========================================

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public String getDescription() {
		return description;
	}

	public void setDescription(String description) {
		this.description = description;
	}

	public String getFilePath() {
		return filePath;
	}

	public void setFilePath(String filePath) {
		this.filePath = filePath;
	}

	public Date getLastRunDate() {
		return lastRunDate;
	}

	public void setLastRunDate(Date lastRunDate) {
		this.lastRunDate = lastRunDate;
	}

	public ReportType getReportType() {
		return reportType;
	}

	public void setReportType(ReportType reportType) {
		this.reportType = reportType;
	}

	public FileType getFileType() {
		return fileType;
	}

	public void setFileType(FileType fileType) {
		this.fileType = fileType;
	}

	public FilterSet getMemberFilterSet() {
		return memberFilterSet;
	}

	public FilterSet getPermissionFilterSet() {
		return permissionFilterSet;
	}

	public FilterSet getFolderFilterSet() {
		return folderFilterSet;
	}

	public String getTemplatePath() {
		return templatePath;
	}

	public void setTemplatePath(String templatePath) {
		this.templatePath = templatePath;
	}

	public void setMemberFilters(List<Filter> memberFilters) {
		this.memberFilters = memberFilters;
	}

	public void setPermissionFilters(List<Filter> permissionFilters) {
		this.permissionFilters = permissionFilters;
	}

	public void setFolderFilters(List<Filter> folderFilters) {
		this.folderFilters = folderFilters;
	}

	public String getEmailAddress() {
		return emailAddress;
	}

	public void setEmailAddress(String emailAddress) {
		this.emailAddress = emailAddress;
	}

	public int getEmailThreshold() {
		return emailThreshold;
	}

	public void setEmailThreshold(int emailThreshold) {
		this.emailThreshold = emailThreshold;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("Report [name=");
		builder.append(name);
		builder.append(", id=");
		builder.append(id);
		builder.append(", description=");
		builder.append(description);
		builder.append(", filePath=");
		builder.append(filePath);
		builder.append(", templatePath=");
		builder.append(templatePath);
		builder.append(", lastRunDate=");
		builder.append(lastRunDate);
		builder.append(", lastFileCount=");
		builder.append(lastCount);
		builder.append(", reportType=");
		builder.append(reportType);
		builder.append(", fileType=");
		builder.append(fileType);
		builder.append(", emailAddress=");
		builder.append(emailAddress);
		builder.append(", emailThreshold=");
		builder.append(emailThreshold);
		builder.append(", useSimplePresentation=");
		builder.append(useSimplePresentation);
		builder.append(", createComparisonReport=");
		builder.append(createComparisonReport);		
		builder.append("]");
		return builder.toString();
	}

	public Integer getLastCount() {
		return lastCount;
	}

	public void setLastCount(Integer lastFileCount) {
		this.lastCount = lastFileCount;
	}

	public String getLastFilePath() {
		return lastFilePath;
	}

	public void setLastFilePath(String lastFilePath) {
		this.lastFilePath = lastFilePath;
	}
	
	public boolean isPolicy() {
		return emailThreshold > 0;
	}

	public String getLastStatusMessage() {
		return lastStatusMessage;
	}

	public void setLastStatusMessage(String lastStatusMessage) {
		this.lastStatusMessage = lastStatusMessage;
	}
	
	@Override
	public int getLastId() {
		return lastId;
	}

	@Override
	public void setLastId(int lastId) {
		Report.lastId = lastId;
	}

	public boolean isUseSimplePresentation() {
		return useSimplePresentation;
	}

	public void setUseSimplePresentation(boolean useSimplePresentation) {
		this.useSimplePresentation = useSimplePresentation;
	}

	public boolean isCreateComparisonReport() {
		return createComparisonReport;
	}

	public void setCreateComparisonReport(boolean createComparisonReport) {
		this.createComparisonReport = createComparisonReport;
	}	
}
