package com.perdemia.permissionanalyzer.data;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class ReportHistory extends DBObject {
	private final static Logger logger = LoggerFactory.getLogger(ReportHistory.class.getName());
		
	/** Table attributes */
	private final static String[] tableAttributes = new String[]{"REPORT_ID", "RUN_DATE", "RESULT_COUNT", "FILE_PATH", "REPORT_TYPE", "FILE_TYPE", "EMAIL_ADDRESS", "THRESHOLD", "STATUS_MESSAGE", "WAS_COMPARISON"};
	
	/** Table fields */
	private Integer reportId;
	private String filePath= "";
	private Date runDate;
	private Integer count;
		
	private Report.ReportType reportType;
	private Report.FileType fileType;
	private String emailAddress;
	private int threshold;
	private String statusMessage;
	private boolean wasComparison;
	
	/**
	 * The last ID that was used for an INSERT of this class in the database. If 0 is returned then the
	 * {@link DBObject#insert()} method will retrieve the max ID from the database.
	 */
	private static int lastId;
	
	/**
	 * Default constructor
	 */
	public ReportHistory(){};
	
	/**
	 * Constructor using a Report object as basis
	 * @param report
	 */
	public ReportHistory(Report report, String statusMessage, String filePath, Date runDate, int count, boolean wasComparison) {
		this.reportId = report.getId();
		this.statusMessage = statusMessage;
		this.filePath = filePath;
		this.runDate = runDate;
		this.count = count;
		this.reportType = report.getReportType();
		this.fileType = report.getFileType();
		this.emailAddress = report.getEmailAddress();
		this.threshold = report.getEmailThreshold();
		this.wasComparison = wasComparison;
	}
	
	/**
	 * Constructor using a Report object as basis
	 * @param report
	 */
	public ReportHistory(Report report) {
		this(report, report.getLastStatusMessage(), report.getLastFilePath(), report.getLastRunDate(), report.getLastCount(), false);
	}
	
	//========================================
	// GENERAL DB ACTIONS
	//========================================
	
	@SuppressWarnings("unchecked")
	public static List<ReportHistory> getAll() throws DBException {
		return (List<ReportHistory>) getAll(ReportHistory.class, "RUN_DATE DESC, REPORT_ID", 0);
	}
	
	public static ReportHistory getById(Integer id) throws DBException {
		return (ReportHistory) getById(ReportHistory.class, id);
	}
		
	@SuppressWarnings("unchecked")
	public static List<ReportHistory> where(String whereClause, Object[] parameters) throws DBException {
		return where(ReportHistory.class, whereClause, parameters, null, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<ReportHistory> where(String whereClause, Object[] parameters, String orderBy) throws DBException {
		return where(ReportHistory.class, whereClause, parameters, orderBy, 0);
	}
	
	@SuppressWarnings("unchecked")
	public static List<ReportHistory> where(String whereClause, Object[] parameters, String orderBy, int limit) throws DBException {
		return where(ReportHistory.class, whereClause, parameters, orderBy, limit);
	}
	
	public static ReportHistory findFirst(String whereClause, Object... parameters) throws DBException {
		return (ReportHistory) findFirst(ReportHistory.class, whereClause, parameters);
	}
	
	public static int delete(Integer id) throws DBException {
		return delete(ReportHistory.class, id);
	}
	
	public static int delete(String whereClause, Object...parameters) throws DBException {
		return delete(ReportHistory.class, whereClause, parameters);
	}
	
	@Override
	protected String[] getTableAttributes(){
		return tableAttributes;
	}
	
	@Override
	protected void buildObject(ResultSet rs) throws SQLException {
		id = rs.getInt("id");
		reportId = rs.getInt(tableAttributes[0]);
		runDate = rs.getTimestamp(tableAttributes[1]);
		count = rs.getInt(tableAttributes[2]);
		filePath = rs.getString(tableAttributes[3]);
		
		reportType = Report.ReportType.values()[rs.getInt(tableAttributes[4])];
		fileType = Report.FileType.values()[rs.getInt(tableAttributes[5])];
		
		// we only have one HTML type now
		if (fileType.equals(Report.FileType.HTML_JQUERY)) {
			fileType = Report.FileType.HTML;
		}
					
		emailAddress = rs.getString(tableAttributes[6]);
		threshold = rs.getInt(tableAttributes[7]);
		statusMessage = rs.getString(tableAttributes[8]);
		wasComparison = rs.getBoolean(tableAttributes[9]);
	}
	
	@Override
	protected void setQueryParameters(PreparedStatement ps) throws SQLException {
		ps.setInt(1, reportId);
		setNullableTimestamp(ps, 2, runDate);
		setNullableInteger(ps, 3, count);
		ps.setString(4, filePath);
		ps.setInt(5, reportType != null ? reportType.ordinal() : null);
		ps.setInt(6, fileType != null ? fileType.ordinal() : null);
		ps.setString(7, emailAddress);
		ps.setInt(8, threshold);
		ps.setString(9, statusMessage);
		ps.setBoolean(10, wasComparison);
	}
	
	//========================================
	// GETTERS AND SETTERS
	//========================================

	public String getFilePath() {
		return filePath;
	}

	public void setFilePath(String filePath) {
		this.filePath = filePath;
	}

	public Date getRunDate() {
		return runDate;
	}

	public void setRunDate(Date lastRunDate) {
		this.runDate = lastRunDate;
	}

	public Report.ReportType getReportType() {
		return reportType;
	}

	public void setReportType(Report.ReportType reportType) {
		this.reportType = reportType;
	}

	public Report.FileType getFileType() {
		return fileType;
	}

	public void setFileType(Report.FileType fileType) {
		this.fileType = fileType;
	}

	public String getEmailAddress() {
		return emailAddress;
	}

	public void setEmailAddress(String emailAddress) {
		this.emailAddress = emailAddress;
	}

	public int getThreshold() {
		return threshold;
	}

	public void setThreshold(int emailThreshold) {
		this.threshold = emailThreshold;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("ReportHistory [reportId=");
		builder.append(reportId);
		builder.append(", id=");
		builder.append(id);
		builder.append(", filePath=");
		builder.append(filePath);
		builder.append(", runDate=");
		builder.append(runDate);
		builder.append(", count=");
		builder.append(count);
		builder.append(", reportType=");
		builder.append(reportType);
		builder.append(", fileType=");
		builder.append(fileType);
		builder.append(", emailAddress=");
		builder.append(emailAddress);
		builder.append(", threshold=");
		builder.append(threshold);
		builder.append(", wasComparison=");
		builder.append(wasComparison);		
		builder.append("]");
		return builder.toString();
	}

	public Integer getCount() {
		return count;
	}

	public void setCount(Integer lastFileCount) {
		this.count = lastFileCount;
	}
	
	public boolean isPolicy() {
		return threshold > 0;
	}

	public String getStatusMessage() {
		return statusMessage;
	}

	public void setStatusMessage(String lastStatusMessage) {
		this.statusMessage = lastStatusMessage;
	}
	
	@Override
	public int getLastId() {
		return lastId;
	}

	@Override
	public void setLastId(int lastId) {
		ReportHistory.lastId = lastId;
	}

	public Integer getReportId() {
		return reportId;
	}

	public void setReportId(Integer reportId) {
		this.reportId = reportId;
	}

	public boolean isWasComparison() {
		return wasComparison;
	}

	public void setWasComparison(boolean wasComparison) {
		this.wasComparison = wasComparison;
	}	
}
