package com.perdemia.permissionanalyzer.data.util;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.security.NoSuchAlgorithmException;
import java.security.spec.InvalidKeySpecException;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.MessageFormat;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.swt.widgets.Display;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.logging.GUILogHandler;
import com.perdemia.permissionanalyzer.application.model.StatusListener;
import com.perdemia.permissionanalyzer.application.preferences.AESEncrypter;
import com.perdemia.permissionanalyzer.application.preferences.PreferenceHelper;
import com.perdemia.permissionanalyzer.application.util.ApplicationUtil;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper.Edition;
import com.perdemia.permissionanalyzer.data.AclItem;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.Domain;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Filter;
import com.perdemia.permissionanalyzer.data.FilterSet;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.Membership;
import com.perdemia.permissionanalyzer.data.PermissionLookup;
import com.perdemia.permissionanalyzer.data.Rdn;
import com.perdemia.permissionanalyzer.data.Report;
import com.perdemia.permissionanalyzer.data.ReportFilterSet;
import com.perdemia.permissionanalyzer.data.ScanLogItem;
import com.perdemia.permissionanalyzer.data.ScanLogItem.Severity;

public class DBUtil {
	private static Logger logger = LoggerFactory.getLogger(DBUtil.class.getName());
	//TODO disable database TRACE level!!!
	private static boolean TRACE_ENABLED = false;
	private static Connection dbConnection;
	private static boolean connectionFailed;
	private static DbType dbType = null;
	private static int dbSchemaVersion = 0;
	
	public enum DbType {
		UNKNOWN, DB2, DERBY, ORACLE, MYSQL, POSTGRESQL, H2, MSSQL
	}
	
	static {
		DriverManager.setLoginTimeout(5);
	}
	
	public static Connection getDBConnection() throws DBException {
		return getDBConnection(true);
	}
		
	public static Connection getDBConnection(boolean useExistingConnection) throws DBException {
		if (connectionFailed) {
			throw new DBException("Database connection failed");
		}
		
		try {
			if (dbConnection == null || dbConnection.isClosed() || !useExistingConnection) {
				logger.info("Opening database connection");
				
				PreferenceHelper prefHelper = PreferenceHelper.getInstance();
			
				logger.info("Creating database connection: ");
				String databasePath = ApplicationUtil.getApplicationPath(); // + "H2DB";
				String connectionString = prefHelper.getDbConnectionString(); // "jdbc:h2:" + databasePath + ";MAX_COMPACT_TIME=2000;ALIAS_COLUMN_NAME=TRUE;CACHE_SIZE=524288;"; //512 MB
				connectionString = connectionString.replace("[app_dir]", databasePath);
				String password = prefHelper.getDefaultDatabasePassword();
				
				logger.info("Using database path: " + databasePath);
				logger.info("Using connection string: " + connectionString);
				
				// license check
				if (!isLocalDB() && !LicenseHelper.getEdition().equals(Edition.ENTERPRISE) && !LicenseHelper.getEdition().equals(Edition.CONSULTANT) 
						&& !LicenseHelper.getEdition().equals(Edition.ADVANCED) && !LicenseHelper.getEdition().equals(Edition.SCAN_AGENT)) {
					String message = "External database support is only available in the Advanced, Enterprise and Scan Agent edition.";
					logger.error(message);
					SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", message);
					connectionFailed = true;
					throw new DBException(message);
				}
				
				if (TRACE_ENABLED && isLocalDB()){
					connectionString += ";TRACE_LEVEL_FILE=3";
				}
				
				// is the embedded database encrypted?
				if (isLocalDB() && prefHelper.isEncryptDatabase()) {
					
					// license check
					if (LicenseHelper.getEdition().equals(Edition.TRIAL) || LicenseHelper.getEdition().equals(Edition.BASIC)) {
						String message = "Database encryption is not available in the Trial or Basic edition.";
						logger.error(message);
						SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", message);
						connectionFailed = true;
						throw new DBException(message);
					}
					
					connectionString += ";CIPHER=AES";
					//encryption password is added to the password in the URL <filepw> <userpw>
					password = AESEncrypter.hashPasswordToString(prefHelper.getMasterPassword()) + " " + password;					
				}
				
				// Register JDBC driver
				Class.forName(prefHelper.getDbConnectionDriver());
				
				// Open a connection
				if (useExistingConnection) {
					if (isLocalDB()) {
						dbConnection = DriverManager.getConnection(connectionString, "permissionanalyzer", password);
					} else {
						dbConnection = DriverManager.getConnection(connectionString, prefHelper.getDbConnectionUsername(), prefHelper.getDbConnectionPassword());
					}
				} else {
					// create a new connection
					Connection newDbConnection = DriverManager.getConnection(connectionString, prefHelper.getDbConnectionUsername(), prefHelper.getDbConnectionPassword());
					return newDbConnection;
				}
				
				// DB2 requires fully qualified table names (including schema) in the WHERE clause,
				// to prevent this, use the statement below
				if (getDbType().equals(DbType.DB2) || getDbType().equals(DbType.DERBY)) {
					DBObject.executeStatement("SET SCHEMA PERMISSIONANALYZER");
				}
				
				// check if the database exists
				try {
					PermissionLookup.tableIsFilled(); // test any DB table
				} catch (DBException ex) {
					if (ex.getMessage().contains("org.h2.jdbc.JdbcSQLException") && ex.getMessage().contains("not found")) {
						// create the database schema
						try {
							logger.info("Creating database schema");
							
							String createSchemaSQL = ApplicationUtil.getFileContent("sql/create_db_h2.sql");
							
							// create the schema in a separate call
							DBObject.executeStatement("CREATE SCHEMA PERMISSIONANALYZER");
							
							createSchemaSQL = createSchemaSQL.substring(createSchemaSQL.indexOf("SET SCHEMA "));
							DBObject.executeStatement(createSchemaSQL);
						} catch (Exception ex2) {
							logger.error("Error while creating database schema", ex2);
							SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while creating database schema: " + ex2.getMessage());
						}
					} else {
						logger.error("The database schema PERMISSIONANALYZER is not found", ex);
						SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "The database schema PERMISSIONANALYZER is not found: " + ex.getMessage());
					}
				}
				
				// check if the permission lookup table has been filled
				try {
					if (!PermissionLookup.tableIsFilled()) {
						logger.info("Filling permission lookup table");
						PermissionLookup.fillTable();
					}
				} catch (DBException ex) {
					logger.error("Error while filling the permission lookup table", ex);
					SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while filling the permission lookup table: " + ex.getMessage());
				}
				
			}
		} catch (SQLException | ClassNotFoundException ex) {
			String message = "Could not open database:\r\n\r\n" + getNestedMessage(ex);
			if (message.toLowerCase().contains("store header")) {
				message += "\r\n\r\nThis probably means that the database has been encrypted and that the given password to decrypt is not correct.";
			}
			logger.error("Could not open database: " + message, ex);
			connectionFailed = true;
			throw new DBException(message, ex);
						
		} catch (NoSuchAlgorithmException | InvalidKeySpecException ex) {
			logger.error("Error while creating encryption key for database", ex);
			connectionFailed = true;
			throw new DBException("Error while creating encryption key for database", ex);
		}
		
		return dbConnection;
	}
	
	public static void closeDBConnection(){
		if (dbConnection != null){
			try {
				if (!dbConnection.isClosed()) {
					DBObject.clearPreparedStatementCache();
					FilterHelper.getInstance().applyNewFilter(false);
					dbConnection.close();
					Member.clearStatistics();
				}
			} catch (SQLException e) {
				logger.error("Exception while closing database connection", e);
			}
		}
		connectionFailed = false;
		clearDbType();
	}
	
	public static void testDbConnection(int dbType, String connectionString, String driver, String username, String password, boolean encryptDatabase) throws Exception {		
		logger.info("Testing database connection");
										
		String databasePath = Platform.getLocation().toFile().getParent() + File.separator;
		connectionString = connectionString.replace("[app_dir]", databasePath);
		
		if (password == null || password.length() == 0) {
			password = PreferenceHelper.getInstance().getDefaultDatabasePassword();
		}
		
		logger.info("Using connection string: " + connectionString);
					
		// is the embedded database encrypted?
		if (isLocalDB(dbType) && encryptDatabase) {
			connectionString += ";CIPHER=AES";
			//encryption password is added to the password in the URL <filepw> <userpw>
			password = AESEncrypter.hashPasswordToString(PreferenceHelper.getInstance().getMasterPassword()) + " " + password;					
		}
		
		// Register JDBC driver
		Class.forName(driver);
		
		// Open a connection
		Connection connection;
		if (isLocalDB(dbType)) {
			connection = DriverManager.getConnection(connectionString, "permissionanalyzer", password);
		} else {
			connection = DriverManager.getConnection(connectionString, username, password);
		}
		
		if (connection == null) {
			throw new Exception("Empty connection returned by the DriverManager.");
		}
	}
	
	public static void main(String[] arg) {
		try {
			Class.forName("org.h2.Driver");
			Connection dbConnection = DriverManager.getConnection("jdbc:h2://C:\\H2DB", "user", "password");
			Statement statement = dbConnection.createStatement();
			//statement.execute("CREATE TABLE TEST (ID INT NOT NULL GENERATED ALWAYS AS IDENTITY CONSTRAINT TEST_PK PRIMARY KEY, NAME VARCHAR(255) NOT NULL)");
			
//			for (int counter=0; counter<1000000; counter++) {
//				statement.execute("INSERT INTO TEST (NAME) VALUES ('testvaluetogeneratesometestdata" + counter + "')");
//			}
			
			statement.execute("SHUTDOWN DEFRAG");
			System.out.println("Closed: " + dbConnection.isClosed()); //true
			dbConnection.close();
			dbConnection = DriverManager.getConnection("jdbc:h2://C:\\H2DB", "user", "password");
			//dbConnection = DriverManager.getConnection("jdbc:h2://C:\\Docs\\Perdemia\\H2DB;CACHE_SIZE=786432;QUERY_CACHE_SIZE=0;MV_STORE=FALSE;MVCC=FALSE;", "permissionanalyzer", "permissionanalyzer");
			System.out.println("Done");
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public static String getNestedMessage(Exception ex){
		Throwable nestedException = ex;
		if (ex.getCause() != null) {
			for (int counter = 0; counter < 5; counter++) {
				nestedException = nestedException.getCause();
				if (nestedException.getCause() == null) {
					break;
				}
			}
			return nestedException.getMessage();
		} else {
			return ex.getMessage();
		}
	}
	
	/**
	 * This method retrieves the version number of the database table and applies the correct changes.
	 * 
	 * @return
	 */
	public static void checkDBChanges() {
		ResultSet rs = null;
		
		try {
			rs = DBObject.executeQuery("SELECT VERSION FROM PERMISSIONANALYZER.INFO", new Object[0]);
			if (rs.next()) {
				dbSchemaVersion = rs.getInt(1);
				
				logger.info("Database schema version: " + dbSchemaVersion);
				
				if (dbSchemaVersion < 19) {
					
					ProgressMonitorDialog progressDialog = new ProgressMonitorDialog(Display.getCurrent().getActiveShell());
					try {
						progressDialog.run(false, false, new IRunnableWithProgress(){

							@Override
							public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
								monitor.beginTask("Upgrading database schema.....", IProgressMonitor.UNKNOWN);
								
								try {
									if (dbSchemaVersion < 2) {
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.FILE RENAME TO FILEINFO");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM ALTER COLUMN FILE_ID RENAME TO FILEINFO_ID");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM ALTER COLUMN DENY RENAME TO IS_DENY");
										
										// Foreign keys have been renamed because Oracle has a name limit of 30 characters
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.FILEINFO DROP CONSTRAINT FILE_PARENT_ID_FILE_ID_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM DROP CONSTRAINT ACLITEM_MEMBER_ID_MEMBER_ID_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM DROP CONSTRAINT ACLITEM_FILE_ID_FILE_ID_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM DROP CONSTRAINT ACLITEM_INHERITED_FROM_FOLDER_ID_FILE_ID_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM DROP CONSTRAINT ACLITEM_PERMISSION_ID_PERMISSIONLOOKUP_ID_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBERSHIP DROP CONSTRAINT MEMBERSHIP_GROUP_ID_MEMBER_ID_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBERSHIP DROP CONSTRAINT MEMBERSHIP_MEMBER_ID_MEMBER_ID_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER DROP CONSTRAINT MEMBER_RDN_ID_RDN_ID_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER DROP CONSTRAINT MEMBER_DOMAIN_ID_DOMAIN_ID_FK");
										
										// re-create the foreign keys
										addForeignKey(FileInfo.class, FileInfo.class, "PARENT_ID", "PARENT_ID", true);
										addForeignKey(AclItem.class, Member.class);
										addForeignKey(AclItem.class, FileInfo.class);
										addForeignKey(AclItem.class, FileInfo.class, "INHERITED_FROM_FOLDER_ID", "INHERITED_FROM", false);
										addForeignKey(AclItem.class, PermissionLookup.class, "PERMISSION_ID", "PERMISSION_ID", false);
										addForeignKey(Membership.class, Member.class, "GROUP_ID", "GROUP_ID", false);
										addForeignKey(Membership.class, Member.class);
										addForeignKey(Member.class, Rdn.class);
										addForeignKey(Member.class, Domain.class);
										
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 2");					
										logger.info("Database upgraded to version 2.");
									}
									
									if (dbSchemaVersion < 3) {
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.FILEINFO ADD IS_DACL_PROTECTED BOOLEAN");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM ADD IS_DIRECTORY BOOLEAN");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD EMAIL_ADDRESS VARCHAR(255)");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD EMAIL_THRESHOLD INT");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 3");					
										logger.info("Database upgraded to version 3.");
									}
									
									if (dbSchemaVersion < 4) {
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER ALTER COLUMN DISPLAY_NAME varchar(1024)");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER ALTER COLUMN DISPLAY_NAME_LOWERCASE varchar(1024)");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 4");					
										logger.info("Database upgraded to version 4.");
									}
									
									if (dbSchemaVersion < 5) {
										DBObject.executeStatement("CREATE INDEX MEMBER_TYPE_I ON PERMISSIONANALYZER.MEMBER(GROUP_TYPE)");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 5");					
										logger.info("Database upgraded to version 5.");
									}
									
									if (dbSchemaVersion < 6) {
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER ADD IS_BUILTIN BOOLEAN");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 6");					
										logger.info("Database upgraded to version 6.");
									}
									
									if (dbSchemaVersion < 7) {
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.RDN ALTER COLUMN RDN varchar(1000)");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 7");					
										logger.info("Database upgraded to version 7.");
									}
									
									if (dbSchemaVersion < 8) {
										if (getDbType().equals(DbType.MSSQL)) {
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.SCANLOGITEM (ID INT IDENTITY(1,1) PRIMARY KEY, SERVER VARCHAR(255) NOT NULL, LOG_DATE DATETIME NOT NULL, Severity INT NOT NULL, MESSAGE VARCHAR(2000) NOT NULL,	DETAILS VARCHAR(4000))");
											DBObject.executeStatement("CREATE INDEX SCANLOGITEM_SERVER_I ON PERMISSIONANALYZER.SCANLOGITEM(SERVER)");					
											DBObject.executeStatement("CREATE INDEX SCANLOGITEM_LOG_DATE_I ON PERMISSIONANALYZER.SCANLOGITEM(LOG_DATE)");
										} else { 
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.SCANLOGITEM (ID INT NOT NULL GENERATED ALWAYS AS IDENTITY CONSTRAINT SCANLOGITEM_PK PRIMARY KEY, SERVER VARCHAR(255) NOT NULL,	LOG_DATE TIMESTAMP NOT NULL, Severity INT NOT NULL,	MESSAGE VARCHAR(2000) NOT NULL, DETAILS VARCHAR(4000))");
											DBObject.executeStatement("CREATE INDEX SCANLOGITEM_SERVER_I ON PERMISSIONANALYZER.SCANLOGITEM(SERVER)");					
											DBObject.executeStatement("CREATE INDEX SCANLOGITEM_LOG_DATE_I ON PERMISSIONANALYZER.SCANLOGITEM(LOG_DATE)");
										}
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 8");
										
										logger.info("Database upgraded to version 8.");
									}
									
									if (dbSchemaVersion < 9) {
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD LAST_COUNT INT");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD LAST_FILE_PATH VARCHAR(3000)");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 9");					
										logger.info("Database upgraded to version 9.");
									}
									
									if (dbSchemaVersion < 10) {
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD LAST_STATUS_MESSAGE VARCHAR(3000)");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 10");					
										logger.info("Database upgraded to version 10.");
									}
													
									if (dbSchemaVersion < 11) {
										if (getDbType().equals(DbType.MSSQL)) {
											ProgressMonitorDialog progressDialog = new ProgressMonitorDialog(Display.getCurrent().getActiveShell());
											try {
												progressDialog.run(true, false, new IRunnableWithProgress(){
					
													@Override
													public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
														try {
															monitor.beginTask("Upgrading database schema...", 12);
															
															// first remove all foreign keys
															removeForeignKey(FileInfo.class, FileInfo.class, "PARENT_ID");
															removeForeignKey(AclItem.class, Member.class);
															removeForeignKey(AclItem.class, FileInfo.class);
															removeForeignKey(AclItem.class, FileInfo.class, "INHERITED_FROM");
															removeForeignKey(AclItem.class, PermissionLookup.class, "PERMISSION_ID");
															removeForeignKey(Membership.class, Member.class, "GROUP_ID");
															removeForeignKey(Membership.class, Member.class);
															removeForeignKey(Member.class, Rdn.class);
															removeForeignKey(Member.class, Domain.class);				
															removeForeignKey(Filter.class, FilterSet.class);
															removeForeignKey(Filter.class, Report.class);
															removeForeignKey(ReportFilterSet.class, Report.class);
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM ALTER COLUMN PERMISSION_ID INT NOT NULL");
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORTFILTERSET DROP CONSTRAINT REPORTFILTERSET_SET_ID_FK");
															
															ResultSet primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__ACLITEM%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
															
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__FILEINFO%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.FILEINFO DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
															
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__DOMAIN%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.DOMAIN DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
															
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__FILTER_%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.FILTER DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
															
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__FILTERSE%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.FILTERSET DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
															
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__MEMBER_%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
															
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__MEMBERSH%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBERSHIP DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
															
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__PERMISS%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.PERMISSIONLOOKUP DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
															
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__RDN%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.RDN DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
													
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__REPORT_%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
																					
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__REPORTFI_%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORTFILTERSET DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
															
															primaryKey = DBObject.executeQuery("SELECT CONSTRAINT_NAME from INFORMATION_SCHEMA.TABLE_CONSTRAINTS WHERE CONSTRAINT_SCHEMA = 'PERMISSIONANALYZER' AND CONSTRAINT_NAME like 'PK__SCANLOG%'");
															primaryKey.next();
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.SCANLOGITEM DROP CONSTRAINT " + primaryKey.getString(1) + ";");
															primaryKey.close();
																				
															String[] tables = new String[]{"FILEINFO", "ACLITEM", "DOMAIN", "FILTER", "FILTERSET", "MEMBER", "MEMBERSHIP", "PERMISSIONLOOKUP", "RDN", "REPORT", "REPORTFILTERSET", "SCANLOGITEM"};
															for (String table : tables) {
																DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER." + table + " ADD TEMPID INT");
																DBObject.executeStatement("UPDATE PERMISSIONANALYZER." + table + " SET TEMPID = ID");
																DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER." + table + " DROP COLUMN ID");
																DBObject.executeStatement("EXEC sp_rename 'PERMISSIONANALYZER." + table + ".TEMPID', 'ID', 'COLUMN'");
																DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER." + table + " ALTER COLUMN ID INT NOT NULL");
																DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER." + table + "	ADD PRIMARY KEY (ID)");
																monitor.worked(1);
															}					
															
															// re-create the foreign keys
															addForeignKey(FileInfo.class, FileInfo.class, "PARENT_ID", "PARENT_ID", false);
															addForeignKey(AclItem.class, Member.class);
															addForeignKey(AclItem.class, FileInfo.class);
															addForeignKey(AclItem.class, FileInfo.class, "INHERITED_FROM_FOLDER_ID", "INHERITED_FROM", false);
															addForeignKey(AclItem.class, PermissionLookup.class, "PERMISSION_ID", "PERMISSION_ID", false);
															addForeignKey(Membership.class, Member.class, "GROUP_ID", "GROUP_ID", false);
															addForeignKey(Membership.class, Member.class);
															addForeignKey(Member.class, Rdn.class);
															addForeignKey(Member.class, Domain.class);
															addForeignKey(Filter.class, FilterSet.class);
															addForeignKey(Filter.class, Report.class);
															addForeignKey(ReportFilterSet.class, Report.class);
															DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORTFILTERSET ADD CONSTRAINT REPORTFILTERSET_SET_ID_FK FOREIGN KEY (FILTERSET_ID) REFERENCES PERMISSIONANALYZER.FILTERSET(ID)");
															
															DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 11");
														} catch (Exception e) {
															logger.error("Error while updating the database", e);
														}
													}
												});
											} catch (InvocationTargetException | InterruptedException ex) {
												logger.error("Error while compressing database.", ex);
												MessageDialog.openError(null, "Error", "Error while compressing the database: " + GUILogHandler.getNestedMessage(ex) + ".\r\nSee the log file for more details.");
											}
										} else if (getDbType().equals(DbType.H2)) {
											
										}
									} 
									
									if (dbSchemaVersion < 12) {
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.FILEINFO ADD OWNER_ID INT");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.FILEINFO ADD CONSTRAINT FILEINFO_OWNER_ID_FK FOREIGN KEY (OWNER_ID) REFERENCES PERMISSIONANALYZER.MEMBER(ID)");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 12");					
										logger.info("Database upgraded to version 12.");
									}
									
									if (dbSchemaVersion < 13) {
										// don't use these two foreign keys anymore, because we cannot make is cascade delete (MSSQL server will complain with a "multi path" error because there
										// is also a foreign key to the same table on the column FILEINFO_ID). If we use a normal foreign key (without cascade delete) then the delete in the scan process
										// will complain about integrity.
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM DROP CONSTRAINT ACLITEM_INHERITED_FROM_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.FILEINFO DROP CONSTRAINT FILEINFO_PARENT_ID_FK");
										
										// create an index on PARENT_ID in the FILEINFO, because this is not indexed anymore now the foreign key is gone
										DBObject.executeStatement("CREATE INDEX PARENTID_I ON PERMISSIONANALYZER.FILEINFO(PARENT_ID)");
										
										// change the foreign key in ACLITEM to a CASCADE DELETE
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM DROP CONSTRAINT ACLITEM_FILEINFO_ID_FK");
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.ACLITEM ADD CONSTRAINT ACLITEM_FILEINFO_ID_FK FOREIGN KEY (FILEINFO_ID) REFERENCES PERMISSIONANALYZER.FILEINFO(ID) ON DELETE CASCADE");
					
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 13");
										logger.info("Database upgraded to version 13.");
									}
									
									if (dbSchemaVersion < 14) {
										// create a new table REPORTHISTORY
										if (getDbType().equals(DbType.ORACLE)) {
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTHISTORY (ID NUMBER(10) PRIMARY KEY,REPORT_ID NUMBER(10) NOT NULL,RUN_DATE DATE,RESULT_COUNT NUMBER(10),FILE_PATH VARCHAR2(3000) NULL,REPORT_TYPE NUMBER(10) NOT NULL,FILE_TYPE NUMBER(10) NOT NULL,EMAIL_ADDRESS VARCHAR2(255),THRESHOLD NUMBER(10),STATUS_MESSAGE VARCHAR2(3000))");
										}
										
										if (getDbType().equals(DbType.MSSQL)) {
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTHISTORY (ID INT PRIMARY KEY,REPORT_ID INT NOT NULL,RUN_DATE DATETIME,RESULT_COUNT INT,FILE_PATH VARCHAR(3000) NULL,REPORT_TYPE INT NOT NULL,FILE_TYPE INT NOT NULL,EMAIL_ADDRESS VARCHAR(255),THRESHOLD INT,STATUS_MESSAGE VARCHAR(3000))");
										}
										
										if (getDbType().equals(DbType.MYSQL) || getDbType().equals(DbType.POSTGRESQL)) {
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTHISTORY (ID INT NOT NULL,REPORT_ID INT NOT NULL,RUN_DATE TIMESTAMP,RESULT_COUNT INT,FILE_PATH VARCHAR(3000) NULL,REPORT_TYPE INT NOT NULL,FILE_TYPE INT NOT NULL,EMAIL_ADDRESS VARCHAR(255),THRESHOLD INT,STATUS_MESSAGE VARCHAR(3000),PRIMARY KEY (ID))");
										}
					
										if (getDbType().equals(DbType.H2) || getDbType().equals(DbType.DERBY) || getDbType().equals(DbType.DB2)) {
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTHISTORY (ID INT NOT NULL CONSTRAINT REPORT_PK PRIMARY KEY,REPORT_ID INT NOT NULL,RUN_DATE TIMESTAMP,RESULT_COUNT INT,FILE_PATH VARCHAR(3000) NULL,REPORT_TYPE INT NOT NULL,FILE_TYPE INT NOT NULL,EMAIL_ADDRESS VARCHAR(255),THRESHOLD INT,STATUS_MESSAGE VARCHAR(3000))");
										}
										
										// foreign key to the REPORT table
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORTHISTORY ADD CONSTRAINT REPORTHISTORY_REPORT_ID_FK FOREIGN KEY (REPORT_ID) REFERENCES PERMISSIONANALYZER.REPORT(ID)");
									
										// create a new column IS_ENABLED in the MEMBER table
										if (getDbType().equals(DbType.ORACLE)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER ADD IS_ENABLED NUMBER(1)");
										}
										
										if (getDbType().equals(DbType.MSSQL)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER ADD IS_ENABLED BIT");
										}
										
										if (getDbType().equals(DbType.DERBY) || getDbType().equals(DbType.DB2)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER ADD IS_ENABLED SMALLINT");
										}
										
										if (getDbType().equals(DbType.H2) || getDbType().equals(DbType.MYSQL) || getDbType().equals(DbType.POSTGRESQL)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.MEMBER ADD IS_ENABLED BOOLEAN");
										}
										
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 14");
										
										logger.info("Database upgraded to version 14.");
									}
									
									if (dbSchemaVersion < 15) {
										DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.SCANLOGITEM ALTER COLUMN MESSAGE varchar(2000) NOT NULL");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 15");					
										logger.info("Database upgraded to version 15.");
									}
									
									if (dbSchemaVersion < 16) {
										if (getDbType().equals(DbType.MSSQL) || getDbType().equals(DbType.DB2) || getDbType().equals(DbType.DERBY) || getDbType().equals(DbType.ORACLE) || getDbType().equals(DbType.POSTGRESQL)) {
											DBObject.executeStatement("CREATE INDEX ACLITEM_MEMBER_ID_I ON PERMISSIONANALYZER.ACLITEM(MEMBER_ID)");
											DBObject.executeStatement("CREATE INDEX ACLITEM_FILEINFO_ID_I ON PERMISSIONANALYZER.ACLITEM(FILEINFO_ID)");
											DBObject.executeStatement("CREATE INDEX ACLITEM_PERMISSION_ID_I ON PERMISSIONANALYZER.ACLITEM(PERMISSION_ID)");
											DBObject.executeStatement("CREATE INDEX MEMBERSHIP_GROUP_ID_I ON PERMISSIONANALYZER.MEMBERSHIP(GROUP_ID)");
											DBObject.executeStatement("CREATE INDEX MEMBERSHIP_MEMBER_ID_I ON PERMISSIONANALYZER.MEMBERSHIP(MEMBER_ID)");
											DBObject.executeStatement("CREATE INDEX FILEINFO_OWNER_ID_I ON PERMISSIONANALYZER.FILEINFO(OWNER_ID)");
										}
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 16");					
										logger.info("Database upgraded to version 16.");
									}
									
									if (dbSchemaVersion < 17) {
										// create a new column USE_SIMPLE_PRESENTATION in the REPORT table
										if (getDbType().equals(DbType.ORACLE)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD USE_SIMPLE_PRESENTATION NUMBER(1)");
										}
										
										if (getDbType().equals(DbType.MSSQL)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD USE_SIMPLE_PRESENTATION BIT");
										}
										
										if (getDbType().equals(DbType.DERBY) || getDbType().equals(DbType.DB2)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD USE_SIMPLE_PRESENTATION SMALLINT");
										}
										
										if (getDbType().equals(DbType.H2) || getDbType().equals(DbType.MYSQL) || getDbType().equals(DbType.POSTGRESQL)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD USE_SIMPLE_PRESENTATION BOOLEAN");
										}
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 17");					
										logger.info("Database upgraded to version 17.");
									}
									
									if (dbSchemaVersion < 18) {
										// create a new table REPORTCOMPARATOR1 and REPORTCOMPARATOR2
										if (getDbType().equals(DbType.ORACLE)) {
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTCOMPARATOR1 (ID NUMBER(10) PRIMARY KEY,REPORT1_ROW_KEY VARCHAR(3000),REPORT1_ROW_VALUE VARCHAR2(3000) NOT NULL)");
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTCOMPARATOR2 (ID NUMBER(10) PRIMARY KEY,REPORT2_ROW_KEY VARCHAR(3000),REPORT2_ROW_VALUE VARCHAR2(3000) NOT NULL)");
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORTHISTORY ADD WAS_COMPARISON NUMBER(1)");
										}
										
										if (getDbType().equals(DbType.MSSQL)) {
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTCOMPARATOR1 (ID INT PRIMARY KEY,REPORT1_ROW_KEY VARCHAR(3000),REPORT1_ROW_VALUE VARCHAR(3000) NOT NULL)");
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTCOMPARATOR2 (ID INT PRIMARY KEY,REPORT2_ROW_KEY VARCHAR(3000),REPORT2_ROW_VALUE VARCHAR(3000) NOT NULL)");
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORTHISTORY ADD WAS_COMPARISON BIT");
										}
										
										if (getDbType().equals(DbType.MYSQL) || getDbType().equals(DbType.POSTGRESQL)) {
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTCOMPARATOR1 (ID INT NOT NULL,REPORT1_ROW_KEY VARCHAR(3000) NOT NULL,REPORT1_ROW_VALUE VARCHAR(3000) NOT NULL,PRIMARY KEY (ID))");
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTCOMPARATOR2 (ID INT NOT NULL,REPORT2_ROW_KEY VARCHAR(3000) NOT NULL,REPORT2_ROW_VALUE VARCHAR(3000) NOT NULL,PRIMARY KEY (ID))");
										}
					
										if (getDbType().equals(DbType.H2) || getDbType().equals(DbType.DERBY) || getDbType().equals(DbType.DB2)) {
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTCOMPARATOR1 (ID INT NOT NULL CONSTRAINT REPORTCOMPARATOR1_PK PRIMARY KEY,REPORT1_ROW_KEY VARCHAR(3000) NOT NULL,REPORT1_ROW_VALUE VARCHAR(3000) NOT NULL)");
											DBObject.executeStatement("CREATE TABLE PERMISSIONANALYZER.REPORTCOMPARATOR2 (ID INT NOT NULL CONSTRAINT REPORTCOMPARATOR2_PK PRIMARY KEY,REPORT2_ROW_KEY VARCHAR(3000) NOT NULL,REPORT2_ROW_VALUE VARCHAR(3000) NOT NULL)");
										}
										
										if (getDbType().equals(DbType.DERBY) || getDbType().equals(DbType.DB2)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORTHISTORY ADD WAS_COMPARISON SMALLINT");
										}
										
										if (getDbType().equals(DbType.H2) || getDbType().equals(DbType.MYSQL) || getDbType().equals(DbType.POSTGRESQL)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORTHISTORY ADD WAS_COMPARISON BOOLEAN");
										}
										
										DBObject.executeStatement("CREATE INDEX REPORTCOMPARATOR1_ROWKEY_I ON PERMISSIONANALYZER.REPORTCOMPARATOR1(REPORT1_ROW_KEY)");
										DBObject.executeStatement("CREATE INDEX REPORTCOMPARATOR2_ROWKEY_I ON PERMISSIONANALYZER.REPORTCOMPARATOR2(REPORT2_ROW_KEY)");
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.REPORTHISTORY SET WAS_COMPARISON = FALSE");
										
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 18");					
										logger.info("Database upgraded to version 18.");
									}
									
									if (dbSchemaVersion < 19) {
										// create a new column CREATE_COMPARISON_REPORT in the REPORT table
										if (getDbType().equals(DbType.ORACLE)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD CREATE_COMPARISON_REPORT NUMBER(1)");
										}
										
										if (getDbType().equals(DbType.MSSQL)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD CREATE_COMPARISON_REPORT BIT");
										}
										
										if (getDbType().equals(DbType.DERBY) || getDbType().equals(DbType.DB2)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD CREATE_COMPARISON_REPORT SMALLINT");
										}
										
										if (getDbType().equals(DbType.H2) || getDbType().equals(DbType.MYSQL) || getDbType().equals(DbType.POSTGRESQL)) {
											DBObject.executeStatement("ALTER TABLE PERMISSIONANALYZER.REPORT ADD CREATE_COMPARISON_REPORT BOOLEAN");
										}
										
										DBObject.executeStatement("UPDATE PERMISSIONANALYZER.INFO SET VERSION = 19");					
										logger.info("Database upgraded to version 19.");
									}									
									
								} catch (Exception e) {
									logger.error("Error while updating the database schema", e);
									SyncMessageDialog.showMessage(MessageDialog.ERROR, "Database schema update", "Error while updating the database schema (current version: " + dbSchemaVersion + "): " + e.getMessage());
								}
							}
						});
					} catch (InvocationTargetException | InterruptedException ex) {
						logger.error("Error while upgrading database schema.", ex);
						MessageDialog.openError(null, "Error", "Error while upgrading database schema: " + GUILogHandler.getNestedMessage(ex) + ".\r\nSee the log file for more details.");
					}
				}
			}
		} catch (Exception e) {
			logger.error("Error while updating the database schema", e);
			SyncMessageDialog.showMessage(MessageDialog.ERROR, "Database schema update", "Error while updating the database schema (current version: " + dbSchemaVersion + "): " + e.getMessage());
		}
		finally {
			DBObject.closeResultSet(rs);
		}
	}
	
	public static void addForeignKey(Class<? extends DBObject> sourceClass, Class<? extends DBObject> foreignClass) throws DBException {
		addForeignKey(sourceClass, foreignClass, foreignClass.getSimpleName().toUpperCase() + "_ID", foreignClass.getSimpleName().toUpperCase() + "_ID", false);
	}
	
	public static void addForeignKey(Class<? extends DBObject> sourceClass, Class<? extends DBObject> foreignClass, String column, String columnShortName, boolean cascadeDelete) throws DBException {
		String query = null;
		
		logger.debug("Adding foreign key for table " + sourceClass.getSimpleName().toUpperCase() + " pointing to table " + foreignClass.getSimpleName().toUpperCase() + " using column " + columnShortName);
		
		try {
			query = MessageFormat.format("ALTER TABLE {0} ADD CONSTRAINT {1}_{2}_FK FOREIGN KEY ({3}) REFERENCES {4}(id) {5}", DBObject.SCHEMA + "." + sourceClass.getSimpleName().toUpperCase(), 
					sourceClass.getSimpleName().toUpperCase(), columnShortName, column, DBObject.SCHEMA + "." + foreignClass.getSimpleName().toUpperCase(),
					cascadeDelete ?  "ON DELETE CASCADE" : "");
			DBObject.executeStatement(query);
			
		} catch (DBException ex) {
			logger.error("Error while adding foreign key using query " + query, ex);
			throw new DBException(query, new Object[]{sourceClass.getSimpleName().toUpperCase(), foreignClass.getSimpleName().toUpperCase()}, ex);
		}
	}
	
	public static void removeForeignKey(Class<? extends DBObject> sourceClass, Class<? extends DBObject> foreignClass) throws DBException {
		removeForeignKey(sourceClass, foreignClass, foreignClass.getSimpleName().toUpperCase() + "_ID");
	}
	
	public static void removeForeignKey(Class<? extends DBObject> sourceClass, Class<? extends DBObject> foreignClass, String columnShortName) throws DBException {
		String query = null;
		
		logger.debug("Dropping foreign key for table " + sourceClass.getSimpleName().toUpperCase() + " pointing to table " + foreignClass.getSimpleName().toUpperCase() + " using column " + columnShortName);
		
		try {
			if (getDbType().equals(DbType.MSSQL) || getDbType().equals(DbType.ORACLE) || getDbType().equals(DbType.POSTGRESQL)) {
				query = "ALTER TABLE {0} DROP CONSTRAINT {1}_{2}_FK";
			} else {
				query = "ALTER TABLE {0} DROP FOREIGN KEY {1}_{2}_FK";
			}
			query = MessageFormat.format(query, DBObject.SCHEMA + "." + sourceClass.getSimpleName().toUpperCase(), sourceClass.getSimpleName().toUpperCase(), columnShortName);
			
			DBObject.executeStatement(query);
			
		} catch (DBException ex) {
			logger.error("Error while removing foreign key using query " + query, ex);
		}
	}
	
	public static void resetIdentityColumn(Class<? extends DBObject> clazz) throws DBException {
		String query = null;
		
		try {			
			if (getDbType().equals(DbType.MSSQL)) {
				query = "DBCC CHECKIDENT ('{0}', RESEED, 1)";
				
			} else if (getDbType().equals(DbType.MYSQL)){
				query = "ALTER TABLE {0} AUTO_INCREMENT = 1";
			
			} else if (getDbType().equals(DbType.POSTGRESQL)) {
				query = "ALTER SEQUENCE {0}_id_seq RESTART WITH 1";
				
			} else if (getDbType().equals(DbType.ORACLE)) {
				query = "DROP SEQUENCE {0}_SEQ";
				query = query.replace("{0}", DBObject.SCHEMA + "." + clazz.getSimpleName().toUpperCase());
				DBObject.executeStatement(query);
				
				query = "CREATE SEQUENCE {0}_SEQ START WITH 1 INCREMENT BY 1";
				query = query.replace("{0}", DBObject.SCHEMA + "." + clazz.getSimpleName().toUpperCase());
				DBObject.executeStatement(query);
				return;
				
			} else {
				query = "ALTER TABLE {0} ALTER COLUMN id RESTART WITH 1";
			}
			
			query = query.replace("{0}", DBObject.SCHEMA + "." + clazz.getSimpleName().toUpperCase());
			DBObject.executeStatement(query);
			
		} catch (DBException ex) {
			logger.error("Error while resetting identity column for table " + clazz.getSimpleName().toUpperCase() + " using query " + query, ex);
		}
	}
	
	@SuppressWarnings("rawtypes")
	public static void updateStatistics() throws DBException {
		String query = null;
		Class[] classes = new Class[] {AclItem.class, FileInfo.class, Member.class, Membership.class};
		
		switch (getDbType()) {
			case H2:
				DBObject.executeStatement("ANALYZE");
				return;
				
			case MSSQL:
				query = "UPDATE STATISTICS {0}";
				break;
				
			case MYSQL:
				query = "ANALYZE TABLE {0}";
				break;
			
			case DB2:
				// using CallableStatement
				try {
					String statement = "CALL SYSPROC.ADMIN_CMD('RUNSTATS ON TABLE PERMISSIONANALYZER.{0} WITH DISTRIBUTION AND INDEXES ALL')";
					for (Class clazz : classes) {
						// note: MessageFormat doesn't work, probably because of the quotes in the String
						CallableStatement cs = DBUtil.getDBConnection().prepareCall(statement.replace("{0}", clazz.getSimpleName().toUpperCase()));
						cs.executeUpdate();
					}

					return;
				} catch (SQLException e) {
					logger.error("Error while running DB2 statistics", e);
					throw new DBException("Error while running DB2 statistics: " + e.getMessage(), e);
				}
			
			case POSTGRESQL: 
				query = "ANALYZE {0}";
				break;
				
			case ORACLE:
				// using CallableStatement
				try {
					CallableStatement cs = DBUtil.getDBConnection().prepareCall ("{call dbms_stats.gather_schema_stats(?, cascade=>TRUE)}");
					cs.setString(1, DBObject.SCHEMA);
					cs.executeUpdate();
					
					return;
				} catch (SQLException e) {
					logger.error("Error while running Oracle schema statistics", e);
					throw new DBException("Error while running Oracle schema statistics: " + e.getMessage(), e);
				}

			case DERBY:
				// using CallableStatement
				try {
					for (Class clazz : classes) {
						CallableStatement cs = DBUtil.getDBConnection().prepareCall ("CALL SYSCS_UTIL.SYSCS_UPDATE_STATISTICS(?, ?, null)");
						cs.setString(1, DBObject.SCHEMA);
						cs.setString(2, clazz.getSimpleName().toUpperCase());
						cs.executeUpdate();
					}

					return;
				} catch (SQLException e) {
					logger.error("Error while running Derby statistics", e);
					throw new DBException("Error while running Derby statistics: " + e.getMessage(), e);
				}
		}
		
		for (Class clazz : classes) {
			DBObject.executeStatement(MessageFormat.format(query, DBObject.SCHEMA + "." + clazz.getSimpleName().toUpperCase()));
		}
	}
	
	public static void truncateTable(Class<? extends DBObject> clazz) throws DBException {
		PreparedStatement ps = null;
		String query = null;
		
		logger.debug("Truncating table " + clazz.getSimpleName().toUpperCase());
		
		try {			
			if (getDbType().equals(DbType.DB2)) {
				query = "TRUNCATE TABLE {0} IMMEDIATE";
			} else {
				query = "TRUNCATE TABLE {0}";
			}
			query = MessageFormat.format(query, DBObject.SCHEMA + "." + clazz.getSimpleName().toUpperCase());
			
			ps = DBObject.getConnection().prepareStatement(query);
				
			ps.execute();
			
			try {
				clazz.newInstance().setLastId(0);
			} catch (InstantiationException | IllegalAccessException e) {
				logger.warn("Could not reset the id generation for table " + clazz.getSimpleName());
			} 
			
		} catch (SQLException ex) {
			logger.error("Error while truncating table using query " + query, ex);
			throw new DBException(query, new Object[]{clazz.getSimpleName().toUpperCase()}, ex);
		} finally {
			DBObject.closeStatement(ps);
		}
	}
	
	public static void truncateAllData(StatusListener statusListener) throws DBException {
		statusListener.updateStatus("Truncating existing info from the database ...");
		statusListener.newStatusMessage(new ScanLogItem(Severity.INFO, "Truncating existing info from the database", null));
		
		// drop foreign keys in order to enable truncating the tables
		removeForeignKey(FileInfo.class, Member.class, "OWNER_ID");
		removeForeignKey(AclItem.class, Member.class);
		removeForeignKey(AclItem.class, FileInfo.class);
		removeForeignKey(AclItem.class, PermissionLookup.class, "PERMISSION_ID");
		removeForeignKey(Membership.class, Member.class, "GROUP_ID");
		removeForeignKey(Membership.class, Member.class);
		removeForeignKey(Member.class, Rdn.class);
		removeForeignKey(Member.class, Domain.class);				
											
		//clear current data
		//TODO only delete the records that are related to the directories to scan
		truncateTable(AclItem.class);
		truncateTable(Membership.class);
		truncateTable(Member.class);
		truncateTable(FileInfo.class);
		truncateTable(Rdn.class);
		truncateTable(Domain.class);
		
		// reset identity columns
		/*
		try{
			logger.debug("Resetting identity columns");
			
			resetIdentityColumn(AclItem.class);
			resetIdentityColumn(Membership.class);
			resetIdentityColumn(Member.class);
			resetIdentityColumn(FileInfo.class);
			resetIdentityColumn(Rdn.class);
			resetIdentityColumn(Domain.class);
		} catch (Exception ex) {
			logger.warn("Could not reset identity column: " + ex.getMessage(), ex);
		}*/
		
		// re-create the foreign keys
		addForeignKey(FileInfo.class, Member.class, "OWNER_ID", "OWNER_ID", false);
		addForeignKey(AclItem.class, Member.class);
		addForeignKey(AclItem.class, FileInfo.class, "FILEINFO_ID", "FILEINFO_ID", true);
		addForeignKey(AclItem.class, PermissionLookup.class, "PERMISSION_ID", "PERMISSION_ID", false);
		addForeignKey(Membership.class, Member.class, "GROUP_ID", "GROUP_ID", false);
		addForeignKey(Membership.class, Member.class);
		addForeignKey(Member.class, Rdn.class);
		addForeignKey(Member.class, Domain.class);
	}
	
	public static void truncateLdapData(final StatusListener statusListener) throws DBException {
		statusListener.updateStatus("Truncating existing ldap info from the database ...");
		statusListener.newStatusMessage(new ScanLogItem(Severity.INFO, "Truncating existing ldap info from the database", null));
		
		// drop foreign keys in order to enable truncating the tables
		removeForeignKey(Membership.class, Member.class, "GROUP_ID");
		removeForeignKey(Membership.class, Member.class);
											
		//clear current data
		truncateTable(Membership.class);
		
		// reset identity columns
		/*
		try{
			logger.debug("Resetting identity columns");
			
			resetIdentityColumn(Membership.class);
			
		} catch (Exception ex) {
			logger.warn("Could not reset identity column: " + ex.getMessage(), ex);
		}*/
		
		// re-create the foreign keys
		addForeignKey(Membership.class, Member.class, "GROUP_ID", "GROUP_ID", false);
		addForeignKey(Membership.class, Member.class);
		
		// add a membership record for all the existing members
		// (this is normally handled in the INSERT method of the Member class)
		// the query insert into permissionanalyzer.membership (group_id, member_id) select id, id from permissionanalyzer.member did not work
		// because we changed the auto increment ID generation into a self managed ID
		try {
			DBObject.getUsingCustomProcessor(DBObject.getPreparedStatement("select id from permissionanalyzer.member"), new CustomProcessor() {
				@Override
				public boolean processResultSet(ResultSet rs) throws SQLException, DBException {
					Membership membership = new Membership(rs.getInt(1), rs.getInt(1));
					membership.save();
					return true;
				}
			});
		} catch (SQLException e) {
			logger.error("Error while running truncating LDAP data", e);
			throw new DBException("Error while truncating LDAP data: " + e.getMessage(), e);
		}
		
		// reset the displayName of all members, so they will be refreshed with new LDAP info
		DBObject.executeStatement("update permissionanalyzer.member set display_name_lowercase = null, display_name = null where rdn_id is not null"); // don't clear local groups/users
	}
	
	public static void truncateDirectoryData(StatusListener statusListener) throws DBException {
		statusListener.updateStatus("Truncating existing directory info from the database ...");
		statusListener.newStatusMessage(new ScanLogItem(Severity.INFO, "Truncating existing directory info from the database", null));
		
		// drop foreign keys in order to enable truncating the tables
		removeForeignKey(AclItem.class, Member.class);
		removeForeignKey(AclItem.class, FileInfo.class);
		removeForeignKey(AclItem.class, PermissionLookup.class, "PERMISSION_ID");
											
		//clear current data
		//TODO only delete the records that are related to the directories to scan
		truncateTable(AclItem.class);
		truncateTable(FileInfo.class);
		
		// reset identity columns
		/*
		try{
			logger.debug("Resetting identity columns");
			
			resetIdentityColumn(AclItem.class);
			resetIdentityColumn(FileInfo.class);
			
		} catch (Exception ex) {
			logger.warn("Could not reset identity column: " + ex.getMessage(), ex);
		}*/
		
		// re-create the foreign keys
		addForeignKey(AclItem.class, Member.class);
		addForeignKey(AclItem.class, FileInfo.class);
		addForeignKey(AclItem.class, PermissionLookup.class, "PERMISSION_ID", "PERMISSION_ID", false);
	}
	
	public static DbType getDbType() throws DBException {
		try {
			if (dbType != null) {
				return dbType;
			} else {
				String productName = DBObject.getConnection().getMetaData().getDatabaseProductName();
				if (productName.startsWith("DB2")) { 
		    		dbType = DbType.DB2; 
		    	} 
		    	else if(productName.equalsIgnoreCase("Apache Derby")){
		    		dbType = DbType.DERBY;
		        }
		    	else if(productName.equalsIgnoreCase("Oracle")){
		    		dbType = DbType.ORACLE;
		        }
		        else if(productName.equalsIgnoreCase("MySQL")){
		        	dbType = DbType.MYSQL;
		        }
		        else if(productName.equalsIgnoreCase("PostgreSQL")){
		        	dbType = DbType.POSTGRESQL;
		        }
		        else if(productName.equalsIgnoreCase("h2")){
		        	dbType = DbType.H2;
		        }
		        else if(productName.equalsIgnoreCase("Microsoft SQL Server")){
		        	dbType = DbType.MSSQL;
		        }
		        else{
		        	dbType = DbType.UNKNOWN;
		        	logger.error("Could not determine database type for product name " + productName);
		        }
				
				logger.info("Determined database type: " + dbType.name());
				return dbType;
			}
		} catch (SQLException ex) {
			logger.error("Error while determining database type", ex);
			throw new DBException(null, null, ex);
		}
	}
	
	public static void clearDbType() {
		dbType = null;
	}
	
	public static boolean isLocalDB() {
		return isLocalDB(PreferenceHelper.getInstance().getDbConnectionType());
	}
	
	private static boolean isLocalDB(int dbConnectionType) {
		return dbConnectionType == 0;
	}
}
