package com.perdemia.permissionanalyzer.data.util;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.Domain;

public class DomainCache {
	private static final Logger logger = LoggerFactory.getLogger(DomainCache.class.getName());
	
	/**
	 * Cache map between domain name and domain id
	 */
	private static final Map<String, Integer> domains = new HashMap<String, Integer>();
	
	/** Cache for the mapping between domain components (DC=TEST,DC=LOCAL) and the netbios name 
	 * @see {@LdapConnectorImpl#getNetBiosDomainName(String)}
	 * */
	private static final Map<String, String> netbiosNames = new HashMap<String, String>();
	
	/**
	 * List of domains that is being held in cache
	 */
	private static List<Domain> domainList; 
	
	/**
	 * Retrieves the domain ID from the cache based on the given domain name. If the name is not found then the cache Map with
	 * all the domains is retrieved from the database again. If the domain name is still not found it is created in the database and added to the cache Map.
	 * 
	 * @param domainName
	 * @return domain ID or null when a exception occurred
	 */
	public static Integer getDomainId(String domainName){
		domainName = domainName.toUpperCase();
		try {
			if (domains.containsKey(domainName)) {
				return domains.get(domainName);
			} else {
				// refresh the list with domains from the database
				domains.clear();
				List<Domain> domainList = getDomains(true);
				for (Domain domain : domainList) {
					domains.put(domain.getDomain(), domain.getId());
				}
				
				// now in the list?
				if (domains.containsKey(domainName)) {
					return domains.get(domainName);
				} else {
					// not in the database yet, create a new Domain record
					Domain domain = new Domain(domainName);
					domain.save();
					domains.put(domainName, domain.getId());
					domainList.add(domain);
					return domain.getId();
				}
			}
		} catch (DBException ex) {
			logger.error("Error while getting domain id from database " + domainName, ex);
			return null;
		}
	}
	
	/**
	 * Clears the Map of the cache and forces a refresh from the database whenever the cache is used.
	 */
	public static void clearCache() {
		domains.clear();
		netbiosNames.clear();
		domainList = null;
	}
	
	public static List<Domain> getDomains(boolean forceRefresh) throws DBException {
		if (domainList == null || forceRefresh) {
			domainList = Domain.getAll();
		}
		return domainList;
	}
	
	public static Domain getDomain(Integer domainId) throws DBException {
		for (Domain domain : getDomains(false)) {
			if (domain.getId().equals(domainId)) {
				return domain;
			}
		}
		return null;
	}

	public static Map<String, String> getNetbiosnames() {
		return netbiosNames;
	}
}
