package com.perdemia.permissionanalyzer.data.util;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.FileInfo;

public class FileCache {
	private static final Logger logger = LoggerFactory.getLogger(FileCache.class.getName());
	
	/** Map with FileInfo objects and Id's */
	private static final Map<Integer, FileInfo> files = new HashMap<Integer, FileInfo>();
	
	/** Map with wildcard search and List of FileInfo objects 
	 * @see {@link FilterHelper#addFilterOnFolders(String, boolean)} 
	 */
	private static final Map<String, List<FileInfo>> wilcardMatches = new HashMap<String, List<FileInfo>>(100);
	
	/** Cache settings */
	private static final int CACHE_SIZE = 10000;
	
	/**
	 * Clears the Map of the cache and forces a refresh from the database whenever the cache is used.
	 */
	public static void clearCache() {
		files.clear();
		wilcardMatches.clear();
	}
	
	public static FileInfo getFile(Integer fileId) throws DBException {
		FileInfo file = files.get(fileId);
		
		if (file != null && logger.isTraceEnabled()) {
			logger.trace("Returning file from cache: " + file.toString());
		} else {
			// not yet in cache
			file = FileInfo.getById(fileId);
			
			if (file != null) {
				// get the corresponding owner
				if (file.getOwnerId() != null) {
					file.setOwner(MemberCache.getMember(file.getOwnerId()));
				}
				
				FileCache.cacheFile(file);
			}
		}
		
		if (file == null) {
			// should never happened, but occurred by one customer
			logger.warn("File ID " + fileId + " not found in the database!");
			logger.warn("Max file ID: " + DBObject.getMaxId(FileInfo.class));
			throw new DBException("File ID " + fileId + " not found in the database!");
		}
		
		return file;
	}
		
	public static void clearFile(Integer fileId) {
		files.remove(fileId);
	}
	
	public static void cacheFile(FileInfo file) {
		if (files.size() > CACHE_SIZE) {
			if (logger.isTraceEnabled()) {
				logger.trace("FileInfo cache full, clear cache");
			}			
			cleanupCache(files);
		}
		files.put(file.getId(), file);
	}
	
	public static List<FileInfo> getWildcardMatches (String wildcardSearch) {
		return wilcardMatches.get(wildcardSearch);
	}
	
	public static List<FileInfo> cacheWildcardSearch (String wildcardSearch, List<FileInfo> files) {
		return wilcardMatches.put(wildcardSearch, files);
	}
	
	/**
	 * Cleans up a cache when the maximum number of items has been reached. It removes half of the entries in the Map.
	 * 
	 * @param map Map to clean
	 */
	private static void cleanupCache(Map map) {
		logger.debug("Clean up half of the File cache, max size reached: " + CACHE_SIZE);
		
		// clear half of the cache
		int clearCount = (CACHE_SIZE / 2);
		int count = 0;
		Iterator iter = map.keySet().iterator();
		
		while (iter.hasNext() && count < clearCount) {
			iter.next();
			iter.remove();
			count++;
		}
	}
}