package com.perdemia.permissionanalyzer.data.util;

import java.sql.PreparedStatement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.FilterHelper.ObjectType;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.Member;

public class MemberCache {
	private static final Logger logger = LoggerFactory.getLogger(MemberCache.class.getName());
	
	/** Maps for member by name and by Id. The name cache is used during the scan process the id cache is used in the overviews */
	private static final Map<Integer, Member> membersById = new HashMap<Integer, Member>(500);
	private static final Map<String, Member> membersByName = new HashMap<String, Member>(500);
	private static final Map<String, Member> membersByCN = new HashMap<String, Member>(500);
	
	
	/** Cache settings */
	private static final int CACHE_SIZE = 50000;
	
	/**
	 * Clears the Map of the cache and forces a refresh from the database whenever the cache is used.
	 */
	public static void clearCache() {
		membersById.clear();
		membersByName.clear();
		membersByCN.clear();
	}
	
	/**
	 * Clears the Map of the cache and forces a refresh from the database whenever the cache is used.
	 */
	public static void clearCacheByCN() {
		membersByCN.clear();
	}
	
	public static Member getMember(Integer memberId) throws DBException {
		Member member = membersById.get(memberId);
		
		if (member != null && logger.isTraceEnabled()) {
			logger.trace("Returning member from cache: " + member.toString());
		} else {
			// not yet in cache
			member = Member.getById(memberId);
			
			if (member != null && member.getId() == null) {
				cacheMemberById(member);
			}
		}
		
		return member;
	}
	
	/** Checks if the accountName can be found in the {@link #membersByName} cache. If not, if will check the database and if it's found
	 * the member will be added to the cache.
	 * 
	 * @param memberName
	 * @param domainId
	 * 
	 * @return null in case the member cannot be found in the cache and in the database
	 * 
	 * @throws DBException
	 */
	public static Member getMemberByName(String memberName, Integer domainId) throws DBException {
		Member member = membersByName.get(memberName + "--" + domainId);
		
		if (member != null && logger.isTraceEnabled()) {
			logger.trace("Returning member from cache: " + member.toString());
		}
		
		if (member == null) {
			// try to get it from the database
			if (domainId == null) {
				member = Member.findFirst("name_lowercase = ? AND domain_id is null", memberName.toLowerCase());
			} else {
				member = Member.findFirst("name_lowercase = ? AND domain_id = ?", memberName.toLowerCase(), domainId);
			}
			
			if (member != null) {
				// add to cache
				cacheMemberByName(member);
			}
		}
		
		return member;
	}
	
	/** Checks if the cn can be found in the {@link #membersByCN} cache. If not, if will check the database and if it's found
	 * the member will be added to the cache.
	 * 
	 * @param cn
	 * @param rdnId
	 * @param accountName optional additional filter, instead of calling getMemberByCN and getMemberByName you can combine the query using this parameter.
	 * @param domainId optional parameter used in combination with accountName
	 * 
	 * @return null in case the member cannot be found in the cache and in the database
	 * 
	 * @throws DBException
	 */
	public static Member getMemberByCN(String cn, int rdnId, String accountName, int domainId) throws DBException {
		Member member = membersByCN.get(cn + "--" + rdnId);
		
		if (member != null && logger.isTraceEnabled()) {
			logger.trace("Returning member from cache: " + member.toString());
		}
		
		if (member == null) {
			// try to get it from the database
			if (accountName != null) {
				member = Member.findFirst("(cn = ? AND rdn_id = ?) or (name = ? and domain_id = ?)", cn, rdnId, accountName, domainId);
			} else {
				member = Member.findFirst("cn = ? AND rdn_id = ?", cn, rdnId);
			}
			
			if (member != null) {
				// add to cache
				cacheMemberByCN(member);
			}
		}
		return member;
	}
	
	public static void clearMember(Integer memberId) {
		membersById.remove(memberId);
	}
	
	public static void clearMemberByName(String memberName) {
		membersByName.remove(memberName);
	}
	
	public static void clearMemberByCN(String cn, int rdnId) {
		membersByCN.remove(cn + "--" + rdnId);
	}
	
	public static void cacheMemberById(Member member) {
		if (membersById.size() > CACHE_SIZE) {
			if (logger.isTraceEnabled()) {
				logger.trace("Member cache full, clear cache");
			}			
			cleanupCache(membersById);
		}
		membersById.put(member.getId(), member);
	}
	
	public static void cacheMemberByName(Member member) {
		if (membersByName.size() > CACHE_SIZE) {
			if (logger.isTraceEnabled()) {
				logger.trace("Member cache full, clear cache");
			}			
			cleanupCache(membersByName);			
		}
		membersByName.put(member.getName(false) + "--" + member.getDomainId(), member);
	}
	
	public static void cacheMemberByCN(Member member) {
		if (membersByCN.size() > CACHE_SIZE) {
			if (logger.isTraceEnabled()) {
				logger.trace("Member cache full, clear cache");
			}			
			cleanupCache(membersByCN);
		}
		membersByCN.put(member.getCn() + "--" + member.getRdnId(), member);
	}
		
	/**
	 * Cleans up a cache when the maximum number of items has been reached. It removes half of the entries in the Map.
	 * 
	 * @param map Map to clean
	 */
	private static void cleanupCache(Map map) {
		logger.debug("Clean up half of the Member cache, max size reached: " + CACHE_SIZE);
		
		// clear half of the cache
		int clearCount = (CACHE_SIZE / 2);
		int count = 0;
		Iterator iter = map.keySet().iterator();
		
		while (iter.hasNext() && count < clearCount) {
			iter.next();
			iter.remove();
			count++;
		}
	}
}