package com.perdemia.permissionanalyzer.data.util;

import java.sql.PreparedStatement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.contentproviders.MainFileTreeContentProvider;
import com.perdemia.permissionanalyzer.application.util.FilterHelper;
import com.perdemia.permissionanalyzer.application.util.FilterHelper.ObjectType;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.Membership;

public class MembershipCache {
	private static final Logger logger = LoggerFactory.getLogger(MembershipCache.class.getName());
	
	/** List of memberships in the cache, in the form of groupId + "-" + memberId */
	private static final List<String> membershipCache = new ArrayList<String>(1000);
	
	/** List of group membership counts used by calculate the ownership ratio,
	 * it is a map between a group id's (1,2,3) and the count of its (nested) members without using the filters from the UI
	 * @see {@link MainFileTreeContentProvider#determineOwnershipRatio(com.perdemia.permissionanalyzer.application.util.FilterHelper, com.perdemia.permissionanalyzer.data.FileInfo)} 
	 */
	private static final Map<String, Integer> groupMembershipCountCache = new HashMap<String, Integer>();
	
	/** List of group membership counts used by calculate the ownership ratio,
	 * it is a map between a group id's (1,2,3) and the count of its (nested) members after applying the filters from the UI
	 * @see {@link MainFileTreeContentProvider#determineOwnershipRatio(com.perdemia.permissionanalyzer.application.util.FilterHelper, com.perdemia.permissionanalyzer.data.FileInfo)} 
	 */
	private static final Map<String, Integer> filteredGroupMembershipCountCache = new HashMap<String, Integer>();
	
	private static final Map<Integer, List<Integer>> memberOf = new HashMap<Integer, List<Integer>>();
	private static final Map<String, List<Member>> filtedGroupMembers = new HashMap<String, List<Member>>();
	private static final Map<String, List<Member>> groupMembers = new HashMap<String, List<Member>>();
	
	/** Cache settings */
	private static final int CACHE_SIZE = 100000;
	
	/**
	 * Clears the List of the cache and forces a refresh from the database whenever the cache is used.
	 */
	public static void clearCache() {
		membershipCache.clear();
		filteredGroupMembershipCountCache.clear();
		groupMembershipCountCache.clear();
		memberOf.clear();
		filtedGroupMembers.clear();
		groupMembers.clear();
	}
	
	public static void clearFilteredMembershipCountCache() {
		filteredGroupMembershipCountCache.clear();
	}
		
	/** Checks if the membership can be found in the {@link #membershipCache} cache. If not, if will check the database and if it's found
	 * the membership will be added to the cache.
	 * 
	 * @param groupId membership's group id
	 * @param memberId membership's member id
	 * 
	 * @return null in case the membership cannot be found in the cache and in the database
	 * 
	 * @throws DBException
	 */
	public static boolean membershipExists (Membership membership) throws DBException {
		boolean exists = membershipCache.contains(membership.getGroupId() + "-" + membership.getMemberId());
		
		if (!exists) {
			// check the database
			// Make sure the index MEMBER_ID is used instead of GROUP_ID. The MEMBER_ID is way more unique than the GROUP_ID, think of the group DOMAIN_USERS.
			// To force this index, we just query on MEMBER_ID. This results in a list of a few groups and further filtering on GROUP_ID is done in the code.
			List<Membership> memberships = Membership.where("MEMBER_ID = ?", new Object[]{membership.getMemberId()});
			Membership existingMembership = null;
			for (Membership nextMembership : memberships) {
				if (nextMembership.getGroupId() == membership.getGroupId()) {
					existingMembership = nextMembership;
					break;
				}
			}
			
			if (existingMembership != null) {
				// found in database
				exists = true;
				cacheMembership(membership);
			}
		}
		
		return exists;
	}
		
	public static void cacheMembership(Membership membership) {
		if (membershipCache.size() > CACHE_SIZE) {
			if (logger.isTraceEnabled()) {
				logger.trace("Membership cache full, clear cache");
			}			
			cleanupCache(membershipCache);			
		}
		membershipCache.add(membership.getGroupId() + "-" + membership.getMemberId());
	}
	
	public static void cacheMembershipCount(String groupIds, int memberCount, boolean isFiltered) {
		if (isFiltered) {
			filteredGroupMembershipCountCache.put(groupIds, memberCount);
		} else {
			groupMembershipCountCache.put(groupIds, memberCount);
		}
	}
	
	public static Integer getMembershipCount(String groupIds, boolean isFiltered) {
		if (isFiltered) {
			return filteredGroupMembershipCountCache.get(groupIds);
		} else {
			return groupMembershipCountCache.get(groupIds);
		}
	}
	
	public static List<Integer> getMemberOf(Member member) throws DBException {
		List<Integer> memberOfIds = memberOf.get(member.getId());
		
		if (memberOfIds == null) {
			List<Member> memberOfList = Member.getMemberOfList(member, null, true, 0, null, 0);
			
			// convert to a list of Integers to limit memory usage and to be able to use the contains(id) of the list
			memberOfIds = new ArrayList<Integer>();
			for (Member group : memberOfList) {
				memberOfIds.add(group.getId());
			}
			
			memberOf.put(member.getId(), memberOfIds);
		}	
		
		return memberOfIds;
	}

	/**
	 * Retrieves the (cached) members of a group considering the configured member filters. If the list of members is not in the cache, then it will be retrieved from the DB
	 * 
	 * @param group
	 * @param includeNestedGroupMemberships
	 * @return
	 * @throws DBException
	 * @see {@link FilerHelper#buildFilterQuery()}
	 */
	public static List<Member> getFiltedGroupMembers(Member group, boolean includeUsers, boolean includeGroups, boolean includeNestedGroupMemberships, boolean allowDuplicateNestedMembers, boolean onlyUseExcludes, FilterHelper filterHelper) throws DBException {
		String cacheKey = group.getId() + String.valueOf(includeUsers) + String.valueOf(includeGroups) + String.valueOf(includeNestedGroupMemberships) + String.valueOf(allowDuplicateNestedMembers);
		List<Member> filteredMembers = filtedGroupMembers.get(cacheKey);
		
		if (filteredMembers == null) {
			String nested = (!includeNestedGroupMemberships ? " AND FILTER_MEMBERSHIP.VIA_GROUP_ID IS NULL" : "");
			String isgroup = (includeUsers != includeGroups ? " AND FILTER_MEMBER.IS_GROUP = " + String.valueOf(includeGroups) : "");
			
			PreparedStatement membersThatMatchFilterPS;
			if (allowDuplicateNestedMembers) {
				membersThatMatchFilterPS = filterHelper.buildFilterQuery(ObjectType.GROUP_MEMBERS, false, onlyUseExcludes, 0, 
						"NAME_LOWERCASE", "FILTER_MEMBERSHIP.GROUP_ID = ? AND FILTER_MEMBERSHIP.GROUP_ID <> FILTER_MEMBERSHIP.MEMBER_ID" + nested + isgroup, group.getId());
			} else { 
				membersThatMatchFilterPS = filterHelper.buildFilterQuery(ObjectType.MEMBER, false, onlyUseExcludes, 0, 
					"NAME_LOWERCASE", "FILTER_MEMBERSHIP.GROUP_ID = ? AND FILTER_MEMBERSHIP.GROUP_ID <> FILTER_MEMBERSHIP.MEMBER_ID" + nested + isgroup, group.getId());
			}
			
			filteredMembers = Member.getByPreparedStatement(membersThatMatchFilterPS);
			filtedGroupMembers.put(cacheKey, filteredMembers);
		}	
		
		return filteredMembers;
	}
	
	public static void clearFiltedGroupMembersCache() {
		filtedGroupMembers.clear();
	}

	/**
	 * Retrieves all the (cached) members of a group. If the list of members is not in the cache, then it will be retrieved from the DB.
	 * 
	 * @param group
	 * @param includeUsers
	 * @param includeGroups
	 * @param includeNestedGroupMemberships
	 * @return
	 * @throws DBException
	 * @see {@link Member#getMembers()}
	 */
	public static List<Member> getGroupMembers(Member group, boolean includeUsers, boolean includeGroups, boolean includeNestedGroupMemberships) throws DBException {
		String cacheKey = group.getId() + String.valueOf(includeUsers) + String.valueOf(includeGroups) + String.valueOf(includeNestedGroupMemberships);
		List<Member> members = groupMembers.get(cacheKey);
		
		if (members == null) {
			// don't use ORDER BY member.name_lowercase because this query might be used with a UNION and Oracle doesn't understand ORDER BY on column name in that case
			members = Member.getMembers(group, null, includeUsers, includeGroups, includeNestedGroupMemberships, 0, "3", 0);
			groupMembers.put(cacheKey, members);
		}
		
		return members;
	}
		
	/**
	 * Cleans up a cache when the maximum number of items has been reached. It removes half of the entries in the List.
	 * 
	 * @param map List to clean
	 */
	private static void cleanupCache(List<String> list) {
		logger.debug("Clean up half of the Membership cache, max size reached: " + CACHE_SIZE);
		
		// clear half of the cache
		int clearCount = (CACHE_SIZE / 2);
		int count = 0;
		Iterator<String> iter = list.iterator();
		
		while (iter.hasNext() && count < clearCount) {
			iter.next();
			iter.remove();
			count++;
		}
	}
}