package com.perdemia.permissionanalyzer.scanner.acl;

import java.io.File;
import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.file.DirectoryStream;
import java.nio.file.FileVisitOption;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.jface.dialogs.MessageDialog;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.model.DirectoryInfo;
import com.perdemia.permissionanalyzer.application.model.DirectoryInfo.SourceType;
import com.perdemia.permissionanalyzer.application.model.StatusListener;
import com.perdemia.permissionanalyzer.application.preferences.PreferenceHelper;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper;
import com.perdemia.permissionanalyzer.application.util.ProgramParameters;
import com.perdemia.permissionanalyzer.application.util.LicenseHelper.Edition;
import com.perdemia.permissionanalyzer.application.util.NetworkShareHelper;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.util.TableCountCache;
import com.perdemia.permissionanalyzer.scanner.groups.LocalGroupHelper;
import com.perdemia.permissionanalyzer.scanner.groups.WinException;

/**
 * Main entry class to start scanning directories and files for their ACL's. This classes processes a List of {@link DirectoryInfo} objects.
 * Depending on the DirectoryInfo settings it will scan local groups, files and inherited permissions. There is a special case here, if the directory
 * path is actually a file, then it will import the ACL information from a file using {@link AclFileVisitor}
 * 
 * @author Reinier
 * @see 
 */
public class AclScanner {
	private StatusListener statusListener;
	private static Logger logger = LoggerFactory.getLogger(AclScanner.class.getName());
	private static String localHostname = null;
	
	/** license limits. The maximum limits are stored as String so that they are obfuscated by Zelix Klassmaster  */
	public static Map<String, String> licenseLimits = new HashMap<String, String>();
		
	static {
		licenseLimits.put("TRIAL_LIMIT_DIRS", "2");
		licenseLimits.put("TRIAL_LIMIT_SUBDIRS", "5");
		licenseLimits.put("TRIAL_LIMIT_SERVERS", "2");
		licenseLimits.put("BASIC_LIMIT_SERVERS", "1");
		licenseLimits.put("STANDARD_LIMIT_SERVERS", "5");
		licenseLimits.put("ADVANCED_LIMIT_SERVERS", "15");
		licenseLimits.put("ENTERPRISE_LIMIT_SERVERS", "2147483647");
		licenseLimits.put("CONSULTANT_LIMIT_SERVERS", "2147483647");
		licenseLimits.put("SCAN_AGENT_LIMIT_SERVERS", "1");
	}
	
	public static int getNumberOfSubDirsForTrial() {
		return Integer.parseInt(licenseLimits.get("TRIAL_LIMIT_SUBDIRS"));
	}
	
	public AclScanner(StatusListener statusListener) {
		this.statusListener = statusListener;
	}

	public void scan(List<DirectoryInfo> directoryList) {
		AclFileVisitor aclFileVisitor = new AclFileVisitor(statusListener, getLocalMachineName(), new JnaAceReader());
		EmcIsilonFileScanner emcIsilonaFileScanner = new EmcIsilonFileScanner(aclFileVisitor);
		PowerShellFileScanner powerShellFileScanner = new PowerShellFileScanner(aclFileVisitor);
		List<String> servers = new ArrayList<String>();
		
		int totalNumberOfDirectories = 0;
		int totalNumberOfFiles = 0;
		int totalNumberOfUsers = 0;
		int totalNumberOfGroups = 0;
		
		// determine the total list of directories (it may contain servers of which we need to find the shares)
		List<DirectoryInfo> allDirectoryList = new ArrayList<DirectoryInfo>();
		for (DirectoryInfo directoryInfo : directoryList) {
			if (directoryInfo.getSourceType().equals(SourceType.SERVER)) {
				if (directoryInfo.isEnabled() || ProgramParameters.isForceScanAllDirectoryContainers() || ProgramParameters.isDirectoryInList(directoryInfo.getPath())) {
					String server = directoryInfo.getPath();
							
					// don't use localhost in the path, because multiple scan agents might scan different servers as 'localhost' into a shared database 
					if (server.equalsIgnoreCase("localhost")) {
						server = AclScanner.getLocalMachineName();
					}
					
					List<String> shares = NetworkShareHelper.getNetworkShares(server, directoryInfo, false);
					
					for (String share : shares) {
						DirectoryInfo dirInfo = new DirectoryInfo();
						dirInfo.setPath("\\\\" + server + "\\" + share);
						dirInfo.setEnabled(true);
						dirInfo.setFollowSymbolicLinks(directoryInfo.isFollowSymbolicLinks());
						dirInfo.setIncludeFiles(directoryInfo.isIncludeFiles());
						dirInfo.setIncludeInheritedPermissions(directoryInfo.isIncludeInheritedPermissions());
						dirInfo.setNumberOfSubdirectories(directoryInfo.getNumberOfSubdirectories());
						dirInfo.setScanLocalGroups(directoryInfo.isScanLocalGroups());
						dirInfo.setSourceType(SourceType.DIRECTORY);
						allDirectoryList.add(dirInfo);
					}
					
					if (shares.isEmpty() && directoryInfo.isScanLocalGroups()) {
						// scan the local group anyway, although no shares have been found
						servers.add(server.toUpperCase());
					}
				}
			} else if ((directoryInfo.getSourceType().equals(SourceType.EMC_ISILON) || directoryInfo.getSourceType().equals(SourceType.POWER_SHELL)) && new File(directoryInfo.getPath()).isDirectory()) {
				// get all text and zip files from this directory
				try {
				DirectoryStream<Path> directoryStream = Files.newDirectoryStream(Paths.get(directoryInfo.getPath()));
				for (Path path : directoryStream) {
	                if (path.toString().toLowerCase().endsWith("txt") || path.toString().toLowerCase().endsWith("zip")) {
	                	DirectoryInfo dirInfo = new DirectoryInfo();
						dirInfo.setPath(path.toString());
						dirInfo.setEnabled(true);
						dirInfo.setScanLocalGroups(false);
						dirInfo.setIncludeFiles(directoryInfo.isIncludeFiles());
						dirInfo.setIncludeInheritedPermissions(directoryInfo.isIncludeInheritedPermissions());
						dirInfo.setSourceType(directoryInfo.getSourceType());
						allDirectoryList.add(dirInfo);
	                }
	            }
				} catch (IOException ex) {
					logger.error("Error while reading txt/zip files for ECM Isilon or PowerShell files", ex);
					SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error reading files", "Error while reading txt/zip files for ECM Isilon or PowerShell files:\r\n\r\n" + ex.getMessage());
				}
			} else {
				
				allDirectoryList.add(directoryInfo);
			}
		}
		
		if (LicenseHelper.getEdition().equals(Edition.TRIAL)) {
			// check if the user edited the preferences on disk, while the trial is limited to 2 directories
			if (allDirectoryList.size() > Integer.parseInt(licenseLimits.get("TRIAL_LIMIT_DIRS"))) {
				SyncMessageDialog.showMessage(MessageDialog.WARNING, "Too many directories", "The trial version is limited to 2 directories, only the first 2 directories will be scanned.");
				logger.warn("Found too many directories to scan in the trial version: " + allDirectoryList.size() + " directories");
				
				// remove all directories except the first two
				while (allDirectoryList.size() > 2) {
					allDirectoryList.remove(allDirectoryList.size() - 1);
				}
			}
			
			for (DirectoryInfo directoryInfo : allDirectoryList) {
				if (directoryInfo.getNumberOfSubdirectories() == 0 || directoryInfo.getNumberOfSubdirectories() > getNumberOfSubDirsForTrial()) {
					SyncMessageDialog.showMessage(MessageDialog.WARNING, "Depth of subdirectories not allowed in trial version", "The trial version has a maximum depth of " + getNumberOfSubDirsForTrial() + " subdirectories, only the first " + getNumberOfSubDirsForTrial() + " levels of directories will be scanned.");
					logger.warn("Depth of subdirectories not allowed in trial version: " + directoryInfo.getNumberOfSubdirectories() + " levels");
					
					directoryInfo.setNumberOfSubdirectories(getNumberOfSubDirsForTrial());
				}
			}
		}
		
		try {
			// scan the local groups and their members of all the selected servers
			for (DirectoryInfo directoryInfo : allDirectoryList) {
				if ((directoryInfo.isEnabled() || ProgramParameters.isForceScanAllDirectoryContainers() || ProgramParameters.isDirectoryInList(directoryInfo.getPath())) && directoryInfo.isScanLocalGroups()) {
					String server;
					if (directoryInfo.getPath().startsWith("\\\\")) { //escaped \\
						int firstBackslash = directoryInfo.getPath().substring(2).indexOf("\\"); //skip the first two backslashes
						if (firstBackslash != -1) {
							server = directoryInfo.getPath().substring(2, firstBackslash + 2); //plus two for the first two backslashes
						} else {
							server = directoryInfo.getPath().substring(2);
						}
					} else {
						server = getLocalMachineName();
					}
					
					server = server.toUpperCase();
					
					if (!servers.contains(server)) {
						servers.add(server);
					}
				}
			}
			
			// check if the number of server exceeds the license limit
			if (servers.size() > Integer.parseInt(licenseLimits.get(LicenseHelper.getEdition().toString() + "_LIMIT_SERVERS"))) {
				throw new Exception(MessageFormat.format("The number of servers ({0}) exceeds the maximum number of servers allowed for this license ({1} servers).", 
						servers.size(), licenseLimits.get(LicenseHelper.getEdition().toString() + "_LIMIT_SERVERS")));
			}
			
			for (String server : servers) {
				try {
					LocalGroupHelper.processLocalGroups(server, statusListener);
					
					totalNumberOfGroups += LocalGroupHelper.getNumberOfGroups();
					totalNumberOfUsers += LocalGroupHelper.getNumberOfUsers();
					
				} catch (DBException | WinException ex) {
					logger.error("Error while scanning local groups on server " + server, ex);
					// error already logged and displayed in the status view on the GUI
				}
			}
			
			// scan all directories
			if (statusListener != null) {
				statusListener.nextStep();
			}
			for (DirectoryInfo directoryInfo : allDirectoryList) {
				if (directoryInfo.isEnabled() || ProgramParameters.isForceScanAllDirectoryContainers() || ProgramParameters.isDirectoryInList(directoryInfo.getPath())) {
					try {
						// reset statistics
						aclFileVisitor.reset();
						aclFileVisitor.setIncludeFiles(directoryInfo.isIncludeFiles());
						aclFileVisitor.setIncludeInheritedPermissions(directoryInfo.isIncludeInheritedPermissions());
						aclFileVisitor.setFollowLinks(directoryInfo.isFollowSymbolicLinks());
												
						// log
						String subDirs = (directoryInfo.getNumberOfSubdirectories() == 0 ? "all" : Integer.toString(directoryInfo.getNumberOfSubdirectories()));
						logger.info(MessageFormat.format("Start scanning directory {0} including {1} levels of subdirectories", directoryInfo.getPath(), subDirs));
						
						Path dirPath = Paths.get(directoryInfo.getPath());
						
						// delete this directory info from the database?
						if (!PreferenceHelper.getInstance().isTruncateData() && !ProgramParameters.isClearData()) {
							// does this directory already exist?
							// use the full path for the first parent
							String fullPath = dirPath.toString();
							if (fullPath.endsWith("\\") || fullPath.endsWith("/")) {
								fullPath = fullPath.substring(0, fullPath.length() - 1);
							}
							
							// Add the server name as suffix in case it a local file, meaning the path doesn't start with
							// '\\' (network share) or '\' a unix import file where the /opt/.. is replaced by Path to \opt\... 
							if (!fullPath.startsWith("\\\\") && !fullPath.startsWith("\\") && getLocalMachineName() != null) {
								fullPath += " (" + getLocalMachineName() + ")";
							}
							
							// we don't use a prepared statement parameter, because the findFirst method will replace a slash to a double slash
							// not sure why :(
							FileInfo existingDir = FileInfo.findFirst("parent_id is null and name_lowercase = '" + fullPath.toLowerCase() + "'");
							
							if (existingDir != null) {
								logger.info("Deleting existing database info for directory " + dirPath);
								DBObject.executeStatement("delete from permissionanalyzer.fileinfo where parent_file_ids like '" + existingDir.getId() + ",%'");
								DBObject.executeStatement("delete from permissionanalyzer.fileinfo where id = " + existingDir.getId());
							}
						}
						
						if (Files.exists(dirPath)) {
							if (directoryInfo.getSourceType().equals(SourceType.POWER_SHELL)) {
								aclFileVisitor.setAceReader(new PowerShellAceReader(powerShellFileScanner));
								powerShellFileScanner.scan(directoryInfo.getPath());
									
							} else if (directoryInfo.getSourceType().equals(SourceType.EMC_ISILON)) {
								aclFileVisitor.setAceReader(new EmcIsilonAceReader(emcIsilonaFileScanner));
								emcIsilonaFileScanner.scan(directoryInfo.getPath());
								
							} else {
								// scan directory and files (if selected)
								aclFileVisitor.setAceReader(new JnaAceReader());
								if (directoryInfo.isFollowSymbolicLinks()) {
									Files.walkFileTree(dirPath, EnumSet.of(FileVisitOption.FOLLOW_LINKS), directoryInfo.getNumberOfSubdirectories() == 0 ? Integer.MAX_VALUE : directoryInfo.getNumberOfSubdirectories(), aclFileVisitor);
								} else {
									Files.walkFileTree(dirPath, EnumSet.noneOf(FileVisitOption.class), directoryInfo.getNumberOfSubdirectories() == 0 ? Integer.MAX_VALUE : directoryInfo.getNumberOfSubdirectories(), aclFileVisitor);
								}
							}
							
							// remember the total number of files/directories scanned
							totalNumberOfDirectories += aclFileVisitor.getNumberOfDirectories();
							totalNumberOfFiles += aclFileVisitor.getNumberOfFiles();
							totalNumberOfGroups += aclFileVisitor.getNumberOfGroups();
							totalNumberOfUsers += aclFileVisitor.getNumberOfUsers();
							
							logger.info(MessageFormat.format("Done scanning directory {0}. Scanned {1} directories, {2} files, {3} groups and {4} users.", directoryInfo.getPath(), aclFileVisitor.getNumberOfDirectories(), aclFileVisitor.getNumberOfFiles(),
									aclFileVisitor.getNumberOfGroups(), aclFileVisitor.getNumberOfUsers()));
							
							if (statusListener != null && statusListener.isCancelled()) {
								break;
							} else {
								if (statusListener != null) {
									statusListener.directoryScanned(directoryInfo, true);
								}
							}
						} else {
							logger.error("Directory not found: " + directoryInfo.getPath());
							
							if (directoryInfo.getSourceType().equals(SourceType.DIRECTORY)) {
								// save the directory name anyway
								FileInfo fileInfo = new FileInfo();
								fileInfo.setName(directoryInfo.getPath());
								fileInfo.setDirectory(true);
								fileInfo.save();
							}
							statusListener.directoryScanned(directoryInfo, false);
						}
					} catch(Exception ex) {
						logger.error("Error while scanning directory " + directoryInfo.getPath(), ex);
						
						if (directoryInfo.getSourceType().equals(SourceType.DIRECTORY)) {
							// save the directory name anyway
							FileInfo fileInfo = new FileInfo();
							fileInfo.setName(directoryInfo.getPath());
							fileInfo.setDirectory(true);
							fileInfo.save();
						}
						
						statusListener.directoryScanned(directoryInfo, false);
					}
				}
			}
			
			logger.info(MessageFormat.format("Scanned a total of {0} directories, {1} files, {2} groups and {3} users.", 
					totalNumberOfDirectories, totalNumberOfFiles, totalNumberOfGroups, totalNumberOfUsers));
			
			if (totalNumberOfDirectories > 0 || totalNumberOfFiles > 0) {
				statusListener.directoriesScanned(totalNumberOfDirectories, totalNumberOfFiles);
			}
			
			// clear the table count cache again, because it might be that the user ran some overviews during the scan
			TableCountCache.clearCache();
			
		} catch (Exception e) {
			logger.error("Error while scanning directories", e);
			// report that all directories have failed
			for (DirectoryInfo directoryInfo : directoryList) {
				if (directoryInfo.isEnabled() || ProgramParameters.isForceScanAllDirectoryContainers()  || ProgramParameters.isDirectoryInList(directoryInfo.getPath())) {
					statusListener.directoryScanned(directoryInfo, false);
				}
			}
		}
	}
	
	public static String getLocalMachineName() {
		if (localHostname == null) {
			try {
				String hostname = InetAddress.getLocalHost().getHostName();
				if (hostname != null && hostname.length() > 0) {
					localHostname = hostname.toUpperCase();
				} else {
					localHostname = "LOCALHOST";
				}
			} catch (UnknownHostException e) {
				logger.warn("Error while getting local machine name: " + e.getMessage());
				localHostname = "LOCALHOST";
			}
		}
		
		return localHostname;
	}
	
	public static void main(String[] args) {
		//ApplicationUtil.openDBConnection();
		
		/*
		AclScanner scanner = new AclScanner(null);
		List<DirectoryInfo> directoryList = new ArrayList<DirectoryInfo>();
		DirectoryInfo directoryInfo = new DirectoryInfo();
		directoryInfo.setNumberOfSubdirectories(0);
		directoryInfo.setPath("C:\\Users\\Reinier\\Test");
		directoryInfo.setEnabled(true);
		directoryList.add(directoryInfo);
		
		scanner.scan(directoryList);
		*/
		
		String test = "computer\\reinier";
		String name = test.substring(test.indexOf("\\") + 1);
		
		// upper case the computer/domain name to be uniform with other parts of the scan
		String computerName = test.substring(0, test.indexOf("\\"));
		System.out.println(computerName.toUpperCase() + "\\" + name);
	}
}
