package com.perdemia.permissionanalyzer.scanner.acl;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.nio.file.Path;
import java.nio.file.attribute.AclEntry;
import java.nio.file.attribute.AclEntryFlag;
import java.nio.file.attribute.AclEntryPermission;
import java.nio.file.attribute.AclEntryType;
import java.nio.file.attribute.AclFileAttributeView;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.data.AclItem;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.Member.GroupType;
import com.perdemia.permissionanalyzer.data.PermissionLookup;
import com.sun.jna.platform.win32.WinNT;

/**
 * This class is deprecated because Java does not check for generic permissions when returning ACL information about a directory using {@link AclFileAttributeView}. Normally you would retrieve ACL info
 * using the following snippet:
 * 
 * <pre>
 * AclFileAttributeView view = Files.getFileAttributeView(dirPath, AclFileAttributeView.class);
 * for (AclEntry aclEntry : view.getAcl()) {
 * 	System.out.println(aclEntry.toString());
 * }
 * </pre>
 * 
 * But unfortunately an empty set of permissions is returned when a directory has one of the generic permission bits enabled (the four high order bits in the access mask). So
 * there are mainly two problems with the getFileAttributeView, first, it uses the old Windows API method getFileSecurity which is not recommended by Microsoft and doesn't
 * determine the inheritance always correct. Second, it is compliant with RFC 3530: Network File System (NFS) version 4 Protocol and therefore doesn't return
 * generic permissions!
 *  
 * There are now three classes to retrieve a List<AclItem> for a given file: 
 * <ol>
 * <li> {@link JavaNioAceReader}: Reflection is used to retrieve the access mask utilizing the JRE classes. After the access mask is retrieved it will check if any generic permission bits are set. Whenever that is the case it will translate the
 * generic permission(s) to the right special permissions and it sets a flag in the {@link AclItem} object that is returned. This method is implemented in {@link #getAclList(Path, boolean)}.
 * The (old) underlying Windows API method used by the JRE is <code>getFileSecurity</code>.
 * <li> {@link JnaAceReader}: The Java Native Access project is used to get the security descriptor of a file using the (recommended) Windows API method <code>GetNamedSecurityInfo</code>. This method
 * is implemented in {@link #getAclListUsingJNA(Path, boolean)}. This method is used when the user want to modify an ACE, in that case this method is called to see where the selected
 * ACE is located in the DACL on the file system. After the ACE index is determined the Windows API method <code>SetNamedSecurityInfo</code> is used to modify the ACE. This will propagate
 * the changes automatically to the underlying child folders/files. Which is not done by the low-level <code>SetFileSecurity</code> used by the JRE's setAcl method!!
 * <li> {@link CustomAceReader}: A special handler to import a file with ACE data.
 * </ol>
 * 
 * This first class performs slightly better then the second one. I guess this has something to do with a slightly different SID lookup in both methods and/or the different Windows API methods.
 * The first class scans 35.000 folders in 1:00 minute, the second class takes 1:20 minutes. BUT, the second class scanned the inheritance better for some of the files. The first class
 * marked all the ACE's in <code>C:\Program Files\Microsoft Office 15\packagesunrisepolicies</code> as explicit, while most of the ACE's are inherited. This was done correctly by the second class.
 * 
 * @author Reinier
 * @deprecated use JnaAceReader
 * @see http://www.codeproject.com/Articles/10042/The-Windows-Access-Control-Model-Part
 */
public final class JavaNioAceReader extends AceReader {
	private final static Logger logger = LoggerFactory.getLogger(JavaNioAceReader.class.getName());
	
	/** Bits for permission flags */
	private static final short OFFSETOF_ACCESS_FLAGS = 1;
	private static final short ACCESS_FLAGS_INHERITED_ACE = 16;
	
	/** Bits for generic permissions in the access mask */
	private static final short OFFSETOF_ACCESS_MASK = 4;
	public static final long GENERIC_ALL_BIT = 268435456; // 2^28;
	public static final long GENERIC_EXECUTE_BIT = 536870912; //2^29;
	public static final long GENERIC_WRITE_BIT = 1073741824; //2^30;
	public static final long GENERIC_READ_BIT = 2147483648L; //2^31; (the constant WinNT.GENERIC_READ is using Integer.MIN_VALUE (-2147483648), but that doesn't seem to work!
	
	/** Corresponding permissions for each generic permission */
	private static final List<Integer> GENERIC_ALL_PERMISSIONS = new ArrayList<Integer>(Arrays.asList(new Integer[]{0,1,2,3,4,5,6,7,8,9,10,11,12})); 
	private static final List<Integer> GENERIC_EXECUTE_PERMISSIONS = new ArrayList<Integer>(Arrays.asList(new Integer[]{5,7,10}));
	private static final List<Integer> GENERIC_WRITE_PERMISSIONS = new ArrayList<Integer>(Arrays.asList(new Integer[]{1,2,4,8,10}));
	private static final List<Integer> GENERIC_READ_PERMISSIONS = new ArrayList<Integer>(Arrays.asList(new Integer[]{0,3,7,10}));
		
	/** Reflection fields */
	private static Field sidStringField;
	private static Field theUnsafe;
	private static Field aceCount;
	
	/** Reflection methods */
	private static Method decode;
	private static Method ConvertStringSidToSid;
	private static Method toWindowsPath;
	private static Method getFinalPath;
	private static Method getFileSecurity;
	private static Method address;
	private static Method GetSecurityDescriptorDacl;
	private static Method GetAce;
	private static Method ConvertSidToStringSid;
	private static Method getInt;
	private static Method getByte;
	private static Method release;
	private static Method GetAclInformation;
	
	/**
	 * Initialize fields and methods of the JDK protected packages using reflection
	 */
	static {
		try {
			//Fields
			sidStringField = Class.forName("sun.nio.fs.WindowsUserPrincipals$User").getDeclaredField("sidString");
			sidStringField.setAccessible(true);
			theUnsafe = Class.forName("sun.misc.Unsafe").getDeclaredField("theUnsafe");
			theUnsafe.setAccessible(true);
			aceCount = Class.forName("sun.nio.fs.WindowsNativeDispatcher$AclInformation").getDeclaredField("aceCount");
			aceCount.setAccessible(true);
			
			//Methods
			decode = Class.forName("sun.nio.fs.WindowsSecurityDescriptor").getDeclaredMethod("decode", Long.TYPE);
			decode.setAccessible(true);
			ConvertStringSidToSid = Class.forName("sun.nio.fs.WindowsNativeDispatcher").getDeclaredMethod("ConvertStringSidToSid", String.class);
			ConvertStringSidToSid.setAccessible(true);
			toWindowsPath = Class.forName("sun.nio.fs.WindowsPath").getDeclaredMethod("toWindowsPath", Path.class);
			toWindowsPath.setAccessible(true);
			getFinalPath = Class.forName("sun.nio.fs.WindowsLinkSupport").getDeclaredMethod("getFinalPath", Class.forName("sun.nio.fs.WindowsPath"), Boolean.TYPE);
			getFinalPath.setAccessible(true);
			getFileSecurity = Class.forName("sun.nio.fs.WindowsAclFileAttributeView").getDeclaredMethod("getFileSecurity", String.class, Integer.TYPE);
			getFileSecurity.setAccessible(true);
			address = Class.forName("sun.nio.fs.NativeBuffer").getDeclaredMethod("address");
			address.setAccessible(true);
			GetSecurityDescriptorDacl = Class.forName("sun.nio.fs.WindowsNativeDispatcher").getDeclaredMethod("GetSecurityDescriptorDacl", Long.TYPE);
			GetSecurityDescriptorDacl.setAccessible(true);
			GetAce = Class.forName("sun.nio.fs.WindowsNativeDispatcher").getDeclaredMethod("GetAce", Long.TYPE, Integer.TYPE);
			GetAce.setAccessible(true);
			ConvertSidToStringSid = Class.forName("sun.nio.fs.WindowsNativeDispatcher").getDeclaredMethod("ConvertSidToStringSid", Long.TYPE);
			ConvertSidToStringSid.setAccessible(true);
			getInt = Class.forName("sun.misc.Unsafe").getDeclaredMethod("getInt", Long.TYPE);
			getInt.setAccessible(true);
			getByte = Class.forName("sun.misc.Unsafe").getDeclaredMethod("getByte", Long.TYPE);
			getByte.setAccessible(true);
			release = Class.forName("sun.nio.fs.NativeBuffer").getDeclaredMethod("release");
			release.setAccessible(true);
			GetAclInformation = Class.forName("sun.nio.fs.WindowsNativeDispatcher").getDeclaredMethod("GetAclInformation", Long.TYPE);
			GetAclInformation.setAccessible(true);
				
		} catch (Exception ex) {
			logger.warn("Exception while initializing NativeAccessController: " + ex.getMessage(), ex);
		}
	}
	
	/**
	 * <p>Gets the access mask of each ACE in the given Path and checks if it has one of the generic permissions bits enabled. If so, it finds the corresponding {@link AclItem} and adds
	 * the special permissions represented by the generic permission(s). 
	 * </p><p>
	 * Because the complete SecurityDescriptor is retrieved of the Path, the method will loop through all ACE's and accepts a list of AclItems as parameter. The SecurityDescriptor of the file
	 * is retrieved using Reflection since Java doesn't return the access mask bits using the standard {@link AclFileAttributeView}!
	 * </p>
	 * @param aclItems
	 * @param path
	 * @param followLinks
	 * 
	 * @deprecated replaced by the method {@link #getAclListUsingJNA(Path, boolean)} because it returns inheritance better in some of the cases and it has better support for modifying ACE's.
	 */
	@Override
	public List<AclItem> getAclList(Path path, boolean followLinks, FileInfo file, WinNT.SECURITY_DESCRIPTOR_RELATIVE secDescriptor) {
		List<AclItem> aclItems = new ArrayList<AclItem>();
		
		Object nativeBuffer = null;
		
		if (logger.isTraceEnabled()) {
			logger.trace("Retrieving ACL information for path: " + path.toString());
		}
			
		try {
			// convert to a Windows path and call the getFinalPath which checks for 'followSymbolicLinks'
			Object windowsPath = toWindowsPath.invoke(null, path);
			String finalPath = (String) getFinalPath.invoke(null, windowsPath, followLinks);
			
			// get the file security
			nativeBuffer = getFileSecurity.invoke(null, finalPath, 4);
			long nativeBufferAddress = (long) address.invoke(nativeBuffer);
			
			// get the DACL and the number of ACE's
			long securityDescriptorDaclAddress = (long) GetSecurityDescriptorDacl.invoke(null, nativeBufferAddress);
			
			int aceCountValue = 0;
			if (securityDescriptorDaclAddress != 0) {
				Object aclInfo = GetAclInformation.invoke(null, securityDescriptorDaclAddress);
				aceCountValue = (int) aceCount.get(aclInfo);
			} else {
				// this address might be null for some files on the root C:\ for example, check for this to prevent a CoreDump and AccessViolation
				throw new Exception("Could not read ACL information of file " + path.toString());
			}
			
			// loop through all the ACE's, decode them, convert them to AclItems and add generic permissions if applicable
			for(int counter = 0; counter < aceCountValue; counter++)
	        {
	            try {
	            	long aceAddress = (long) GetAce.invoke(null, securityDescriptorDaclAddress, counter);
	            	
	    			AclEntry decodedAclItem = (AclEntry) decode.invoke(null, aceAddress);	    			
	    			Object unsafe = theUnsafe.get(null);
	    			int mask = (int) getInt.invoke(unsafe, aceAddress + OFFSETOF_ACCESS_MASK);
	    			byte flags = (byte) getByte.invoke(unsafe, aceAddress + OFFSETOF_ACCESS_FLAGS);
	    			    				    				
    				// convert to a AclItem which can be saved in the database
    				AclItem aclItem = buildAclItem(decodedAclItem, mask, path);
        											
					// inherited ace flag (not provided by the JRE)
    				aclItem.setFlagInheritedAce((flags & ACCESS_FLAGS_INHERITED_ACE) > 0); 
    				aclItem.setDirectory(path.toFile().isDirectory());
    				
					if (logger.isTraceEnabled()) {
						logger.trace("Retrieved ACE information, new AclItem: " + aclItem.toString());
					}
					
					aclItems.add(aclItem);
	            } catch (Exception ex) {
	            	logger.error("Error while getting generic permissions for path " + path.toString() + " and ACE index " + counter, ex);
	            }
	        }
		} catch(Throwable ex) {
			logger.error("Error while reading permissions for path " + path.toString(), ex);
		} finally {
			try {
				if (nativeBuffer != null) {
					release.invoke(nativeBuffer);
				}
			} catch (Exception ex){}
		}
		
		return aclItems;
	}
	
	/**
	 * Creates a new {@link AclItem} object based on a {@link AclEntry} object. The given object is a JRE typed object given by the {@link AclFileAttributeView} view and the
	 * type that is returned a {@link DBObject} object which can be saved in the database.
	 *   
	 * @param aclEntry ACE information given by the JRE
	 * @param mask access mask of the ACE used to determine the GENERIC permissions which are not returned by the JRE
	 * 
	 * @return ACE information in a AceItem object ready to be saved in the database
	 * 
	 * @deprecated used by the deprecated method {@link #getAclList(Path, boolean)} and is being replaced by {@link #buildAclItem(com.sun.jna.platform.win32.WinNT.ACCESS_ACEStructure, Path)}.
	 */
	private AclItem buildAclItem(AclEntry aclEntry, int mask, Path path) {
		// create a new AclItem
		AclItem aclItem = new AclItem();
				
		Member member = new Member();
		member.setDisplayName(member.getName(false));
		aclItem.setMember(member);
		
		// user or group?
		String sid = null;
		try {
			if (aclEntry.principal().getClass().getSimpleName().equals("Group")) {
				member.setGroupType(GroupType.COMPUTER_LOCAL);
				member.setGroup(true);
			}
			
			sid = String.valueOf(sidStringField.get(aclEntry.principal())).toUpperCase();
			
			// we now check for BUILT-IN accounts. We convert those account names to <server>\\Administrators instead of BUILT-IN\Administrators.
			member.setName(aclEntry.principal().getName(), sid, null, getMachineNameFromPath(path.toString()));
									
		} catch (IllegalArgumentException | IllegalAccessException e) {
			logger.warn("Could not determine SID for member " + aclEntry.principal().getName() + ": " + e.getMessage());
		}
		
		// no built-in group, or failed determining if it was a built-in group?
		if (member.getName(false).length() == 0) {
			String name = aclEntry.principal().getName();
			if (name.startsWith("\\")) {
				name = name.substring(1);
			}
			member.setName(name);
		}
							
		// flags
		aclItem.setDeny(aclEntry.type().equals(AclEntryType.DENY));
		
		for (AclEntryFlag flag : aclEntry.flags()) {
			if (flag.equals(AclEntryFlag.DIRECTORY_INHERIT)) {
				aclItem.setFlagDirectoryInherit(true);
			}
			if (flag.equals(AclEntryFlag.FILE_INHERIT)) {
				aclItem.setFlagFileInherit(true);
			}
			if (flag.equals(AclEntryFlag.INHERIT_ONLY)) {
				aclItem.setFlagInheritOnly(true);
			}
			if (flag.equals(AclEntryFlag.NO_PROPAGATE_INHERIT)) {
				aclItem.setFlagNoPropagateInherit(true);
			}
		}
													
		// Determine the permission value (only unique items using Set, see next part generic permissions)
		Set<Integer> permissionOrdinals = new HashSet<Integer>();
		for (AclEntryPermission aclPermission : aclEntry.permissions()) {
			// skip permission Synchronize
			if (!aclPermission.equals(AclEntryPermission.SYNCHRONIZE)) {
				permissionOrdinals.add(aclPermission.ordinal());
			}
		}
		
		// generic permissions?
		boolean genericAll = ((mask & GENERIC_ALL_BIT) > 0);
		boolean genericExecute = ((mask & GENERIC_EXECUTE_BIT) > 0);
		boolean genericWrite = ((mask & GENERIC_WRITE_BIT) > 0);
		boolean genericRead = ((mask & GENERIC_READ_BIT) > 0);
		
		// add the generic permissions (again only unique items using Set)
		if (genericAll) {
			aclItem.setGenericAll(true);
			permissionOrdinals.addAll(GENERIC_ALL_PERMISSIONS);
		}
		if (genericExecute) {
			aclItem.setGenericExecute(true);
			permissionOrdinals.addAll(GENERIC_EXECUTE_PERMISSIONS);
		}
		if (genericWrite) {
			aclItem.setGenericWrite(true);
			permissionOrdinals.addAll(GENERIC_WRITE_PERMISSIONS);
		}
		if (genericRead) {
			aclItem.setGenericRead(true);
			permissionOrdinals.addAll(GENERIC_READ_PERMISSIONS);
		}
		
		aclItem.setPermissionId(PermissionLookup.getIdBySpecialPermissions(permissionOrdinals));
				
		return aclItem;
	}

	@Override
	public boolean isCurrentFileDaclProtected() {
		throw new UnsupportedOperationException();
	}
}
