package com.perdemia.permissionanalyzer.scanner.acl;

import java.nio.file.AccessDeniedException;
import java.nio.file.Path;
import java.nio.file.attribute.AclEntry;
import java.nio.file.attribute.AclFileAttributeView;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.data.AclItem;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.Member.GroupType;
import com.perdemia.permissionanalyzer.data.PermissionLookup;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SpecialPermission;
import com.sun.jna.platform.win32.Advapi32Util;
import com.sun.jna.platform.win32.WinNT;
import com.sun.jna.platform.win32.WinNT.ACCESS_ACEStructure;

/**
 * This class is written because Java does not check for generic permissions when returning ACL information about a directory using {@link AclFileAttributeView}. Normally you would retrieve ACL info
 * using the following snippet:
 * 
 * <pre>
 * AclFileAttributeView view = Files.getFileAttributeView(dirPath, AclFileAttributeView.class);
 * for (AclEntry aclEntry : view.getAcl()) {
 * 	System.out.println(aclEntry.toString());
 * }
 * </pre>
 * 
 * But unfortunately an empty set of permissions is returned when a directory has one of the generic permission bits enabled (the four high order bits in the access mask). So
 * there are mainly two problems with the getFileAttributeView, first, it uses the old Windows API method getFileSecurity which is not recommended by Microsoft and doesn't
 * determine the inheritance always correct. Second, it is compliant with RFC 3530: Network File System (NFS) version 4 Protocol and therefore doesn't return
 * generic permissions!
 *  
 * There are now three classes to retrieve a List<AclItem> for a given file:  
 * <ol>
 * <li> {@link JavaNioAceReader}: Reflection is used to retrieve the access mask utilizing the JRE classes. After the access mask is retrieved it will check if any generic permission bits are set. Whenever that is the case it will translate the
 * generic permission(s) to the right special permissions and it sets a flag in the {@link AclItem} object that is returned. This method is implemented in {@link #getAclList(Path, boolean)}.
 * The (old) underlying Windows API method used by the JRE is <code>getFileSecurity</code>.
 * <li> {@link JnaAceReader}: The Java Native Access project is used to get the security descriptor of a file using the (recommended) Windows API method <code>GetNamedSecurityInfo</code>. This method
 * is implemented in {@link #getAclListUsingJNA(Path, boolean)}. This method is used when the user want to modify an ACE, in that case this method is called to see where the selected
 * ACE is located in the DACL on the file system. After the ACE index is determined the Windows API method <code>SetNamedSecurityInfo</code> is used to modify the ACE. This will propagate
 * the changes automatically to the underlying child folders/files. Which is not done by the low-level <code>SetFileSecurity</code> used by the JRE's setAcl method!!
 * <li> {@link CustomAceReader}: A special handler to import a file with ACE data.
 * </ol>
 * 
 * This first class performs slightly better then the second one. I guess this has something to do with a slightly different SID lookup in both methods and/or the different Windows API methods.
 * The first class scans 35.000 folders in 1:00 minute, the second class takes 1:20 minutes. BUT, the second class scanned the inheritance better for some of the files. The first class
 * marked all the ACE's in <code>C:\Program Files\Microsoft Office 15\packagesunrisepolicies</code> as explicit, while most of the ACE's are inherited. This was done correctly by the second class.
 * 
 * @author Reinier
 * @see http://www.codeproject.com/Articles/10042/The-Windows-Access-Control-Model-Part
 */
public final class JnaAceReader extends AceReader {
	private final static Logger logger = LoggerFactory.getLogger(JnaAceReader.class.getName());
	
	/** Bits for generic permissions in the access mask */
	public static final long GENERIC_ALL_BIT = 268435456; // 2^28;
	public static final long GENERIC_EXECUTE_BIT = 536870912; //2^29;
	public static final long GENERIC_WRITE_BIT = 1073741824; //2^30;
	public static final long GENERIC_READ_BIT = 2147483648L; //2^31; (the constant WinNT.GENERIC_READ is using Integer.MIN_VALUE (-2147483648), but that doesn't seem to work!
	
	private static boolean sidResolveErrorDisplayed = false;
		
	/**
	 * <p>Gets the access mask of each ACE in the given Path and checks if it has one of the generic permissions bits enabled. If so, it finds the corresponding {@link AclItem} and adds
	 * the special permissions represented by the generic permission(s). 
	 * </p><p>
	 * Because the complete SecurityDescriptor is retrieved of the Path, the method will loop through all ACE's and accepts a list of AclItems as parameter. The SecurityDescriptor of the file
	 * is retrieved using Reflection since Java doesn't return the access mask bits using the standard {@link AclFileAttributeView}!
	 * </p>
	 * @param aclItems
	 * @param path
	 * @param followLinks
	 * @param file optional FileInfo object that is being updated with the SE_DACL_PROTECTED flag from the path's security descriptor
	 */
	@Override
	public List<AclItem> getAclList(Path path, boolean followLinks, FileInfo file, WinNT.SECURITY_DESCRIPTOR_RELATIVE secDescriptor) throws AccessDeniedException {
		
		// DACL
		WinNT.ACL dacl = secDescriptor.getDiscretionaryACL();
		
		if (dacl == null) {
			throw new AccessDeniedException("Error while processing path " + path.toFile().getAbsolutePath() + ": Access denied");
		}

		// List of ACE's
		//WinNT.ACCESS_ACEStructure[] aces = dacl.getACEStructures();
		WinNT.ACE_HEADER[] aceHeaders = dacl.getACEs();
				
		// List of AclItems for the DB (mapped from the actual directory ACE's)
		List<AclItem> aclItems = new ArrayList<AclItem>();
		
		if (logger.isTraceEnabled()) {
			logger.trace("Retrieving ACL information for path: " + path.toString());
		}
				
		for (int counter = 0; counter < aceHeaders.length; counter++) {
	    	try {		    				    				
	    		WinNT.ACE_HEADER header = aceHeaders[counter];
	    		 if(header instanceof ACCESS_ACEStructure) {
                     ACCESS_ACEStructure ace = (ACCESS_ACEStructure) header;
	    		
					// convert to a AclItem which can be saved in the database
					AclItem aclItem = buildAclItem(ace, path);
					if (file != null) {
						aclItem.setDirectory(file.isDirectory()); //copy flag, see ModifyACEHandler.changeAclItem
					}
					
					if (logger.isTraceEnabled()) {
						logger.trace("Retrieved ACE information, new AclItem: " + aclItem.toString());
					}
					
					aclItems.add(aclItem);
	    		 } else {
	    			 logger.trace("ACE header type not recognized: " + header.AceType);
	    		 }
            } catch (Exception ex) {
            	logger.error("Error while getting generic permissions for path " + path.toString() + " and ACE index " + counter + "", ex);
            }
		}
		
		return aclItems;
	}
	
	/**
	 * Creates a new {@link AclItem} object based on a {@link AclEntry} object. The given object is a JRE typed object given by the {@link AclFileAttributeView} view and the
	 * type that is returned a {@link DBObject} object which can be saved in the database.
	 *   
	 * @param aclEntry ACE information given by the JRE
	 * @param mask access mask of the ACE used to determine the GENERIC permissions which are not returned by the JRE
	 * 
	 * @return ACE information in a AceItem object ready to be saved in the database
	 */
	private AclItem buildAclItem(WinNT.ACCESS_ACEStructure ace, Path path) {
		// create a new AclItem
		AclItem aclItem = new AclItem();
				
		Member member = new Member();
		member.setDisplayName(member.getName(false));
		aclItem.setMember(member);
		
		// get the Account info based on the ACE SID
		try {
			Advapi32Util.Account account = Advapi32Util.getAccountBySid(ace.getSID());
			
			if (account.accountType == WinNT.SID_NAME_USE.SidTypeWellKnownGroup || 
				account.accountType == WinNT.SID_NAME_USE.SidTypeGroup || 
				account.accountType == WinNT.SID_NAME_USE.SidTypeAlias) {
				member.setGroupType(GroupType.GLOBAL); //assumption, we don't know at this point
				member.setGroup(true);
			}
			
			// we now check for BUILT-IN accounts. We convert those account names to <server>\\Administrators instead of BUILT-IN\Administrators.
			member.setName(account.name, account.sidString, account.domain, getMachineNameFromPath(path.toString()));
			
		} catch (Exception ex) {
			if (!sidResolveErrorDisplayed) {
				logger.warn("Could not translate SID " + ace.getSID().getSidString() + " to account name on file/folder " + path.toString(), ex);
				sidResolveErrorDisplayed = true;
			} else {
				logger.debug("Could not translate SID " + ace.getSID().getSidString() + " to account name on file/folder " + path.toString(), ex);
			}
			member.setName(ace.getSidString());
		}
							
		// flags
		aclItem.setDeny(ace instanceof WinNT.ACCESS_DENIED_ACE);
		aclItem.setFlagDirectoryInherit((ace.AceFlags & WinNT.CONTAINER_INHERIT_ACE) > 0);
		aclItem.setFlagFileInherit((ace.AceFlags & WinNT.OBJECT_INHERIT_ACE) > 0);
		aclItem.setFlagInheritOnly((ace.AceFlags & WinNT.INHERIT_ONLY_ACE) > 0);
		aclItem.setFlagNoPropagateInherit((ace.AceFlags & WinNT.NO_PROPAGATE_INHERIT_ACE) > 0);
		aclItem.setFlagInheritedAce((ace.AceFlags & WinNT.INHERITED_ACE) > 0);
														
		// Determine the permission value (only unique items using Set, see next part generic permissions)
		Set<Integer> permissionOrdinals = new HashSet<Integer>();
		for (SpecialPermission specialPermission : SpecialPermission.values()) {
			// skip permission Synchronize
			if (!specialPermission.equals(SpecialPermission.SYNCHRONIZE)) {
				if ((specialPermission.getBitValue() & ace.Mask) > 0) {
					permissionOrdinals.add(specialPermission.ordinal());
				}
			}
		}
		
		// generic permissions?
		boolean genericAll = ((ace.Mask & GENERIC_ALL_BIT) > 0);
		boolean genericExecute = ((ace.Mask & GENERIC_EXECUTE_BIT) > 0);
		boolean genericWrite = ((ace.Mask & GENERIC_WRITE_BIT) > 0);
		boolean genericRead = ((ace.Mask & GENERIC_READ_BIT) > 0);
		
		// add the generic permissions (again only unique items using Set)
		if (genericAll) {
			aclItem.setGenericAll(true);
			permissionOrdinals.addAll(GENERIC_ALL_PERMISSIONS);
		}
		if (genericExecute) {
			aclItem.setGenericExecute(true);
			permissionOrdinals.addAll(GENERIC_EXECUTE_PERMISSIONS);
		}
		if (genericWrite) {
			aclItem.setGenericWrite(true);
			permissionOrdinals.addAll(GENERIC_WRITE_PERMISSIONS);
		}
		if (genericRead) {
			aclItem.setGenericRead(true);
			permissionOrdinals.addAll(GENERIC_READ_PERMISSIONS);
		}
		
		aclItem.setPermissionId(PermissionLookup.getIdBySpecialPermissions(permissionOrdinals));
				
		return aclItem;
	}

	@Override
	public boolean isCurrentFileDaclProtected() {
		throw new UnsupportedOperationException();
	}
}
