package com.perdemia.permissionanalyzer.scanner.acl;

import java.nio.file.AccessDeniedException;
import java.nio.file.Path;
import java.nio.file.attribute.AclFileAttributeView;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.jface.dialogs.MessageDialog;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.data.AclItem;
import com.perdemia.permissionanalyzer.data.FileInfo;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.PermissionLookup;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SimplePermission;
import com.perdemia.permissionanalyzer.data.PermissionLookup.SpecialPermission;
import com.sun.jna.platform.win32.WinNT;

/**
 * This class is written to read the ACE data from a text file which can be imported in the application.
 * 
 * @author Reinier
 * @see @{link JnaAceReader}
 * @see {@link JavaNioAceReader}
 * @see {@link PowerShellAceReader}
 */
public final class PowerShellAceReader extends AceReader {
	private final static Logger logger = LoggerFactory.getLogger(PowerShellAceReader.class.getName());
	private PowerShellFileScanner importScanner;
	
	public PowerShellAceReader(PowerShellFileScanner importScanner) {
		this.importScanner = importScanner;
	}
	
	/**
	 * <p>Gets the access mask of each ACE in the given Path and checks if it has one of the generic permissions bits enabled. If so, it finds the corresponding {@link AclItem} and adds
	 * the special permissions represented by the generic permission(s). 
	 * </p><p>
	 * Because the complete SecurityDescriptor is retrieved of the Path, the method will loop through all ACE's and accepts a list of AclItems as parameter. The SecurityDescriptor of the file
	 * is retrieved using Reflection since Java doesn't return the access mask bits using the standard {@link AclFileAttributeView}!
	 * </p>
	 * @param aclItems
	 * @param path
	 * @param followLinks
	 * @param file optional FileInfo object that is being updated with the SE_DACL_PROTECTED flag from the path's security descriptor
	 */
	public List<AclItem> getAclList(Path path, boolean followLinks, FileInfo file, WinNT.SECURITY_DESCRIPTOR_RELATIVE secDescriptor) throws AccessDeniedException {
		// List of ACE's from the import file (the ones that belong to the current file)
		List<String> aces = importScanner.getAces();
		
		// List of AclItems for the DB
		List<AclItem> aclItems = new ArrayList<AclItem>();
				
		for (String ace : aces) {
			AclItem aclItem = new AclItem();
			aclItem.setMember(new Member());
			
			try {				
				// remove the prefix Access : 
				ace = ace.substring(9);
				
				if (!ace.isEmpty()) {
					String[] aceFields = ace.split(",");
					
					// Determine the permission value (only unique items using Set, see next part generic permissions)
					Set<Integer> permissionOrdinals = new HashSet<Integer>();
										
					for (int counter = 0; counter < aceFields.length; counter++) {
						String aceField = aceFields[counter].trim();
						switch(aceField) {
							case "Deny":
								aclItem.setDeny(true); break;
							case "Allow":
								aclItem.setDeny(false); break;
								
							// ACE INHERITED
							case "True":
								aclItem.setFlagInheritedAce(true); break;
							case "False":
								aclItem.setFlagInheritedAce(false); break;
								
							// NOT SUPPORTED IN THE SAMPLE POWERSHELL SCRIPT, BUT JUST IN CASE:
							case "User":
								aclItem.getMember(false).setGroup(false);
							case "Group":
								aclItem.getMember(false).setGroup(true);
								
							// FLAGS
							case "ContainerInherit":
								aclItem.setFlagDirectoryInherit(true); break;
							case "ObjectInherit":
								aclItem.setFlagFileInherit(true); break;
							case "InheritOnly":
								aclItem.setFlagInheritOnly(true); break;
							case "NoPropagateInherit":
								aclItem.setFlagNoPropagateInherit(true); break;
								
							// SIMPLE PERMISSIONS
							case "Read":
								permissionOrdinals.addAll(Arrays.asList(PermissionLookup.permissionMappings.get(SimplePermission.READ)));
								break;
							case "ReadAndExecute":
								permissionOrdinals.addAll(Arrays.asList(PermissionLookup.permissionMappings.get(SimplePermission.READ_EXECUTE)));
								break;
							case "ListDirectory":
								permissionOrdinals.addAll(Arrays.asList(PermissionLookup.permissionMappings.get(SimplePermission.LIST_FOLDER_CONTENT)));
								break;
							case "Write":
								permissionOrdinals.addAll(Arrays.asList(PermissionLookup.permissionMappings.get(SimplePermission.WRITE)));
								break;
							case "Modify":
								permissionOrdinals.addAll(Arrays.asList(PermissionLookup.permissionMappings.get(SimplePermission.MODIFY)));
								break;
							case "FullControl":
								permissionOrdinals.addAll(Arrays.asList(PermissionLookup.permissionMappings.get(SimplePermission.FULL)));
								break;
								
							// SPECIAL PERMISSIONS
							case "AppendData":
								permissionOrdinals.add(SpecialPermission.APPEND_DATA.ordinal());
								break;
							case "ChangePermissions":
								permissionOrdinals.add(SpecialPermission.WRITE_ACL.ordinal());
								break;
							case "CreateDirectories":
								permissionOrdinals.add(SpecialPermission.APPEND_DATA.ordinal());
								break;
							case "CreateFiles":
								permissionOrdinals.add(SpecialPermission.WRITE_DATA.ordinal());
								break;
							case "Delete":
								permissionOrdinals.add(SpecialPermission.DELETE.ordinal());
								break;
							case "DeleteSubdirectoriesAndFiles":
								permissionOrdinals.add(SpecialPermission.DELETE_CHILD.ordinal());
								break;
							case "ExecuteFile":
								permissionOrdinals.add(SpecialPermission.EXECUTE.ordinal());
								break;
							case "ReadAttributes":
								permissionOrdinals.add(SpecialPermission.READ_ATTRIBUTES.ordinal());
								break;
							case "ReadData":
								permissionOrdinals.add(SpecialPermission.READ_DATA.ordinal());
								break;
							case "ReadExtendedAttributes":
								permissionOrdinals.add(SpecialPermission.READ_NAMED_ATTRS.ordinal());
								break;
							case "ReadPermissions":
								permissionOrdinals.add(SpecialPermission.READ_ACL.ordinal());
								break;
							case "TakeOwnership":
								permissionOrdinals.add(SpecialPermission.WRITE_OWNER.ordinal());
								break;
							case "Traverse":
								permissionOrdinals.add(SpecialPermission.EXECUTE.ordinal());
								break;
							case "WriteAttributes":
								permissionOrdinals.add(SpecialPermission.WRITE_ATTRIBUTES.ordinal());
								break;
							case "WriteData":
								permissionOrdinals.add(SpecialPermission.WRITE_DATA.ordinal());
								break;
							case "WriteExtendedAttributes":
								permissionOrdinals.add(SpecialPermission.WRITE_NAMED_ATTRS.ordinal());
								break;
							case "Synchronize":
								break; // no action
								
							// GENERIC PERMISSIONS
							case "268435456": case "-268435456":
								aclItem.setGenericAll(true);
								permissionOrdinals.addAll(GENERIC_ALL_PERMISSIONS);
								break;
							case "1610612736": case "-1610612736": 
								aclItem.setGenericExecute(true);
								aclItem.setGenericRead(true);
								permissionOrdinals.addAll(GENERIC_READ_PERMISSIONS);
								permissionOrdinals.addAll(GENERIC_EXECUTE_PERMISSIONS);
								break;
							case "2147483648": case "-2147483648":
								aclItem.setGenericRead(true);
								permissionOrdinals.addAll(GENERIC_READ_PERMISSIONS);
							case "536870912": case "-536870912":
								aclItem.setGenericExecute(true);
								permissionOrdinals.addAll(GENERIC_EXECUTE_PERMISSIONS);
							case "1073741824": case "-1073741824":
								aclItem.setGenericRead(true);
								permissionOrdinals.addAll(GENERIC_WRITE_PERMISSIONS);
							case "536805376": case "-536805376":
								aclItem.setGenericExecute(true);
								aclItem.setGenericRead(true);
								aclItem.setGenericWrite(true);
								permissionOrdinals.addAll(GENERIC_READ_PERMISSIONS);
								permissionOrdinals.addAll(GENERIC_EXECUTE_PERMISSIONS);
								permissionOrdinals.addAll(GENERIC_WRITE_PERMISSIONS);
								break;
							
							case "None":
								break; // no action
								
							default:
								// ACCOUNT NAME
								if (counter == 0) {
									aclItem.getMember(false).setName(aceField);
									break;
								} else {
									logger.warn("Unknown ACE attribute found: " + aceField);
								}
						}
					}
					
					aclItem.setPermissionId(PermissionLookup.getIdBySpecialPermissions(permissionOrdinals));
										
					if (file != null) {
						aclItem.setDirectory(file.isDirectory());
					}
					
					aclItems.add(aclItem);
					if (logger.isTraceEnabled()) {
						logger.trace("Retrieved ACE information, new AclItem: " + aclItem.toString());
					}
				}
			} catch(Exception ex) {
				logger.error("Error while processing ace " + ace, ex);
				SyncMessageDialog.showMessage(MessageDialog.ERROR, "Error", "Error while processing ace " + ace + ": " + ex.getMessage());
			}
		}
				
		return aclItems;
	}
	
	@Override
	public boolean isCurrentFileDaclProtected() {
		return false;
	}
}
