package com.perdemia.permissionanalyzer.scanner.groups;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.text.MessageFormat;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.model.StatusListener;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.Member.GroupType;
import com.perdemia.permissionanalyzer.data.Membership;
import com.perdemia.permissionanalyzer.data.util.DomainCache;
import com.perdemia.permissionanalyzer.data.util.MemberCache;
import com.perdemia.permissionanalyzer.scanner.acl.JnaAceReader;
import com.sun.jna.platform.win32.Netapi32Util;
import com.sun.jna.platform.win32.Netapi32Util.Group;
import com.sun.jna.platform.win32.Netapi32Util.LocalGroupMember;
import com.sun.jna.platform.win32.Netapi32Util.LocalGroupMember.SID_NAME_USE;
import com.sun.jna.platform.win32.Win32Exception;

/**
 * Takes care of the communication with Netapi32 to retrieve local groups from a server and members from these local groups.
 * 
 * @author Reinier
 *
 */
public class LocalGroupHelper {
	private static Logger logger = LoggerFactory.getLogger(LocalGroupHelper.class.toString());
	
	private static int numberOfGroups = 0;
	private static int numberOfUsers = 0;
	
	/**
	 * <p>Retrieves all the local groups and the members from the given <code>server</code>.  Local group and group members are converted to
	 * {@link Member} objects and saved in the database. For each membership a new {@link Membership} record is created in the DB.
	 * </p><p>
	 * Once a <code>Member</code> is saved in the database it is hold in cache, so that no further DB communication is required for that member. It is important
	 * to keep the cache until all servers are scanned. (different servers can have the same domain groups in their local groups and we don't want duplicate entries in the
	 * DB).</p>
	 * 
	 * @param server server name or ip address
	 * 
	 * @see {@link #clearCache()}
	 */
	public static void processLocalGroups(String server, StatusListener statusListener) throws DBException, WinException {
		logger.debug("Processing local groups for server " + server);
		
		try {
			numberOfGroups = 0;
			numberOfUsers = 0;
			
			statusListener.updateStatus("Retrieving local groups from server " + server);
			Group[] groups = Netapi32Util.getLocalGroups(server);
			
			for (Group group : groups) {
				numberOfGroups++;
				
				if (statusListener.isCancelled()) {
					return;
				}
				
				// save this local group as Member in the database
				if (logger.isTraceEnabled()) {
					logger.trace("Found local group: " + group.name);
				}
				String groupName = group.name;
												
				if (!groupName.contains("\\")) {
					groupName = server + "\\" + groupName;
				}
				Member localGroup = new Member(groupName);
				localGroup = MemberCache.getMemberByName(localGroup.getName(false), localGroup.getDomainId());
				
				if (localGroup == null) {
					localGroup = new Member(groupName);
					localGroup.setGroupType(GroupType.COMPUTER_LOCAL);
					localGroup.setGroup(true);
					localGroup.setDisplayName(localGroup.getName(false));
					localGroup.setBuiltIn(JnaAceReader.isLocalBuiltinAccountByName(localGroup.getName(false)));
					localGroup.save();
								
					// It may occur that the license limit of the number of users/groups has been reached
					// in that case stop scanning groups for this server
					if (localGroup.getId() == null) {
						break; // quit
					} else {
						// add to cache, remember ID and continue
						MemberCache.cacheMemberByName(localGroup);
					}
					
					statusListener.groupsScanned(1);
				}
				
				// process the members of this local group
				statusListener.updateStatus("Processing members of local group " + group.name + " on server " + server);
				List<LocalGroupMember> members = null;
				try {
					members = Netapi32Util.getLocalGroupMembers(group.name, server);
				} catch (Win32Exception ex) {
					if (ex.getHR().intValue() == 0) {  //no members are found
						continue;
					} else {
						logger.error("Error while processing members of local group " + group.name + " on server " + server, ex);
						continue;
						//throw ex;
					}
				}
				
				for (LocalGroupMember localMember : members) {
					
					if (statusListener.isCancelled()) {
						return;
					}
					
					if (logger.isTraceEnabled()) {
						logger.trace("Processing next local group member: " + localMember.getName());
					}
					
					// determine the account name and domain object
					String accountName;
					int domainId;
					if (localMember.getName().contains("\\")) {
						accountName = localMember.getName().substring(localMember.getName().indexOf("\\") + 1);
						String domainName = localMember.getName().substring(0, localMember.getName().indexOf("\\"));
						domainId = DomainCache.getDomainId(domainName);
						
					} else {
						domainId = DomainCache.getDomainId(server.toUpperCase());
						accountName = localMember.getName();
					}
					
					// retrieve from cache
					Member member = MemberCache.getMemberByName(accountName, domainId);
					
					if (member == null) {
						// not in cache or database yet, create a new member and save in DB
						member = new Member(accountName);
						member.setDomainId(domainId);
						member.setDisplayName(member.getName(false));
						
						if (localMember.getSidType().equals(SID_NAME_USE.SidTypeGroup) ||
							localMember.getSidType().equals(SID_NAME_USE.SidTypeWellKnownGroup) ||
							localMember.getSidType().equals(SID_NAME_USE.SidTypeAlias)) { // alias for local group
							//group
							member.setGroup(true);
							member.setGroupType(GroupType.GLOBAL); // group type may be overridden in the LDAP scan
							
						} else if (localMember.getSidType().equals(SID_NAME_USE.SidTypeUser)) {
							//user, abuse the group type field no to support the filter "Include AD groups and users"
							member.setGroupType(GroupType.COMPUTER_LOCAL); // group type may be overridden in the LDAP scan
							
						} else if (localMember.getSidType().equals(SID_NAME_USE.SidTypeUnknown) ||
							localMember.getSidType().equals(SID_NAME_USE.SidTypeInvalid)) {
							
							// report error in GUI and log file
							logger.error(MessageFormat.format("Could not translate SID to name for member {0} ({1}) in local group {2} on server {3}.", localMember.getSid(), localMember.getName(), group.name, server));
							continue;
							
						} else if (localMember.getSidType().equals(SID_NAME_USE.SidTypeDeletedAccount) ){
							if (logger.isTraceEnabled()) {
								logger.trace("Skipped local group member because it is marked as deleted: " + localMember.getName());
							}
							continue;
						} else {
							// computer, label or domain SID Type >> ignore
							if (logger.isTraceEnabled()) {
								logger.trace("Skipped local group member because it is a computer, label or domain object: " + localMember.getName() + " - " + localMember.getSidType());
							}
							continue;
						}
						
						member.save();
						
						// It may occur that the license limit of the number of users/groups has been reached
						// in that case stop scanning groups for this server
						if (member.getId() != null) {
							// add to cache, remember ID and continue
							MemberCache.cacheMemberByName(member);
						}
						
						if (member.isGroup()) {
							statusListener.groupsScanned(1);
							numberOfGroups++;
						} else {
							statusListener.usersScanned(1);
							numberOfUsers++;
						}
					} else {
						if (logger.isTraceEnabled()) {
							logger.trace("Retrieved local group member from cache: " + localMember.getName());
						}
					}
					
					// create the membership in the DB
					if (member != null && member.getId() != null) {
						Membership membership = new Membership(localGroup.getId(), member.getId());
						membership.save();
					}
				}
			}
			logger.info(MessageFormat.format("Done scanning local groups on server {0}. Found {1} local groups and {2} users.", server, numberOfGroups, numberOfUsers));
			
		} catch (Win32Exception ex) {
			logger.error("Error while scanning local groups on server " + server, ex);
			throw new WinException(ex.getMessage(), ex);
		}
	}
		
	public static void main(String[] args) {
		try {
			System.out.println(InetAddress.getLocalHost().getHostName());
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
		/*
		try {
			List<LocalGroupMember> members = Netapi32Util.getLocalGroupMembers("blabla", null);
		} catch (Win32Exception ex) {
			System.out.println(ex.getHR().intValue());
			ex.printStackTrace();
		}*/
		
		/*
		try {
			processLocalGroups("localhost2", null);
		} catch (DBException | WinException e) {
			e.printStackTrace();
		}
	     */
	     
	     // \\WIN2K3R2EE
		/*
	     System.out.println(Netapi32Util.getDCName());
	     DomainTrust[] trusts = Netapi32Util.getDomainTrusts();
	     for (DomainTrust trust : trusts) {
	    	 System.out.println("DnsDomainName: " + trust.DnsDomainName); //testdomain.local
	    	 System.out.println("NetbiosDomainName: " + trust.NetbiosDomainName); //TESTDOMAIN
	     }*/
	}

	public static int getNumberOfGroups() {
		return numberOfGroups;
	}

	public static int getNumberOfUsers() {
		return numberOfUsers;
	}
}
