package com.perdemia.permissionanalyzer.scanner.groups;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.model.StatusListener;
import com.perdemia.permissionanalyzer.data.DBException;
import com.perdemia.permissionanalyzer.data.DBObject;
import com.perdemia.permissionanalyzer.data.Member;
import com.perdemia.permissionanalyzer.data.Membership;
import com.perdemia.permissionanalyzer.data.util.CustomProcessor;
import com.perdemia.permissionanalyzer.data.util.DBUtil;
import com.perdemia.permissionanalyzer.data.util.DBUtil.DbType;

/**
 * Determines nested group membership and completes the table MEMBERSHIP. This table does not only contain direct
 * memberships, but also the nested group memberships.
 * 
 * @author Reinier
 *
 */
public class NestedGroupHelper {
	private static Logger logger = LoggerFactory.getLogger(NestedGroupHelper.class.toString());
	
	/** Groups can occur in multiple nested group membership, just process them once. */
	private static List<Integer> processedGroupIds;
	
	private static Map<Integer, List<Membership>> cache = new HashMap<Integer, List<Membership>>();
	
	/**
	 * <p>Loops through all the MEMBER_ID's in the table MEMBER. For each member it determines all the nested group memberships and 
	 * adds them as new record (GROUP_ID, MEMBER_ID, VIA_GROUP_ID) in the table.
	 * </p>
	 * 
	 */
	public static void processNestedGroupMemberships(final StatusListener statusListener) throws DBException {
		logger.debug("Processing nested group membership");
		PreparedStatement ps = null;
		Connection connection = null;
		
		try {
			if (DBUtil.getDbType().equals(DbType.MYSQL)) {
				// MySQL cannot handle two open RecordSets on one connection. Since we loop through the record set to save memory and
				// we query new information in this loop, we need a new DB connection in case of MySQL
				connection = DBUtil.getDBConnection(false);
				ps = connection.prepareStatement(Member.getAllQuery(), ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);
				ps.setFetchSize(Integer.MIN_VALUE);
				
			} else {
				// use the existing connection and create a new PreparedStatement
				connection = DBUtil.getDBConnection();
				ps = connection.prepareStatement(Member.getAllQuery());
			}
			
			DBObject.getUsingCustomProcessor(ps, new CustomProcessor(){
				@Override
				public boolean processResultSet(ResultSet rs) throws SQLException, DBException {
					int userId = rs.getInt("id");
					String username = rs.getString("NAME");
					
					if (statusListener != null) {
						statusListener.updateStatus("Determining nested group membership for " + username);
					}
					
					processedGroupIds = new ArrayList<Integer>();
					
					// get all direct group membership
					List<Membership> memberships = Membership.where("MEMBER_ID = ? AND VIA_GROUP_ID IS NULL", new Object[]{userId});
					
					// loop through all direct memberships, and get the (nested) members of each group
					for (Membership membership : memberships) {
						processNestedGroupMembership(membership.getGroupId(), userId);
					}
					
					// continue? (meaning not cancelled)
					return statusListener == null || !statusListener.isCancelled();				
				}
			});
			
		} catch (SQLException ex) {
			throw new DBException("SQL error while determining nested group memberships", ex);
			
		} finally {
			cache.clear();
			
			// In case of MySQL a new connection has been created, close it
			if (DBUtil.getDbType().equals(DbType.MYSQL) && connection != null) {
				try {
					connection.close();
				} catch (SQLException e) {
					logger.error("Error while closing database connection in NestedGroupHelper", e);
				}
			}
		}
		
		logger.info("Done determining nested group membership.");
	}
	
	/**
	 * Recursive method to loop through all the nested group memberships of a user.
	 * 
	 * @param groupId next groupId to process
	 * @param userId userId
	 * 
	 * @throws DBException
	 */
	private static void processNestedGroupMembership(int groupId, int userId) throws DBException {
		
		// the membership table contains rows with the same group_id and member_id (to be included in some of the filters). Skips those rows.
		if (groupId != userId && !processedGroupIds.contains(groupId)) {
			
			// process once
			processedGroupIds.add(groupId);
				
			// get all direct group memberships of the given groupId 
			if (cache.get(groupId) == null) {
				
				// cache limit reached?
				if (cache.size() > 10000) {
					cache.clear();
				}
				
				List<Membership> memberOfGroups = Membership.where("MEMBER_ID = ? AND VIA_GROUP_ID IS NULL", new Object[]{groupId});
				cache.put(groupId, memberOfGroups);
			}
			
			List<Membership> memberOfGroups = cache.get(groupId);
			
			// loop through all direct memberships, and get the members of each group
			for (Membership membership : memberOfGroups) {
				
				if (membership.getGroupId() != userId && membership.getGroupId() != groupId) {

					// create new record
					Membership newMembership = new Membership(membership.getGroupId(), userId, groupId);
					newMembership.save();
					
					// get all nested group membership by calling this method again
					processNestedGroupMembership(membership.getGroupId(), userId);
				}
			}
		}
	}
}
