package com.perdemia.permissionanalyzer.scanner.ldap;

import org.eclipse.jface.dialogs.MessageDialog;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perdemia.permissionanalyzer.application.jface.dialogs.SyncMessageDialog;
import com.perdemia.permissionanalyzer.application.model.LdapConnectionInfo;

public class LdapFactory {
	private static Logger logger = LoggerFactory.getLogger(LdapFactory.class.getName());
	
	/**
	 * Returns a instance to the LDAP connector implementation. This method serves as single point of definition, so we can easily switch to
	 * the stub implementation ({@link LdapConnectorStub}).
	 * 
	 * @return an instance to a {@link LdapConnector}}
	 */
	public static LdapConnector getLdapConnector() {
		return new LdapConnectorImpl();
	}
	
	/**
	 * Opens the LDAP connection and handles error messages.
	 * 
	 * @param connectionInfo connection details
	 * @param ldapConnector ldapConnector implementation to use
	 * 
	 * @return <code>true</code> in case the connection was opened, <code>false</code> in case an exception occured
	 */
	public static boolean openConnection(LdapConnectionInfo connectionInfo, LdapConnector ldapConnector) {
		try {
			ldapConnector.openConnection(connectionInfo);
		} catch (Exception ex) {
			logger.error("Error while opening the LDAP connection.", ex);
			logger.info(connectionInfo.toString());
			String detailMessage = ex.toString();
			String userFriendlyMessage = getUserFriendlyLdapError(ex.toString());
			
			if (userFriendlyMessage != null) {
				detailMessage = detailMessage + "\r\n\r\n" + userFriendlyMessage;
			}
			
			SyncMessageDialog.showMessage(MessageDialog.ERROR, "LDAP connection error", "Error while opening the LDAP connection:\r\n\r\n" + detailMessage);
			return false;
		}
		
		return true;
	}
	
	public static String getUserFriendlyLdapError(String ldapError) {
		if (ldapError.contains("Connection refused")) {
			return "Connection refused means the server be reached, but that the port was blocked or invalid. Perhaps a firewall is blocking the connection or the hostport is not correct.";
		}
		if (ldapError.contains("Connection timed out")) {
			return "Connection timed out means the server could not be reached. Perhaps a firewall is blocking the connection or the hostname is not correct.";
		}
		if (ldapError.contains("AcceptSecurityContext")) {
			return "Please make sure the user credentials are corret and that the username includes the domain, like domainadmin@mydomain.com";
		}
		if (ldapError.contains("SSLHandshakeException")) {
			return "Please make sure the server has SSL enabled and that the port number is the correct SSL port.";
		}
		return null;
	}
}
